/* Generated by re2c 1.1.1 on Fri May 27 20:11:39 2022 */
#line 1 "tokenizer.re"
#include <limits>

#include <lfortran/parser/parser_exception.h>
#include <lfortran/parser/tokenizer.h>
#include <lfortran/parser/parser.tab.hh>
#include <libasr/bigint.h>

namespace LFortran
{

void lex_format(unsigned char *&cur, Location &loc,
        unsigned char *&start);

void Tokenizer::set_string(const std::string &str)
{
    // The input string must be NULL terminated, otherwise the tokenizer will
    // not detect the end of string. After C++11, the std::string is guaranteed
    // to end with \0, but we check this here just in case.
    LFORTRAN_ASSERT(str[str.size()] == '\0');
    cur = (unsigned char *)(&str[0]);
    string_start = cur;
    cur_line = cur;
    line_num = 1;
}

template<int base>
bool adddgt(uint64_t &u, uint64_t d)
{
    if (u > (std::numeric_limits<uint64_t>::max() - d) / base) {
        return false;
    }
    u = u * base + d;
    return true;
}

bool lex_dec(const unsigned char *s, const unsigned char *e, uint64_t &u)
{
    for (u = 0; s < e; ++s) {
        if (!adddgt<10>(u, *s - 0x30u)) {
            return false;
        }
    }
    return true;
}

// Tokenizes integer of the kind 1234_ikind into `u` and `suffix`
// s ... the start of the integer
// e ... the character after the end
bool lex_int(const unsigned char *s, const unsigned char *e, uint64_t &u,
    Str &suffix)
{
    for (u = 0; s < e; ++s) {
        if (*s == '_') {
            s++;
            suffix.p = (char*) s;
            suffix.n = e-s;
            return true;
        } else if (!adddgt<10>(u, *s - 0x30u)) {
            return false;
        }
    }
    suffix.p = nullptr;
    suffix.n = 0;
    return true;
}

void lex_int_large(Allocator &al, const unsigned char *s,
    const unsigned char *e, BigInt::BigInt &u, Str &suffix)
{
    uint64_t ui;
    if (lex_int(s, e, ui, suffix)) {
        if (ui <= BigInt::MAX_SMALL_INT) {
            u.from_smallint(ui);
            return;
        }
    }
    const unsigned char *start = s;
    for (; s < e; ++s) {
        if (*s == '_') {
            s++;
            suffix.p = (char*) s;
            suffix.n = e-s;

            Str num;
            num.p = (char*)start;
            num.n = s-start-1;
            u.from_largeint(al, num);
            return;
        }
    }
    suffix.p = nullptr;
    suffix.n = 0;
    Str num;
    num.p = (char*)start;
    num.n = e-start;
    u.from_largeint(al, num);
}

uint64_t parse_int(const unsigned char *s)
{
    while (*s == ' ') s++;
    uint64_t u;
    for (u = 0; ; ++s) {
        if (*s >= '0' && *s <= '9') {
            if (!adddgt<10>(u, *s - 0x30u)) {
                return false;
            }
        } else {
            return u;
        }
    }
}

#define KW(x) token(yylval.string); RET(KW_##x);
#define RET(x) token_loc(loc); last_token=yytokentype::x; return yytokentype::x;
#define WARN_REL(x) add_rel_warning(diagnostics, yytokentype::TK_##x);

void Tokenizer::add_rel_warning(diag::Diagnostics &diagnostics, int rel_token) const {
    static const std::map<int, std::pair<std::string, std::string>> m = {
        {yytokentype::TK_EQ, {"==", ".eq."}},
        {yytokentype::TK_NE, {"/=", ".ne."}},
        {yytokentype::TK_LT, {"<",  ".lt."}},
        {yytokentype::TK_GT, {">",  ".gt."}},
        {yytokentype::TK_LE, {"<=", ".le."}},
        {yytokentype::TK_GE, {">=", ".ge."}},
    };
    const std::string rel_new = m.at(rel_token).first;
    const std::string rel_old = m.at(rel_token).second;
    Location loc;
    token_loc(loc);
    diagnostics.tokenizer_style_label(
        "Use '" + rel_new + "' instead of '" + rel_old + "'",
        {loc},
        "help: write this as '" + rel_new + "'");
}

int Tokenizer::lex(Allocator &al, YYSTYPE &yylval, Location &loc, diag::Diagnostics &diagnostics)
{
    if (enddo_state == 1) {
        enddo_state = 2;
        KW(END_DO)
    } else if (enddo_state == 2) {
        enddo_insert_count--;
        if (enddo_insert_count > 0) {
            enddo_state = 1;
        } else {
            enddo_state = 0;
            token_loc(loc); line_num++; cur_line=cur;
            last_token = yytokentype::TK_NEWLINE;
        }
        return yytokentype::TK_NEWLINE;
    }
    for (;;) {
        tok = cur;

        /*
        Re2c has excellent documentation at:

        https://re2c.org/manual/manual_c.html

        The first paragraph there explains the basics:

        * If multiple rules match, the longest match takes precedence
        * If multiple rules match the same string, the earlier rule takes
          precedence
        * Default rule `*` should always be defined, it has the lowest priority
          regardless of its place and matches any code unit
        * We use the "Sentinel character" method for end of input:
            * The end of the input text is denoted with a null character \x00
            * Thus the null character cannot be part of the input otherwise
            * There is one rule to match \x00 to end the parser
            * No other rule is allowed to match \x00, otherwise the re2c block
              would parse past the end of the string and segfaults
            * A special case of the previous point are negated character
              ranges, such as [^"\x00], where one must include \x00 in it to
              ensure this rule does not match \x00 (all other rules simply do
              not mention \x00)
            * See the "Handling the end of input" section in the re2c
              documentation for more info

        The re2c block interacts with the rest of the code via just one pointer
        variable `cur`. On entering the re2c block, the `cur` variable must
        point to the first character of the token to be tokenized by the block.
        The re2c block below then executes on its own until a rule is matched:
        the action in {} is then executed. In that action `cur` points to the
        first character of the next token.

        Before the re2c block we save the current `cur` into `tok`, so that we
        can use `tok` and `cur` in the action in {} to extract the token that
        corresponds to the rule that got matched:

        * `tok` points to the first character of the token
        * `cur-1` points to the last character of the token
        * `cur` points to the first character of the next token
        * `cur-tok` is the length of the token

        In the action, we do one of:

        * call `continue` which executes another cycle in the for loop (which
          will parse the next token); we use this to skip a token
        * call `return` which returns from this function; we return a token
        * throw an exception (terminates the tokenizer)

        In the first two cases, `cur` points to first character of the next
        token, which becomes `tok` at the next iteration of the loop (either
        right away after `continue` or after the `lex` function is called again
        after `return`).

        See the manual for more details.
        */


        // These two variables are needed by the re2c block below internally,
        // initialization is not needed. One can think of them as local
        // variables of the re2c block.
        unsigned char *mar, *ctxmar;
        
#line 221 "tokenizer.cpp"
{
	unsigned char yych;
	unsigned int yyaccept = 0;
	static const unsigned char yybm[] = {
		/* table 1 .. 8: 0 */
		  0,   2,   2,   2,   2,   2,   2,   2, 
		  2,   3,   3,   3,   2,   3,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  3,   2,   2,   2,   2,   2,   2,   0, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		242, 242, 210, 210, 210, 210, 210, 210, 
		146, 146,   2,   2,   2,   2,   2,   2, 
		  2, 150, 150, 150, 150, 150, 150,  22, 
		 22,  22,  22,  22,  22,  22,  22,  22, 
		 22,  22,  22,  22,  22,  22,  22,  22, 
		 22,  22,  22,   2,   2,   2,   2,  26, 
		  2, 150, 150, 150, 150, 150, 150,  22, 
		 22,  22,  22,  22,  22,  22,  22,  22, 
		 22,  22,  22,  22,  22,  22,  22,  22, 
		 22,  22,  22,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		  2,   2,   2,   2,   2,   2,   2,   2, 
		/* table 9 .. 13: 256 */
		  0, 192, 192, 192, 192, 192, 192, 192, 
		192, 200, 128, 200, 192, 200, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		200, 192,  64, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		240, 240, 240, 240, 240, 240, 240, 240, 
		240, 240, 192, 192, 192, 192, 192, 192, 
		192, 224, 224, 224, 224, 224, 224, 224, 
		224, 224, 224, 224, 224, 224, 224, 224, 
		224, 224, 224, 224, 224, 224, 224, 224, 
		224, 224, 224, 192, 192, 192, 192, 192, 
		192, 224, 224, 224, 224, 224, 224, 224, 
		224, 224, 224, 224, 224, 224, 224, 224, 
		224, 224, 224, 224, 224, 224, 224, 224, 
		224, 224, 224, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
	};
	yych = *cur;
	if (yybm[256+yych] & 8) {
		goto yy6;
	}
	switch (yych) {
	case 0x00:	goto yy2;
	case '\t':
	case '\n':	goto yy9;
	case ' ':
	case '!':	goto yy11;
	case '"':	goto yy12;
	case '#':	goto yy13;
	case '%':	goto yy14;
	case '&':	goto yy16;
	case '\'':	goto yy17;
	case '(':	goto yy18;
	case ')':	goto yy20;
	case '*':	goto yy22;
	case '+':	goto yy24;
	case ',':	goto yy26;
	case '-':	goto yy28;
	case '.':	goto yy30;
	case '/':	goto yy31;
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':	goto yy33;
	case ':':	goto yy36;
	case ';':	goto yy38;
	case '<':	goto yy40;
	case '=':	goto yy42;
	case '>':	goto yy44;
	case 'A':
	case 'a':	goto yy46;
	case 'B':
	case 'b':	goto yy48;
	case 'C':
	case 'c':	goto yy49;
	case 'D':
	case 'd':	goto yy50;
	case 'E':
	case 'e':	goto yy51;
	case 'F':
	case 'f':	goto yy52;
	case 'G':
	case 'g':	goto yy53;
	case 'H':
	case 'J':
	case 'X':
	case 'Y':
	case '_':
	case 'h':
	case 'j':
	case 'x':
	case 'y':	goto yy54;
	case 'I':
	case 'i':	goto yy56;
	case 'K':
	case 'k':	goto yy57;
	case 'L':
	case 'l':	goto yy58;
	case 'M':
	case 'm':	goto yy59;
	case 'N':
	case 'n':	goto yy60;
	case 'O':
	case 'o':	goto yy61;
	case 'P':
	case 'p':	goto yy62;
	case 'Q':
	case 'q':	goto yy63;
	case 'R':
	case 'r':	goto yy64;
	case 'S':
	case 's':	goto yy65;
	case 'T':
	case 't':	goto yy66;
	case 'U':
	case 'u':	goto yy67;
	case 'V':
	case 'v':	goto yy68;
	case 'W':
	case 'w':	goto yy69;
	case 'Z':
	case 'z':	goto yy70;
	case '[':	goto yy71;
	case ']':	goto yy73;
	case '|':	goto yy75;
	default:	goto yy4;
	}
yy2:
	++cur;
#line 250 "tokenizer.re"
	{ RET(END_OF_FILE); }
#line 393 "tokenizer.cpp"
yy4:
	++cur;
yy5:
#line 241 "tokenizer.re"
	{ token_loc(loc);
                std::string t = token();
                throw parser_local::TokenizerError(diag::Diagnostic(
                    "Token '" + t + "' is not recognized",
                    diag::Level::Error, diag::Stage::Tokenizer, {
                        diag::Label("token not recognized", {loc})
                    })
                );
            }
#line 407 "tokenizer.cpp"
yy6:
	yych = *++cur;
	if (yybm[256+yych] & 8) {
		goto yy6;
	}
#line 251 "tokenizer.re"
	{ continue; }
#line 415 "tokenizer.cpp"
yy9:
	++cur;
#line 502 "tokenizer.re"
	{
                if (enddo_newline_process) {
                    enddo_newline_process = false;
                    enddo_state = 1;
                    return yytokentype::TK_NEWLINE;
                } else {
                    enddo_newline_process = false;
                    enddo_insert_count = 0;
                    token_loc(loc); line_num++; cur_line=cur;
                    last_token = yytokentype::TK_NEWLINE;
                    return yytokentype::TK_NEWLINE;
                }
            }
#line 432 "tokenizer.cpp"
yy11:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= 0x00) goto yy5;
	goto yy78;
yy12:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= 0x00) goto yy5;
	goto yy83;
yy13:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= 0x00) goto yy5;
	goto yy87;
yy14:
	++cur;
#line 535 "tokenizer.re"
	{ RET(TK_PERCENT) }
#line 452 "tokenizer.cpp"
yy16:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= '\f') {
		if (yych <= 0x08) goto yy5;
		if (yych <= '\v') goto yy91;
		goto yy5;
	} else {
		if (yych <= '\r') goto yy91;
		if (yych <= 0x1F) goto yy5;
		if (yych <= '!') goto yy91;
		goto yy5;
	}
yy17:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= 0x00) goto yy5;
	goto yy98;
yy18:
	yych = *++cur;
	if (yych == '/') {
		ctxmar = cur;
		goto yy101;
	}
#line 517 "tokenizer.re"
	{ RET(TK_LPAREN) }
#line 479 "tokenizer.cpp"
yy20:
	++cur;
#line 525 "tokenizer.re"
	{ RET(TK_RPAREN) }
#line 484 "tokenizer.cpp"
yy22:
	yych = *++cur;
	if (yych == '*') goto yy102;
#line 537 "tokenizer.re"
	{ RET(TK_STAR) }
#line 490 "tokenizer.cpp"
yy24:
	++cur;
#line 529 "tokenizer.re"
	{ RET(TK_PLUS) }
#line 495 "tokenizer.cpp"
yy26:
	++cur;
#line 536 "tokenizer.re"
	{ RET(TK_COMMA) }
#line 500 "tokenizer.cpp"
yy28:
	++cur;
#line 530 "tokenizer.re"
	{ RET(TK_MINUS) }
#line 505 "tokenizer.cpp"
yy30:
	yyaccept = 0;
	yych = *(mar = ++cur);
	switch (yych) {
	case '.':	goto yy104;
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':	goto yy106;
	case 'A':
	case 'a':	goto yy109;
	case 'B':
	case 'C':
	case 'D':
	case 'H':
	case 'I':
	case 'J':
	case 'K':
	case 'M':
	case 'P':
	case 'Q':
	case 'R':
	case 'S':
	case 'U':
	case 'V':
	case 'W':
	case 'Y':
	case 'Z':
	case 'b':
	case 'c':
	case 'd':
	case 'h':
	case 'i':
	case 'j':
	case 'k':
	case 'm':
	case 'p':
	case 'q':
	case 'r':
	case 's':
	case 'u':
	case 'v':
	case 'w':
	case 'y':
	case 'z':	goto yy110;
	case 'E':
	case 'e':	goto yy112;
	case 'F':
	case 'f':	goto yy113;
	case 'G':
	case 'g':	goto yy114;
	case 'L':
	case 'l':	goto yy115;
	case 'N':
	case 'n':	goto yy116;
	case 'O':
	case 'o':	goto yy117;
	case 'T':
	case 't':	goto yy118;
	case 'X':
	case 'x':	goto yy119;
	default:	goto yy5;
	}
yy31:
	yych = *++cur;
	if (yych <= '.') {
		if (yych == ')') goto yy120;
	} else {
		if (yych <= '/') goto yy122;
		if (yych == '=') goto yy124;
	}
#line 534 "tokenizer.re"
	{ RET(TK_SLASH) }
#line 585 "tokenizer.cpp"
yy33:
	yyaccept = 1;
	yych = *(mar = ++cur);
	if (yybm[256+yych] & 16) {
		goto yy33;
	}
	if (yych <= ' ') {
		if (yych <= '\v') {
			if (yych == '\t') {
				ctxmar = cur;
				goto yy126;
			}
			if (yych >= '\v') {
				ctxmar = cur;
				goto yy126;
			}
		} else {
			if (yych == '\r') {
				ctxmar = cur;
				goto yy126;
			}
			if (yych >= ' ') {
				ctxmar = cur;
				goto yy126;
			}
		}
	} else {
		if (yych <= 'E') {
			if (yych == '.') {
				ctxmar = cur;
				goto yy128;
			}
			if (yych >= 'D') goto yy129;
		} else {
			if (yych <= '_') {
				if (yych >= '_') goto yy130;
			} else {
				if (yych <= 'c') goto yy35;
				if (yych <= 'e') goto yy129;
			}
		}
	}
yy35:
#line 622 "tokenizer.re"
	{
                lex_int_large(al, tok, cur,
                    yylval.int_suffix.int_n,
                    yylval.int_suffix.int_kind);
                RET(TK_INTEGER)
            }
#line 636 "tokenizer.cpp"
yy36:
	yych = *++cur;
	if (yych == ':') goto yy131;
#line 532 "tokenizer.re"
	{ RET(TK_COLON) }
#line 642 "tokenizer.cpp"
yy38:
	++cur;
#line 533 "tokenizer.re"
	{ RET(TK_SEMICOLON) }
#line 647 "tokenizer.cpp"
yy40:
	yych = *++cur;
	if (yych == '=') goto yy133;
#line 554 "tokenizer.re"
	{ RET(TK_LT) }
#line 653 "tokenizer.cpp"
yy42:
	yych = *++cur;
	if (yych <= '<') goto yy43;
	if (yych <= '=') goto yy135;
	if (yych <= '>') goto yy137;
yy43:
#line 531 "tokenizer.re"
	{ RET(TK_EQUAL) }
#line 662 "tokenizer.cpp"
yy44:
	yych = *++cur;
	if (yych == '=') goto yy139;
#line 560 "tokenizer.re"
	{ RET(TK_GT) }
#line 668 "tokenizer.cpp"
yy46:
	yych = *++cur;
	if (yych <= 'S') {
		if (yych <= 'K') {
			if (yych == 'B') goto yy141;
			goto yy55;
		} else {
			if (yych <= 'L') goto yy142;
			if (yych <= 'R') goto yy55;
			goto yy143;
		}
	} else {
		if (yych <= 'k') {
			if (yych == 'b') goto yy141;
			goto yy55;
		} else {
			if (yych <= 'l') goto yy142;
			if (yych == 's') goto yy143;
			goto yy55;
		}
	}
yy47:
#line 664 "tokenizer.re"
	{ token(yylval.string); RET(TK_NAME) }
#line 693 "tokenizer.cpp"
yy48:
	yyaccept = 2;
	yych = *(mar = ++cur);
	if (yych <= 'I') {
		if (yych <= '\'') {
			if (yych == '"') goto yy146;
			if (yych <= '&') goto yy55;
			goto yy147;
		} else {
			if (yych == 'A') goto yy148;
			if (yych <= 'H') goto yy55;
			goto yy149;
		}
	} else {
		if (yych <= 'a') {
			if (yych == 'L') goto yy150;
			if (yych <= '`') goto yy55;
			goto yy148;
		} else {
			if (yych <= 'i') {
				if (yych <= 'h') goto yy55;
				goto yy149;
			} else {
				if (yych == 'l') goto yy150;
				goto yy55;
			}
		}
	}
yy49:
	yych = *++cur;
	switch (yych) {
	case 'A':
	case 'a':	goto yy151;
	case 'H':
	case 'h':	goto yy152;
	case 'L':
	case 'l':	goto yy153;
	case 'O':
	case 'o':	goto yy154;
	case 'R':
	case 'r':	goto yy155;
	case 'Y':
	case 'y':	goto yy156;
	default:	goto yy55;
	}
yy50:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych <= 'E') {
			if (yych == 'A') goto yy157;
			if (yych <= 'D') goto yy55;
			goto yy158;
		} else {
			if (yych == 'I') goto yy159;
			if (yych <= 'N') goto yy55;
			goto yy160;
		}
	} else {
		if (yych <= 'e') {
			if (yych == 'a') goto yy157;
			if (yych <= 'd') goto yy55;
			goto yy158;
		} else {
			if (yych <= 'i') {
				if (yych <= 'h') goto yy55;
				goto yy159;
			} else {
				if (yych == 'o') goto yy160;
				goto yy55;
			}
		}
	}
yy51:
	yych = *++cur;
	switch (yych) {
	case 'L':
	case 'l':	goto yy162;
	case 'N':
	case 'n':	goto yy163;
	case 'Q':
	case 'q':	goto yy164;
	case 'R':
	case 'r':	goto yy165;
	case 'V':
	case 'v':	goto yy166;
	case 'X':
	case 'x':	goto yy167;
	default:	goto yy55;
	}
yy52:
	yych = *++cur;
	switch (yych) {
	case 'I':
	case 'i':	goto yy168;
	case 'L':
	case 'l':	goto yy169;
	case 'O':
	case 'o':	goto yy170;
	case 'U':
	case 'u':	goto yy171;
	default:	goto yy55;
	}
yy53:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych == 'E') goto yy172;
		if (yych <= 'N') goto yy55;
		goto yy173;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy55;
			goto yy172;
		} else {
			if (yych == 'o') goto yy173;
			goto yy55;
		}
	}
yy54:
	yych = *++cur;
yy55:
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
	goto yy47;
yy56:
	yych = *++cur;
	if (yych <= 'S') {
		if (yych <= 'L') {
			if (yych == 'F') goto yy175;
			goto yy55;
		} else {
			if (yych <= 'M') goto yy177;
			if (yych <= 'N') goto yy178;
			if (yych <= 'R') goto yy55;
			goto yy180;
		}
	} else {
		if (yych <= 'm') {
			if (yych == 'f') goto yy175;
			if (yych <= 'l') goto yy55;
			goto yy177;
		} else {
			if (yych <= 'n') goto yy178;
			if (yych == 's') goto yy180;
			goto yy55;
		}
	}
yy57:
	yych = *++cur;
	if (yych == 'I') goto yy182;
	if (yych == 'i') goto yy182;
	goto yy55;
yy58:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych == 'E') goto yy183;
		if (yych <= 'N') goto yy55;
		goto yy184;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy55;
			goto yy183;
		} else {
			if (yych == 'o') goto yy184;
			goto yy55;
		}
	}
yy59:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych == 'E') goto yy185;
		if (yych <= 'N') goto yy55;
		goto yy186;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy55;
			goto yy185;
		} else {
			if (yych == 'o') goto yy186;
			goto yy55;
		}
	}
yy60:
	yych = *++cur;
	if (yych <= 'U') {
		if (yych <= 'E') {
			if (yych == 'A') goto yy187;
			if (yych <= 'D') goto yy55;
			goto yy188;
		} else {
			if (yych == 'O') goto yy189;
			if (yych <= 'T') goto yy55;
			goto yy190;
		}
	} else {
		if (yych <= 'e') {
			if (yych == 'a') goto yy187;
			if (yych <= 'd') goto yy55;
			goto yy188;
		} else {
			if (yych <= 'o') {
				if (yych <= 'n') goto yy55;
				goto yy189;
			} else {
				if (yych == 'u') goto yy190;
				goto yy55;
			}
		}
	}
yy61:
	yyaccept = 2;
	yych = *(mar = ++cur);
	if (yych <= 'P') {
		if (yych <= '\'') {
			if (yych == '"') goto yy191;
			if (yych <= '&') goto yy55;
			goto yy192;
		} else {
			if (yych == 'N') goto yy193;
			if (yych <= 'O') goto yy55;
			goto yy194;
		}
	} else {
		if (yych <= 'n') {
			if (yych == 'U') goto yy195;
			if (yych <= 'm') goto yy55;
			goto yy193;
		} else {
			if (yych <= 'p') {
				if (yych <= 'o') goto yy55;
				goto yy194;
			} else {
				if (yych == 'u') goto yy195;
				goto yy55;
			}
		}
	}
yy62:
	yych = *++cur;
	if (yych <= 'U') {
		if (yych <= 'O') {
			if (yych == 'A') goto yy196;
			if (yych <= 'N') goto yy55;
			goto yy197;
		} else {
			if (yych == 'R') goto yy198;
			if (yych <= 'T') goto yy55;
			goto yy199;
		}
	} else {
		if (yych <= 'o') {
			if (yych == 'a') goto yy196;
			if (yych <= 'n') goto yy55;
			goto yy197;
		} else {
			if (yych <= 'r') {
				if (yych <= 'q') goto yy55;
				goto yy198;
			} else {
				if (yych == 'u') goto yy199;
				goto yy55;
			}
		}
	}
yy63:
	yych = *++cur;
	if (yych == 'U') goto yy200;
	if (yych == 'u') goto yy200;
	goto yy55;
yy64:
	yych = *++cur;
	if (yych <= 'E') {
		if (yych == 'A') goto yy201;
		if (yych <= 'D') goto yy55;
		goto yy202;
	} else {
		if (yych <= 'a') {
			if (yych <= '`') goto yy55;
			goto yy201;
		} else {
			if (yych == 'e') goto yy202;
			goto yy55;
		}
	}
yy65:
	yych = *++cur;
	switch (yych) {
	case 'A':
	case 'a':	goto yy203;
	case 'E':
	case 'e':	goto yy204;
	case 'H':
	case 'h':	goto yy205;
	case 'O':
	case 'o':	goto yy206;
	case 'T':
	case 't':	goto yy207;
	case 'U':
	case 'u':	goto yy208;
	case 'Y':
	case 'y':	goto yy209;
	default:	goto yy55;
	}
yy66:
	yych = *++cur;
	switch (yych) {
	case 'A':
	case 'a':	goto yy210;
	case 'E':
	case 'e':	goto yy211;
	case 'H':
	case 'h':	goto yy212;
	case 'O':
	case 'o':	goto yy213;
	case 'Y':
	case 'y':	goto yy215;
	default:	goto yy55;
	}
yy67:
	yych = *++cur;
	if (yych <= 'S') {
		if (yych == 'N') goto yy216;
		if (yych <= 'R') goto yy55;
		goto yy217;
	} else {
		if (yych <= 'n') {
			if (yych <= 'm') goto yy55;
			goto yy216;
		} else {
			if (yych == 's') goto yy217;
			goto yy55;
		}
	}
yy68:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych == 'A') goto yy218;
		if (yych <= 'N') goto yy55;
		goto yy219;
	} else {
		if (yych <= 'a') {
			if (yych <= '`') goto yy55;
			goto yy218;
		} else {
			if (yych == 'o') goto yy219;
			goto yy55;
		}
	}
yy69:
	yych = *++cur;
	if (yych <= 'R') {
		if (yych <= 'G') {
			if (yych == 'A') goto yy220;
			goto yy55;
		} else {
			if (yych <= 'H') goto yy221;
			if (yych <= 'Q') goto yy55;
			goto yy222;
		}
	} else {
		if (yych <= 'g') {
			if (yych == 'a') goto yy220;
			goto yy55;
		} else {
			if (yych <= 'h') goto yy221;
			if (yych == 'r') goto yy222;
			goto yy55;
		}
	}
yy70:
	yyaccept = 2;
	yych = *(mar = ++cur);
	if (yych == '"') goto yy223;
	if (yych == '\'') goto yy224;
	goto yy55;
yy71:
	++cur;
yy72:
#line 526 "tokenizer.re"
	{ RET(TK_LBRACKET) }
#line 1075 "tokenizer.cpp"
yy73:
	++cur;
#line 527 "tokenizer.re"
	{ RET(TK_RBRACKET) }
#line 1080 "tokenizer.cpp"
yy75:
	++cur;
#line 538 "tokenizer.re"
	{ RET(TK_VBAR) }
#line 1085 "tokenizer.cpp"
yy77:
	yych = *++cur;
yy78:
	if (yybm[256+yych] & 64) {
		goto yy77;
	}
	if (yych >= 0x01) goto yy80;
yy79:
	cur = mar;
	if (yyaccept <= 8) {
		if (yyaccept <= 4) {
			if (yyaccept <= 2) {
				if (yyaccept <= 1) {
					if (yyaccept == 0) {
						goto yy5;
					} else {
						goto yy35;
					}
				} else {
					goto yy47;
				}
			} else {
				if (yyaccept == 3) {
					goto yy85;
				} else {
					goto yy94;
				}
			}
		} else {
			if (yyaccept <= 6) {
				if (yyaccept == 5) {
					goto yy100;
				} else {
					goto yy72;
				}
			} else {
				if (yyaccept == 7) {
					goto yy108;
				} else {
					goto yy161;
				}
			}
		}
	} else {
		if (yyaccept <= 13) {
			if (yyaccept <= 11) {
				if (yyaccept <= 10) {
					if (yyaccept == 9) {
						goto yy179;
					} else {
						goto yy294;
					}
				} else {
					goto yy757;
				}
			} else {
				if (yyaccept == 12) {
					goto yy904;
				} else {
					goto yy980;
				}
			}
		} else {
			if (yyaccept <= 15) {
				if (yyaccept == 14) {
					goto yy1079;
				} else {
					goto yy1165;
				}
			} else {
				if (yyaccept == 16) {
					goto yy1283;
				} else {
					goto yy1285;
				}
			}
		}
	}
yy80:
	++cur;
#line 640 "tokenizer.re"
	{
                line_num++; cur_line=cur;
                token(yylval.string);
                yylval.string.n--;
                token_loc(loc);
                if (last_token == yytokentype::TK_NEWLINE) {
                    return yytokentype::TK_COMMENT;
                } else {
                    last_token=yytokentype::TK_NEWLINE;
                    return yytokentype::TK_EOLCOMMENT;
                }
            }
#line 1179 "tokenizer.cpp"
yy82:
	yych = *++cur;
yy83:
	if (yybm[256+yych] & 128) {
		goto yy82;
	}
	if (yych <= 0x00) goto yy79;
	yyaccept = 3;
	yych = *(mar = ++cur);
	if (yych == '"') goto yy82;
yy85:
#line 660 "tokenizer.re"
	{ token_str(yylval.string); RET(TK_STRING) }
#line 1193 "tokenizer.cpp"
yy86:
	yych = *++cur;
yy87:
	if (yych <= 0x00) goto yy79;
	if (yych != '\n') goto yy86;
	++cur;
#line 654 "tokenizer.re"
	{ line_num++; cur_line=cur; continue; }
#line 1202 "tokenizer.cpp"
yy90:
	yych = *++cur;
yy91:
	if (yych <= '\f') {
		if (yych <= '\t') {
			if (yych <= 0x08) goto yy79;
			goto yy90;
		} else {
			if (yych <= '\n') goto yy92;
			if (yych <= '\v') goto yy90;
			goto yy79;
		}
	} else {
		if (yych <= 0x1F) {
			if (yych <= '\r') goto yy90;
			goto yy79;
		} else {
			if (yych <= ' ') goto yy90;
			if (yych <= '!') goto yy95;
			goto yy79;
		}
	}
yy92:
	yyaccept = 4;
	yych = *(mar = ++cur);
	if (yybm[0+yych] & 1) {
		goto yy92;
	}
	if (yych <= 0x1F) goto yy94;
	if (yych <= '!') goto yy95;
	if (yych == '&') goto yy225;
yy94:
#line 636 "tokenizer.re"
	{
                line_num++; cur_line=cur; continue;
            }
#line 1239 "tokenizer.cpp"
yy95:
	yych = *++cur;
	if (yych <= 0x00) goto yy79;
	if (yych == '\n') goto yy92;
	goto yy95;
yy97:
	yych = *++cur;
yy98:
	if (yybm[0+yych] & 2) {
		goto yy97;
	}
	if (yych <= 0x00) goto yy79;
	yyaccept = 5;
	yych = *(mar = ++cur);
	if (yych == '\'') goto yy97;
yy100:
#line 661 "tokenizer.re"
	{ token_str(yylval.string); RET(TK_STRING) }
#line 1258 "tokenizer.cpp"
yy101:
	yyaccept = 6;
	yych = *(mar = ++cur);
	if (yych <= 0x1F) {
		if (yych <= '\n') {
			if (yych == '\t') goto yy226;
			goto yy72;
		} else {
			if (yych == '\f') goto yy72;
			if (yych <= '\r') goto yy226;
			goto yy72;
		}
	} else {
		if (yych <= '.') {
			if (yych <= ' ') goto yy226;
			if (yych == ')') goto yy228;
			goto yy72;
		} else {
			if (yych <= '/') goto yy230;
			if (yych == '=') goto yy231;
			goto yy72;
		}
	}
yy102:
	++cur;
#line 543 "tokenizer.re"
	{ RET(TK_POW) }
#line 1286 "tokenizer.cpp"
yy104:
	++cur;
#line 541 "tokenizer.re"
	{ RET(TK_DBL_DOT) }
#line 1291 "tokenizer.cpp"
yy106:
	yyaccept = 7;
	yych = *(mar = ++cur);
yy107:
	if (yych <= 'E') {
		if (yych <= '/') goto yy108;
		if (yych <= '9') goto yy106;
		if (yych >= 'D') goto yy129;
	} else {
		if (yych <= '_') {
			if (yych >= '_') goto yy233;
		} else {
			if (yych <= 'c') goto yy108;
			if (yych <= 'e') goto yy129;
		}
	}
yy108:
#line 593 "tokenizer.re"
	{ token(yylval.string); RET(TK_REAL) }
#line 1311 "tokenizer.cpp"
yy109:
	yych = *++cur;
	if (yych == 'N') goto yy236;
	if (yych == 'n') goto yy236;
	goto yy111;
yy110:
	yych = *++cur;
yy111:
	if (yybm[0+yych] & 4) {
		goto yy110;
	}
	if (yych == '.') goto yy234;
	goto yy79;
yy112:
	yych = *++cur;
	if (yych == 'Q') goto yy237;
	if (yych == 'q') goto yy237;
	goto yy111;
yy113:
	yych = *++cur;
	if (yych == 'A') goto yy238;
	if (yych == 'a') goto yy238;
	goto yy111;
yy114:
	yych = *++cur;
	if (yych <= 'T') {
		if (yych == 'E') goto yy239;
		if (yych <= 'S') goto yy111;
		goto yy240;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy111;
			goto yy239;
		} else {
			if (yych == 't') goto yy240;
			goto yy111;
		}
	}
yy115:
	yych = *++cur;
	if (yych <= 'T') {
		if (yych == 'E') goto yy241;
		if (yych <= 'S') goto yy111;
		goto yy242;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy111;
			goto yy241;
		} else {
			if (yych == 't') goto yy242;
			goto yy111;
		}
	}
yy116:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych == 'E') goto yy243;
		if (yych <= 'N') goto yy111;
		goto yy244;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy111;
			goto yy243;
		} else {
			if (yych == 'o') goto yy244;
			goto yy111;
		}
	}
yy117:
	yych = *++cur;
	if (yych == 'R') goto yy245;
	if (yych == 'r') goto yy245;
	goto yy111;
yy118:
	yych = *++cur;
	if (yych == 'R') goto yy246;
	if (yych == 'r') goto yy246;
	goto yy111;
yy119:
	yych = *++cur;
	if (yych == 'O') goto yy247;
	if (yych == 'o') goto yy247;
	goto yy111;
yy120:
	++cur;
#line 528 "tokenizer.re"
	{ RET(TK_RBRACKET_OLD) }
#line 1399 "tokenizer.cpp"
yy122:
	++cur;
#line 544 "tokenizer.re"
	{ RET(TK_CONCAT) }
#line 1404 "tokenizer.cpp"
yy124:
	++cur;
#line 551 "tokenizer.re"
	{ RET(TK_NE) }
#line 1409 "tokenizer.cpp"
yy126:
	yych = *++cur;
	if (yych <= 0x1F) {
		if (yych <= '\n') {
			if (yych == '\t') goto yy126;
			goto yy79;
		} else {
			if (yych == '\f') goto yy79;
			if (yych <= '\r') goto yy126;
			goto yy79;
		}
	} else {
		if (yych <= '^') {
			if (yych <= ' ') goto yy126;
			if (yych <= '@') goto yy79;
			if (yych <= 'Z') goto yy248;
			goto yy79;
		} else {
			if (yych == '`') goto yy79;
			if (yych <= 'z') goto yy248;
			goto yy79;
		}
	}
yy128:
	yyaccept = 7;
	yych = *(mar = ++cur);
	if (yych <= 'Z') {
		if (yych <= '@') goto yy107;
		if (yych <= 'C') goto yy251;
		if (yych <= 'E') goto yy253;
		goto yy251;
	} else {
		if (yych <= 'c') {
			if (yych <= '`') goto yy107;
			goto yy251;
		} else {
			if (yych <= 'e') goto yy253;
			if (yych <= 'z') goto yy251;
			goto yy107;
		}
	}
yy129:
	yych = *++cur;
	if (yych <= ',') {
		if (yych == '+') goto yy254;
		goto yy79;
	} else {
		if (yych <= '-') goto yy254;
		if (yych <= '/') goto yy79;
		if (yych <= '9') goto yy255;
		goto yy79;
	}
yy130:
	yych = *++cur;
	if (yych <= '9') {
		if (yych <= '&') {
			if (yych == '"') goto yy82;
			goto yy79;
		} else {
			if (yych <= '\'') goto yy97;
			if (yych <= '/') goto yy79;
			goto yy257;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy79;
			if (yych <= 'Z') goto yy259;
			goto yy79;
		} else {
			if (yych == '`') goto yy79;
			if (yych <= 'z') goto yy259;
			goto yy79;
		}
	}
yy131:
	++cur;
#line 542 "tokenizer.re"
	{ RET(TK_DBL_COLON) }
#line 1488 "tokenizer.cpp"
yy133:
	++cur;
#line 557 "tokenizer.re"
	{ RET(TK_LE) }
#line 1493 "tokenizer.cpp"
yy135:
	++cur;
#line 548 "tokenizer.re"
	{ RET(TK_EQ) }
#line 1498 "tokenizer.cpp"
yy137:
	++cur;
#line 545 "tokenizer.re"
	{ RET(TK_ARROW) }
#line 1503 "tokenizer.cpp"
yy139:
	++cur;
#line 563 "tokenizer.re"
	{ RET(TK_GE) }
#line 1508 "tokenizer.cpp"
yy141:
	yych = *++cur;
	if (yych == 'S') goto yy261;
	if (yych == 's') goto yy261;
	goto yy55;
yy142:
	yych = *++cur;
	if (yych == 'L') goto yy262;
	if (yych == 'l') goto yy262;
	goto yy55;
yy143:
	yych = *++cur;
	if (yych <= 'Y') {
		if (yych == 'S') goto yy264;
		if (yych <= 'X') goto yy55;
		goto yy265;
	} else {
		if (yych <= 's') {
			if (yych <= 'r') goto yy55;
			goto yy264;
		} else {
			if (yych == 'y') goto yy265;
			goto yy55;
		}
	}
yy144:
	yyaccept = 2;
	yych = *(mar = ++cur);
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych <= '&') {
		if (yych == '"') goto yy82;
		goto yy47;
	} else {
		if (yych <= '\'') goto yy97;
		if (yych == '_') goto yy144;
		goto yy47;
	}
yy146:
	yych = *++cur;
	if (yybm[0+yych] & 32) {
		goto yy266;
	}
	goto yy79;
yy147:
	yych = *++cur;
	if (yych <= '/') goto yy79;
	if (yych <= '1') goto yy268;
	goto yy79;
yy148:
	yych = *++cur;
	if (yych == 'C') goto yy270;
	if (yych == 'c') goto yy270;
	goto yy55;
yy149:
	yych = *++cur;
	if (yych == 'N') goto yy271;
	if (yych == 'n') goto yy271;
	goto yy55;
yy150:
	yych = *++cur;
	if (yych == 'O') goto yy272;
	if (yych == 'o') goto yy272;
	goto yy55;
yy151:
	yych = *++cur;
	if (yych <= 'S') {
		if (yych == 'L') goto yy273;
		if (yych <= 'R') goto yy55;
		goto yy274;
	} else {
		if (yych <= 'l') {
			if (yych <= 'k') goto yy55;
			goto yy273;
		} else {
			if (yych == 's') goto yy274;
			goto yy55;
		}
	}
yy152:
	yych = *++cur;
	if (yych == 'A') goto yy275;
	if (yych == 'a') goto yy275;
	goto yy55;
yy153:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych == 'A') goto yy276;
		if (yych <= 'N') goto yy55;
		goto yy277;
	} else {
		if (yych <= 'a') {
			if (yych <= '`') goto yy55;
			goto yy276;
		} else {
			if (yych == 'o') goto yy277;
			goto yy55;
		}
	}
yy154:
	yych = *++cur;
	if (yych <= 'N') {
		if (yych <= 'D') {
			if (yych <= 'C') goto yy55;
			goto yy278;
		} else {
			if (yych <= 'L') goto yy55;
			if (yych <= 'M') goto yy279;
			goto yy280;
		}
	} else {
		if (yych <= 'l') {
			if (yych == 'd') goto yy278;
			goto yy55;
		} else {
			if (yych <= 'm') goto yy279;
			if (yych <= 'n') goto yy280;
			goto yy55;
		}
	}
yy155:
	yych = *++cur;
	if (yych == 'I') goto yy281;
	if (yych == 'i') goto yy281;
	goto yy55;
yy156:
	yych = *++cur;
	if (yych == 'C') goto yy282;
	if (yych == 'c') goto yy282;
	goto yy55;
yy157:
	yych = *++cur;
	if (yych == 'T') goto yy283;
	if (yych == 't') goto yy283;
	goto yy55;
yy158:
	yych = *++cur;
	if (yych <= 'F') {
		if (yych == 'A') goto yy284;
		if (yych <= 'E') goto yy55;
		goto yy285;
	} else {
		if (yych <= 'a') {
			if (yych <= '`') goto yy55;
			goto yy284;
		} else {
			if (yych == 'f') goto yy285;
			goto yy55;
		}
	}
yy159:
	yych = *++cur;
	if (yych == 'M') goto yy286;
	if (yych == 'm') goto yy286;
	goto yy55;
yy160:
	yyaccept = 8;
	yych = *(mar = ++cur);
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= '@') {
		if (yych <= '\f') {
			if (yych <= '\t') {
				if (yych >= '\t') {
					ctxmar = cur;
					goto yy287;
				}
			} else {
				if (yych == '\v') {
					ctxmar = cur;
					goto yy287;
				}
			}
		} else {
			if (yych <= ' ') {
				if (yych <= '\r') {
					ctxmar = cur;
					goto yy287;
				}
				if (yych >= ' ') {
					ctxmar = cur;
					goto yy287;
				}
			} else {
				if (yych <= '/') goto yy161;
				if (yych <= '9') goto yy54;
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'V') {
				if (yych == 'U') goto yy289;
				goto yy54;
			} else {
				if (yych <= 'W') goto yy290;
				if (yych <= 'Z') goto yy54;
			}
		} else {
			if (yych <= 'v') {
				if (yych == 'u') goto yy289;
				goto yy54;
			} else {
				if (yych <= 'w') goto yy290;
				if (yych <= 'z') goto yy54;
			}
		}
	}
yy161:
#line 293 "tokenizer.re"
	{ KW(DO) }
#line 1721 "tokenizer.cpp"
yy162:
	yych = *++cur;
	if (yych <= 'S') {
		if (yych == 'E') goto yy291;
		if (yych <= 'R') goto yy55;
		goto yy292;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy55;
			goto yy291;
		} else {
			if (yych == 's') goto yy292;
			goto yy55;
		}
	}
yy163:
	yych = *++cur;
	if (yych <= 'U') {
		if (yych <= 'D') {
			if (yych <= 'C') goto yy55;
			goto yy293;
		} else {
			if (yych <= 'S') goto yy55;
			if (yych <= 'T') goto yy295;
			goto yy296;
		}
	} else {
		if (yych <= 's') {
			if (yych == 'd') goto yy293;
			goto yy55;
		} else {
			if (yych <= 't') goto yy295;
			if (yych <= 'u') goto yy296;
			goto yy55;
		}
	}
yy164:
	yych = *++cur;
	if (yych == 'U') goto yy297;
	if (yych == 'u') goto yy297;
	goto yy55;
yy165:
	yych = *++cur;
	if (yych == 'R') goto yy298;
	if (yych == 'r') goto yy298;
	goto yy55;
yy166:
	yych = *++cur;
	if (yych == 'E') goto yy299;
	if (yych == 'e') goto yy299;
	goto yy55;
yy167:
	yych = *++cur;
	if (yych <= 'T') {
		if (yych == 'I') goto yy300;
		if (yych <= 'S') goto yy55;
		goto yy301;
	} else {
		if (yych <= 'i') {
			if (yych <= 'h') goto yy55;
			goto yy300;
		} else {
			if (yych == 't') goto yy301;
			goto yy55;
		}
	}
yy168:
	yych = *++cur;
	if (yych <= 'N') {
		if (yych == 'L') goto yy302;
		if (yych <= 'M') goto yy55;
		goto yy303;
	} else {
		if (yych <= 'l') {
			if (yych <= 'k') goto yy55;
			goto yy302;
		} else {
			if (yych == 'n') goto yy303;
			goto yy55;
		}
	}
yy169:
	yych = *++cur;
	if (yych == 'U') goto yy304;
	if (yych == 'u') goto yy304;
	goto yy55;
yy170:
	yych = *++cur;
	if (yych == 'R') goto yy305;
	if (yych == 'r') goto yy305;
	goto yy55;
yy171:
	yych = *++cur;
	if (yych == 'N') goto yy306;
	if (yych == 'n') goto yy306;
	goto yy55;
yy172:
	yych = *++cur;
	if (yych == 'N') goto yy307;
	if (yych == 'n') goto yy307;
	goto yy55;
yy173:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'T') {
		if (yych <= '9') {
			if (yych >= '0') goto yy54;
		} else {
			if (yych <= '@') goto yy174;
			if (yych <= 'S') goto yy54;
			goto yy308;
		}
	} else {
		if (yych <= 's') {
			if (yych <= 'Z') goto yy54;
			if (yych >= 'a') goto yy54;
		} else {
			if (yych <= 't') goto yy308;
			if (yych <= 'z') goto yy54;
		}
	}
yy174:
#line 409 "tokenizer.re"
	{ KW(GO) }
#line 1848 "tokenizer.cpp"
yy175:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 411 "tokenizer.re"
	{ KW(IF) }
#line 1857 "tokenizer.cpp"
yy177:
	yych = *++cur;
	if (yych <= 'P') {
		if (yych == 'A') goto yy309;
		if (yych <= 'O') goto yy55;
		goto yy310;
	} else {
		if (yych <= 'a') {
			if (yych <= '`') goto yy55;
			goto yy309;
		} else {
			if (yych == 'p') goto yy310;
			goto yy55;
		}
	}
yy178:
	yyaccept = 9;
	yych = *(mar = ++cur);
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'O') {
		if (yych <= 0x1F) {
			if (yych <= '\n') {
				if (yych == '\t') goto yy311;
			} else {
				if (yych == '\f') goto yy179;
				if (yych <= '\r') goto yy311;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ' ') goto yy311;
				if (yych <= '/') goto yy179;
				if (yych <= '9') goto yy54;
			} else {
				if (yych == 'C') goto yy313;
				if (yych <= 'N') goto yy54;
				goto yy314;
			}
		}
	} else {
		if (yych <= 'c') {
			if (yych <= 'T') {
				if (yych == 'Q') goto yy315;
				if (yych <= 'S') goto yy54;
				goto yy316;
			} else {
				if (yych <= 'Z') goto yy54;
				if (yych <= '`') goto yy179;
				if (yych <= 'b') goto yy54;
				goto yy313;
			}
		} else {
			if (yych <= 'q') {
				if (yych == 'o') goto yy314;
				if (yych <= 'p') goto yy54;
				goto yy315;
			} else {
				if (yych == 't') goto yy316;
				if (yych <= 'z') goto yy54;
			}
		}
	}
yy179:
#line 416 "tokenizer.re"
	{ KW(IN) }
#line 1924 "tokenizer.cpp"
yy180:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 425 "tokenizer.re"
	{ KW(IS) }
#line 1933 "tokenizer.cpp"
yy182:
	yych = *++cur;
	if (yych == 'N') goto yy317;
	if (yych == 'n') goto yy317;
	goto yy55;
yy183:
	yych = *++cur;
	if (yych == 'N') goto yy318;
	if (yych == 'n') goto yy318;
	goto yy55;
yy184:
	yych = *++cur;
	if (yych <= 'G') {
		if (yych == 'C') goto yy320;
		if (yych <= 'F') goto yy55;
		goto yy321;
	} else {
		if (yych <= 'c') {
			if (yych <= 'b') goto yy55;
			goto yy320;
		} else {
			if (yych == 'g') goto yy321;
			goto yy55;
		}
	}
yy185:
	yych = *++cur;
	if (yych == 'M') goto yy322;
	if (yych == 'm') goto yy322;
	goto yy55;
yy186:
	yych = *++cur;
	if (yych <= 'L') {
		if (yych == 'D') goto yy323;
		if (yych <= 'K') goto yy55;
		goto yy324;
	} else {
		if (yych <= 'd') {
			if (yych <= 'c') goto yy55;
			goto yy323;
		} else {
			if (yych == 'l') goto yy324;
			goto yy55;
		}
	}
yy187:
	yych = *++cur;
	if (yych == 'M') goto yy325;
	if (yych == 'm') goto yy325;
	goto yy55;
yy188:
	yych = *++cur;
	if (yych == 'W') goto yy326;
	if (yych == 'w') goto yy326;
	goto yy55;
yy189:
	yych = *++cur;
	if (yych <= 'P') {
		if (yych == 'N') goto yy327;
		if (yych <= 'O') goto yy55;
		goto yy328;
	} else {
		if (yych <= 'n') {
			if (yych <= 'm') goto yy55;
			goto yy327;
		} else {
			if (yych == 'p') goto yy328;
			goto yy55;
		}
	}
yy190:
	yych = *++cur;
	if (yych == 'L') goto yy329;
	if (yych == 'l') goto yy329;
	goto yy55;
yy191:
	yych = *++cur;
	if (yybm[0+yych] & 64) {
		goto yy330;
	}
	goto yy79;
yy192:
	yych = *++cur;
	if (yych <= '/') goto yy79;
	if (yych <= '7') goto yy332;
	goto yy79;
yy193:
	yych = *++cur;
	if (yych == 'L') goto yy334;
	if (yych == 'l') goto yy334;
	goto yy55;
yy194:
	yych = *++cur;
	if (yych <= 'T') {
		if (yych == 'E') goto yy335;
		if (yych <= 'S') goto yy55;
		goto yy336;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy55;
			goto yy335;
		} else {
			if (yych == 't') goto yy336;
			goto yy55;
		}
	}
yy195:
	yych = *++cur;
	if (yych == 'T') goto yy337;
	if (yych == 't') goto yy337;
	goto yy55;
yy196:
	yych = *++cur;
	if (yych <= 'S') {
		if (yych <= 'Q') goto yy55;
		if (yych <= 'R') goto yy339;
		goto yy340;
	} else {
		if (yych <= 'q') goto yy55;
		if (yych <= 'r') goto yy339;
		if (yych <= 's') goto yy340;
		goto yy55;
	}
yy197:
	yych = *++cur;
	if (yych <= 'S') {
		if (yych == 'I') goto yy341;
		if (yych <= 'R') goto yy55;
		goto yy342;
	} else {
		if (yych <= 'i') {
			if (yych <= 'h') goto yy55;
			goto yy341;
		} else {
			if (yych == 's') goto yy342;
			goto yy55;
		}
	}
yy198:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych <= 'H') {
			if (yych == 'E') goto yy343;
			goto yy55;
		} else {
			if (yych <= 'I') goto yy344;
			if (yych <= 'N') goto yy55;
			goto yy345;
		}
	} else {
		if (yych <= 'h') {
			if (yych == 'e') goto yy343;
			goto yy55;
		} else {
			if (yych <= 'i') goto yy344;
			if (yych == 'o') goto yy345;
			goto yy55;
		}
	}
yy199:
	yych = *++cur;
	if (yych <= 'R') {
		if (yych == 'B') goto yy346;
		if (yych <= 'Q') goto yy55;
		goto yy347;
	} else {
		if (yych <= 'b') {
			if (yych <= 'a') goto yy55;
			goto yy346;
		} else {
			if (yych == 'r') goto yy347;
			goto yy55;
		}
	}
yy200:
	yych = *++cur;
	if (yych == 'I') goto yy348;
	if (yych == 'i') goto yy348;
	goto yy55;
yy201:
	yych = *++cur;
	if (yych == 'N') goto yy349;
	if (yych == 'n') goto yy349;
	goto yy55;
yy202:
	yych = *++cur;
	switch (yych) {
	case 'A':
	case 'a':	goto yy350;
	case 'C':
	case 'c':	goto yy351;
	case 'D':
	case 'd':	goto yy352;
	case 'S':
	case 's':	goto yy353;
	case 'T':
	case 't':	goto yy354;
	case 'W':
	case 'w':	goto yy355;
	default:	goto yy55;
	}
yy203:
	yych = *++cur;
	if (yych == 'V') goto yy356;
	if (yych == 'v') goto yy356;
	goto yy55;
yy204:
	yych = *++cur;
	if (yych <= 'Q') {
		if (yych == 'L') goto yy357;
		if (yych <= 'P') goto yy55;
		goto yy358;
	} else {
		if (yych <= 'l') {
			if (yych <= 'k') goto yy55;
			goto yy357;
		} else {
			if (yych == 'q') goto yy358;
			goto yy55;
		}
	}
yy205:
	yych = *++cur;
	if (yych == 'A') goto yy359;
	if (yych == 'a') goto yy359;
	goto yy55;
yy206:
	yych = *++cur;
	if (yych == 'U') goto yy360;
	if (yych == 'u') goto yy360;
	goto yy55;
yy207:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych == 'A') goto yy361;
		if (yych <= 'N') goto yy55;
		goto yy362;
	} else {
		if (yych <= 'a') {
			if (yych <= '`') goto yy55;
			goto yy361;
		} else {
			if (yych == 'o') goto yy362;
			goto yy55;
		}
	}
yy208:
	yych = *++cur;
	if (yych == 'B') goto yy363;
	if (yych == 'b') goto yy363;
	goto yy55;
yy209:
	yych = *++cur;
	if (yych == 'N') goto yy364;
	if (yych == 'n') goto yy364;
	goto yy55;
yy210:
	yych = *++cur;
	if (yych == 'R') goto yy365;
	if (yych == 'r') goto yy365;
	goto yy55;
yy211:
	yych = *++cur;
	if (yych == 'A') goto yy366;
	if (yych == 'a') goto yy366;
	goto yy55;
yy212:
	yych = *++cur;
	if (yych == 'E') goto yy367;
	if (yych == 'e') goto yy367;
	goto yy55;
yy213:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 490 "tokenizer.re"
	{ KW(TO) }
#line 2213 "tokenizer.cpp"
yy215:
	yych = *++cur;
	if (yych == 'P') goto yy368;
	if (yych == 'p') goto yy368;
	goto yy55;
yy216:
	yych = *++cur;
	if (yych == 'F') goto yy369;
	if (yych == 'f') goto yy369;
	goto yy55;
yy217:
	yych = *++cur;
	if (yych == 'E') goto yy370;
	if (yych == 'e') goto yy370;
	goto yy55;
yy218:
	yych = *++cur;
	if (yych == 'L') goto yy372;
	if (yych == 'l') goto yy372;
	goto yy55;
yy219:
	yych = *++cur;
	if (yych == 'L') goto yy373;
	if (yych == 'l') goto yy373;
	goto yy55;
yy220:
	yych = *++cur;
	if (yych == 'I') goto yy374;
	if (yych == 'i') goto yy374;
	goto yy55;
yy221:
	yych = *++cur;
	if (yych <= 'I') {
		if (yych == 'E') goto yy375;
		if (yych <= 'H') goto yy55;
		goto yy376;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy55;
			goto yy375;
		} else {
			if (yych == 'i') goto yy376;
			goto yy55;
		}
	}
yy222:
	yych = *++cur;
	if (yych == 'I') goto yy377;
	if (yych == 'i') goto yy377;
	goto yy55;
yy223:
	yych = *++cur;
	if (yych == '"') goto yy79;
	goto yy379;
yy224:
	yych = *++cur;
	if (yych == '\'') goto yy79;
	goto yy381;
yy225:
	++cur;
	goto yy94;
yy226:
	yych = *++cur;
	if (yych <= '\f') {
		if (yych <= '\t') {
			if (yych <= 0x08) goto yy79;
			goto yy226;
		} else {
			if (yych == '\v') goto yy226;
			goto yy79;
		}
	} else {
		if (yych <= ' ') {
			if (yych <= '\r') goto yy226;
			if (yych <= 0x1F) goto yy79;
			goto yy226;
		} else {
			if (yych == ')') goto yy382;
			goto yy79;
		}
	}
yy228:
	++cur;
	cur -= 2;
#line 519 "tokenizer.re"
	{ RET(TK_LPAREN) }
#line 2300 "tokenizer.cpp"
yy230:
	yych = *++cur;
	if (yych == ')') goto yy386;
	goto yy385;
yy231:
	++cur;
	cur -= 2;
#line 518 "tokenizer.re"
	{ RET(TK_LPAREN) }
#line 2310 "tokenizer.cpp"
yy233:
	yych = *++cur;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy79;
		if (yych <= '9') goto yy388;
		if (yych <= '@') goto yy79;
		goto yy390;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy79;
			goto yy390;
		} else {
			if (yych <= '`') goto yy79;
			if (yych <= 'z') goto yy390;
			goto yy79;
		}
	}
yy234:
	++cur;
#line 663 "tokenizer.re"
	{ token(yylval.string); RET(TK_DEF_OP) }
#line 2332 "tokenizer.cpp"
yy236:
	yych = *++cur;
	if (yych == 'D') goto yy392;
	if (yych == 'd') goto yy392;
	goto yy111;
yy237:
	yych = *++cur;
	if (yych <= 'U') {
		if (yych == '.') goto yy393;
		goto yy111;
	} else {
		if (yych <= 'V') goto yy395;
		if (yych == 'v') goto yy395;
		goto yy111;
	}
yy238:
	yych = *++cur;
	if (yych == 'L') goto yy396;
	if (yych == 'l') goto yy396;
	goto yy111;
yy239:
	yych = *++cur;
	if (yych == '.') goto yy397;
	goto yy111;
yy240:
	yych = *++cur;
	if (yych == '.') goto yy399;
	goto yy111;
yy241:
	yych = *++cur;
	if (yych == '.') goto yy401;
	goto yy111;
yy242:
	yych = *++cur;
	if (yych == '.') goto yy403;
	goto yy111;
yy243:
	yych = *++cur;
	if (yych <= 'P') {
		if (yych == '.') goto yy405;
		goto yy111;
	} else {
		if (yych <= 'Q') goto yy407;
		if (yych == 'q') goto yy407;
		goto yy111;
	}
yy244:
	yych = *++cur;
	if (yych == 'T') goto yy408;
	if (yych == 't') goto yy408;
	goto yy111;
yy245:
	yych = *++cur;
	if (yych == '.') goto yy409;
	goto yy111;
yy246:
	yych = *++cur;
	if (yych == 'U') goto yy411;
	if (yych == 'u') goto yy411;
	goto yy111;
yy247:
	yych = *++cur;
	if (yych == 'R') goto yy412;
	if (yych == 'r') goto yy412;
	goto yy111;
yy248:
	yych = *++cur;
	if (yybm[0+yych] & 16) {
		goto yy248;
	}
	cur = ctxmar;
#line 594 "tokenizer.re"
	{
                if (last_token == yytokentype::TK_NEWLINE) {
                    uint64_t u;
                    if (lex_int(tok, cur, u, yylval.int_suffix.int_kind)) {
                            yylval.n = u;
                            if (enddo_label_stack[enddo_label_stack.size()-1] == u) {
                                while (enddo_label_stack[enddo_label_stack.size()-1] == u) {
                                    enddo_label_stack.pop_back();
                                    enddo_insert_count++;
                                }
                                enddo_newline_process = true;
                            } else {
                                enddo_newline_process = false;
                            }
                            RET(TK_LABEL)
                    } else {
                        token_loc(loc);
                        std::string t = token();
                        throw LFortran::parser_local::TokenizerError("Integer '" + t + "' too large",
                            loc);
                    }
                } else {
                    lex_int_large(al, tok, cur,
                        yylval.int_suffix.int_n,
                        yylval.int_suffix.int_kind);
                    RET(TK_INTEGER)
                }
            }
#line 2433 "tokenizer.cpp"
yy251:
	yych = *++cur;
yy252:
	if (yych <= '@') {
		if (yych == '.') goto yy413;
		goto yy79;
	} else {
		if (yych <= 'Z') goto yy251;
		if (yych <= '`') goto yy79;
		if (yych <= 'z') goto yy251;
		goto yy79;
	}
yy253:
	yych = *++cur;
	if (yych <= ',') {
		if (yych != '+') goto yy252;
	} else {
		if (yych <= '-') goto yy254;
		if (yych <= '/') goto yy252;
		if (yych <= '9') goto yy255;
		goto yy252;
	}
yy254:
	yych = *++cur;
	if (yych <= '/') goto yy79;
	if (yych >= ':') goto yy79;
yy255:
	yyaccept = 7;
	yych = *(mar = ++cur);
	if (yych <= '/') goto yy108;
	if (yych <= '9') goto yy255;
	if (yych == '_') goto yy233;
	goto yy108;
yy257:
	yyaccept = 1;
	yych = *(mar = ++cur);
	if (yych <= '\r') {
		if (yych <= '\n') {
			if (yych == '\t') {
				ctxmar = cur;
				goto yy126;
			}
			goto yy35;
		} else {
			if (yych == '\f') goto yy35;
			ctxmar = cur;
			goto yy126;
		}
	} else {
		if (yych <= '-') {
			if (yych == ' ') {
				ctxmar = cur;
				goto yy126;
			}
			goto yy35;
		} else {
			if (yych <= '.') {
				ctxmar = cur;
				goto yy415;
			}
			if (yych <= '/') goto yy35;
			if (yych <= '9') goto yy257;
			goto yy35;
		}
	}
yy259:
	yyaccept = 1;
	yych = *(mar = ++cur);
	if (yych <= '-') {
		if (yych <= '\v') {
			if (yych == '\t') {
				ctxmar = cur;
				goto yy126;
			}
			if (yych <= '\n') goto yy35;
			ctxmar = cur;
			goto yy126;
		} else {
			if (yych <= '\r') {
				if (yych <= '\f') goto yy35;
				ctxmar = cur;
				goto yy126;
			} else {
				if (yych == ' ') {
					ctxmar = cur;
					goto yy126;
				}
				goto yy35;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '/') {
				if (yych <= '.') {
					ctxmar = cur;
					goto yy415;
				}
				goto yy35;
			} else {
				if (yych <= '9') goto yy259;
				if (yych <= '@') goto yy35;
				goto yy259;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy35;
				goto yy259;
			} else {
				if (yych <= '`') goto yy35;
				if (yych <= 'z') goto yy259;
				goto yy35;
			}
		}
	}
yy261:
	yych = *++cur;
	if (yych == 'T') goto yy416;
	if (yych == 't') goto yy416;
	goto yy55;
yy262:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'O') {
		if (yych <= '9') {
			if (yych >= '0') goto yy54;
		} else {
			if (yych <= '@') goto yy263;
			if (yych <= 'N') goto yy54;
			goto yy417;
		}
	} else {
		if (yych <= 'n') {
			if (yych <= 'Z') goto yy54;
			if (yych >= 'a') goto yy54;
		} else {
			if (yych <= 'o') goto yy417;
			if (yych <= 'z') goto yy54;
		}
	}
yy263:
#line 255 "tokenizer.re"
	{ KW(ALL) }
#line 2578 "tokenizer.cpp"
yy264:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych == 'I') goto yy418;
		if (yych <= 'N') goto yy55;
		goto yy419;
	} else {
		if (yych <= 'i') {
			if (yych <= 'h') goto yy55;
			goto yy418;
		} else {
			if (yych == 'o') goto yy419;
			goto yy55;
		}
	}
yy265:
	yych = *++cur;
	if (yych == 'N') goto yy420;
	if (yych == 'n') goto yy420;
	goto yy55;
yy266:
	yych = *++cur;
	if (yybm[0+yych] & 32) {
		goto yy266;
	}
	if (yych == '"') goto yy421;
	goto yy79;
yy268:
	yych = *++cur;
	if (yych == '\'') goto yy423;
	if (yych <= '/') goto yy79;
	if (yych <= '1') goto yy268;
	goto yy79;
yy270:
	yych = *++cur;
	if (yych == 'K') goto yy425;
	if (yych == 'k') goto yy425;
	goto yy55;
yy271:
	yych = *++cur;
	if (yych == 'D') goto yy426;
	if (yych == 'd') goto yy426;
	goto yy55;
yy272:
	yych = *++cur;
	if (yych == 'C') goto yy428;
	if (yych == 'c') goto yy428;
	goto yy55;
yy273:
	yych = *++cur;
	if (yych == 'L') goto yy429;
	if (yych == 'l') goto yy429;
	goto yy55;
yy274:
	yych = *++cur;
	if (yych == 'E') goto yy431;
	if (yych == 'e') goto yy431;
	goto yy55;
yy275:
	yych = *++cur;
	if (yych <= 'R') {
		if (yych == 'N') goto yy433;
		if (yych <= 'Q') goto yy55;
		goto yy434;
	} else {
		if (yych <= 'n') {
			if (yych <= 'm') goto yy55;
			goto yy433;
		} else {
			if (yych == 'r') goto yy434;
			goto yy55;
		}
	}
yy276:
	yych = *++cur;
	if (yych == 'S') goto yy435;
	if (yych == 's') goto yy435;
	goto yy55;
yy277:
	yych = *++cur;
	if (yych == 'S') goto yy436;
	if (yych == 's') goto yy436;
	goto yy55;
yy278:
	yych = *++cur;
	if (yych == 'I') goto yy437;
	if (yych == 'i') goto yy437;
	goto yy55;
yy279:
	yych = *++cur;
	if (yych <= 'P') {
		if (yych == 'M') goto yy438;
		if (yych <= 'O') goto yy55;
		goto yy439;
	} else {
		if (yych <= 'm') {
			if (yych <= 'l') goto yy55;
			goto yy438;
		} else {
			if (yych == 'p') goto yy439;
			goto yy55;
		}
	}
yy280:
	yych = *++cur;
	if (yych <= 'T') {
		if (yych == 'C') goto yy440;
		if (yych <= 'S') goto yy55;
		goto yy441;
	} else {
		if (yych <= 'c') {
			if (yych <= 'b') goto yy55;
			goto yy440;
		} else {
			if (yych == 't') goto yy441;
			goto yy55;
		}
	}
yy281:
	yych = *++cur;
	if (yych == 'T') goto yy442;
	if (yych == 't') goto yy442;
	goto yy55;
yy282:
	yych = *++cur;
	if (yych == 'L') goto yy443;
	if (yych == 'l') goto yy443;
	goto yy55;
yy283:
	yych = *++cur;
	if (yych == 'A') goto yy444;
	if (yych == 'a') goto yy444;
	goto yy55;
yy284:
	yych = *++cur;
	if (yych == 'L') goto yy446;
	if (yych == 'l') goto yy446;
	goto yy55;
yy285:
	yych = *++cur;
	if (yych <= 'E') {
		if (yych == 'A') goto yy447;
		if (yych <= 'D') goto yy55;
		goto yy448;
	} else {
		if (yych <= 'a') {
			if (yych <= '`') goto yy55;
			goto yy447;
		} else {
			if (yych == 'e') goto yy448;
			goto yy55;
		}
	}
yy286:
	yych = *++cur;
	if (yych == 'E') goto yy449;
	if (yych == 'e') goto yy449;
	goto yy55;
yy287:
	yych = *++cur;
	if (yych <= '\f') {
		if (yych <= '\t') {
			if (yych <= 0x08) goto yy79;
			goto yy287;
		} else {
			if (yych == '\v') goto yy287;
			goto yy79;
		}
	} else {
		if (yych <= ' ') {
			if (yych <= '\r') goto yy287;
			if (yych <= 0x1F) goto yy79;
			goto yy287;
		} else {
			if (yych <= '/') goto yy79;
			if (yych <= '9') goto yy450;
			goto yy79;
		}
	}
yy289:
	yych = *++cur;
	if (yych == 'B') goto yy453;
	if (yych == 'b') goto yy453;
	goto yy55;
yy290:
	yych = *++cur;
	if (yych == 'H') goto yy454;
	if (yych == 'h') goto yy454;
	goto yy55;
yy291:
	yych = *++cur;
	if (yych == 'M') goto yy455;
	if (yych == 'm') goto yy455;
	goto yy55;
yy292:
	yych = *++cur;
	if (yych == 'E') goto yy456;
	if (yych == 'e') goto yy456;
	goto yy55;
yy293:
	yyaccept = 10;
	yych = *(mar = ++cur);
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	switch (yych) {
	case '\t':
	case '\v':
	case '\r':	goto yy458;
	case ' ':	goto yy460;
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	case 'G':
	case 'H':
	case 'J':
	case 'K':
	case 'L':
	case 'N':
	case 'O':
	case 'Q':
	case 'R':
	case 'U':
	case 'V':
	case 'X':
	case 'Y':
	case 'Z':
	case 'g':
	case 'h':
	case 'j':
	case 'k':
	case 'l':
	case 'n':
	case 'o':
	case 'q':
	case 'r':
	case 'u':
	case 'v':
	case 'x':
	case 'y':
	case 'z':	goto yy54;
	case 'A':
	case 'a':	goto yy461;
	case 'B':
	case 'b':	goto yy462;
	case 'C':
	case 'c':	goto yy463;
	case 'D':
	case 'd':	goto yy464;
	case 'E':
	case 'e':	goto yy465;
	case 'F':
	case 'f':	goto yy466;
	case 'I':
	case 'i':	goto yy467;
	case 'M':
	case 'm':	goto yy468;
	case 'P':
	case 'p':	goto yy469;
	case 'S':
	case 's':	goto yy470;
	case 'T':
	case 't':	goto yy471;
	case 'W':
	case 'w':	goto yy472;
	default:	goto yy294;
	}
yy294:
#line 302 "tokenizer.re"
	{ KW(END) }
#line 2856 "tokenizer.cpp"
yy295:
	yych = *++cur;
	if (yych == 'R') goto yy473;
	if (yych == 'r') goto yy473;
	goto yy55;
yy296:
	yych = *++cur;
	if (yych == 'M') goto yy474;
	if (yych == 'm') goto yy474;
	goto yy55;
yy297:
	yych = *++cur;
	if (yych == 'I') goto yy476;
	if (yych == 'i') goto yy476;
	goto yy55;
yy298:
	yych = *++cur;
	if (yych <= 'O') {
		if (yych == 'M') goto yy477;
		if (yych <= 'N') goto yy55;
		goto yy478;
	} else {
		if (yych <= 'm') {
			if (yych <= 'l') goto yy55;
			goto yy477;
		} else {
			if (yych == 'o') goto yy478;
			goto yy55;
		}
	}
yy299:
	yych = *++cur;
	if (yych == 'N') goto yy479;
	if (yych == 'n') goto yy479;
	goto yy55;
yy300:
	yych = *++cur;
	if (yych == 'T') goto yy480;
	if (yych == 't') goto yy480;
	goto yy55;
yy301:
	yych = *++cur;
	if (yych == 'E') goto yy482;
	if (yych == 'e') goto yy482;
	goto yy55;
yy302:
	yych = *++cur;
	if (yych == 'E') goto yy483;
	if (yych == 'e') goto yy483;
	goto yy55;
yy303:
	yych = *++cur;
	if (yych == 'A') goto yy485;
	if (yych == 'a') goto yy485;
	goto yy55;
yy304:
	yych = *++cur;
	if (yych == 'S') goto yy486;
	if (yych == 's') goto yy486;
	goto yy55;
yy305:
	yych = *++cur;
	if (yych <= 'M') {
		if (yych == 'A') goto yy487;
		if (yych <= 'L') goto yy55;
		goto yy488;
	} else {
		if (yych <= 'a') {
			if (yych <= '`') goto yy55;
			goto yy487;
		} else {
			if (yych == 'm') goto yy488;
			goto yy55;
		}
	}
yy306:
	yych = *++cur;
	if (yych == 'C') goto yy490;
	if (yych == 'c') goto yy490;
	goto yy55;
yy307:
	yych = *++cur;
	if (yych == 'E') goto yy491;
	if (yych == 'e') goto yy491;
	goto yy55;
yy308:
	yych = *++cur;
	if (yych == 'O') goto yy492;
	if (yych == 'o') goto yy492;
	goto yy55;
yy309:
	yych = *++cur;
	if (yych == 'G') goto yy494;
	if (yych == 'g') goto yy494;
	goto yy55;
yy310:
	yych = *++cur;
	if (yych <= 'U') {
		if (yych <= 'N') {
			if (yych == 'L') goto yy495;
			goto yy55;
		} else {
			if (yych <= 'O') goto yy496;
			if (yych <= 'T') goto yy55;
			goto yy497;
		}
	} else {
		if (yych <= 'n') {
			if (yych == 'l') goto yy495;
			goto yy55;
		} else {
			if (yych <= 'o') goto yy496;
			if (yych == 'u') goto yy497;
			goto yy55;
		}
	}
yy311:
	yych = *++cur;
	if (yych <= '\r') {
		if (yych <= '\n') {
			if (yych == '\t') goto yy311;
			goto yy79;
		} else {
			if (yych == '\f') goto yy79;
			goto yy311;
		}
	} else {
		if (yych <= 'N') {
			if (yych == ' ') goto yy311;
			goto yy79;
		} else {
			if (yych <= 'O') goto yy498;
			if (yych == 'o') goto yy498;
			goto yy79;
		}
	}
yy313:
	yych = *++cur;
	if (yych == 'L') goto yy499;
	if (yych == 'l') goto yy499;
	goto yy55;
yy314:
	yych = *++cur;
	if (yych == 'U') goto yy500;
	if (yych == 'u') goto yy500;
	goto yy55;
yy315:
	yych = *++cur;
	if (yych == 'U') goto yy501;
	if (yych == 'u') goto yy501;
	goto yy55;
yy316:
	yych = *++cur;
	if (yych <= 'R') {
		if (yych == 'E') goto yy502;
		if (yych <= 'Q') goto yy55;
		goto yy503;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy55;
			goto yy502;
		} else {
			if (yych == 'r') goto yy503;
			goto yy55;
		}
	}
yy317:
	yych = *++cur;
	if (yych == 'D') goto yy504;
	if (yych == 'd') goto yy504;
	goto yy55;
yy318:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 427 "tokenizer.re"
	{ KW(LEN) }
#line 3036 "tokenizer.cpp"
yy320:
	yych = *++cur;
	if (yych == 'A') goto yy506;
	if (yych == 'a') goto yy506;
	goto yy55;
yy321:
	yych = *++cur;
	if (yych == 'I') goto yy507;
	if (yych == 'i') goto yy507;
	goto yy55;
yy322:
	yych = *++cur;
	if (yych == 'O') goto yy508;
	if (yych == 'o') goto yy508;
	goto yy55;
yy323:
	yych = *++cur;
	if (yych == 'U') goto yy509;
	if (yych == 'u') goto yy509;
	goto yy55;
yy324:
	yych = *++cur;
	if (yych == 'D') goto yy510;
	if (yych == 'd') goto yy510;
	goto yy55;
yy325:
	yych = *++cur;
	if (yych == 'E') goto yy512;
	if (yych == 'e') goto yy512;
	goto yy55;
yy326:
	yych = *++cur;
	if (yych == '_') goto yy514;
	goto yy55;
yy327:
	yych = *++cur;
	if (yych <= '^') {
		if (yych == 'E') goto yy515;
		goto yy55;
	} else {
		if (yych <= '_') goto yy517;
		if (yych == 'e') goto yy515;
		goto yy55;
	}
yy328:
	yych = *++cur;
	if (yych == 'A') goto yy518;
	if (yych == 'a') goto yy518;
	goto yy55;
yy329:
	yych = *++cur;
	if (yych == 'L') goto yy519;
	if (yych == 'l') goto yy519;
	goto yy55;
yy330:
	yych = *++cur;
	if (yybm[0+yych] & 64) {
		goto yy330;
	}
	if (yych == '"') goto yy520;
	goto yy79;
yy332:
	yych = *++cur;
	if (yych == '\'') goto yy522;
	if (yych <= '/') goto yy79;
	if (yych <= '7') goto yy332;
	goto yy79;
yy334:
	yych = *++cur;
	if (yych == 'Y') goto yy524;
	if (yych == 'y') goto yy524;
	goto yy55;
yy335:
	yych = *++cur;
	if (yych <= 'R') {
		if (yych == 'N') goto yy526;
		if (yych <= 'Q') goto yy55;
		goto yy528;
	} else {
		if (yych <= 'n') {
			if (yych <= 'm') goto yy55;
			goto yy526;
		} else {
			if (yych == 'r') goto yy528;
			goto yy55;
		}
	}
yy336:
	yych = *++cur;
	if (yych == 'I') goto yy529;
	if (yych == 'i') goto yy529;
	goto yy55;
yy337:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 447 "tokenizer.re"
	{ KW(OUT) }
#line 3137 "tokenizer.cpp"
yy339:
	yych = *++cur;
	if (yych == 'A') goto yy530;
	if (yych == 'a') goto yy530;
	goto yy55;
yy340:
	yych = *++cur;
	if (yych == 'S') goto yy531;
	if (yych == 's') goto yy531;
	goto yy55;
yy341:
	yych = *++cur;
	if (yych == 'N') goto yy533;
	if (yych == 'n') goto yy533;
	goto yy55;
yy342:
	yych = *++cur;
	if (yych == 'T') goto yy534;
	if (yych == 't') goto yy534;
	goto yy55;
yy343:
	yych = *++cur;
	if (yych == 'C') goto yy536;
	if (yych == 'c') goto yy536;
	goto yy55;
yy344:
	yych = *++cur;
	if (yych <= 'V') {
		if (yych == 'N') goto yy537;
		if (yych <= 'U') goto yy55;
		goto yy538;
	} else {
		if (yych <= 'n') {
			if (yych <= 'm') goto yy55;
			goto yy537;
		} else {
			if (yych == 'v') goto yy538;
			goto yy55;
		}
	}
yy345:
	yych = *++cur;
	if (yych <= 'T') {
		if (yych <= 'F') {
			if (yych == 'C') goto yy539;
			goto yy55;
		} else {
			if (yych <= 'G') goto yy540;
			if (yych <= 'S') goto yy55;
			goto yy541;
		}
	} else {
		if (yych <= 'f') {
			if (yych == 'c') goto yy539;
			goto yy55;
		} else {
			if (yych <= 'g') goto yy540;
			if (yych == 't') goto yy541;
			goto yy55;
		}
	}
yy346:
	yych = *++cur;
	if (yych == 'L') goto yy542;
	if (yych == 'l') goto yy542;
	goto yy55;
yy347:
	yych = *++cur;
	if (yych == 'E') goto yy543;
	if (yych == 'e') goto yy543;
	goto yy55;
yy348:
	yych = *++cur;
	if (yych == 'E') goto yy545;
	if (yych == 'e') goto yy545;
	goto yy55;
yy349:
	yych = *++cur;
	if (yych == 'K') goto yy546;
	if (yych == 'k') goto yy546;
	goto yy55;
yy350:
	yych = *++cur;
	if (yych <= 'L') {
		if (yych == 'D') goto yy548;
		if (yych <= 'K') goto yy55;
		goto yy550;
	} else {
		if (yych <= 'd') {
			if (yych <= 'c') goto yy55;
			goto yy548;
		} else {
			if (yych == 'l') goto yy550;
			goto yy55;
		}
	}
yy351:
	yych = *++cur;
	if (yych == 'U') goto yy552;
	if (yych == 'u') goto yy552;
	goto yy55;
yy352:
	yych = *++cur;
	if (yych == 'U') goto yy553;
	if (yych == 'u') goto yy553;
	goto yy55;
yy353:
	yych = *++cur;
	if (yych == 'U') goto yy554;
	if (yych == 'u') goto yy554;
	goto yy55;
yy354:
	yych = *++cur;
	if (yych == 'U') goto yy555;
	if (yych == 'u') goto yy555;
	goto yy55;
yy355:
	yych = *++cur;
	if (yych == 'I') goto yy556;
	if (yych == 'i') goto yy556;
	goto yy55;
yy356:
	yych = *++cur;
	if (yych == 'E') goto yy557;
	if (yych == 'e') goto yy557;
	goto yy55;
yy357:
	yych = *++cur;
	if (yych == 'E') goto yy559;
	if (yych == 'e') goto yy559;
	goto yy55;
yy358:
	yych = *++cur;
	if (yych == 'U') goto yy560;
	if (yych == 'u') goto yy560;
	goto yy55;
yy359:
	yych = *++cur;
	if (yych == 'R') goto yy561;
	if (yych == 'r') goto yy561;
	goto yy55;
yy360:
	yych = *++cur;
	if (yych == 'R') goto yy562;
	if (yych == 'r') goto yy562;
	goto yy55;
yy361:
	yych = *++cur;
	if (yych == 'T') goto yy563;
	if (yych == 't') goto yy563;
	goto yy55;
yy362:
	yych = *++cur;
	if (yych == 'P') goto yy565;
	if (yych == 'p') goto yy565;
	goto yy55;
yy363:
	yych = *++cur;
	if (yych <= 'R') {
		if (yych == 'M') goto yy567;
		if (yych <= 'Q') goto yy55;
		goto yy568;
	} else {
		if (yych <= 'm') {
			if (yych <= 'l') goto yy55;
			goto yy567;
		} else {
			if (yych == 'r') goto yy568;
			goto yy55;
		}
	}
yy364:
	yych = *++cur;
	if (yych == 'C') goto yy569;
	if (yych == 'c') goto yy569;
	goto yy55;
yy365:
	yych = *++cur;
	if (yych == 'G') goto yy571;
	if (yych == 'g') goto yy571;
	goto yy55;
yy366:
	yych = *++cur;
	if (yych == 'M') goto yy572;
	if (yych == 'm') goto yy572;
	goto yy55;
yy367:
	yych = *++cur;
	if (yych == 'N') goto yy574;
	if (yych == 'n') goto yy574;
	goto yy55;
yy368:
	yych = *++cur;
	if (yych == 'E') goto yy576;
	if (yych == 'e') goto yy576;
	goto yy55;
yy369:
	yych = *++cur;
	if (yych == 'O') goto yy578;
	if (yych == 'o') goto yy578;
	goto yy55;
yy370:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 493 "tokenizer.re"
	{ KW(USE) }
#line 3347 "tokenizer.cpp"
yy372:
	yych = *++cur;
	if (yych == 'U') goto yy579;
	if (yych == 'u') goto yy579;
	goto yy55;
yy373:
	yych = *++cur;
	if (yych == 'A') goto yy580;
	if (yych == 'a') goto yy580;
	goto yy55;
yy374:
	yych = *++cur;
	if (yych == 'T') goto yy581;
	if (yych == 't') goto yy581;
	goto yy55;
yy375:
	yych = *++cur;
	if (yych == 'R') goto yy583;
	if (yych == 'r') goto yy583;
	goto yy55;
yy376:
	yych = *++cur;
	if (yych == 'L') goto yy584;
	if (yych == 'l') goto yy584;
	goto yy55;
yy377:
	yych = *++cur;
	if (yych == 'T') goto yy585;
	if (yych == 't') goto yy585;
	goto yy55;
yy378:
	yych = *++cur;
yy379:
	if (yybm[0+yych] & 128) {
		goto yy378;
	}
	if (yych == '"') goto yy586;
	goto yy79;
yy380:
	yych = *++cur;
yy381:
	if (yych <= '9') {
		if (yych == '\'') goto yy588;
		if (yych <= '/') goto yy79;
		goto yy380;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy79;
			goto yy380;
		} else {
			if (yych <= '`') goto yy79;
			if (yych <= 'f') goto yy380;
			goto yy79;
		}
	}
yy382:
	++cur;
	cur = ctxmar;
#line 521 "tokenizer.re"
	{ RET(TK_LPAREN) }
#line 3408 "tokenizer.cpp"
yy384:
	yych = *++cur;
yy385:
	if (yych <= '\f') {
		if (yych <= '\t') {
			if (yych <= 0x08) goto yy79;
			goto yy384;
		} else {
			if (yych == '\v') goto yy384;
			goto yy79;
		}
	} else {
		if (yych <= ' ') {
			if (yych <= '\r') goto yy384;
			if (yych <= 0x1F) goto yy79;
			goto yy384;
		} else {
			if (yych == ')') goto yy590;
			goto yy79;
		}
	}
yy386:
	++cur;
	cur -= 3;
#line 524 "tokenizer.re"
	{ RET(TK_LPAREN) }
#line 3435 "tokenizer.cpp"
yy388:
	yych = *++cur;
	if (yych <= '/') goto yy108;
	if (yych <= '9') goto yy388;
	goto yy108;
yy390:
	yych = *++cur;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy108;
		if (yych <= '9') goto yy390;
		if (yych <= '@') goto yy108;
		goto yy390;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy108;
			goto yy390;
		} else {
			if (yych <= '`') goto yy108;
			if (yych <= 'z') goto yy390;
			goto yy108;
		}
	}
yy392:
	yych = *++cur;
	if (yych == '.') goto yy592;
	goto yy111;
yy393:
	++cur;
#line 549 "tokenizer.re"
	{ WARN_REL(EQ) RET(TK_EQ) }
#line 3466 "tokenizer.cpp"
yy395:
	yych = *++cur;
	if (yych == '.') goto yy594;
	goto yy111;
yy396:
	yych = *++cur;
	if (yych == 'S') goto yy596;
	if (yych == 's') goto yy596;
	goto yy111;
yy397:
	++cur;
#line 564 "tokenizer.re"
	{ WARN_REL(GE) RET(TK_GE) }
#line 3480 "tokenizer.cpp"
yy399:
	++cur;
#line 561 "tokenizer.re"
	{ WARN_REL(GT) RET(TK_GT) }
#line 3485 "tokenizer.cpp"
yy401:
	++cur;
#line 558 "tokenizer.re"
	{ WARN_REL(LE) RET(TK_LE) }
#line 3490 "tokenizer.cpp"
yy403:
	++cur;
#line 555 "tokenizer.re"
	{ WARN_REL(LT) RET(TK_LT) }
#line 3495 "tokenizer.cpp"
yy405:
	++cur;
#line 552 "tokenizer.re"
	{ WARN_REL(NE) RET(TK_NE) }
#line 3500 "tokenizer.cpp"
yy407:
	yych = *++cur;
	if (yych == 'V') goto yy597;
	if (yych == 'v') goto yy597;
	goto yy111;
yy408:
	yych = *++cur;
	if (yych == '.') goto yy598;
	goto yy111;
yy409:
	++cur;
#line 570 "tokenizer.re"
	{ RET(TK_OR) }
#line 3514 "tokenizer.cpp"
yy411:
	yych = *++cur;
	if (yych == 'E') goto yy600;
	if (yych == 'e') goto yy600;
	goto yy111;
yy412:
	yych = *++cur;
	if (yych == '.') goto yy601;
	goto yy111;
yy413:
	++cur;
	cur = ctxmar;
#line 585 "tokenizer.re"
	{
                lex_int_large(al, tok, cur,
                    yylval.int_suffix.int_n,
                    yylval.int_suffix.int_kind);
                RET(TK_INTEGER)
            }
#line 3534 "tokenizer.cpp"
yy415:
	yych = *++cur;
	if (yych == '.') goto yy79;
	goto yy252;
yy416:
	yych = *++cur;
	if (yych == 'R') goto yy603;
	if (yych == 'r') goto yy603;
	goto yy55;
yy417:
	yych = *++cur;
	if (yych == 'C') goto yy604;
	if (yych == 'c') goto yy604;
	goto yy55;
yy418:
	yych = *++cur;
	if (yych == 'G') goto yy605;
	if (yych == 'g') goto yy605;
	goto yy55;
yy419:
	yych = *++cur;
	if (yych == 'C') goto yy606;
	if (yych == 'c') goto yy606;
	goto yy55;
yy420:
	yych = *++cur;
	if (yych == 'C') goto yy607;
	if (yych == 'c') goto yy607;
	goto yy55;
yy421:
	++cur;
#line 629 "tokenizer.re"
	{ token(yylval.string); RET(TK_BOZ_CONSTANT) }
#line 3568 "tokenizer.cpp"
yy423:
	++cur;
#line 630 "tokenizer.re"
	{ token(yylval.string); RET(TK_BOZ_CONSTANT) }
#line 3573 "tokenizer.cpp"
yy425:
	yych = *++cur;
	if (yych == 'S') goto yy608;
	if (yych == 's') goto yy608;
	goto yy55;
yy426:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 263 "tokenizer.re"
	{ KW(BIND) }
#line 3587 "tokenizer.cpp"
yy428:
	yych = *++cur;
	if (yych == 'K') goto yy609;
	if (yych == 'k') goto yy609;
	goto yy55;
yy429:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 265 "tokenizer.re"
	{ KW(CALL) }
#line 3601 "tokenizer.cpp"
yy431:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 266 "tokenizer.re"
	{ KW(CASE) }
#line 3610 "tokenizer.cpp"
yy433:
	yych = *++cur;
	if (yych == 'G') goto yy611;
	if (yych == 'g') goto yy611;
	goto yy55;
yy434:
	yych = *++cur;
	if (yych == 'A') goto yy612;
	if (yych == 'a') goto yy612;
	goto yy55;
yy435:
	yych = *++cur;
	if (yych == 'S') goto yy613;
	if (yych == 's') goto yy613;
	goto yy55;
yy436:
	yych = *++cur;
	if (yych == 'E') goto yy615;
	if (yych == 'e') goto yy615;
	goto yy55;
yy437:
	yych = *++cur;
	if (yych == 'M') goto yy617;
	if (yych == 'm') goto yy617;
	goto yy55;
yy438:
	yych = *++cur;
	if (yych == 'O') goto yy618;
	if (yych == 'o') goto yy618;
	goto yy55;
yy439:
	yych = *++cur;
	if (yych == 'L') goto yy619;
	if (yych == 'l') goto yy619;
	goto yy55;
yy440:
	yych = *++cur;
	if (yych == 'U') goto yy620;
	if (yych == 'u') goto yy620;
	goto yy55;
yy441:
	yych = *++cur;
	if (yych <= 'I') {
		if (yych == 'A') goto yy621;
		if (yych <= 'H') goto yy55;
		goto yy622;
	} else {
		if (yych <= 'a') {
			if (yych <= '`') goto yy55;
			goto yy621;
		} else {
			if (yych == 'i') goto yy622;
			goto yy55;
		}
	}
yy442:
	yych = *++cur;
	if (yych == 'I') goto yy623;
	if (yych == 'i') goto yy623;
	goto yy55;
yy443:
	yych = *++cur;
	if (yych == 'E') goto yy624;
	if (yych == 'e') goto yy624;
	goto yy55;
yy444:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 281 "tokenizer.re"
	{ KW(DATA) }
#line 3684 "tokenizer.cpp"
yy446:
	yych = *++cur;
	if (yych == 'L') goto yy626;
	if (yych == 'l') goto yy626;
	goto yy55;
yy447:
	yych = *++cur;
	if (yych == 'U') goto yy627;
	if (yych == 'u') goto yy627;
	goto yy55;
yy448:
	yych = *++cur;
	if (yych == 'R') goto yy628;
	if (yych == 'r') goto yy628;
	goto yy55;
yy449:
	yych = *++cur;
	if (yych == 'N') goto yy629;
	if (yych == 'n') goto yy629;
	goto yy55;
yy450:
	yych = *++cur;
	if (yych <= '/') goto yy452;
	if (yych <= '9') goto yy450;
yy452:
	cur = ctxmar;
#line 286 "tokenizer.re"
	{
                // This is a label do statement, we have to match the
                // corresponding continue base "end do".
                uint64_t n = parse_int(cur);
                enddo_label_stack.push_back(n);
                KW(DO);
            }
#line 3719 "tokenizer.cpp"
yy453:
	yych = *++cur;
	if (yych == 'L') goto yy630;
	if (yych == 'l') goto yy630;
	goto yy55;
yy454:
	yych = *++cur;
	if (yych == 'I') goto yy631;
	if (yych == 'i') goto yy631;
	goto yy55;
yy455:
	yych = *++cur;
	if (yych == 'E') goto yy632;
	if (yych == 'e') goto yy632;
	goto yy55;
yy456:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'W') {
		if (yych <= '@') {
			if (yych <= '/') goto yy457;
			if (yych <= '9') goto yy54;
		} else {
			if (yych == 'I') goto yy633;
			if (yych <= 'V') goto yy54;
			goto yy634;
		}
	} else {
		if (yych <= 'i') {
			if (yych <= 'Z') goto yy54;
			if (yych <= '`') goto yy457;
			if (yych <= 'h') goto yy54;
			goto yy633;
		} else {
			if (yych == 'w') goto yy634;
			if (yych <= 'z') goto yy54;
		}
	}
yy457:
#line 298 "tokenizer.re"
	{ KW(ELSE) }
#line 3763 "tokenizer.cpp"
yy458:
	yych = *++cur;
yy459:
	switch (yych) {
	case '\t':
	case '\v':
	case '\r':
	case ' ':	goto yy458;
	case 'A':
	case 'a':	goto yy635;
	case 'B':
	case 'b':	goto yy636;
	case 'C':
	case 'c':	goto yy637;
	case 'D':
	case 'd':	goto yy638;
	case 'E':
	case 'e':	goto yy639;
	case 'F':
	case 'f':	goto yy640;
	case 'I':
	case 'i':	goto yy641;
	case 'M':
	case 'm':	goto yy642;
	case 'P':
	case 'p':	goto yy643;
	case 'S':
	case 's':	goto yy644;
	case 'T':
	case 't':	goto yy645;
	case 'W':
	case 'w':	goto yy646;
	default:	goto yy79;
	}
yy460:
	yych = *++cur;
	if (yych == 'F') goto yy647;
	if (yych == 'f') goto yy647;
	goto yy459;
yy461:
	yych = *++cur;
	if (yych == 'S') goto yy648;
	if (yych == 's') goto yy648;
	goto yy55;
yy462:
	yych = *++cur;
	if (yych == 'L') goto yy649;
	if (yych == 'l') goto yy649;
	goto yy55;
yy463:
	yych = *++cur;
	if (yych == 'R') goto yy650;
	if (yych == 'r') goto yy650;
	goto yy55;
yy464:
	yych = *++cur;
	if (yych == 'O') goto yy651;
	if (yych == 'o') goto yy651;
	goto yy55;
yy465:
	yych = *++cur;
	if (yych == 'N') goto yy653;
	if (yych == 'n') goto yy653;
	goto yy55;
yy466:
	yych = *++cur;
	if (yych <= 'U') {
		if (yych <= 'N') {
			if (yych == 'I') goto yy654;
			goto yy55;
		} else {
			if (yych <= 'O') goto yy655;
			if (yych <= 'T') goto yy55;
			goto yy656;
		}
	} else {
		if (yych <= 'n') {
			if (yych == 'i') goto yy654;
			goto yy55;
		} else {
			if (yych <= 'o') goto yy655;
			if (yych == 'u') goto yy656;
			goto yy55;
		}
	}
yy467:
	yych = *++cur;
	if (yych <= 'N') {
		if (yych == 'F') goto yy657;
		if (yych <= 'M') goto yy55;
		goto yy659;
	} else {
		if (yych <= 'f') {
			if (yych <= 'e') goto yy55;
			goto yy657;
		} else {
			if (yych == 'n') goto yy659;
			goto yy55;
		}
	}
yy468:
	yych = *++cur;
	if (yych == 'O') goto yy660;
	if (yych == 'o') goto yy660;
	goto yy55;
yy469:
	yych = *++cur;
	if (yych == 'R') goto yy661;
	if (yych == 'r') goto yy661;
	goto yy55;
yy470:
	yych = *++cur;
	if (yych <= 'U') {
		if (yych == 'E') goto yy662;
		if (yych <= 'T') goto yy55;
		goto yy663;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy55;
			goto yy662;
		} else {
			if (yych == 'u') goto yy663;
			goto yy55;
		}
	}
yy471:
	yych = *++cur;
	if (yych <= 'Y') {
		if (yych == 'E') goto yy664;
		if (yych <= 'X') goto yy55;
		goto yy665;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy55;
			goto yy664;
		} else {
			if (yych == 'y') goto yy665;
			goto yy55;
		}
	}
yy472:
	yych = *++cur;
	if (yych == 'H') goto yy666;
	if (yych == 'h') goto yy666;
	goto yy55;
yy473:
	yych = *++cur;
	if (yych == 'Y') goto yy667;
	if (yych == 'y') goto yy667;
	goto yy55;
yy474:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'E') {
		if (yych <= '9') {
			if (yych >= '0') goto yy54;
		} else {
			if (yych <= '@') goto yy475;
			if (yych <= 'D') goto yy54;
			goto yy669;
		}
	} else {
		if (yych <= 'd') {
			if (yych <= 'Z') goto yy54;
			if (yych >= 'a') goto yy54;
		} else {
			if (yych <= 'e') goto yy669;
			if (yych <= 'z') goto yy54;
		}
	}
yy475:
#line 379 "tokenizer.re"
	{ KW(ENUM) }
#line 3939 "tokenizer.cpp"
yy476:
	yych = *++cur;
	if (yych == 'V') goto yy670;
	if (yych == 'v') goto yy670;
	goto yy55;
yy477:
	yych = *++cur;
	if (yych == 'S') goto yy671;
	if (yych == 's') goto yy671;
	goto yy55;
yy478:
	yych = *++cur;
	if (yych == 'R') goto yy672;
	if (yych == 'r') goto yy672;
	goto yy55;
yy479:
	yych = *++cur;
	if (yych == 'T') goto yy674;
	if (yych == 't') goto yy674;
	goto yy55;
yy480:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 385 "tokenizer.re"
	{ KW(EXIT) }
#line 3968 "tokenizer.cpp"
yy482:
	yych = *++cur;
	if (yych <= 'R') {
		if (yych == 'N') goto yy676;
		if (yych <= 'Q') goto yy55;
		goto yy677;
	} else {
		if (yych <= 'n') {
			if (yych <= 'm') goto yy55;
			goto yy676;
		} else {
			if (yych == 'r') goto yy677;
			goto yy55;
		}
	}
yy483:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 388 "tokenizer.re"
	{ KW(FILE) }
#line 3992 "tokenizer.cpp"
yy485:
	yych = *++cur;
	if (yych == 'L') goto yy678;
	if (yych == 'l') goto yy678;
	goto yy55;
yy486:
	yych = *++cur;
	if (yych == 'H') goto yy680;
	if (yych == 'h') goto yy680;
	goto yy55;
yy487:
	yych = *++cur;
	if (yych == 'L') goto yy682;
	if (yych == 'l') goto yy682;
	goto yy55;
yy488:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'T') {
		if (yych <= '@') {
			if (yych <= '/') goto yy489;
			if (yych <= '9') goto yy54;
		} else {
			if (yych <= 'A') goto yy683;
			if (yych <= 'S') goto yy54;
			goto yy684;
		}
	} else {
		if (yych <= 'a') {
			if (yych <= 'Z') goto yy54;
			if (yych >= 'a') goto yy683;
		} else {
			if (yych == 't') goto yy684;
			if (yych <= 'z') goto yy54;
		}
	}
yy489:
#line 405 "tokenizer.re"
	{ KW(FORM) }
#line 4034 "tokenizer.cpp"
yy490:
	yych = *++cur;
	if (yych == 'T') goto yy685;
	if (yych == 't') goto yy685;
	goto yy55;
yy491:
	yych = *++cur;
	if (yych == 'R') goto yy686;
	if (yych == 'r') goto yy686;
	goto yy55;
yy492:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 410 "tokenizer.re"
	{ KW(GOTO) }
#line 4053 "tokenizer.cpp"
yy494:
	yych = *++cur;
	if (yych == 'E') goto yy687;
	if (yych == 'e') goto yy687;
	goto yy55;
yy495:
	yych = *++cur;
	if (yych == 'I') goto yy688;
	if (yych == 'i') goto yy688;
	goto yy55;
yy496:
	yych = *++cur;
	if (yych == 'R') goto yy689;
	if (yych == 'r') goto yy689;
	goto yy55;
yy497:
	yych = *++cur;
	if (yych == 'R') goto yy690;
	if (yych == 'r') goto yy690;
	goto yy55;
yy498:
	yych = *++cur;
	if (yych == 'U') goto yy691;
	if (yych == 'u') goto yy691;
	goto yy79;
yy499:
	yych = *++cur;
	if (yych == 'U') goto yy692;
	if (yych == 'u') goto yy692;
	goto yy55;
yy500:
	yych = *++cur;
	if (yych == 'T') goto yy693;
	if (yych == 't') goto yy693;
	goto yy55;
yy501:
	yych = *++cur;
	if (yych == 'I') goto yy695;
	if (yych == 'i') goto yy695;
	goto yy55;
yy502:
	yych = *++cur;
	if (yych <= 'R') {
		if (yych <= 'M') {
			if (yych == 'G') goto yy696;
			goto yy55;
		} else {
			if (yych <= 'N') goto yy697;
			if (yych <= 'Q') goto yy55;
			goto yy698;
		}
	} else {
		if (yych <= 'm') {
			if (yych == 'g') goto yy696;
			goto yy55;
		} else {
			if (yych <= 'n') goto yy697;
			if (yych == 'r') goto yy698;
			goto yy55;
		}
	}
yy503:
	yych = *++cur;
	if (yych == 'I') goto yy699;
	if (yych == 'i') goto yy699;
	goto yy55;
yy504:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 426 "tokenizer.re"
	{ KW(KIND) }
#line 4128 "tokenizer.cpp"
yy506:
	yych = *++cur;
	if (yych == 'L') goto yy700;
	if (yych == 'l') goto yy700;
	goto yy55;
yy507:
	yych = *++cur;
	if (yych == 'C') goto yy702;
	if (yych == 'c') goto yy702;
	goto yy55;
yy508:
	yych = *++cur;
	if (yych == 'R') goto yy703;
	if (yych == 'r') goto yy703;
	goto yy55;
yy509:
	yych = *++cur;
	if (yych == 'L') goto yy704;
	if (yych == 'l') goto yy704;
	goto yy55;
yy510:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 433 "tokenizer.re"
	{ KW(MOLD) }
#line 4157 "tokenizer.cpp"
yy512:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'L') {
		if (yych <= '9') {
			if (yych >= '0') goto yy54;
		} else {
			if (yych <= '@') goto yy513;
			if (yych <= 'K') goto yy54;
			goto yy705;
		}
	} else {
		if (yych <= 'k') {
			if (yych <= 'Z') goto yy54;
			if (yych >= 'a') goto yy54;
		} else {
			if (yych <= 'l') goto yy705;
			if (yych <= 'z') goto yy54;
		}
	}
yy513:
#line 434 "tokenizer.re"
	{ KW(NAME) }
#line 4183 "tokenizer.cpp"
yy514:
	yyaccept = 2;
	yych = *(mar = ++cur);
	if (yych <= '\'') {
		if (yych == '"') goto yy82;
		if (yych <= '&') goto yy55;
		goto yy97;
	} else {
		if (yych <= 'I') {
			if (yych <= 'H') goto yy55;
			goto yy706;
		} else {
			if (yych == 'i') goto yy706;
			goto yy55;
		}
	}
yy515:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 441 "tokenizer.re"
	{ KW(NONE) }
#line 4208 "tokenizer.cpp"
yy517:
	yyaccept = 2;
	yych = *(mar = ++cur);
	if (yych <= 'O') {
		if (yych <= '\'') {
			if (yych == '"') goto yy82;
			if (yych <= '&') goto yy55;
			goto yy97;
		} else {
			if (yych == 'I') goto yy707;
			if (yych <= 'N') goto yy55;
			goto yy708;
		}
	} else {
		if (yych <= 'i') {
			if (yych == 'R') goto yy709;
			if (yych <= 'h') goto yy55;
			goto yy707;
		} else {
			if (yych <= 'o') {
				if (yych <= 'n') goto yy55;
				goto yy708;
			} else {
				if (yych == 'r') goto yy709;
				goto yy55;
			}
		}
	}
yy518:
	yych = *++cur;
	if (yych == 'S') goto yy710;
	if (yych == 's') goto yy710;
	goto yy55;
yy519:
	yych = *++cur;
	if (yych == 'I') goto yy711;
	if (yych == 'i') goto yy711;
	goto yy55;
yy520:
	++cur;
#line 631 "tokenizer.re"
	{ token(yylval.string); RET(TK_BOZ_CONSTANT) }
#line 4251 "tokenizer.cpp"
yy522:
	++cur;
#line 632 "tokenizer.re"
	{ token(yylval.string); RET(TK_BOZ_CONSTANT) }
#line 4256 "tokenizer.cpp"
yy524:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 443 "tokenizer.re"
	{ KW(ONLY) }
#line 4265 "tokenizer.cpp"
yy526:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 444 "tokenizer.re"
	{ KW(OPEN) }
#line 4274 "tokenizer.cpp"
yy528:
	yych = *++cur;
	if (yych == 'A') goto yy712;
	if (yych == 'a') goto yy712;
	goto yy55;
yy529:
	yych = *++cur;
	if (yych == 'O') goto yy713;
	if (yych == 'o') goto yy713;
	goto yy55;
yy530:
	yych = *++cur;
	if (yych == 'M') goto yy714;
	if (yych == 'm') goto yy714;
	goto yy55;
yy531:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 449 "tokenizer.re"
	{ KW(PASS) }
#line 4298 "tokenizer.cpp"
yy533:
	yych = *++cur;
	if (yych == 'T') goto yy715;
	if (yych == 't') goto yy715;
	goto yy55;
yy534:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 451 "tokenizer.re"
	{ KW(POST) }
#line 4312 "tokenizer.cpp"
yy536:
	yych = *++cur;
	if (yych == 'I') goto yy716;
	if (yych == 'i') goto yy716;
	goto yy55;
yy537:
	yych = *++cur;
	if (yych == 'T') goto yy717;
	if (yych == 't') goto yy717;
	goto yy55;
yy538:
	yych = *++cur;
	if (yych == 'A') goto yy719;
	if (yych == 'a') goto yy719;
	goto yy55;
yy539:
	yych = *++cur;
	if (yych == 'E') goto yy720;
	if (yych == 'e') goto yy720;
	goto yy55;
yy540:
	yych = *++cur;
	if (yych == 'R') goto yy721;
	if (yych == 'r') goto yy721;
	goto yy55;
yy541:
	yych = *++cur;
	if (yych == 'E') goto yy722;
	if (yych == 'e') goto yy722;
	goto yy55;
yy542:
	yych = *++cur;
	if (yych == 'I') goto yy723;
	if (yych == 'i') goto yy723;
	goto yy55;
yy543:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 459 "tokenizer.re"
	{ KW(PURE) }
#line 4356 "tokenizer.cpp"
yy545:
	yych = *++cur;
	if (yych == 'T') goto yy724;
	if (yych == 't') goto yy724;
	goto yy55;
yy546:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 461 "tokenizer.re"
	{ KW(RANK) }
#line 4370 "tokenizer.cpp"
yy548:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 462 "tokenizer.re"
	{ KW(READ) }
#line 4379 "tokenizer.cpp"
yy550:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 463 "tokenizer.re"
	{KW(REAL) }
#line 4388 "tokenizer.cpp"
yy552:
	yych = *++cur;
	if (yych == 'R') goto yy726;
	if (yych == 'r') goto yy726;
	goto yy55;
yy553:
	yych = *++cur;
	if (yych == 'C') goto yy727;
	if (yych == 'c') goto yy727;
	goto yy55;
yy554:
	yych = *++cur;
	if (yych == 'L') goto yy728;
	if (yych == 'l') goto yy728;
	goto yy55;
yy555:
	yych = *++cur;
	if (yych == 'R') goto yy729;
	if (yych == 'r') goto yy729;
	goto yy55;
yy556:
	yych = *++cur;
	if (yych == 'N') goto yy730;
	if (yych == 'n') goto yy730;
	goto yy55;
yy557:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 469 "tokenizer.re"
	{ KW(SAVE) }
#line 4422 "tokenizer.cpp"
yy559:
	yych = *++cur;
	if (yych == 'C') goto yy731;
	if (yych == 'c') goto yy731;
	goto yy55;
yy560:
	yych = *++cur;
	if (yych == 'E') goto yy732;
	if (yych == 'e') goto yy732;
	goto yy55;
yy561:
	yych = *++cur;
	if (yych == 'E') goto yy733;
	if (yych == 'e') goto yy733;
	goto yy55;
yy562:
	yych = *++cur;
	if (yych == 'C') goto yy734;
	if (yych == 'c') goto yy734;
	goto yy55;
yy563:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 477 "tokenizer.re"
	{ KW(STAT) }
#line 4451 "tokenizer.cpp"
yy565:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 478 "tokenizer.re"
	{ KW(STOP) }
#line 4460 "tokenizer.cpp"
yy567:
	yych = *++cur;
	if (yych == 'O') goto yy735;
	if (yych == 'o') goto yy735;
	goto yy55;
yy568:
	yych = *++cur;
	if (yych == 'O') goto yy736;
	if (yych == 'o') goto yy736;
	goto yy55;
yy569:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'T') {
		if (yych <= 'H') {
			if (yych <= '9') {
				if (yych >= '0') goto yy54;
			} else {
				if (yych <= '@') goto yy570;
				if (yych <= 'A') goto yy737;
				goto yy54;
			}
		} else {
			if (yych <= 'L') {
				if (yych <= 'I') goto yy738;
				goto yy54;
			} else {
				if (yych <= 'M') goto yy739;
				if (yych <= 'S') goto yy54;
				goto yy740;
			}
		}
	} else {
		if (yych <= 'i') {
			if (yych <= '`') {
				if (yych <= 'Z') goto yy54;
			} else {
				if (yych <= 'a') goto yy737;
				if (yych <= 'h') goto yy54;
				goto yy738;
			}
		} else {
			if (yych <= 's') {
				if (yych == 'm') goto yy739;
				goto yy54;
			} else {
				if (yych <= 't') goto yy740;
				if (yych <= 'z') goto yy54;
			}
		}
	}
yy570:
#line 481 "tokenizer.re"
	{ KW(SYNC) }
#line 4517 "tokenizer.cpp"
yy571:
	yych = *++cur;
	if (yych == 'E') goto yy741;
	if (yych == 'e') goto yy741;
	goto yy55;
yy572:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy742;
#line 487 "tokenizer.re"
	{ KW(TEAM) }
#line 4531 "tokenizer.cpp"
yy574:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 489 "tokenizer.re"
	{ KW(THEN) }
#line 4540 "tokenizer.cpp"
yy576:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 491 "tokenizer.re"
	{ KW(TYPE) }
#line 4549 "tokenizer.cpp"
yy578:
	yych = *++cur;
	if (yych == 'R') goto yy743;
	if (yych == 'r') goto yy743;
	goto yy55;
yy579:
	yych = *++cur;
	if (yych == 'E') goto yy744;
	if (yych == 'e') goto yy744;
	goto yy55;
yy580:
	yych = *++cur;
	if (yych == 'T') goto yy746;
	if (yych == 't') goto yy746;
	goto yy55;
yy581:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 496 "tokenizer.re"
	{ KW(WAIT) }
#line 4573 "tokenizer.cpp"
yy583:
	yych = *++cur;
	if (yych == 'E') goto yy747;
	if (yych == 'e') goto yy747;
	goto yy55;
yy584:
	yych = *++cur;
	if (yych == 'E') goto yy749;
	if (yych == 'e') goto yy749;
	goto yy55;
yy585:
	yych = *++cur;
	if (yych == 'E') goto yy751;
	if (yych == 'e') goto yy751;
	goto yy55;
yy586:
	++cur;
#line 633 "tokenizer.re"
	{ token(yylval.string); RET(TK_BOZ_CONSTANT) }
#line 4593 "tokenizer.cpp"
yy588:
	++cur;
#line 634 "tokenizer.re"
	{ token(yylval.string); RET(TK_BOZ_CONSTANT) }
#line 4598 "tokenizer.cpp"
yy590:
	++cur;
	cur = ctxmar;
#line 523 "tokenizer.re"
	{ RET(TK_LPAREN) }
#line 4604 "tokenizer.cpp"
yy592:
	++cur;
#line 569 "tokenizer.re"
	{ RET(TK_AND) }
#line 4609 "tokenizer.cpp"
yy594:
	++cur;
#line 572 "tokenizer.re"
	{ RET(TK_EQV) }
#line 4614 "tokenizer.cpp"
yy596:
	yych = *++cur;
	if (yych == 'E') goto yy753;
	if (yych == 'e') goto yy753;
	goto yy111;
yy597:
	yych = *++cur;
	if (yych == '.') goto yy754;
	goto yy111;
yy598:
	++cur;
#line 568 "tokenizer.re"
	{ RET(TK_NOT) }
#line 4628 "tokenizer.cpp"
yy600:
	yych = *++cur;
	if (yych == '.') goto yy756;
	goto yy111;
yy601:
	++cur;
#line 571 "tokenizer.re"
	{ RET(TK_XOR) }
#line 4637 "tokenizer.cpp"
yy603:
	yych = *++cur;
	if (yych == 'A') goto yy758;
	if (yych == 'a') goto yy758;
	goto yy55;
yy604:
	yych = *++cur;
	if (yych == 'A') goto yy759;
	if (yych == 'a') goto yy759;
	goto yy55;
yy605:
	yych = *++cur;
	if (yych == 'N') goto yy760;
	if (yych == 'n') goto yy760;
	goto yy55;
yy606:
	yych = *++cur;
	if (yych == 'I') goto yy762;
	if (yych == 'i') goto yy762;
	goto yy55;
yy607:
	yych = *++cur;
	if (yych == 'H') goto yy763;
	if (yych == 'h') goto yy763;
	goto yy55;
yy608:
	yych = *++cur;
	if (yych == 'P') goto yy764;
	if (yych == 'p') goto yy764;
	goto yy55;
yy609:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 264 "tokenizer.re"
	{ KW(BLOCK) }
#line 4676 "tokenizer.cpp"
yy611:
	yych = *++cur;
	if (yych == 'E') goto yy765;
	if (yych == 'e') goto yy765;
	goto yy55;
yy612:
	yych = *++cur;
	if (yych == 'C') goto yy767;
	if (yych == 'c') goto yy767;
	goto yy55;
yy613:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 270 "tokenizer.re"
	{ KW(CLASS) }
#line 4695 "tokenizer.cpp"
yy615:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 271 "tokenizer.re"
	{ KW(CLOSE) }
#line 4704 "tokenizer.cpp"
yy617:
	yych = *++cur;
	if (yych == 'E') goto yy768;
	if (yych == 'e') goto yy768;
	goto yy55;
yy618:
	yych = *++cur;
	if (yych == 'N') goto yy769;
	if (yych == 'n') goto yy769;
	goto yy55;
yy619:
	yych = *++cur;
	if (yych == 'E') goto yy771;
	if (yych == 'e') goto yy771;
	goto yy55;
yy620:
	yych = *++cur;
	if (yych == 'R') goto yy772;
	if (yych == 'r') goto yy772;
	goto yy55;
yy621:
	yych = *++cur;
	if (yych == 'I') goto yy773;
	if (yych == 'i') goto yy773;
	goto yy55;
yy622:
	yych = *++cur;
	if (yych <= 'N') {
		if (yych == 'G') goto yy774;
		if (yych <= 'M') goto yy55;
		goto yy775;
	} else {
		if (yych <= 'g') {
			if (yych <= 'f') goto yy55;
			goto yy774;
		} else {
			if (yych == 'n') goto yy775;
			goto yy55;
		}
	}
yy623:
	yych = *++cur;
	if (yych == 'C') goto yy776;
	if (yych == 'c') goto yy776;
	goto yy55;
yy624:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 280 "tokenizer.re"
	{ KW(CYCLE) }
#line 4758 "tokenizer.cpp"
yy626:
	yych = *++cur;
	if (yych == 'O') goto yy777;
	if (yych == 'o') goto yy777;
	goto yy55;
yy627:
	yych = *++cur;
	if (yych == 'L') goto yy778;
	if (yych == 'l') goto yy778;
	goto yy55;
yy628:
	yych = *++cur;
	if (yych == 'R') goto yy779;
	if (yych == 'r') goto yy779;
	goto yy55;
yy629:
	yych = *++cur;
	if (yych == 'S') goto yy780;
	if (yych == 's') goto yy780;
	goto yy55;
yy630:
	yych = *++cur;
	if (yych == 'E') goto yy781;
	if (yych == 'e') goto yy781;
	goto yy55;
yy631:
	yych = *++cur;
	if (yych == 'L') goto yy783;
	if (yych == 'l') goto yy783;
	goto yy55;
yy632:
	yych = *++cur;
	if (yych == 'N') goto yy784;
	if (yych == 'n') goto yy784;
	goto yy55;
yy633:
	yych = *++cur;
	if (yych == 'F') goto yy785;
	if (yych == 'f') goto yy785;
	goto yy55;
yy634:
	yych = *++cur;
	if (yych == 'H') goto yy787;
	if (yych == 'h') goto yy787;
	goto yy55;
yy635:
	yych = *++cur;
	if (yych == 'S') goto yy788;
	if (yych == 's') goto yy788;
	goto yy79;
yy636:
	yych = *++cur;
	if (yych == 'L') goto yy789;
	if (yych == 'l') goto yy789;
	goto yy79;
yy637:
	yych = *++cur;
	if (yych == 'R') goto yy790;
	if (yych == 'r') goto yy790;
	goto yy79;
yy638:
	yych = *++cur;
	if (yych == 'O') goto yy791;
	if (yych == 'o') goto yy791;
	goto yy79;
yy639:
	yych = *++cur;
	if (yych == 'N') goto yy793;
	if (yych == 'n') goto yy793;
	goto yy79;
yy640:
	yych = *++cur;
	if (yych <= 'U') {
		if (yych == 'O') goto yy794;
		if (yych <= 'T') goto yy79;
		goto yy795;
	} else {
		if (yych <= 'o') {
			if (yych <= 'n') goto yy79;
			goto yy794;
		} else {
			if (yych == 'u') goto yy795;
			goto yy79;
		}
	}
yy641:
	yych = *++cur;
	if (yych <= 'N') {
		if (yych == 'F') goto yy796;
		if (yych <= 'M') goto yy79;
		goto yy798;
	} else {
		if (yych <= 'f') {
			if (yych <= 'e') goto yy79;
			goto yy796;
		} else {
			if (yych == 'n') goto yy798;
			goto yy79;
		}
	}
yy642:
	yych = *++cur;
	if (yych == 'O') goto yy799;
	if (yych == 'o') goto yy799;
	goto yy79;
yy643:
	yych = *++cur;
	if (yych == 'R') goto yy800;
	if (yych == 'r') goto yy800;
	goto yy79;
yy644:
	yych = *++cur;
	if (yych <= 'U') {
		if (yych == 'E') goto yy801;
		if (yych <= 'T') goto yy79;
		goto yy802;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy79;
			goto yy801;
		} else {
			if (yych == 'u') goto yy802;
			goto yy79;
		}
	}
yy645:
	yych = *++cur;
	if (yych <= 'Y') {
		if (yych == 'E') goto yy803;
		if (yych <= 'X') goto yy79;
		goto yy804;
	} else {
		if (yych <= 'e') {
			if (yych <= 'd') goto yy79;
			goto yy803;
		} else {
			if (yych == 'y') goto yy804;
			goto yy79;
		}
	}
yy646:
	yych = *++cur;
	if (yych == 'H') goto yy805;
	if (yych == 'h') goto yy805;
	goto yy79;
yy647:
	yych = *++cur;
	if (yych <= 'U') {
		if (yych <= 'N') {
			if (yych == 'I') goto yy806;
			goto yy79;
		} else {
			if (yych <= 'O') goto yy794;
			if (yych <= 'T') goto yy79;
			goto yy795;
		}
	} else {
		if (yych <= 'n') {
			if (yych == 'i') goto yy806;
			goto yy79;
		} else {
			if (yych <= 'o') goto yy794;
			if (yych == 'u') goto yy795;
			goto yy79;
		}
	}
yy648:
	yych = *++cur;
	if (yych == 'S') goto yy807;
	if (yych == 's') goto yy807;
	goto yy55;
yy649:
	yych = *++cur;
	if (yych == 'O') goto yy808;
	if (yych == 'o') goto yy808;
	goto yy55;
yy650:
	yych = *++cur;
	if (yych == 'I') goto yy809;
	if (yych == 'i') goto yy809;
	goto yy55;
yy651:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 364 "tokenizer.re"
	{
                if (enddo_newline_process) {
                    KW(CONTINUE)
                } else {
                    KW(ENDDO)
                }
            }
#line 4954 "tokenizer.cpp"
yy653:
	yych = *++cur;
	if (yych == 'U') goto yy810;
	if (yych == 'u') goto yy810;
	goto yy55;
yy654:
	yych = *++cur;
	if (yych == 'L') goto yy811;
	if (yych == 'l') goto yy811;
	goto yy55;
yy655:
	yych = *++cur;
	if (yych == 'R') goto yy812;
	if (yych == 'r') goto yy812;
	goto yy55;
yy656:
	yych = *++cur;
	if (yych == 'N') goto yy813;
	if (yych == 'n') goto yy813;
	goto yy55;
yy657:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 349 "tokenizer.re"
	{ KW(ENDIF) }
#line 4983 "tokenizer.cpp"
yy659:
	yych = *++cur;
	if (yych == 'T') goto yy814;
	if (yych == 't') goto yy814;
	goto yy55;
yy660:
	yych = *++cur;
	if (yych == 'D') goto yy815;
	if (yych == 'd') goto yy815;
	goto yy55;
yy661:
	yych = *++cur;
	if (yych == 'O') goto yy816;
	if (yych == 'o') goto yy816;
	goto yy55;
yy662:
	yych = *++cur;
	if (yych == 'L') goto yy817;
	if (yych == 'l') goto yy817;
	goto yy55;
yy663:
	yych = *++cur;
	if (yych == 'B') goto yy818;
	if (yych == 'b') goto yy818;
	goto yy55;
yy664:
	yych = *++cur;
	if (yych == 'A') goto yy819;
	if (yych == 'a') goto yy819;
	goto yy55;
yy665:
	yych = *++cur;
	if (yych == 'P') goto yy820;
	if (yych == 'p') goto yy820;
	goto yy55;
yy666:
	yych = *++cur;
	if (yych == 'E') goto yy821;
	if (yych == 'e') goto yy821;
	goto yy55;
yy667:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 378 "tokenizer.re"
	{ KW(ENTRY) }
#line 5032 "tokenizer.cpp"
yy669:
	yych = *++cur;
	if (yych == 'R') goto yy822;
	if (yych == 'r') goto yy822;
	goto yy55;
yy670:
	yych = *++cur;
	if (yych == 'A') goto yy823;
	if (yych == 'a') goto yy823;
	goto yy55;
yy671:
	yych = *++cur;
	if (yych == 'G') goto yy824;
	if (yych == 'g') goto yy824;
	goto yy55;
yy672:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 383 "tokenizer.re"
	{ KW(ERROR) }
#line 5056 "tokenizer.cpp"
yy674:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 384 "tokenizer.re"
	{ KW(EVENT) }
#line 5065 "tokenizer.cpp"
yy676:
	yych = *++cur;
	if (yych == 'D') goto yy826;
	if (yych == 'd') goto yy826;
	goto yy55;
yy677:
	yych = *++cur;
	if (yych == 'N') goto yy827;
	if (yych == 'n') goto yy827;
	goto yy55;
yy678:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 389 "tokenizer.re"
	{ KW(FINAL) }
#line 5084 "tokenizer.cpp"
yy680:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 390 "tokenizer.re"
	{ KW(FLUSH) }
#line 5093 "tokenizer.cpp"
yy682:
	yych = *++cur;
	if (yych == 'L') goto yy828;
	if (yych == 'l') goto yy828;
	goto yy55;
yy683:
	yych = *++cur;
	if (yych == 'T') goto yy830;
	if (yych == 't') goto yy830;
	goto yy55;
yy684:
	yych = *++cur;
	if (yych == 'E') goto yy832;
	if (yych == 'e') goto yy832;
	goto yy55;
yy685:
	yych = *++cur;
	if (yych == 'I') goto yy833;
	if (yych == 'i') goto yy833;
	goto yy55;
yy686:
	yych = *++cur;
	if (yych == 'I') goto yy834;
	if (yych == 'i') goto yy834;
	goto yy55;
yy687:
	yych = *++cur;
	if (yych == 'S') goto yy835;
	if (yych == 's') goto yy835;
	goto yy55;
yy688:
	yych = *++cur;
	if (yych == 'C') goto yy837;
	if (yych == 'c') goto yy837;
	goto yy55;
yy689:
	yych = *++cur;
	if (yych == 'T') goto yy838;
	if (yych == 't') goto yy838;
	goto yy55;
yy690:
	yych = *++cur;
	if (yych == 'E') goto yy840;
	if (yych == 'e') goto yy840;
	goto yy55;
yy691:
	yych = *++cur;
	if (yych == 'T') goto yy842;
	if (yych == 't') goto yy842;
	goto yy79;
yy692:
	yych = *++cur;
	if (yych == 'D') goto yy844;
	if (yych == 'd') goto yy844;
	goto yy55;
yy693:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 418 "tokenizer.re"
	{ KW(INOUT) }
#line 5157 "tokenizer.cpp"
yy695:
	yych = *++cur;
	if (yych == 'R') goto yy845;
	if (yych == 'r') goto yy845;
	goto yy55;
yy696:
	yych = *++cur;
	if (yych == 'E') goto yy846;
	if (yych == 'e') goto yy846;
	goto yy55;
yy697:
	yych = *++cur;
	if (yych == 'T') goto yy847;
	if (yych == 't') goto yy847;
	goto yy55;
yy698:
	yych = *++cur;
	if (yych == 'F') goto yy849;
	if (yych == 'f') goto yy849;
	goto yy55;
yy699:
	yych = *++cur;
	if (yych == 'N') goto yy850;
	if (yych == 'n') goto yy850;
	goto yy55;
yy700:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy851;
#line 428 "tokenizer.re"
	{ KW(LOCAL) }
#line 5191 "tokenizer.cpp"
yy702:
	yych = *++cur;
	if (yych == 'A') goto yy852;
	if (yych == 'a') goto yy852;
	goto yy55;
yy703:
	yych = *++cur;
	if (yych == 'Y') goto yy853;
	if (yych == 'y') goto yy853;
	goto yy55;
yy704:
	yych = *++cur;
	if (yych == 'E') goto yy855;
	if (yych == 'e') goto yy855;
	goto yy55;
yy705:
	yych = *++cur;
	if (yych == 'I') goto yy857;
	if (yych == 'i') goto yy857;
	goto yy55;
yy706:
	yych = *++cur;
	if (yych == 'N') goto yy858;
	if (yych == 'n') goto yy858;
	goto yy55;
yy707:
	yych = *++cur;
	if (yych == 'N') goto yy859;
	if (yych == 'n') goto yy859;
	goto yy55;
yy708:
	yych = *++cur;
	if (yych == 'V') goto yy860;
	if (yych == 'v') goto yy860;
	goto yy55;
yy709:
	yych = *++cur;
	if (yych == 'E') goto yy861;
	if (yych == 'e') goto yy861;
	goto yy55;
yy710:
	yych = *++cur;
	if (yych == 'S') goto yy862;
	if (yych == 's') goto yy862;
	goto yy55;
yy711:
	yych = *++cur;
	if (yych == 'F') goto yy864;
	if (yych == 'f') goto yy864;
	goto yy55;
yy712:
	yych = *++cur;
	if (yych == 'T') goto yy865;
	if (yych == 't') goto yy865;
	goto yy55;
yy713:
	yych = *++cur;
	if (yych == 'N') goto yy866;
	if (yych == 'n') goto yy866;
	goto yy55;
yy714:
	yych = *++cur;
	if (yych == 'E') goto yy867;
	if (yych == 'e') goto yy867;
	goto yy55;
yy715:
	yych = *++cur;
	if (yych == 'E') goto yy868;
	if (yych == 'e') goto yy868;
	goto yy55;
yy716:
	yych = *++cur;
	if (yych == 'S') goto yy869;
	if (yych == 's') goto yy869;
	goto yy55;
yy717:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 453 "tokenizer.re"
	{ KW(PRINT) }
#line 5275 "tokenizer.cpp"
yy719:
	yych = *++cur;
	if (yych == 'T') goto yy870;
	if (yych == 't') goto yy870;
	goto yy55;
yy720:
	yych = *++cur;
	if (yych == 'D') goto yy871;
	if (yych == 'd') goto yy871;
	goto yy55;
yy721:
	yych = *++cur;
	if (yych == 'A') goto yy872;
	if (yych == 'a') goto yy872;
	goto yy55;
yy722:
	yych = *++cur;
	if (yych == 'C') goto yy873;
	if (yych == 'c') goto yy873;
	goto yy55;
yy723:
	yych = *++cur;
	if (yych == 'C') goto yy874;
	if (yych == 'c') goto yy874;
	goto yy55;
yy724:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 460 "tokenizer.re"
	{ KW(QUIET) }
#line 5309 "tokenizer.cpp"
yy726:
	yych = *++cur;
	if (yych == 'S') goto yy876;
	if (yych == 's') goto yy876;
	goto yy55;
yy727:
	yych = *++cur;
	if (yych == 'E') goto yy877;
	if (yych == 'e') goto yy877;
	goto yy55;
yy728:
	yych = *++cur;
	if (yych == 'T') goto yy879;
	if (yych == 't') goto yy879;
	goto yy55;
yy729:
	yych = *++cur;
	if (yych == 'N') goto yy881;
	if (yych == 'n') goto yy881;
	goto yy55;
yy730:
	yych = *++cur;
	if (yych == 'D') goto yy883;
	if (yych == 'd') goto yy883;
	goto yy55;
yy731:
	yych = *++cur;
	if (yych == 'T') goto yy885;
	if (yych == 't') goto yy885;
	goto yy55;
yy732:
	yych = *++cur;
	if (yych == 'N') goto yy887;
	if (yych == 'n') goto yy887;
	goto yy55;
yy733:
	yych = *++cur;
	if (yych == 'D') goto yy888;
	if (yych == 'd') goto yy888;
	goto yy55;
yy734:
	yych = *++cur;
	if (yych == 'E') goto yy890;
	if (yych == 'e') goto yy890;
	goto yy55;
yy735:
	yych = *++cur;
	if (yych == 'D') goto yy892;
	if (yych == 'd') goto yy892;
	goto yy55;
yy736:
	yych = *++cur;
	if (yych == 'U') goto yy893;
	if (yych == 'u') goto yy893;
	goto yy55;
yy737:
	yych = *++cur;
	if (yych == 'L') goto yy894;
	if (yych == 'l') goto yy894;
	goto yy55;
yy738:
	yych = *++cur;
	if (yych == 'M') goto yy895;
	if (yych == 'm') goto yy895;
	goto yy55;
yy739:
	yych = *++cur;
	if (yych == 'E') goto yy896;
	if (yych == 'e') goto yy896;
	goto yy55;
yy740:
	yych = *++cur;
	if (yych == 'E') goto yy897;
	if (yych == 'e') goto yy897;
	goto yy55;
yy741:
	yych = *++cur;
	if (yych == 'T') goto yy898;
	if (yych == 't') goto yy898;
	goto yy55;
yy742:
	yyaccept = 2;
	yych = *(mar = ++cur);
	if (yych <= '\'') {
		if (yych == '"') goto yy82;
		if (yych <= '&') goto yy55;
		goto yy97;
	} else {
		if (yych <= 'N') {
			if (yych <= 'M') goto yy55;
			goto yy900;
		} else {
			if (yych == 'n') goto yy900;
			goto yy55;
		}
	}
yy743:
	yych = *++cur;
	if (yych == 'M') goto yy901;
	if (yych == 'm') goto yy901;
	goto yy55;
yy744:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 494 "tokenizer.re"
	{ KW(VALUE) }
#line 5419 "tokenizer.cpp"
yy746:
	yych = *++cur;
	if (yych == 'I') goto yy902;
	if (yych == 'i') goto yy902;
	goto yy55;
yy747:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 497 "tokenizer.re"
	{ KW(WHERE) }
#line 5433 "tokenizer.cpp"
yy749:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 498 "tokenizer.re"
	{ KW(WHILE) }
#line 5442 "tokenizer.cpp"
yy751:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 499 "tokenizer.re"
	{ KW(WRITE) }
#line 5451 "tokenizer.cpp"
yy753:
	yych = *++cur;
	if (yych == '.') goto yy903;
	goto yy111;
yy754:
	++cur;
#line 573 "tokenizer.re"
	{ RET(TK_NEQV) }
#line 5460 "tokenizer.cpp"
yy756:
	yyaccept = 11;
	yych = *(mar = ++cur);
	if (yych == '_') goto yy905;
yy757:
#line 577 "tokenizer.re"
	{ RET(TK_TRUE) }
#line 5468 "tokenizer.cpp"
yy758:
	yych = *++cur;
	if (yych == 'C') goto yy906;
	if (yych == 'c') goto yy906;
	goto yy55;
yy759:
	yych = *++cur;
	if (yych == 'T') goto yy907;
	if (yych == 't') goto yy907;
	goto yy55;
yy760:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'M') {
		if (yych <= '9') {
			if (yych >= '0') goto yy54;
		} else {
			if (yych <= '@') goto yy761;
			if (yych <= 'L') goto yy54;
			goto yy908;
		}
	} else {
		if (yych <= 'l') {
			if (yych <= 'Z') goto yy54;
			if (yych >= 'a') goto yy54;
		} else {
			if (yych <= 'm') goto yy908;
			if (yych <= 'z') goto yy54;
		}
	}
yy761:
#line 258 "tokenizer.re"
	{ KW(ASSIGN) }
#line 5504 "tokenizer.cpp"
yy762:
	yych = *++cur;
	if (yych == 'A') goto yy909;
	if (yych == 'a') goto yy909;
	goto yy55;
yy763:
	yych = *++cur;
	if (yych == 'R') goto yy910;
	if (yych == 'r') goto yy910;
	goto yy55;
yy764:
	yych = *++cur;
	if (yych == 'A') goto yy911;
	if (yych == 'a') goto yy911;
	goto yy55;
yy765:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'T') {
		if (yych <= '9') {
			if (yych >= '0') goto yy54;
		} else {
			if (yych <= '@') goto yy766;
			if (yych <= 'S') goto yy54;
			goto yy912;
		}
	} else {
		if (yych <= 's') {
			if (yych <= 'Z') goto yy54;
			if (yych >= 'a') goto yy54;
		} else {
			if (yych <= 't') goto yy912;
			if (yych <= 'z') goto yy54;
		}
	}
yy766:
#line 267 "tokenizer.re"
	{ KW(CHANGE) }
#line 5545 "tokenizer.cpp"
yy767:
	yych = *++cur;
	if (yych == 'T') goto yy913;
	if (yych == 't') goto yy913;
	goto yy55;
yy768:
	yych = *++cur;
	if (yych == 'N') goto yy914;
	if (yych == 'n') goto yy914;
	goto yy55;
yy769:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 273 "tokenizer.re"
	{ KW(COMMON) }
#line 5564 "tokenizer.cpp"
yy771:
	yych = *++cur;
	if (yych == 'X') goto yy915;
	if (yych == 'x') goto yy915;
	goto yy55;
yy772:
	yych = *++cur;
	if (yych == 'R') goto yy917;
	if (yych == 'r') goto yy917;
	goto yy55;
yy773:
	yych = *++cur;
	if (yych == 'N') goto yy918;
	if (yych == 'n') goto yy918;
	goto yy55;
yy774:
	yych = *++cur;
	if (yych == 'U') goto yy919;
	if (yych == 'u') goto yy919;
	goto yy55;
yy775:
	yych = *++cur;
	if (yych == 'U') goto yy920;
	if (yych == 'u') goto yy920;
	goto yy55;
yy776:
	yych = *++cur;
	if (yych == 'A') goto yy921;
	if (yych == 'a') goto yy921;
	goto yy55;
yy777:
	yych = *++cur;
	if (yych == 'C') goto yy922;
	if (yych == 'c') goto yy922;
	goto yy55;
yy778:
	yych = *++cur;
	if (yych == 'T') goto yy923;
	if (yych == 't') goto yy923;
	goto yy55;
yy779:
	yych = *++cur;
	if (yych == 'E') goto yy925;
	if (yych == 'e') goto yy925;
	goto yy55;
yy780:
	yych = *++cur;
	if (yych == 'I') goto yy926;
	if (yych == 'i') goto yy926;
	goto yy55;
yy781:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'P') {
		if (yych <= '9') {
			if (yych >= '0') goto yy54;
		} else {
			if (yych <= '@') goto yy782;
			if (yych <= 'O') goto yy54;
			goto yy927;
		}
	} else {
		if (yych <= 'o') {
			if (yych <= 'Z') goto yy54;
			if (yych >= 'a') goto yy54;
		} else {
			if (yych <= 'p') goto yy927;
			if (yych <= 'z') goto yy54;
		}
	}
yy782:
#line 295 "tokenizer.re"
	{ KW(DOUBLE) }
#line 5640 "tokenizer.cpp"
yy783:
	yych = *++cur;
	if (yych == 'E') goto yy928;
	if (yych == 'e') goto yy928;
	goto yy55;
yy784:
	yych = *++cur;
	if (yych == 'T') goto yy930;
	if (yych == 't') goto yy930;
	goto yy55;
yy785:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 299 "tokenizer.re"
	{ KW(ELSEIF) }
#line 5659 "tokenizer.cpp"
yy787:
	yych = *++cur;
	if (yych == 'E') goto yy931;
	if (yych == 'e') goto yy931;
	goto yy55;
yy788:
	yych = *++cur;
	if (yych == 'S') goto yy932;
	if (yych == 's') goto yy932;
	goto yy79;
yy789:
	yych = *++cur;
	if (yych == 'O') goto yy933;
	if (yych == 'o') goto yy933;
	goto yy79;
yy790:
	yych = *++cur;
	if (yych == 'I') goto yy934;
	if (yych == 'i') goto yy934;
	goto yy79;
yy791:
	++cur;
#line 357 "tokenizer.re"
	{
                if (enddo_newline_process) {
                    KW(CONTINUE)
                } else {
                    KW(END_DO)
                }
            }
#line 5690 "tokenizer.cpp"
yy793:
	yych = *++cur;
	if (yych == 'U') goto yy935;
	if (yych == 'u') goto yy935;
	goto yy79;
yy794:
	yych = *++cur;
	if (yych == 'R') goto yy936;
	if (yych == 'r') goto yy936;
	goto yy79;
yy795:
	yych = *++cur;
	if (yych == 'N') goto yy937;
	if (yych == 'n') goto yy937;
	goto yy79;
yy796:
	++cur;
#line 348 "tokenizer.re"
	{ KW(END_IF) }
#line 5710 "tokenizer.cpp"
yy798:
	yych = *++cur;
	if (yych == 'T') goto yy938;
	if (yych == 't') goto yy938;
	goto yy79;
yy799:
	yych = *++cur;
	if (yych == 'D') goto yy939;
	if (yych == 'd') goto yy939;
	goto yy79;
yy800:
	yych = *++cur;
	if (yych == 'O') goto yy940;
	if (yych == 'o') goto yy940;
	goto yy79;
yy801:
	yych = *++cur;
	if (yych == 'L') goto yy941;
	if (yych == 'l') goto yy941;
	goto yy79;
yy802:
	yych = *++cur;
	if (yych == 'B') goto yy942;
	if (yych == 'b') goto yy942;
	goto yy79;
yy803:
	yych = *++cur;
	if (yych == 'A') goto yy943;
	if (yych == 'a') goto yy943;
	goto yy79;
yy804:
	yych = *++cur;
	if (yych == 'P') goto yy944;
	if (yych == 'p') goto yy944;
	goto yy79;
yy805:
	yych = *++cur;
	if (yych == 'E') goto yy945;
	if (yych == 'e') goto yy945;
	goto yy79;
yy806:
	yych = *++cur;
	if (yych == 'L') goto yy946;
	if (yych == 'l') goto yy946;
	goto yy79;
yy807:
	yych = *++cur;
	if (yych == 'O') goto yy947;
	if (yych == 'o') goto yy947;
	goto yy55;
yy808:
	yych = *++cur;
	if (yych == 'C') goto yy948;
	if (yych == 'c') goto yy948;
	goto yy55;
yy809:
	yych = *++cur;
	if (yych == 'T') goto yy949;
	if (yych == 't') goto yy949;
	goto yy55;
yy810:
	yych = *++cur;
	if (yych == 'M') goto yy950;
	if (yych == 'm') goto yy950;
	goto yy55;
yy811:
	yych = *++cur;
	if (yych == 'E') goto yy952;
	if (yych == 'e') goto yy952;
	goto yy55;
yy812:
	yych = *++cur;
	if (yych == 'A') goto yy954;
	if (yych == 'a') goto yy954;
	goto yy55;
yy813:
	yych = *++cur;
	if (yych == 'C') goto yy955;
	if (yych == 'c') goto yy955;
	goto yy55;
yy814:
	yych = *++cur;
	if (yych == 'E') goto yy956;
	if (yych == 'e') goto yy956;
	goto yy55;
yy815:
	yych = *++cur;
	if (yych == 'U') goto yy957;
	if (yych == 'u') goto yy957;
	goto yy55;
yy816:
	yych = *++cur;
	if (yych <= 'G') {
		if (yych == 'C') goto yy958;
		if (yych <= 'F') goto yy55;
		goto yy959;
	} else {
		if (yych <= 'c') {
			if (yych <= 'b') goto yy55;
			goto yy958;
		} else {
			if (yych == 'g') goto yy959;
			goto yy55;
		}
	}
yy817:
	yych = *++cur;
	if (yych == 'E') goto yy960;
	if (yych == 'e') goto yy960;
	goto yy55;
yy818:
	yych = *++cur;
	if (yych <= 'R') {
		if (yych == 'M') goto yy961;
		if (yych <= 'Q') goto yy55;
		goto yy962;
	} else {
		if (yych <= 'm') {
			if (yych <= 'l') goto yy55;
			goto yy961;
		} else {
			if (yych == 'r') goto yy962;
			goto yy55;
		}
	}
yy819:
	yych = *++cur;
	if (yych == 'M') goto yy963;
	if (yych == 'm') goto yy963;
	goto yy55;
yy820:
	yych = *++cur;
	if (yych == 'E') goto yy965;
	if (yych == 'e') goto yy965;
	goto yy55;
yy821:
	yych = *++cur;
	if (yych == 'R') goto yy967;
	if (yych == 'r') goto yy967;
	goto yy55;
yy822:
	yych = *++cur;
	if (yych == 'A') goto yy968;
	if (yych == 'a') goto yy968;
	goto yy55;
yy823:
	yych = *++cur;
	if (yych == 'L') goto yy969;
	if (yych == 'l') goto yy969;
	goto yy55;
yy824:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 382 "tokenizer.re"
	{ KW(ERRMSG) }
#line 5869 "tokenizer.cpp"
yy826:
	yych = *++cur;
	if (yych == 'S') goto yy970;
	if (yych == 's') goto yy970;
	goto yy55;
yy827:
	yych = *++cur;
	if (yych == 'A') goto yy972;
	if (yych == 'a') goto yy972;
	goto yy55;
yy828:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 391 "tokenizer.re"
	{ KW(FORALL) }
#line 5888 "tokenizer.cpp"
yy830:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'T') {
		if (yych <= '9') {
			if (yych >= '0') goto yy54;
		} else {
			if (yych <= '@') goto yy831;
			if (yych <= 'S') goto yy54;
			goto yy973;
		}
	} else {
		if (yych <= 's') {
			if (yych <= 'Z') goto yy54;
			if (yych >= 'a') goto yy54;
		} else {
			if (yych <= 't') goto yy973;
			if (yych <= 'z') goto yy54;
		}
	}
yy831:
#line 392 "tokenizer.re"
	{
                if (last_token == yytokentype::TK_LABEL) {
                    unsigned char *start;
                    lex_format(cur, loc, start);
                    yylval.string.p = (char*) start;
                    yylval.string.n = cur-start-1;
                    RET(TK_FORMAT)
                } else {
                    token(yylval.string);
                    RET(TK_NAME)
                }
            }
#line 5925 "tokenizer.cpp"
yy832:
	yych = *++cur;
	if (yych == 'A') goto yy974;
	if (yych == 'a') goto yy974;
	goto yy55;
yy833:
	yych = *++cur;
	if (yych == 'O') goto yy975;
	if (yych == 'o') goto yy975;
	goto yy55;
yy834:
	yych = *++cur;
	if (yych == 'C') goto yy976;
	if (yych == 'c') goto yy976;
	goto yy55;
yy835:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 412 "tokenizer.re"
	{ KW(IMAGES) }
#line 5949 "tokenizer.cpp"
yy837:
	yych = *++cur;
	if (yych == 'I') goto yy978;
	if (yych == 'i') goto yy978;
	goto yy55;
yy838:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 414 "tokenizer.re"
	{ KW(IMPORT) }
#line 5963 "tokenizer.cpp"
yy840:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 415 "tokenizer.re"
	{ KW(IMPURE) }
#line 5972 "tokenizer.cpp"
yy842:
	++cur;
#line 419 "tokenizer.re"
	{ KW(IN_OUT) }
#line 5977 "tokenizer.cpp"
yy844:
	yych = *++cur;
	if (yych == 'E') goto yy979;
	if (yych == 'e') goto yy979;
	goto yy55;
yy845:
	yych = *++cur;
	if (yych == 'E') goto yy981;
	if (yych == 'e') goto yy981;
	goto yy55;
yy846:
	yych = *++cur;
	if (yych == 'R') goto yy983;
	if (yych == 'r') goto yy983;
	goto yy55;
yy847:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 422 "tokenizer.re"
	{ KW(INTENT) }
#line 6001 "tokenizer.cpp"
yy849:
	yych = *++cur;
	if (yych == 'A') goto yy985;
	if (yych == 'a') goto yy985;
	goto yy55;
yy850:
	yych = *++cur;
	if (yych == 'S') goto yy986;
	if (yych == 's') goto yy986;
	goto yy55;
yy851:
	yyaccept = 2;
	yych = *(mar = ++cur);
	if (yych <= '\'') {
		if (yych == '"') goto yy82;
		if (yych <= '&') goto yy55;
		goto yy97;
	} else {
		if (yych <= 'I') {
			if (yych <= 'H') goto yy55;
			goto yy987;
		} else {
			if (yych == 'i') goto yy987;
			goto yy55;
		}
	}
yy852:
	yych = *++cur;
	if (yych == 'L') goto yy988;
	if (yych == 'l') goto yy988;
	goto yy55;
yy853:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 431 "tokenizer.re"
	{ KW(MEMORY) }
#line 6041 "tokenizer.cpp"
yy855:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 432 "tokenizer.re"
	{ KW(MODULE) }
#line 6050 "tokenizer.cpp"
yy857:
	yych = *++cur;
	if (yych == 'S') goto yy990;
	if (yych == 's') goto yy990;
	goto yy55;
yy858:
	yych = *++cur;
	if (yych == 'D') goto yy991;
	if (yych == 'd') goto yy991;
	goto yy55;
yy859:
	yych = *++cur;
	if (yych == 'T') goto yy992;
	if (yych == 't') goto yy992;
	goto yy55;
yy860:
	yych = *++cur;
	if (yych == 'E') goto yy993;
	if (yych == 'e') goto yy993;
	goto yy55;
yy861:
	yych = *++cur;
	if (yych == 'C') goto yy994;
	if (yych == 'c') goto yy994;
	goto yy55;
yy862:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 437 "tokenizer.re"
	{ KW(NOPASS) }
#line 6084 "tokenizer.cpp"
yy864:
	yych = *++cur;
	if (yych == 'Y') goto yy995;
	if (yych == 'y') goto yy995;
	goto yy55;
yy865:
	yych = *++cur;
	if (yych == 'O') goto yy997;
	if (yych == 'o') goto yy997;
	goto yy55;
yy866:
	yych = *++cur;
	if (yych == 'A') goto yy998;
	if (yych == 'a') goto yy998;
	goto yy55;
yy867:
	yych = *++cur;
	if (yych == 'T') goto yy999;
	if (yych == 't') goto yy999;
	goto yy55;
yy868:
	yych = *++cur;
	if (yych == 'R') goto yy1000;
	if (yych == 'r') goto yy1000;
	goto yy55;
yy869:
	yych = *++cur;
	if (yych == 'I') goto yy1002;
	if (yych == 'i') goto yy1002;
	goto yy55;
yy870:
	yych = *++cur;
	if (yych == 'E') goto yy1003;
	if (yych == 'e') goto yy1003;
	goto yy55;
yy871:
	yych = *++cur;
	if (yych == 'U') goto yy1005;
	if (yych == 'u') goto yy1005;
	goto yy55;
yy872:
	yych = *++cur;
	if (yych == 'M') goto yy1006;
	if (yych == 'm') goto yy1006;
	goto yy55;
yy873:
	yych = *++cur;
	if (yych == 'T') goto yy1008;
	if (yych == 't') goto yy1008;
	goto yy55;
yy874:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 458 "tokenizer.re"
	{ KW(PUBLIC) }
#line 6143 "tokenizer.cpp"
yy876:
	yych = *++cur;
	if (yych == 'I') goto yy1009;
	if (yych == 'i') goto yy1009;
	goto yy55;
yy877:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 465 "tokenizer.re"
	{ KW(REDUCE) }
#line 6157 "tokenizer.cpp"
yy879:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 466 "tokenizer.re"
	{ KW(RESULT) }
#line 6166 "tokenizer.cpp"
yy881:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 467 "tokenizer.re"
	{ KW(RETURN) }
#line 6175 "tokenizer.cpp"
yy883:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 468 "tokenizer.re"
	{ KW(REWIND) }
#line 6184 "tokenizer.cpp"
yy885:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= 'T') {
		if (yych <= 'B') {
			if (yych <= '/') goto yy886;
			if (yych <= '9') goto yy54;
			if (yych >= 'A') goto yy54;
		} else {
			if (yych <= 'Q') {
				if (yych <= 'C') goto yy1010;
				goto yy54;
			} else {
				if (yych <= 'R') goto yy1011;
				if (yych <= 'S') goto yy54;
				goto yy1012;
			}
		}
	} else {
		if (yych <= 'q') {
			if (yych <= '`') {
				if (yych <= 'Z') goto yy54;
			} else {
				if (yych == 'c') goto yy1010;
				goto yy54;
			}
		} else {
			if (yych <= 's') {
				if (yych <= 'r') goto yy1011;
				goto yy54;
			} else {
				if (yych <= 't') goto yy1012;
				if (yych <= 'z') goto yy54;
			}
		}
	}
yy886:
#line 470 "tokenizer.re"
	{ KW(SELECT) }
#line 6226 "tokenizer.cpp"
yy887:
	yych = *++cur;
	if (yych == 'C') goto yy1013;
	if (yych == 'c') goto yy1013;
	goto yy55;
yy888:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 475 "tokenizer.re"
	{ KW(SHARED) }
#line 6240 "tokenizer.cpp"
yy890:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 476 "tokenizer.re"
	{ KW(SOURCE) }
#line 6249 "tokenizer.cpp"
yy892:
	yych = *++cur;
	if (yych == 'U') goto yy1014;
	if (yych == 'u') goto yy1014;
	goto yy55;
yy893:
	yych = *++cur;
	if (yych == 'T') goto yy1015;
	if (yych == 't') goto yy1015;
	goto yy55;
yy894:
	yych = *++cur;
	if (yych == 'L') goto yy1016;
	if (yych == 'l') goto yy1016;
	goto yy55;
yy895:
	yych = *++cur;
	if (yych == 'A') goto yy1018;
	if (yych == 'a') goto yy1018;
	goto yy55;
yy896:
	yych = *++cur;
	if (yych == 'M') goto yy1019;
	if (yych == 'm') goto yy1019;
	goto yy55;
yy897:
	yych = *++cur;
	if (yych == 'A') goto yy1020;
	if (yych == 'a') goto yy1020;
	goto yy55;
yy898:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 486 "tokenizer.re"
	{ KW(TARGET) }
#line 6288 "tokenizer.cpp"
yy900:
	yych = *++cur;
	if (yych == 'U') goto yy1021;
	if (yych == 'u') goto yy1021;
	goto yy55;
yy901:
	yych = *++cur;
	if (yych == 'A') goto yy1022;
	if (yych == 'a') goto yy1022;
	goto yy55;
yy902:
	yych = *++cur;
	if (yych == 'L') goto yy1023;
	if (yych == 'l') goto yy1023;
	goto yy55;
yy903:
	yyaccept = 12;
	yych = *(mar = ++cur);
	if (yych == '_') goto yy1024;
yy904:
#line 578 "tokenizer.re"
	{ RET(TK_FALSE) }
#line 6311 "tokenizer.cpp"
yy905:
	yych = *++cur;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy79;
		if (yych <= '9') goto yy1025;
		if (yych <= '@') goto yy79;
		goto yy1027;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy79;
			goto yy1027;
		} else {
			if (yych <= '`') goto yy79;
			if (yych <= 'z') goto yy1027;
			goto yy79;
		}
	}
yy906:
	yych = *++cur;
	if (yych == 'T') goto yy1029;
	if (yych == 't') goto yy1029;
	goto yy55;
yy907:
	yych = *++cur;
	if (yych <= 'E') {
		if (yych == 'A') goto yy1031;
		if (yych <= 'D') goto yy55;
		goto yy1032;
	} else {
		if (yych <= 'a') {
			if (yych <= '`') goto yy55;
			goto yy1031;
		} else {
			if (yych == 'e') goto yy1032;
			goto yy55;
		}
	}
yy908:
	yych = *++cur;
	if (yych == 'E') goto yy1034;
	if (yych == 'e') goto yy1034;
	goto yy55;
yy909:
	yych = *++cur;
	if (yych == 'T') goto yy1035;
	if (yych == 't') goto yy1035;
	goto yy55;
yy910:
	yych = *++cur;
	if (yych == 'O') goto yy1036;
	if (yych == 'o') goto yy1036;
	goto yy55;
yy911:
	yych = *++cur;
	if (yych == 'C') goto yy1037;
	if (yych == 'c') goto yy1037;
	goto yy55;
yy912:
	yych = *++cur;
	if (yych == 'E') goto yy1038;
	if (yych == 'e') goto yy1038;
	goto yy55;
yy913:
	yych = *++cur;
	if (yych == 'E') goto yy1039;
	if (yych == 'e') goto yy1039;
	goto yy55;
yy914:
	yych = *++cur;
	if (yych == 'S') goto yy1040;
	if (yych == 's') goto yy1040;
	goto yy55;
yy915:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 274 "tokenizer.re"
	{ KW(COMPLEX) }
#line 6392 "tokenizer.cpp"
yy917:
	yych = *++cur;
	if (yych == 'E') goto yy1041;
	if (yych == 'e') goto yy1041;
	goto yy55;
yy918:
	yych = *++cur;
	if (yych == 'S') goto yy1042;
	if (yych == 's') goto yy1042;
	goto yy55;
yy919:
	yych = *++cur;
	if (yych == 'O') goto yy1044;
	if (yych == 'o') goto yy1044;
	goto yy55;
yy920:
	yych = *++cur;
	if (yych == 'E') goto yy1045;
	if (yych == 'e') goto yy1045;
	goto yy55;
yy921:
	yych = *++cur;
	if (yych == 'L') goto yy1047;
	if (yych == 'l') goto yy1047;
	goto yy55;
yy922:
	yych = *++cur;
	if (yych == 'A') goto yy1049;
	if (yych == 'a') goto yy1049;
	goto yy55;
yy923:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 283 "tokenizer.re"
	{ KW(DEFAULT) }
#line 6431 "tokenizer.cpp"
yy925:
	yych = *++cur;
	if (yych == 'D') goto yy1050;
	if (yych == 'd') goto yy1050;
	goto yy55;
yy926:
	yych = *++cur;
	if (yych == 'O') goto yy1052;
	if (yych == 'o') goto yy1052;
	goto yy55;
yy927:
	yych = *++cur;
	if (yych == 'R') goto yy1053;
	if (yych == 'r') goto yy1053;
	goto yy55;
yy928:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 294 "tokenizer.re"
	{ KW(DOWHILE) }
#line 6455 "tokenizer.cpp"
yy930:
	yych = *++cur;
	if (yych == 'A') goto yy1054;
	if (yych == 'a') goto yy1054;
	goto yy55;
yy931:
	yych = *++cur;
	if (yych == 'R') goto yy1055;
	if (yych == 'r') goto yy1055;
	goto yy55;
yy932:
	yych = *++cur;
	if (yych == 'O') goto yy1056;
	if (yych == 'o') goto yy1056;
	goto yy79;
yy933:
	yych = *++cur;
	if (yych == 'C') goto yy1057;
	if (yych == 'c') goto yy1057;
	goto yy79;
yy934:
	yych = *++cur;
	if (yych == 'T') goto yy1058;
	if (yych == 't') goto yy1058;
	goto yy79;
yy935:
	yych = *++cur;
	if (yych == 'M') goto yy1059;
	if (yych == 'm') goto yy1059;
	goto yy79;
yy936:
	yych = *++cur;
	if (yych == 'A') goto yy1061;
	if (yych == 'a') goto yy1061;
	goto yy79;
yy937:
	yych = *++cur;
	if (yych == 'C') goto yy1062;
	if (yych == 'c') goto yy1062;
	goto yy79;
yy938:
	yych = *++cur;
	if (yych == 'E') goto yy1063;
	if (yych == 'e') goto yy1063;
	goto yy79;
yy939:
	yych = *++cur;
	if (yych == 'U') goto yy1064;
	if (yych == 'u') goto yy1064;
	goto yy79;
yy940:
	yych = *++cur;
	if (yych <= 'G') {
		if (yych == 'C') goto yy1065;
		if (yych <= 'F') goto yy79;
		goto yy1066;
	} else {
		if (yych <= 'c') {
			if (yych <= 'b') goto yy79;
			goto yy1065;
		} else {
			if (yych == 'g') goto yy1066;
			goto yy79;
		}
	}
yy941:
	yych = *++cur;
	if (yych == 'E') goto yy1067;
	if (yych == 'e') goto yy1067;
	goto yy79;
yy942:
	yych = *++cur;
	if (yych <= 'R') {
		if (yych == 'M') goto yy1068;
		if (yych <= 'Q') goto yy79;
		goto yy1069;
	} else {
		if (yych <= 'm') {
			if (yych <= 'l') goto yy79;
			goto yy1068;
		} else {
			if (yych == 'r') goto yy1069;
			goto yy79;
		}
	}
yy943:
	yych = *++cur;
	if (yych == 'M') goto yy1070;
	if (yych == 'm') goto yy1070;
	goto yy79;
yy944:
	yych = *++cur;
	if (yych == 'E') goto yy1072;
	if (yych == 'e') goto yy1072;
	goto yy79;
yy945:
	yych = *++cur;
	if (yych == 'R') goto yy1074;
	if (yych == 'r') goto yy1074;
	goto yy79;
yy946:
	yych = *++cur;
	if (yych == 'E') goto yy1075;
	if (yych == 'e') goto yy1075;
	goto yy79;
yy947:
	yych = *++cur;
	if (yych == 'C') goto yy1077;
	if (yych == 'c') goto yy1077;
	goto yy55;
yy948:
	yych = *++cur;
	if (yych == 'K') goto yy1078;
	if (yych == 'k') goto yy1078;
	goto yy55;
yy949:
	yych = *++cur;
	if (yych == 'I') goto yy1080;
	if (yych == 'i') goto yy1080;
	goto yy55;
yy950:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 331 "tokenizer.re"
	{ KW(ENDENUM) }
#line 6584 "tokenizer.cpp"
yy952:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 376 "tokenizer.re"
	{ KW(ENDFILE) }
#line 6593 "tokenizer.cpp"
yy954:
	yych = *++cur;
	if (yych == 'L') goto yy1081;
	if (yych == 'l') goto yy1081;
	goto yy55;
yy955:
	yych = *++cur;
	if (yych == 'T') goto yy1082;
	if (yych == 't') goto yy1082;
	goto yy55;
yy956:
	yych = *++cur;
	if (yych == 'R') goto yy1083;
	if (yych == 'r') goto yy1083;
	goto yy55;
yy957:
	yych = *++cur;
	if (yych == 'L') goto yy1084;
	if (yych == 'l') goto yy1084;
	goto yy55;
yy958:
	yych = *++cur;
	if (yych == 'E') goto yy1085;
	if (yych == 'e') goto yy1085;
	goto yy55;
yy959:
	yych = *++cur;
	if (yych == 'R') goto yy1086;
	if (yych == 'r') goto yy1086;
	goto yy55;
yy960:
	yych = *++cur;
	if (yych == 'C') goto yy1087;
	if (yych == 'c') goto yy1087;
	goto yy55;
yy961:
	yych = *++cur;
	if (yych == 'O') goto yy1088;
	if (yych == 'o') goto yy1088;
	goto yy55;
yy962:
	yych = *++cur;
	if (yych == 'O') goto yy1089;
	if (yych == 'o') goto yy1089;
	goto yy55;
yy963:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 343 "tokenizer.re"
	{ KW(ENDTEAM) }
#line 6647 "tokenizer.cpp"
yy965:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 355 "tokenizer.re"
	{ KW(ENDTYPE) }
#line 6656 "tokenizer.cpp"
yy967:
	yych = *++cur;
	if (yych == 'E') goto yy1090;
	if (yych == 'e') goto yy1090;
	goto yy55;
yy968:
	yych = *++cur;
	if (yych == 'T') goto yy1092;
	if (yych == 't') goto yy1092;
	goto yy55;
yy969:
	yych = *++cur;
	if (yych == 'E') goto yy1093;
	if (yych == 'e') goto yy1093;
	goto yy55;
yy970:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 386 "tokenizer.re"
	{ KW(EXTENDS) }
#line 6680 "tokenizer.cpp"
yy972:
	yych = *++cur;
	if (yych == 'L') goto yy1094;
	if (yych == 'l') goto yy1094;
	goto yy55;
yy973:
	yych = *++cur;
	if (yych == 'E') goto yy1096;
	if (yych == 'e') goto yy1096;
	goto yy55;
yy974:
	yych = *++cur;
	if (yych == 'M') goto yy1097;
	if (yych == 'm') goto yy1097;
	goto yy55;
yy975:
	yych = *++cur;
	if (yych == 'N') goto yy1099;
	if (yych == 'n') goto yy1099;
	goto yy55;
yy976:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 408 "tokenizer.re"
	{ KW(GENERIC) }
#line 6709 "tokenizer.cpp"
yy978:
	yych = *++cur;
	if (yych == 'T') goto yy1101;
	if (yych == 't') goto yy1101;
	goto yy55;
yy979:
	yyaccept = 13;
	yych = *(mar = ++cur);
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych <= '\f') {
		if (yych <= '\t') {
			if (yych >= '\t') goto yy1103;
		} else {
			if (yych == '\v') goto yy1103;
		}
	} else {
		if (yych <= ' ') {
			if (yych <= '\r') goto yy1103;
			if (yych >= ' ') goto yy1103;
		} else {
			if (yych == '_') goto yy144;
		}
	}
yy980:
#line 417 "tokenizer.re"
	{ KW(INCLUDE) }
#line 6738 "tokenizer.cpp"
yy981:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 420 "tokenizer.re"
	{ KW(INQUIRE) }
#line 6747 "tokenizer.cpp"
yy983:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 421 "tokenizer.re"
	{ KW(INTEGER) }
#line 6756 "tokenizer.cpp"
yy985:
	yych = *++cur;
	if (yych == 'C') goto yy1105;
	if (yych == 'c') goto yy1105;
	goto yy55;
yy986:
	yych = *++cur;
	if (yych == 'I') goto yy1106;
	if (yych == 'i') goto yy1106;
	goto yy55;
yy987:
	yych = *++cur;
	if (yych == 'N') goto yy1107;
	if (yych == 'n') goto yy1107;
	goto yy55;
yy988:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 430 "tokenizer.re"
	{ KW(LOGICAL) }
#line 6780 "tokenizer.cpp"
yy990:
	yych = *++cur;
	if (yych == 'T') goto yy1108;
	if (yych == 't') goto yy1108;
	goto yy55;
yy991:
	yych = *++cur;
	if (yych == 'E') goto yy1110;
	if (yych == 'e') goto yy1110;
	goto yy55;
yy992:
	yych = *++cur;
	if (yych == 'R') goto yy1111;
	if (yych == 'r') goto yy1111;
	goto yy55;
yy993:
	yych = *++cur;
	if (yych == 'R') goto yy1112;
	if (yych == 'r') goto yy1112;
	goto yy55;
yy994:
	yych = *++cur;
	if (yych == 'U') goto yy1113;
	if (yych == 'u') goto yy1113;
	goto yy55;
yy995:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 442 "tokenizer.re"
	{ KW(NULLIFY) }
#line 6814 "tokenizer.cpp"
yy997:
	yych = *++cur;
	if (yych == 'R') goto yy1114;
	if (yych == 'r') goto yy1114;
	goto yy55;
yy998:
	yych = *++cur;
	if (yych == 'L') goto yy1116;
	if (yych == 'l') goto yy1116;
	goto yy55;
yy999:
	yych = *++cur;
	if (yych == 'E') goto yy1118;
	if (yych == 'e') goto yy1118;
	goto yy55;
yy1000:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 450 "tokenizer.re"
	{ KW(POINTER) }
#line 6838 "tokenizer.cpp"
yy1002:
	yych = *++cur;
	if (yych == 'O') goto yy1119;
	if (yych == 'o') goto yy1119;
	goto yy55;
yy1003:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 454 "tokenizer.re"
	{ KW(PRIVATE) }
#line 6852 "tokenizer.cpp"
yy1005:
	yych = *++cur;
	if (yych == 'R') goto yy1120;
	if (yych == 'r') goto yy1120;
	goto yy55;
yy1006:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 456 "tokenizer.re"
	{ KW(PROGRAM) }
#line 6866 "tokenizer.cpp"
yy1008:
	yych = *++cur;
	if (yych == 'E') goto yy1121;
	if (yych == 'e') goto yy1121;
	goto yy55;
yy1009:
	yych = *++cur;
	if (yych == 'V') goto yy1122;
	if (yych == 'v') goto yy1122;
	goto yy55;
yy1010:
	yych = *++cur;
	if (yych == 'A') goto yy1123;
	if (yych == 'a') goto yy1123;
	goto yy55;
yy1011:
	yych = *++cur;
	if (yych == 'A') goto yy1124;
	if (yych == 'a') goto yy1124;
	goto yy55;
yy1012:
	yych = *++cur;
	if (yych == 'Y') goto yy1125;
	if (yych == 'y') goto yy1125;
	goto yy55;
yy1013:
	yych = *++cur;
	if (yych == 'E') goto yy1126;
	if (yych == 'e') goto yy1126;
	goto yy55;
yy1014:
	yych = *++cur;
	if (yych == 'L') goto yy1128;
	if (yych == 'l') goto yy1128;
	goto yy55;
yy1015:
	yych = *++cur;
	if (yych == 'I') goto yy1129;
	if (yych == 'i') goto yy1129;
	goto yy55;
yy1016:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 482 "tokenizer.re"
	{ KW(SYNC_ALL) }
#line 6915 "tokenizer.cpp"
yy1018:
	yych = *++cur;
	if (yych == 'G') goto yy1130;
	if (yych == 'g') goto yy1130;
	goto yy55;
yy1019:
	yych = *++cur;
	if (yych == 'O') goto yy1131;
	if (yych == 'o') goto yy1131;
	goto yy55;
yy1020:
	yych = *++cur;
	if (yych == 'M') goto yy1132;
	if (yych == 'm') goto yy1132;
	goto yy55;
yy1021:
	yych = *++cur;
	if (yych == 'M') goto yy1134;
	if (yych == 'm') goto yy1134;
	goto yy55;
yy1022:
	yych = *++cur;
	if (yych == 'T') goto yy1135;
	if (yych == 't') goto yy1135;
	goto yy55;
yy1023:
	yych = *++cur;
	if (yych == 'E') goto yy1136;
	if (yych == 'e') goto yy1136;
	goto yy55;
yy1024:
	yych = *++cur;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy79;
		if (yych <= '9') goto yy1138;
		if (yych <= '@') goto yy79;
		goto yy1140;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy79;
			goto yy1140;
		} else {
			if (yych <= '`') goto yy79;
			if (yych <= 'z') goto yy1140;
			goto yy79;
		}
	}
yy1025:
	yych = *++cur;
	if (yych <= '/') goto yy757;
	if (yych <= '9') goto yy1025;
	goto yy757;
yy1027:
	yych = *++cur;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy757;
		if (yych <= '9') goto yy1027;
		if (yych <= '@') goto yy757;
		goto yy1027;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy757;
			goto yy1027;
		} else {
			if (yych <= '`') goto yy757;
			if (yych <= 'z') goto yy1027;
			goto yy757;
		}
	}
yy1029:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 254 "tokenizer.re"
	{ KW(ABSTRACT) }
#line 6993 "tokenizer.cpp"
yy1031:
	yych = *++cur;
	if (yych == 'B') goto yy1142;
	if (yych == 'b') goto yy1142;
	goto yy55;
yy1032:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 257 "tokenizer.re"
	{ KW(ALLOCATE) }
#line 7007 "tokenizer.cpp"
yy1034:
	yych = *++cur;
	if (yych == 'N') goto yy1143;
	if (yych == 'n') goto yy1143;
	goto yy55;
yy1035:
	yych = *++cur;
	if (yych == 'E') goto yy1144;
	if (yych == 'e') goto yy1144;
	goto yy55;
yy1036:
	yych = *++cur;
	if (yych == 'N') goto yy1146;
	if (yych == 'n') goto yy1146;
	goto yy55;
yy1037:
	yych = *++cur;
	if (yych == 'E') goto yy1147;
	if (yych == 'e') goto yy1147;
	goto yy55;
yy1038:
	yych = *++cur;
	if (yych == 'A') goto yy1149;
	if (yych == 'a') goto yy1149;
	goto yy55;
yy1039:
	yych = *++cur;
	if (yych == 'R') goto yy1150;
	if (yych == 'r') goto yy1150;
	goto yy55;
yy1040:
	yych = *++cur;
	if (yych == 'I') goto yy1152;
	if (yych == 'i') goto yy1152;
	goto yy55;
yy1041:
	yych = *++cur;
	if (yych == 'N') goto yy1153;
	if (yych == 'n') goto yy1153;
	goto yy55;
yy1042:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 276 "tokenizer.re"
	{ KW(CONTAINS) }
#line 7056 "tokenizer.cpp"
yy1044:
	yych = *++cur;
	if (yych == 'U') goto yy1154;
	if (yych == 'u') goto yy1154;
	goto yy55;
yy1045:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 278 "tokenizer.re"
	{ KW(CONTINUE) }
#line 7070 "tokenizer.cpp"
yy1047:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 279 "tokenizer.re"
	{ KW(CRITICAL) }
#line 7079 "tokenizer.cpp"
yy1049:
	yych = *++cur;
	if (yych == 'T') goto yy1155;
	if (yych == 't') goto yy1155;
	goto yy55;
yy1050:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 284 "tokenizer.re"
	{ KW(DEFERRED) }
#line 7093 "tokenizer.cpp"
yy1052:
	yych = *++cur;
	if (yych == 'N') goto yy1156;
	if (yych == 'n') goto yy1156;
	goto yy55;
yy1053:
	yych = *++cur;
	if (yych == 'E') goto yy1158;
	if (yych == 'e') goto yy1158;
	goto yy55;
yy1054:
	yych = *++cur;
	if (yych == 'L') goto yy1159;
	if (yych == 'l') goto yy1159;
	goto yy55;
yy1055:
	yych = *++cur;
	if (yych == 'E') goto yy1161;
	if (yych == 'e') goto yy1161;
	goto yy55;
yy1056:
	yych = *++cur;
	if (yych == 'C') goto yy1163;
	if (yych == 'c') goto yy1163;
	goto yy79;
yy1057:
	yych = *++cur;
	if (yych == 'K') goto yy1164;
	if (yych == 'k') goto yy1164;
	goto yy79;
yy1058:
	yych = *++cur;
	if (yych == 'I') goto yy1166;
	if (yych == 'i') goto yy1166;
	goto yy79;
yy1059:
	++cur;
#line 330 "tokenizer.re"
	{ KW(END_ENUM) }
#line 7133 "tokenizer.cpp"
yy1061:
	yych = *++cur;
	if (yych == 'L') goto yy1167;
	if (yych == 'l') goto yy1167;
	goto yy79;
yy1062:
	yych = *++cur;
	if (yych == 'T') goto yy1168;
	if (yych == 't') goto yy1168;
	goto yy79;
yy1063:
	yych = *++cur;
	if (yych == 'R') goto yy1169;
	if (yych == 'r') goto yy1169;
	goto yy79;
yy1064:
	yych = *++cur;
	if (yych == 'L') goto yy1170;
	if (yych == 'l') goto yy1170;
	goto yy79;
yy1065:
	yych = *++cur;
	if (yych == 'E') goto yy1171;
	if (yych == 'e') goto yy1171;
	goto yy79;
yy1066:
	yych = *++cur;
	if (yych == 'R') goto yy1172;
	if (yych == 'r') goto yy1172;
	goto yy79;
yy1067:
	yych = *++cur;
	if (yych == 'C') goto yy1173;
	if (yych == 'c') goto yy1173;
	goto yy79;
yy1068:
	yych = *++cur;
	if (yych == 'O') goto yy1174;
	if (yych == 'o') goto yy1174;
	goto yy79;
yy1069:
	yych = *++cur;
	if (yych == 'O') goto yy1175;
	if (yych == 'o') goto yy1175;
	goto yy79;
yy1070:
	++cur;
#line 342 "tokenizer.re"
	{ KW(END_TEAM) }
#line 7183 "tokenizer.cpp"
yy1072:
	++cur;
#line 354 "tokenizer.re"
	{ KW(END_TYPE) }
#line 7188 "tokenizer.cpp"
yy1074:
	yych = *++cur;
	if (yych == 'E') goto yy1176;
	if (yych == 'e') goto yy1176;
	goto yy79;
yy1075:
	++cur;
#line 375 "tokenizer.re"
	{ KW(END_FILE) }
#line 7198 "tokenizer.cpp"
yy1077:
	yych = *++cur;
	if (yych == 'I') goto yy1178;
	if (yych == 'i') goto yy1178;
	goto yy55;
yy1078:
	yyaccept = 14;
	yych = *(mar = ++cur);
	if (yybm[0+yych] & 8) {
		goto yy144;
	}
	if (yych <= '/') {
		if (yych <= '\v') {
			if (yych == '\t') goto yy1179;
			if (yych >= '\v') goto yy1179;
		} else {
			if (yych <= '\r') {
				if (yych >= '\r') goto yy1179;
			} else {
				if (yych == ' ') goto yy1179;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '@') {
				if (yych <= '9') goto yy54;
			} else {
				if (yych == 'D') goto yy1181;
				goto yy54;
			}
		} else {
			if (yych <= 'c') {
				if (yych >= 'a') goto yy54;
			} else {
				if (yych <= 'd') goto yy1181;
				if (yych <= 'z') goto yy54;
			}
		}
	}
yy1079:
#line 314 "tokenizer.re"
	{ KW(ENDBLOCK) }
#line 7241 "tokenizer.cpp"
yy1080:
	yych = *++cur;
	if (yych == 'C') goto yy1182;
	if (yych == 'c') goto yy1182;
	goto yy55;
yy1081:
	yych = *++cur;
	if (yych == 'L') goto yy1183;
	if (yych == 'l') goto yy1183;
	goto yy55;
yy1082:
	yych = *++cur;
	if (yych == 'I') goto yy1185;
	if (yych == 'i') goto yy1185;
	goto yy55;
yy1083:
	yych = *++cur;
	if (yych == 'F') goto yy1186;
	if (yych == 'f') goto yy1186;
	goto yy55;
yy1084:
	yych = *++cur;
	if (yych == 'E') goto yy1187;
	if (yych == 'e') goto yy1187;
	goto yy55;
yy1085:
	yych = *++cur;
	if (yych == 'D') goto yy1189;
	if (yych == 'd') goto yy1189;
	goto yy55;
yy1086:
	yych = *++cur;
	if (yych == 'A') goto yy1190;
	if (yych == 'a') goto yy1190;
	goto yy55;
yy1087:
	yych = *++cur;
	if (yych == 'T') goto yy1191;
	if (yych == 't') goto yy1191;
	goto yy55;
yy1088:
	yych = *++cur;
	if (yych == 'D') goto yy1193;
	if (yych == 'd') goto yy1193;
	goto yy55;
yy1089:
	yych = *++cur;
	if (yych == 'U') goto yy1194;
	if (yych == 'u') goto yy1194;
	goto yy55;
yy1090:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 373 "tokenizer.re"
	{ KW(ENDWHERE) }
#line 7300 "tokenizer.cpp"
yy1092:
	yych = *++cur;
	if (yych == 'O') goto yy1195;
	if (yych == 'o') goto yy1195;
	goto yy55;
yy1093:
	yych = *++cur;
	if (yych == 'N') goto yy1196;
	if (yych == 'n') goto yy1196;
	goto yy55;
yy1094:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 387 "tokenizer.re"
	{ KW(EXTERNAL) }
#line 7319 "tokenizer.cpp"
yy1096:
	yych = *++cur;
	if (yych == 'D') goto yy1197;
	if (yych == 'd') goto yy1197;
	goto yy55;
yy1097:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 406 "tokenizer.re"
	{ KW(FORM_TEAM) }
#line 7333 "tokenizer.cpp"
yy1099:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 407 "tokenizer.re"
	{ KW(FUNCTION) }
#line 7342 "tokenizer.cpp"
yy1101:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 413 "tokenizer.re"
	{ KW(IMPLICIT) }
#line 7351 "tokenizer.cpp"
yy1103:
	yych = *++cur;
	if (yych <= '"') {
		if (yych <= '\f') {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy79;
				goto yy1103;
			} else {
				if (yych == '\v') goto yy1103;
				goto yy79;
			}
		} else {
			if (yych <= 0x1F) {
				if (yych <= '\r') goto yy1103;
				goto yy79;
			} else {
				if (yych <= ' ') goto yy1103;
				if (yych <= '!') goto yy79;
				goto yy1199;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '\'') {
				if (yych <= '&') goto yy79;
				goto yy1201;
			} else {
				if (yych <= '/') goto yy79;
				if (yych <= '9') goto yy1203;
				goto yy79;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy1205;
				if (yych <= '^') goto yy79;
				goto yy1205;
			} else {
				if (yych <= '`') goto yy79;
				if (yych <= 'z') goto yy1205;
				goto yy79;
			}
		}
	}
yy1105:
	yych = *++cur;
	if (yych == 'E') goto yy1207;
	if (yych == 'e') goto yy1207;
	goto yy55;
yy1106:
	yych = *++cur;
	if (yych == 'C') goto yy1209;
	if (yych == 'c') goto yy1209;
	goto yy55;
yy1107:
	yych = *++cur;
	if (yych == 'I') goto yy1211;
	if (yych == 'i') goto yy1211;
	goto yy55;
yy1108:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 435 "tokenizer.re"
	{ KW(NAMELIST) }
#line 7418 "tokenizer.cpp"
yy1110:
	yych = *++cur;
	if (yych == 'X') goto yy1212;
	if (yych == 'x') goto yy1212;
	goto yy55;
yy1111:
	yych = *++cur;
	if (yych == 'I') goto yy1214;
	if (yych == 'i') goto yy1214;
	goto yy55;
yy1112:
	yych = *++cur;
	if (yych == 'R') goto yy1215;
	if (yych == 'r') goto yy1215;
	goto yy55;
yy1113:
	yych = *++cur;
	if (yych == 'R') goto yy1216;
	if (yych == 'r') goto yy1216;
	goto yy55;
yy1114:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 445 "tokenizer.re"
	{ KW(OPERATOR) }
#line 7447 "tokenizer.cpp"
yy1116:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 446 "tokenizer.re"
	{ KW(OPTIONAL) }
#line 7456 "tokenizer.cpp"
yy1118:
	yych = *++cur;
	if (yych == 'R') goto yy1217;
	if (yych == 'r') goto yy1217;
	goto yy55;
yy1119:
	yych = *++cur;
	if (yych == 'N') goto yy1219;
	if (yych == 'n') goto yy1219;
	goto yy55;
yy1120:
	yych = *++cur;
	if (yych == 'E') goto yy1221;
	if (yych == 'e') goto yy1221;
	goto yy55;
yy1121:
	yych = *++cur;
	if (yych == 'D') goto yy1223;
	if (yych == 'd') goto yy1223;
	goto yy55;
yy1122:
	yych = *++cur;
	if (yych == 'E') goto yy1225;
	if (yych == 'e') goto yy1225;
	goto yy55;
yy1123:
	yych = *++cur;
	if (yych == 'S') goto yy1227;
	if (yych == 's') goto yy1227;
	goto yy55;
yy1124:
	yych = *++cur;
	if (yych == 'N') goto yy1228;
	if (yych == 'n') goto yy1228;
	goto yy55;
yy1125:
	yych = *++cur;
	if (yych == 'P') goto yy1229;
	if (yych == 'p') goto yy1229;
	goto yy55;
yy1126:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 474 "tokenizer.re"
	{ KW(SEQUENCE) }
#line 7505 "tokenizer.cpp"
yy1128:
	yych = *++cur;
	if (yych == 'E') goto yy1230;
	if (yych == 'e') goto yy1230;
	goto yy55;
yy1129:
	yych = *++cur;
	if (yych == 'N') goto yy1232;
	if (yych == 'n') goto yy1232;
	goto yy55;
yy1130:
	yych = *++cur;
	if (yych == 'E') goto yy1233;
	if (yych == 'e') goto yy1233;
	goto yy55;
yy1131:
	yych = *++cur;
	if (yych == 'R') goto yy1234;
	if (yych == 'r') goto yy1234;
	goto yy55;
yy1132:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 485 "tokenizer.re"
	{ KW(SYNC_TEAM) }
#line 7534 "tokenizer.cpp"
yy1134:
	yych = *++cur;
	if (yych == 'B') goto yy1235;
	if (yych == 'b') goto yy1235;
	goto yy55;
yy1135:
	yych = *++cur;
	if (yych == 'T') goto yy1236;
	if (yych == 't') goto yy1236;
	goto yy55;
yy1136:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 495 "tokenizer.re"
	{ KW(VOLATILE) }
#line 7553 "tokenizer.cpp"
yy1138:
	yych = *++cur;
	if (yych <= '/') goto yy904;
	if (yych <= '9') goto yy1138;
	goto yy904;
yy1140:
	yych = *++cur;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy904;
		if (yych <= '9') goto yy1140;
		if (yych <= '@') goto yy904;
		goto yy1140;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy904;
			goto yy1140;
		} else {
			if (yych <= '`') goto yy904;
			if (yych <= 'z') goto yy1140;
			goto yy904;
		}
	}
yy1142:
	yych = *++cur;
	if (yych == 'L') goto yy1237;
	if (yych == 'l') goto yy1237;
	goto yy55;
yy1143:
	yych = *++cur;
	if (yych == 'T') goto yy1238;
	if (yych == 't') goto yy1238;
	goto yy55;
yy1144:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 260 "tokenizer.re"
	{ KW(ASSOCIATE) }
#line 7594 "tokenizer.cpp"
yy1146:
	yych = *++cur;
	if (yych == 'O') goto yy1240;
	if (yych == 'o') goto yy1240;
	goto yy55;
yy1147:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 262 "tokenizer.re"
	{ KW(BACKSPACE) }
#line 7608 "tokenizer.cpp"
yy1149:
	yych = *++cur;
	if (yych == 'M') goto yy1241;
	if (yych == 'm') goto yy1241;
	goto yy55;
yy1150:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 269 "tokenizer.re"
	{ KW(CHARACTER) }
#line 7622 "tokenizer.cpp"
yy1152:
	yych = *++cur;
	if (yych == 'O') goto yy1243;
	if (yych == 'o') goto yy1243;
	goto yy55;
yy1153:
	yych = *++cur;
	if (yych == 'T') goto yy1244;
	if (yych == 't') goto yy1244;
	goto yy55;
yy1154:
	yych = *++cur;
	if (yych == 'S') goto yy1246;
	if (yych == 's') goto yy1246;
	goto yy55;
yy1155:
	yych = *++cur;
	if (yych == 'E') goto yy1248;
	if (yych == 'e') goto yy1248;
	goto yy55;
yy1156:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 285 "tokenizer.re"
	{ KW(DIMENSION) }
#line 7651 "tokenizer.cpp"
yy1158:
	yych = *++cur;
	if (yych == 'C') goto yy1250;
	if (yych == 'c') goto yy1250;
	goto yy55;
yy1159:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 297 "tokenizer.re"
	{ KW(ELEMENTAL) }
#line 7665 "tokenizer.cpp"
yy1161:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 300 "tokenizer.re"
	{ KW(ELSEWHERE) }
#line 7674 "tokenizer.cpp"
yy1163:
	yych = *++cur;
	if (yych == 'I') goto yy1251;
	if (yych == 'i') goto yy1251;
	goto yy79;
yy1164:
	yyaccept = 15;
	yych = *(mar = ++cur);
	if (yych <= '\r') {
		if (yych <= '\n') {
			if (yych == '\t') goto yy1252;
		} else {
			if (yych != '\f') goto yy1252;
		}
	} else {
		if (yych <= 'C') {
			if (yych == ' ') goto yy1252;
		} else {
			if (yych <= 'D') goto yy1254;
			if (yych == 'd') goto yy1254;
		}
	}
yy1165:
#line 313 "tokenizer.re"
	{ KW(END_BLOCK) }
#line 7700 "tokenizer.cpp"
yy1166:
	yych = *++cur;
	if (yych == 'C') goto yy1255;
	if (yych == 'c') goto yy1255;
	goto yy79;
yy1167:
	yych = *++cur;
	if (yych == 'L') goto yy1256;
	if (yych == 'l') goto yy1256;
	goto yy79;
yy1168:
	yych = *++cur;
	if (yych == 'I') goto yy1258;
	if (yych == 'i') goto yy1258;
	goto yy79;
yy1169:
	yych = *++cur;
	if (yych == 'F') goto yy1259;
	if (yych == 'f') goto yy1259;
	goto yy79;
yy1170:
	yych = *++cur;
	if (yych == 'E') goto yy1260;
	if (yych == 'e') goto yy1260;
	goto yy79;
yy1171:
	yych = *++cur;
	if (yych == 'D') goto yy1262;
	if (yych == 'd') goto yy1262;
	goto yy79;
yy1172:
	yych = *++cur;
	if (yych == 'A') goto yy1263;
	if (yych == 'a') goto yy1263;
	goto yy79;
yy1173:
	yych = *++cur;
	if (yych == 'T') goto yy1264;
	if (yych == 't') goto yy1264;
	goto yy79;
yy1174:
	yych = *++cur;
	if (yych == 'D') goto yy1266;
	if (yych == 'd') goto yy1266;
	goto yy79;
yy1175:
	yych = *++cur;
	if (yych == 'U') goto yy1267;
	if (yych == 'u') goto yy1267;
	goto yy79;
yy1176:
	++cur;
#line 372 "tokenizer.re"
	{ KW(END_WHERE) }
#line 7755 "tokenizer.cpp"
yy1178:
	yych = *++cur;
	if (yych == 'A') goto yy1268;
	if (yych == 'a') goto yy1268;
	goto yy55;
yy1179:
	yych = *++cur;
	if (yych <= '\r') {
		if (yych <= '\n') {
			if (yych == '\t') goto yy1179;
			goto yy79;
		} else {
			if (yych == '\f') goto yy79;
			goto yy1179;
		}
	} else {
		if (yych <= 'C') {
			if (yych == ' ') goto yy1179;
			goto yy79;
		} else {
			if (yych <= 'D') goto yy1269;
			if (yych == 'd') goto yy1269;
			goto yy79;
		}
	}
yy1181:
	yych = *++cur;
	if (yych == 'A') goto yy1270;
	if (yych == 'a') goto yy1270;
	goto yy55;
yy1182:
	yych = *++cur;
	if (yych == 'A') goto yy1271;
	if (yych == 'a') goto yy1271;
	goto yy55;
yy1183:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 346 "tokenizer.re"
	{ KW(ENDFORALL) }
#line 7799 "tokenizer.cpp"
yy1185:
	yych = *++cur;
	if (yych == 'O') goto yy1272;
	if (yych == 'o') goto yy1272;
	goto yy55;
yy1186:
	yych = *++cur;
	if (yych == 'A') goto yy1273;
	if (yych == 'a') goto yy1273;
	goto yy55;
yy1187:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 308 "tokenizer.re"
	{ KW(ENDMODULE) }
#line 7818 "tokenizer.cpp"
yy1189:
	yych = *++cur;
	if (yych == 'U') goto yy1274;
	if (yych == 'u') goto yy1274;
	goto yy55;
yy1190:
	yych = *++cur;
	if (yych == 'M') goto yy1275;
	if (yych == 'm') goto yy1275;
	goto yy55;
yy1191:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 334 "tokenizer.re"
	{ KW(ENDSELECT) }
#line 7837 "tokenizer.cpp"
yy1193:
	yych = *++cur;
	if (yych == 'U') goto yy1277;
	if (yych == 'u') goto yy1277;
	goto yy55;
yy1194:
	yych = *++cur;
	if (yych == 'T') goto yy1278;
	if (yych == 't') goto yy1278;
	goto yy55;
yy1195:
	yych = *++cur;
	if (yych == 'R') goto yy1279;
	if (yych == 'r') goto yy1279;
	goto yy55;
yy1196:
	yych = *++cur;
	if (yych == 'C') goto yy1281;
	if (yych == 'c') goto yy1281;
	goto yy55;
yy1197:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 404 "tokenizer.re"
	{ KW(FORMATTED) }
#line 7866 "tokenizer.cpp"
yy1199:
	yych = *++cur;
	if (yych <= 0x00) goto yy79;
	if (yych == '"') goto yy1282;
	goto yy1199;
yy1201:
	yych = *++cur;
	if (yych <= 0x00) goto yy79;
	if (yych == '\'') goto yy1284;
	goto yy1201;
yy1203:
	yych = *++cur;
	if (yych <= '/') goto yy79;
	if (yych <= '9') goto yy1203;
	if (yych == '_') goto yy1286;
	goto yy79;
yy1205:
	yych = *++cur;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy79;
		if (yych <= '9') goto yy1205;
		if (yych <= '@') goto yy79;
		goto yy1205;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy79;
			goto yy1287;
		} else {
			if (yych <= '`') goto yy79;
			if (yych <= 'z') goto yy1205;
			goto yy79;
		}
	}
yy1207:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 423 "tokenizer.re"
	{ KW(INTERFACE) }
#line 7908 "tokenizer.cpp"
yy1209:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 424 "tokenizer.re"
	{ KW(INTRINSIC) }
#line 7917 "tokenizer.cpp"
yy1211:
	yych = *++cur;
	if (yych == 'T') goto yy1289;
	if (yych == 't') goto yy1289;
	goto yy55;
yy1212:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 436 "tokenizer.re"
	{ KW(NEW_INDEX) }
#line 7931 "tokenizer.cpp"
yy1214:
	yych = *++cur;
	if (yych == 'N') goto yy1291;
	if (yych == 'n') goto yy1291;
	goto yy55;
yy1215:
	yych = *++cur;
	if (yych == 'I') goto yy1292;
	if (yych == 'i') goto yy1292;
	goto yy55;
yy1216:
	yych = *++cur;
	if (yych == 'S') goto yy1293;
	if (yych == 's') goto yy1293;
	goto yy55;
yy1217:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 448 "tokenizer.re"
	{ KW(PARAMETER) }
#line 7955 "tokenizer.cpp"
yy1219:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 452 "tokenizer.re"
	{ KW(PRECISION) }
#line 7964 "tokenizer.cpp"
yy1221:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 455 "tokenizer.re"
	{ KW(PROCEDURE) }
#line 7973 "tokenizer.cpp"
yy1223:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 457 "tokenizer.re"
	{ KW(PROTECTED) }
#line 7982 "tokenizer.cpp"
yy1225:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 464 "tokenizer.re"
	{ KW(RECURSIVE) }
#line 7991 "tokenizer.cpp"
yy1227:
	yych = *++cur;
	if (yych == 'E') goto yy1294;
	if (yych == 'e') goto yy1294;
	goto yy55;
yy1228:
	yych = *++cur;
	if (yych == 'K') goto yy1296;
	if (yych == 'k') goto yy1296;
	goto yy55;
yy1229:
	yych = *++cur;
	if (yych == 'E') goto yy1298;
	if (yych == 'e') goto yy1298;
	goto yy55;
yy1230:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 479 "tokenizer.re"
	{ KW(SUBMODULE) }
#line 8015 "tokenizer.cpp"
yy1232:
	yych = *++cur;
	if (yych == 'E') goto yy1300;
	if (yych == 'e') goto yy1300;
	goto yy55;
yy1233:
	yych = *++cur;
	if (yych == 'S') goto yy1302;
	if (yych == 's') goto yy1302;
	goto yy55;
yy1234:
	yych = *++cur;
	if (yych == 'Y') goto yy1304;
	if (yych == 'y') goto yy1304;
	goto yy55;
yy1235:
	yych = *++cur;
	if (yych == 'E') goto yy1306;
	if (yych == 'e') goto yy1306;
	goto yy55;
yy1236:
	yych = *++cur;
	if (yych == 'E') goto yy1307;
	if (yych == 'e') goto yy1307;
	goto yy55;
yy1237:
	yych = *++cur;
	if (yych == 'E') goto yy1308;
	if (yych == 'e') goto yy1308;
	goto yy55;
yy1238:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 259 "tokenizer.re"
	{ KW(ASSIGNMENT) }
#line 8054 "tokenizer.cpp"
yy1240:
	yych = *++cur;
	if (yych == 'U') goto yy1310;
	if (yych == 'u') goto yy1310;
	goto yy55;
yy1241:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 268 "tokenizer.re"
	{ KW(CHANGE_TEAM) }
#line 8068 "tokenizer.cpp"
yy1243:
	yych = *++cur;
	if (yych == 'N') goto yy1311;
	if (yych == 'n') goto yy1311;
	goto yy55;
yy1244:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 275 "tokenizer.re"
	{ KW(CONCURRENT) }
#line 8082 "tokenizer.cpp"
yy1246:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 277 "tokenizer.re"
	{ KW(CONTIGUOUS) }
#line 8091 "tokenizer.cpp"
yy1248:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 282 "tokenizer.re"
	{ KW(DEALLOCATE) }
#line 8100 "tokenizer.cpp"
yy1250:
	yych = *++cur;
	if (yych == 'I') goto yy1313;
	if (yych == 'i') goto yy1313;
	goto yy55;
yy1251:
	yych = *++cur;
	if (yych == 'A') goto yy1314;
	if (yych == 'a') goto yy1314;
	goto yy79;
yy1252:
	yych = *++cur;
	if (yych <= '\r') {
		if (yych <= '\n') {
			if (yych == '\t') goto yy1252;
			goto yy79;
		} else {
			if (yych == '\f') goto yy79;
			goto yy1252;
		}
	} else {
		if (yych <= 'C') {
			if (yych == ' ') goto yy1252;
			goto yy79;
		} else {
			if (yych <= 'D') goto yy1315;
			if (yych == 'd') goto yy1315;
			goto yy79;
		}
	}
yy1254:
	yych = *++cur;
	if (yych == 'A') goto yy1316;
	if (yych == 'a') goto yy1316;
	goto yy79;
yy1255:
	yych = *++cur;
	if (yych == 'A') goto yy1317;
	if (yych == 'a') goto yy1317;
	goto yy79;
yy1256:
	++cur;
#line 345 "tokenizer.re"
	{ KW(END_FORALL) }
#line 8145 "tokenizer.cpp"
yy1258:
	yych = *++cur;
	if (yych == 'O') goto yy1318;
	if (yych == 'o') goto yy1318;
	goto yy79;
yy1259:
	yych = *++cur;
	if (yych == 'A') goto yy1319;
	if (yych == 'a') goto yy1319;
	goto yy79;
yy1260:
	++cur;
#line 307 "tokenizer.re"
	{ KW(END_MODULE) }
#line 8160 "tokenizer.cpp"
yy1262:
	yych = *++cur;
	if (yych == 'U') goto yy1320;
	if (yych == 'u') goto yy1320;
	goto yy79;
yy1263:
	yych = *++cur;
	if (yych == 'M') goto yy1321;
	if (yych == 'm') goto yy1321;
	goto yy79;
yy1264:
	++cur;
#line 333 "tokenizer.re"
	{ KW(END_SELECT) }
#line 8175 "tokenizer.cpp"
yy1266:
	yych = *++cur;
	if (yych == 'U') goto yy1323;
	if (yych == 'u') goto yy1323;
	goto yy79;
yy1267:
	yych = *++cur;
	if (yych == 'T') goto yy1324;
	if (yych == 't') goto yy1324;
	goto yy79;
yy1268:
	yych = *++cur;
	if (yych == 'T') goto yy1325;
	if (yych == 't') goto yy1325;
	goto yy55;
yy1269:
	yych = *++cur;
	if (yych == 'A') goto yy1326;
	if (yych == 'a') goto yy1326;
	goto yy79;
yy1270:
	yych = *++cur;
	if (yych == 'T') goto yy1327;
	if (yych == 't') goto yy1327;
	goto yy55;
yy1271:
	yych = *++cur;
	if (yych == 'L') goto yy1328;
	if (yych == 'l') goto yy1328;
	goto yy55;
yy1272:
	yych = *++cur;
	if (yych == 'N') goto yy1330;
	if (yych == 'n') goto yy1330;
	goto yy55;
yy1273:
	yych = *++cur;
	if (yych == 'C') goto yy1332;
	if (yych == 'c') goto yy1332;
	goto yy55;
yy1274:
	yych = *++cur;
	if (yych == 'R') goto yy1333;
	if (yych == 'r') goto yy1333;
	goto yy55;
yy1275:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 305 "tokenizer.re"
	{ KW(ENDPROGRAM) }
#line 8229 "tokenizer.cpp"
yy1277:
	yych = *++cur;
	if (yych == 'L') goto yy1334;
	if (yych == 'l') goto yy1334;
	goto yy55;
yy1278:
	yych = *++cur;
	if (yych == 'I') goto yy1335;
	if (yych == 'i') goto yy1335;
	goto yy55;
yy1279:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 380 "tokenizer.re"
	{ KW(ENUMERATOR) }
#line 8248 "tokenizer.cpp"
yy1281:
	yych = *++cur;
	if (yych == 'E') goto yy1336;
	if (yych == 'e') goto yy1336;
	goto yy55;
yy1282:
	yyaccept = 16;
	yych = *(mar = ++cur);
	if (yych == '"') goto yy1199;
yy1283:
#line 657 "tokenizer.re"
	{ continue; }
#line 8261 "tokenizer.cpp"
yy1284:
	yyaccept = 17;
	yych = *(mar = ++cur);
	if (yych == '\'') goto yy1201;
yy1285:
#line 658 "tokenizer.re"
	{ continue; }
#line 8269 "tokenizer.cpp"
yy1286:
	yych = *++cur;
	if (yych == '"') goto yy1199;
	if (yych == '\'') goto yy1201;
	goto yy79;
yy1287:
	yych = *++cur;
	if (yych <= '9') {
		if (yych <= '&') {
			if (yych == '"') goto yy1199;
			goto yy79;
		} else {
			if (yych <= '\'') goto yy1201;
			if (yych <= '/') goto yy79;
			goto yy1205;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy79;
			if (yych <= 'Z') goto yy1205;
			goto yy79;
		} else {
			if (yych <= '_') goto yy1287;
			if (yych <= '`') goto yy79;
			if (yych <= 'z') goto yy1205;
			goto yy79;
		}
	}
yy1289:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 429 "tokenizer.re"
	{ KW(LOCAL_INIT) }
#line 8306 "tokenizer.cpp"
yy1291:
	yych = *++cur;
	if (yych == 'S') goto yy1338;
	if (yych == 's') goto yy1338;
	goto yy55;
yy1292:
	yych = *++cur;
	if (yych == 'D') goto yy1339;
	if (yych == 'd') goto yy1339;
	goto yy55;
yy1293:
	yych = *++cur;
	if (yych == 'I') goto yy1340;
	if (yych == 'i') goto yy1340;
	goto yy55;
yy1294:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 471 "tokenizer.re"
	{ KW(SELECT_CASE) }
#line 8330 "tokenizer.cpp"
yy1296:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 472 "tokenizer.re"
	{ KW(SELECT_RANK) }
#line 8339 "tokenizer.cpp"
yy1298:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 473 "tokenizer.re"
	{ KW(SELECT_TYPE) }
#line 8348 "tokenizer.cpp"
yy1300:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 480 "tokenizer.re"
	{ KW(SUBROUTINE) }
#line 8357 "tokenizer.cpp"
yy1302:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 483 "tokenizer.re"
	{ KW(SYNC_IMAGES) }
#line 8366 "tokenizer.cpp"
yy1304:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 484 "tokenizer.re"
	{ KW(SYNC_MEMORY) }
#line 8375 "tokenizer.cpp"
yy1306:
	yych = *++cur;
	if (yych == 'R') goto yy1341;
	if (yych == 'r') goto yy1341;
	goto yy55;
yy1307:
	yych = *++cur;
	if (yych == 'D') goto yy1343;
	if (yych == 'd') goto yy1343;
	goto yy55;
yy1308:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 256 "tokenizer.re"
	{ KW(ALLOCATABLE) }
#line 8394 "tokenizer.cpp"
yy1310:
	yych = *++cur;
	if (yych == 'S') goto yy1345;
	if (yych == 's') goto yy1345;
	goto yy55;
yy1311:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 272 "tokenizer.re"
	{ KW(CODIMENSION) }
#line 8408 "tokenizer.cpp"
yy1313:
	yych = *++cur;
	if (yych == 'S') goto yy1347;
	if (yych == 's') goto yy1347;
	goto yy55;
yy1314:
	yych = *++cur;
	if (yych == 'T') goto yy1348;
	if (yych == 't') goto yy1348;
	goto yy79;
yy1315:
	yych = *++cur;
	if (yych == 'A') goto yy1349;
	if (yych == 'a') goto yy1349;
	goto yy79;
yy1316:
	yych = *++cur;
	if (yych == 'T') goto yy1350;
	if (yych == 't') goto yy1350;
	goto yy79;
yy1317:
	yych = *++cur;
	if (yych == 'L') goto yy1351;
	if (yych == 'l') goto yy1351;
	goto yy79;
yy1318:
	yych = *++cur;
	if (yych == 'N') goto yy1353;
	if (yych == 'n') goto yy1353;
	goto yy79;
yy1319:
	yych = *++cur;
	if (yych == 'C') goto yy1355;
	if (yych == 'c') goto yy1355;
	goto yy79;
yy1320:
	yych = *++cur;
	if (yych == 'R') goto yy1356;
	if (yych == 'r') goto yy1356;
	goto yy79;
yy1321:
	++cur;
#line 304 "tokenizer.re"
	{ KW(END_PROGRAM) }
#line 8453 "tokenizer.cpp"
yy1323:
	yych = *++cur;
	if (yych == 'L') goto yy1357;
	if (yych == 'l') goto yy1357;
	goto yy79;
yy1324:
	yych = *++cur;
	if (yych == 'I') goto yy1358;
	if (yych == 'i') goto yy1358;
	goto yy79;
yy1325:
	yych = *++cur;
	if (yych == 'E') goto yy1359;
	if (yych == 'e') goto yy1359;
	goto yy55;
yy1326:
	yych = *++cur;
	if (yych == 'T') goto yy1361;
	if (yych == 't') goto yy1361;
	goto yy79;
yy1327:
	yych = *++cur;
	if (yych == 'A') goto yy1362;
	if (yych == 'a') goto yy1362;
	goto yy55;
yy1328:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 340 "tokenizer.re"
	{ KW(ENDCRITICAL) }
#line 8487 "tokenizer.cpp"
yy1330:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 325 "tokenizer.re"
	{ KW(ENDFUNCTION) }
#line 8496 "tokenizer.cpp"
yy1332:
	yych = *++cur;
	if (yych == 'E') goto yy1364;
	if (yych == 'e') goto yy1364;
	goto yy55;
yy1333:
	yych = *++cur;
	if (yych == 'E') goto yy1366;
	if (yych == 'e') goto yy1366;
	goto yy55;
yy1334:
	yych = *++cur;
	if (yych == 'E') goto yy1368;
	if (yych == 'e') goto yy1368;
	goto yy55;
yy1335:
	yych = *++cur;
	if (yych == 'N') goto yy1370;
	if (yych == 'n') goto yy1370;
	goto yy55;
yy1336:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 381 "tokenizer.re"
	{ KW(EQUIVALENCE) }
#line 8525 "tokenizer.cpp"
yy1338:
	yych = *++cur;
	if (yych == 'I') goto yy1371;
	if (yych == 'i') goto yy1371;
	goto yy55;
yy1339:
	yych = *++cur;
	if (yych == 'A') goto yy1372;
	if (yych == 'a') goto yy1372;
	goto yy55;
yy1340:
	yych = *++cur;
	if (yych == 'V') goto yy1373;
	if (yych == 'v') goto yy1373;
	goto yy55;
yy1341:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 488 "tokenizer.re"
	{ KW(TEAM_NUMBER) }
#line 8549 "tokenizer.cpp"
yy1343:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 492 "tokenizer.re"
	{ KW(UNFORMATTED) }
#line 8558 "tokenizer.cpp"
yy1345:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 261 "tokenizer.re"
	{ KW(ASYNCHRONOUS) }
#line 8567 "tokenizer.cpp"
yy1347:
	yych = *++cur;
	if (yych == 'I') goto yy1374;
	if (yych == 'i') goto yy1374;
	goto yy55;
yy1348:
	yych = *++cur;
	if (yych == 'E') goto yy1375;
	if (yych == 'e') goto yy1375;
	goto yy79;
yy1349:
	yych = *++cur;
	if (yych == 'T') goto yy1377;
	if (yych == 't') goto yy1377;
	goto yy79;
yy1350:
	yych = *++cur;
	if (yych == 'A') goto yy1378;
	if (yych == 'a') goto yy1378;
	goto yy79;
yy1351:
	++cur;
#line 339 "tokenizer.re"
	{ KW(END_CRITICAL) }
#line 8592 "tokenizer.cpp"
yy1353:
	++cur;
#line 324 "tokenizer.re"
	{ KW(END_FUNCTION) }
#line 8597 "tokenizer.cpp"
yy1355:
	yych = *++cur;
	if (yych == 'E') goto yy1380;
	if (yych == 'e') goto yy1380;
	goto yy79;
yy1356:
	yych = *++cur;
	if (yych == 'E') goto yy1382;
	if (yych == 'e') goto yy1382;
	goto yy79;
yy1357:
	yych = *++cur;
	if (yych == 'E') goto yy1384;
	if (yych == 'e') goto yy1384;
	goto yy79;
yy1358:
	yych = *++cur;
	if (yych == 'N') goto yy1386;
	if (yych == 'n') goto yy1386;
	goto yy79;
yy1359:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 337 "tokenizer.re"
	{ KW(ENDASSOCIATE) }
#line 8626 "tokenizer.cpp"
yy1361:
	yych = *++cur;
	if (yych == 'A') goto yy1387;
	if (yych == 'a') goto yy1387;
	goto yy79;
yy1362:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 319 "tokenizer.re"
	{ KW(ENDBLOCKDATA) }
#line 8640 "tokenizer.cpp"
yy1364:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 352 "tokenizer.re"
	{ KW(ENDINTERFACE) }
#line 8649 "tokenizer.cpp"
yy1366:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 328 "tokenizer.re"
	{ KW(ENDPROCEDURE) }
#line 8658 "tokenizer.cpp"
yy1368:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 311 "tokenizer.re"
	{ KW(ENDSUBMODULE) }
#line 8667 "tokenizer.cpp"
yy1370:
	yych = *++cur;
	if (yych == 'E') goto yy1389;
	if (yych == 'e') goto yy1389;
	goto yy55;
yy1371:
	yych = *++cur;
	if (yych == 'C') goto yy1391;
	if (yych == 'c') goto yy1391;
	goto yy55;
yy1372:
	yych = *++cur;
	if (yych == 'B') goto yy1393;
	if (yych == 'b') goto yy1393;
	goto yy55;
yy1373:
	yych = *++cur;
	if (yych == 'E') goto yy1394;
	if (yych == 'e') goto yy1394;
	goto yy55;
yy1374:
	yych = *++cur;
	if (yych == 'O') goto yy1396;
	if (yych == 'o') goto yy1396;
	goto yy55;
yy1375:
	++cur;
#line 336 "tokenizer.re"
	{ KW(END_ASSOCIATE) }
#line 8697 "tokenizer.cpp"
yy1377:
	yych = *++cur;
	if (yych == 'A') goto yy1397;
	if (yych == 'a') goto yy1397;
	goto yy79;
yy1378:
	++cur;
#line 318 "tokenizer.re"
	{ KW(END_BLOCK_DATA) }
#line 8707 "tokenizer.cpp"
yy1380:
	++cur;
#line 351 "tokenizer.re"
	{ KW(END_INTERFACE) }
#line 8712 "tokenizer.cpp"
yy1382:
	++cur;
#line 327 "tokenizer.re"
	{ KW(END_PROCEDURE) }
#line 8717 "tokenizer.cpp"
yy1384:
	++cur;
#line 310 "tokenizer.re"
	{ KW(END_SUBMODULE) }
#line 8722 "tokenizer.cpp"
yy1386:
	yych = *++cur;
	if (yych == 'E') goto yy1399;
	if (yych == 'e') goto yy1399;
	goto yy79;
yy1387:
	++cur;
#line 317 "tokenizer.re"
	{ KW(END_BLOCK_DATA) }
#line 8732 "tokenizer.cpp"
yy1389:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 322 "tokenizer.re"
	{ KW(ENDSUBROUTINE) }
#line 8741 "tokenizer.cpp"
yy1391:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 438 "tokenizer.re"
	{ KW(NON_INTRINSIC) }
#line 8750 "tokenizer.cpp"
yy1393:
	yych = *++cur;
	if (yych == 'L') goto yy1401;
	if (yych == 'l') goto yy1401;
	goto yy55;
yy1394:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 440 "tokenizer.re"
	{ KW(NON_RECURSIVE) }
#line 8764 "tokenizer.cpp"
yy1396:
	yych = *++cur;
	if (yych == 'N') goto yy1402;
	if (yych == 'n') goto yy1402;
	goto yy55;
yy1397:
	++cur;
#line 316 "tokenizer.re"
	{ KW(END_BLOCK_DATA) }
#line 8774 "tokenizer.cpp"
yy1399:
	++cur;
#line 321 "tokenizer.re"
	{ KW(END_SUBROUTINE) }
#line 8779 "tokenizer.cpp"
yy1401:
	yych = *++cur;
	if (yych == 'E') goto yy1404;
	if (yych == 'e') goto yy1404;
	goto yy55;
yy1402:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 296 "tokenizer.re"
	{ KW(DOUBLE_PRECISION) }
#line 8793 "tokenizer.cpp"
yy1404:
	yych = *++cur;
	if (yybm[256+yych] & 32) {
		goto yy54;
	}
	if (yych == '_') goto yy144;
#line 439 "tokenizer.re"
	{ KW(NON_OVERRIDABLE) }
#line 8802 "tokenizer.cpp"
}
#line 665 "tokenizer.re"

    }
}

std::string token(unsigned char *tok, unsigned char* cur)
{
    return std::string((char *)tok, cur - tok);
}

void token_loc(Location &loc)
{
    loc.first = 1;
    loc.last = 1;
}

void lex_format(unsigned char *&cur, Location &loc,
        unsigned char *&start) {
    int num_paren = 0;
    for (;;) {
        unsigned char *tok = cur;
        unsigned char *mar;
        
#line 8827 "tokenizer.cpp"
{
	unsigned char yych;
	unsigned int yyaccept = 0;
	static const unsigned char yybm[] = {
		  0, 208, 208, 208, 208, 208, 208, 208, 
		208, 244, 176, 244, 208, 244, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		244, 208, 192, 208, 208, 208, 208,  80, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		216, 216, 216, 216, 216, 216, 216, 216, 
		216, 216, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
		208, 208, 208, 208, 208, 208, 208, 208, 
	};
	yych = *cur;
	if (yybm[0+yych] & 4) {
		goto yy1412;
	}
	switch (yych) {
	case 0x00:	goto yy1408;
	case '"':	goto yy1415;
	case '&':	goto yy1416;
	case '\'':	goto yy1417;
	case '(':	goto yy1418;
	case ')':	goto yy1420;
	case '*':	goto yy1422;
	case ',':	goto yy1423;
	case '/':
	case ':':	goto yy1425;
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':	goto yy1427;
	case 'A':
	case 'a':	goto yy1428;
	case 'B':
	case 'I':
	case 'O':
	case 'Z':
	case 'b':
	case 'i':
	case 'o':
	case 'z':	goto yy1431;
	case 'D':
	case 'F':
	case 'd':
	case 'f':	goto yy1432;
	case 'E':
	case 'e':	goto yy1433;
	case 'G':
	case 'g':	goto yy1434;
	case 'L':
	case 'l':	goto yy1435;
	case 'P':
	case 'p':	goto yy1436;
	case 'T':
	case 't':	goto yy1437;
	case 'X':
	case 'x':	goto yy1438;
	default:	goto yy1410;
	}
yy1408:
	++cur;
#line 758 "tokenizer.re"
	{
                token_loc(loc);
                std::string t = token(tok, cur);
                throw LFortran::parser_local::TokenizerError(
                    "End of file not expected in `format` statement '" + t + "'", loc);
            }
#line 8927 "tokenizer.cpp"
yy1410:
	++cur;
yy1411:
#line 724 "tokenizer.re"
	{
                token_loc(loc);
                std::string t = token(tok, cur);
                throw LFortran::parser_local::TokenizerError("Token '" + t
                    + "' is not recognized in `format` statement", loc);
            }
#line 8938 "tokenizer.cpp"
yy1412:
	yych = *++cur;
	if (yybm[0+yych] & 4) {
		goto yy1412;
	}
#line 764 "tokenizer.re"
	{ continue; }
#line 8946 "tokenizer.cpp"
yy1415:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= 0x00) goto yy1411;
	goto yy1440;
yy1416:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= '\f') {
		if (yych <= 0x08) goto yy1411;
		if (yych <= '\v') goto yy1445;
		goto yy1411;
	} else {
		if (yych <= '\r') goto yy1445;
		if (yych <= 0x1F) goto yy1411;
		if (yych <= '!') goto yy1445;
		goto yy1411;
	}
yy1417:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= 0x00) goto yy1411;
	goto yy1452;
yy1418:
	++cur;
#line 730 "tokenizer.re"
	{
                if (num_paren == 0) {
                    num_paren++;
                    start = cur;
                    continue;
                } else {
                    cur--;
                    unsigned char *tmp;
                    lex_format(cur, loc, tmp);
                    continue;
                }
            }
#line 8985 "tokenizer.cpp"
yy1420:
	++cur;
#line 754 "tokenizer.re"
	{
                LFORTRAN_ASSERT(num_paren == 1);
                return;
            }
#line 8993 "tokenizer.cpp"
yy1422:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= '\f') {
		if (yych <= '\t') {
			if (yych <= 0x08) goto yy1411;
			goto yy1455;
		} else {
			if (yych == '\v') goto yy1455;
			goto yy1411;
		}
	} else {
		if (yych <= ' ') {
			if (yych <= '\r') goto yy1455;
			if (yych <= 0x1F) goto yy1411;
			goto yy1455;
		} else {
			if (yych == '(') goto yy1457;
			goto yy1411;
		}
	}
yy1423:
	++cur;
#line 765 "tokenizer.re"
	{ continue; }
#line 9019 "tokenizer.cpp"
yy1425:
	++cur;
yy1426:
#line 770 "tokenizer.re"
	{ continue; }
#line 9025 "tokenizer.cpp"
yy1427:
	yyaccept = 0;
	yych = *(mar = ++cur);
	switch (yych) {
	case '\t':
	case '\v':
	case '\r':
	case ' ':
	case '(':
	case '/':
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	case 'A':
	case 'B':
	case 'D':
	case 'E':
	case 'F':
	case 'G':
	case 'I':
	case 'L':
	case 'O':
	case 'P':
	case 'X':
	case 'Z':
	case 'a':
	case 'b':
	case 'd':
	case 'e':
	case 'f':
	case 'g':
	case 'i':
	case 'l':
	case 'o':
	case 'p':
	case 'x':
	case 'z':	goto yy1464;
	default:	goto yy1411;
	}
yy1428:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy1428;
	}
yy1430:
#line 769 "tokenizer.re"
	{ continue; }
#line 9080 "tokenizer.cpp"
yy1431:
	yych = *++cur;
	if (yych <= '/') goto yy1411;
	if (yych <= '9') goto yy1470;
	goto yy1411;
yy1432:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= '/') goto yy1411;
	if (yych <= '9') goto yy1472;
	goto yy1411;
yy1433:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= 'W') {
		if (yych <= 'M') {
			if (yych <= '/') goto yy1411;
			if (yych <= '9') goto yy1474;
			goto yy1411;
		} else {
			if (yych <= 'N') goto yy1476;
			if (yych == 'S') goto yy1476;
			goto yy1411;
		}
	} else {
		if (yych <= 'r') {
			if (yych <= 'X') goto yy1476;
			if (yych == 'n') goto yy1476;
			goto yy1411;
		} else {
			if (yych <= 's') goto yy1476;
			if (yych == 'x') goto yy1476;
			goto yy1411;
		}
	}
yy1434:
	yych = *++cur;
	if (yych <= '/') goto yy1411;
	if (yych <= '9') goto yy1477;
	goto yy1411;
yy1435:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy1428;
	}
	goto yy1411;
yy1436:
	yyaccept = 1;
	yych = *(mar = ++cur);
	if (yych <= 'D') goto yy1430;
	if (yych <= 'F') goto yy1466;
	if (yych <= 'd') goto yy1430;
	if (yych <= 'f') goto yy1466;
	goto yy1430;
yy1437:
	yyaccept = 0;
	yych = *(mar = ++cur);
	if (yych <= 'Q') {
		if (yych <= '9') {
			if (yych <= '/') goto yy1411;
			goto yy1479;
		} else {
			if (yych == 'L') goto yy1481;
			goto yy1411;
		}
	} else {
		if (yych <= 'l') {
			if (yych <= 'R') goto yy1481;
			if (yych <= 'k') goto yy1411;
			goto yy1481;
		} else {
			if (yych == 'r') goto yy1481;
			goto yy1411;
		}
	}
yy1438:
	++cur;
	goto yy1430;
yy1439:
	yych = *++cur;
yy1440:
	if (yybm[0+yych] & 16) {
		goto yy1439;
	}
	if (yych >= 0x01) goto yy1442;
yy1441:
	cur = mar;
	if (yyaccept <= 2) {
		if (yyaccept <= 1) {
			if (yyaccept == 0) {
				goto yy1411;
			} else {
				goto yy1430;
			}
		} else {
			goto yy1443;
		}
	} else {
		if (yyaccept == 3) {
			goto yy1448;
		} else {
			goto yy1454;
		}
	}
yy1442:
	yyaccept = 2;
	yych = *(mar = ++cur);
	if (yych == '"') goto yy1439;
yy1443:
#line 767 "tokenizer.re"
	{ continue; }
#line 9192 "tokenizer.cpp"
yy1444:
	yych = *++cur;
yy1445:
	if (yych <= '\f') {
		if (yych <= '\t') {
			if (yych <= 0x08) goto yy1441;
			goto yy1444;
		} else {
			if (yych <= '\n') goto yy1446;
			if (yych <= '\v') goto yy1444;
			goto yy1441;
		}
	} else {
		if (yych <= 0x1F) {
			if (yych <= '\r') goto yy1444;
			goto yy1441;
		} else {
			if (yych <= ' ') goto yy1444;
			if (yych <= '!') goto yy1449;
			goto yy1441;
		}
	}
yy1446:
	yyaccept = 3;
	yych = *(mar = ++cur);
	if (yybm[0+yych] & 32) {
		goto yy1446;
	}
	if (yych <= 0x1F) goto yy1448;
	if (yych <= '!') goto yy1449;
	if (yych == '&') goto yy1482;
yy1448:
#line 766 "tokenizer.re"
	{ continue; }
#line 9227 "tokenizer.cpp"
yy1449:
	yych = *++cur;
	if (yybm[0+yych] & 64) {
		goto yy1449;
	}
	if (yych <= 0x00) goto yy1441;
	goto yy1446;
yy1451:
	yych = *++cur;
yy1452:
	if (yybm[0+yych] & 128) {
		goto yy1451;
	}
	if (yych <= 0x00) goto yy1441;
	yyaccept = 4;
	yych = *(mar = ++cur);
	if (yych == '\'') goto yy1451;
yy1454:
#line 768 "tokenizer.re"
	{ continue; }
#line 9248 "tokenizer.cpp"
yy1455:
	yych = *++cur;
	if (yych <= '\f') {
		if (yych <= '\t') {
			if (yych <= 0x08) goto yy1441;
			goto yy1455;
		} else {
			if (yych == '\v') goto yy1455;
			goto yy1441;
		}
	} else {
		if (yych <= ' ') {
			if (yych <= '\r') goto yy1455;
			if (yych <= 0x1F) goto yy1441;
			goto yy1455;
		} else {
			if (yych != '(') goto yy1441;
		}
	}
yy1457:
	++cur;
#line 748 "tokenizer.re"
	{
                cur--;
                unsigned char *tmp;
                lex_format(cur, loc, tmp);
                continue;
            }
#line 9277 "tokenizer.cpp"
yy1459:
	yych = *++cur;
	if (yych <= '\f') {
		if (yych <= '\t') {
			if (yych <= 0x08) goto yy1441;
			goto yy1459;
		} else {
			if (yych == '\v') goto yy1459;
			goto yy1441;
		}
	} else {
		if (yych <= ' ') {
			if (yych <= '\r') goto yy1459;
			if (yych <= 0x1F) goto yy1441;
			goto yy1459;
		} else {
			if (yych != '(') goto yy1441;
		}
	}
yy1461:
	++cur;
#line 742 "tokenizer.re"
	{
                cur--;
                unsigned char *tmp;
                lex_format(cur, loc, tmp);
                continue;
            }
#line 9306 "tokenizer.cpp"
yy1463:
	yych = *++cur;
yy1464:
	switch (yych) {
	case '\t':
	case '\v':
	case '\r':
	case ' ':	goto yy1459;
	case '(':	goto yy1461;
	case '/':	goto yy1425;
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':	goto yy1463;
	case 'A':
	case 'a':	goto yy1428;
	case 'B':
	case 'I':
	case 'O':
	case 'Z':
	case 'b':
	case 'i':
	case 'o':
	case 'z':	goto yy1465;
	case 'D':
	case 'F':
	case 'd':
	case 'f':	goto yy1466;
	case 'E':
	case 'e':	goto yy1467;
	case 'G':
	case 'g':	goto yy1468;
	case 'L':
	case 'l':	goto yy1469;
	case 'P':
	case 'p':	goto yy1436;
	case 'X':
	case 'x':	goto yy1438;
	default:	goto yy1441;
	}
yy1465:
	yych = *++cur;
	if (yych <= '/') goto yy1441;
	if (yych <= '9') goto yy1470;
	goto yy1441;
yy1466:
	yych = *++cur;
	if (yych <= '/') goto yy1441;
	if (yych <= '9') goto yy1472;
	goto yy1441;
yy1467:
	yych = *++cur;
	if (yych <= 'W') {
		if (yych <= 'M') {
			if (yych <= '/') goto yy1441;
			if (yych <= '9') goto yy1474;
			goto yy1441;
		} else {
			if (yych <= 'N') goto yy1476;
			if (yych == 'S') goto yy1476;
			goto yy1441;
		}
	} else {
		if (yych <= 'r') {
			if (yych <= 'X') goto yy1476;
			if (yych == 'n') goto yy1476;
			goto yy1441;
		} else {
			if (yych <= 's') goto yy1476;
			if (yych == 'x') goto yy1476;
			goto yy1441;
		}
	}
yy1468:
	yych = *++cur;
	if (yych <= '/') goto yy1441;
	if (yych <= '9') goto yy1477;
	goto yy1441;
yy1469:
	yych = *++cur;
	if (yybm[0+yych] & 8) {
		goto yy1428;
	}
	goto yy1441;
yy1470:
	yyaccept = 1;
	yych = *(mar = ++cur);
	if (yych == '.') goto yy1469;
	if (yych <= '/') goto yy1430;
	if (yych <= '9') goto yy1470;
	goto yy1430;
yy1472:
	yych = *++cur;
	if (yych == '.') goto yy1469;
	if (yych <= '/') goto yy1441;
	if (yych <= '9') goto yy1472;
	goto yy1441;
yy1474:
	yych = *++cur;
	if (yych == '.') goto yy1483;
	if (yych <= '/') goto yy1441;
	if (yych <= '9') goto yy1474;
	goto yy1441;
yy1476:
	yych = *++cur;
	if (yych <= '/') goto yy1441;
	if (yych <= '9') goto yy1474;
	goto yy1441;
yy1477:
	yyaccept = 1;
	yych = *(mar = ++cur);
	if (yych == '.') goto yy1483;
	if (yych <= '/') goto yy1430;
	if (yych <= '9') goto yy1477;
	goto yy1430;
yy1479:
	yych = *++cur;
	if (yych <= '/') goto yy1426;
	if (yych <= '9') goto yy1479;
	goto yy1426;
yy1481:
	yych = *++cur;
	if (yych <= '/') goto yy1441;
	if (yych <= '9') goto yy1479;
	goto yy1441;
yy1482:
	++cur;
	goto yy1448;
yy1483:
	yych = *++cur;
	if (yych <= '/') goto yy1441;
	if (yych >= ':') goto yy1441;
yy1484:
	yyaccept = 1;
	yych = *(mar = ++cur);
	if (yych <= 'D') {
		if (yych <= '/') goto yy1430;
		if (yych <= '9') goto yy1484;
		goto yy1430;
	} else {
		if (yych <= 'E') goto yy1469;
		if (yych == 'e') goto yy1469;
		goto yy1430;
	}
}
#line 771 "tokenizer.re"

    }
}

} // namespace LFortran
