#ifndef LFORTRAN_ASR_H
#define LFORTRAN_ASR_H

// Generated by grammar/asdl_cpp.py

#include <libasr/alloc.h>
#include <libasr/location.h>
#include <libasr/colors.h>
#include <libasr/containers.h>
#include <libasr/exception.h>
#include <libasr/asr_scopes.h>


namespace LFortran::ASR {

enum asrType
{
    unit, symbol, stmt, expr, ttype, attribute, tbind, case_stmt
};

struct asr_t
{
    asrType type;
    Location loc;
};


template <class T, class U>
inline bool is_a(const U &x)
{
    return T::class_type == x.type;
}

// Cast one level down

template <class T, class U>
static inline T* down_cast(const U *f)
{
    LFORTRAN_ASSERT(f != nullptr);
    LFORTRAN_ASSERT(is_a<T>(*f));
    return (T*)f;
}

// Cast two levels down

template <class T>
static inline T* down_cast2(const asr_t *f)
{
    typedef typename T::parent_type ptype;
    ptype *t = down_cast<ptype>(f);
    return down_cast<T>(t);
}


/******************************************************************************/
// Forward declarations

struct unit_t; // Sum
struct symbol_t; // Sum
enum storage_typeType // Simple Sum
{ // Types
    Default, Save, Parameter, Allocatable
};
enum accessType // Simple Sum
{ // Types
    Public, Private
};
enum intentType // Simple Sum
{ // Types
    Local, In, Out, InOut, ReturnVar, Unspecified
};
enum deftypeType // Simple Sum
{ // Types
    Implementation, Interface
};
enum presenceType // Simple Sum
{ // Types
    Required, Optional
};
enum abiType // Simple Sum
{ // Types
    Source, LFortranModule, GFortranModule, BindC, Interactive, Intrinsic
};
struct stmt_t; // Sum
struct expr_t; // Sum
struct ttype_t; // Sum
enum boolopType // Simple Sum
{ // Types
    And, Or, Xor, NEqv, Eqv
};
enum binopType // Simple Sum
{ // Types
    Add, Sub, Mul, Div, Pow
};
enum unaryopType // Simple Sum
{ // Types
    Invert, Not, UAdd, USub
};
enum stropType // Simple Sum
{ // Types
    Repeat
};
enum cmpopType // Simple Sum
{ // Types
    Eq, NotEq, Lt, LtE, Gt, GtE
};
enum integerbozType // Simple Sum
{ // Types
    Binary, Hex, Octal
};
enum arrayboundType // Simple Sum
{ // Types
    LBound, UBound
};
enum cast_kindType // Simple Sum
{ // Types
    RealToInteger, IntegerToReal, RealToReal, IntegerToInteger, RealToComplex, IntegerToComplex, IntegerToLogical, ComplexToComplex, ComplexToReal, LogicalToInteger
};
struct dimension_t; // Product
struct alloc_arg_t; // Product
struct attribute_t; // Sum
struct attribute_arg_t; // Product
struct call_arg_t; // Product
struct tbind_t; // Sum
struct array_index_t; // Product
struct do_loop_head_t; // Product
struct case_stmt_t; // Sum


/******************************************************************************/
// Products declarations

struct dimension_t // Product
{
    Location loc;
    expr_t* m_start;
    expr_t* m_end;
};
struct alloc_arg_t // Product
{
    Location loc;
    symbol_t* m_a;
    dimension_t* m_dims; size_t n_dims; // Sequence
};
struct attribute_arg_t // Product
{
    Location loc;
    char* m_arg;
};
struct call_arg_t // Product
{
    Location loc;
    expr_t* m_value;
};
struct array_index_t // Product
{
    Location loc;
    expr_t* m_left;
    expr_t* m_right;
    expr_t* m_step;
};
struct do_loop_head_t // Product
{
    Location loc;
    expr_t* m_v;
    expr_t* m_start;
    expr_t* m_end;
    expr_t* m_increment;
};


/******************************************************************************/
// Sums declarations

enum unitType // Types
{
    TranslationUnit
};

struct unit_t // Sum
{
    const static asrType class_type = asrType::unit;
    asr_t base;
    unitType type;
};

    struct TranslationUnit_t // Constructor
    {
        const static unitType class_type = unitType::TranslationUnit;
        typedef unit_t parent_type;
        unit_t base;
        SymbolTable* m_global_scope;
        asr_t** m_items; size_t n_items; // Sequence
    };
    static inline asr_t* make_TranslationUnit_t(Allocator &al, const Location &a_loc, SymbolTable* a_global_scope, asr_t** a_items, size_t n_items) {
        TranslationUnit_t *n;
        n = al.make_new<TranslationUnit_t>();
        n->base.type = unitType::TranslationUnit;
        n->base.base.type = asrType::unit;
        n->base.base.loc = a_loc;
        n->m_global_scope = a_global_scope;
        a_global_scope->asr_owner = (asr_t*)n;
        n->m_items = a_items;
        n->n_items = n_items;
        return (asr_t*)n;
    }



enum symbolType // Types
{
    Program, Module, Subroutine, Function, GenericProcedure, CustomOperator, ExternalSymbol, DerivedType, Variable, ClassType, ClassProcedure, AssociateBlock, Block
};

struct symbol_t // Sum
{
    const static asrType class_type = asrType::symbol;
    asr_t base;
    symbolType type;
};

    struct Program_t // Constructor
    {
        const static symbolType class_type = symbolType::Program;
        typedef symbol_t parent_type;
        symbol_t base;
        SymbolTable* m_symtab;
        char* m_name;
        char** m_dependencies; size_t n_dependencies; // Sequence
        stmt_t** m_body; size_t n_body; // Sequence
    };
    static inline asr_t* make_Program_t(Allocator &al, const Location &a_loc, SymbolTable* a_symtab, char* a_name, char** a_dependencies, size_t n_dependencies, stmt_t** a_body, size_t n_body) {
        Program_t *n;
        n = al.make_new<Program_t>();
        n->base.type = symbolType::Program;
        n->base.base.type = asrType::symbol;
        n->base.base.loc = a_loc;
        n->m_symtab = a_symtab;
        a_symtab->asr_owner = (asr_t*)n;
        n->m_name = a_name;
        n->m_dependencies = a_dependencies;
        n->n_dependencies = n_dependencies;
        n->m_body = a_body;
        n->n_body = n_body;
        return (asr_t*)n;
    }

    struct Module_t // Constructor
    {
        const static symbolType class_type = symbolType::Module;
        typedef symbol_t parent_type;
        symbol_t base;
        SymbolTable* m_symtab;
        char* m_name;
        char** m_dependencies; size_t n_dependencies; // Sequence
        bool m_loaded_from_mod;
        bool m_intrinsic;
    };
    static inline asr_t* make_Module_t(Allocator &al, const Location &a_loc, SymbolTable* a_symtab, char* a_name, char** a_dependencies, size_t n_dependencies, bool a_loaded_from_mod, bool a_intrinsic) {
        Module_t *n;
        n = al.make_new<Module_t>();
        n->base.type = symbolType::Module;
        n->base.base.type = asrType::symbol;
        n->base.base.loc = a_loc;
        n->m_symtab = a_symtab;
        a_symtab->asr_owner = (asr_t*)n;
        n->m_name = a_name;
        n->m_dependencies = a_dependencies;
        n->n_dependencies = n_dependencies;
        n->m_loaded_from_mod = a_loaded_from_mod;
        n->m_intrinsic = a_intrinsic;
        return (asr_t*)n;
    }

    struct Subroutine_t // Constructor
    {
        const static symbolType class_type = symbolType::Subroutine;
        typedef symbol_t parent_type;
        symbol_t base;
        SymbolTable* m_symtab;
        char* m_name;
        expr_t** m_args; size_t n_args; // Sequence
        stmt_t** m_body; size_t n_body; // Sequence
        abiType m_abi;
        accessType m_access;
        deftypeType m_deftype;
        char* m_bindc_name;
        bool m_pure;
        bool m_module;
    };
    static inline asr_t* make_Subroutine_t(Allocator &al, const Location &a_loc, SymbolTable* a_symtab, char* a_name, expr_t** a_args, size_t n_args, stmt_t** a_body, size_t n_body, abiType a_abi, accessType a_access, deftypeType a_deftype, char* a_bindc_name, bool a_pure, bool a_module) {
        Subroutine_t *n;
        n = al.make_new<Subroutine_t>();
        n->base.type = symbolType::Subroutine;
        n->base.base.type = asrType::symbol;
        n->base.base.loc = a_loc;
        n->m_symtab = a_symtab;
        a_symtab->asr_owner = (asr_t*)n;
        n->m_name = a_name;
        n->m_args = a_args;
        n->n_args = n_args;
        n->m_body = a_body;
        n->n_body = n_body;
        n->m_abi = a_abi;
        n->m_access = a_access;
        n->m_deftype = a_deftype;
        n->m_bindc_name = a_bindc_name;
        n->m_pure = a_pure;
        n->m_module = a_module;
        return (asr_t*)n;
    }

    struct Function_t // Constructor
    {
        const static symbolType class_type = symbolType::Function;
        typedef symbol_t parent_type;
        symbol_t base;
        SymbolTable* m_symtab;
        char* m_name;
        expr_t** m_args; size_t n_args; // Sequence
        stmt_t** m_body; size_t n_body; // Sequence
        expr_t* m_return_var;
        abiType m_abi;
        accessType m_access;
        deftypeType m_deftype;
        char* m_bindc_name;
    };
    static inline asr_t* make_Function_t(Allocator &al, const Location &a_loc, SymbolTable* a_symtab, char* a_name, expr_t** a_args, size_t n_args, stmt_t** a_body, size_t n_body, expr_t* a_return_var, abiType a_abi, accessType a_access, deftypeType a_deftype, char* a_bindc_name) {
        Function_t *n;
        n = al.make_new<Function_t>();
        n->base.type = symbolType::Function;
        n->base.base.type = asrType::symbol;
        n->base.base.loc = a_loc;
        n->m_symtab = a_symtab;
        a_symtab->asr_owner = (asr_t*)n;
        n->m_name = a_name;
        n->m_args = a_args;
        n->n_args = n_args;
        n->m_body = a_body;
        n->n_body = n_body;
        n->m_return_var = a_return_var;
        n->m_abi = a_abi;
        n->m_access = a_access;
        n->m_deftype = a_deftype;
        n->m_bindc_name = a_bindc_name;
        return (asr_t*)n;
    }

    struct GenericProcedure_t // Constructor
    {
        const static symbolType class_type = symbolType::GenericProcedure;
        typedef symbol_t parent_type;
        symbol_t base;
        SymbolTable* m_parent_symtab;
        char* m_name;
        symbol_t** m_procs; size_t n_procs; // Sequence
        accessType m_access;
    };
    static inline asr_t* make_GenericProcedure_t(Allocator &al, const Location &a_loc, SymbolTable* a_parent_symtab, char* a_name, symbol_t** a_procs, size_t n_procs, accessType a_access) {
        GenericProcedure_t *n;
        n = al.make_new<GenericProcedure_t>();
        n->base.type = symbolType::GenericProcedure;
        n->base.base.type = asrType::symbol;
        n->base.base.loc = a_loc;
        n->m_parent_symtab = a_parent_symtab;
        n->m_name = a_name;
        n->m_procs = a_procs;
        n->n_procs = n_procs;
        n->m_access = a_access;
        return (asr_t*)n;
    }

    struct CustomOperator_t // Constructor
    {
        const static symbolType class_type = symbolType::CustomOperator;
        typedef symbol_t parent_type;
        symbol_t base;
        SymbolTable* m_parent_symtab;
        char* m_name;
        symbol_t** m_procs; size_t n_procs; // Sequence
        accessType m_access;
    };
    static inline asr_t* make_CustomOperator_t(Allocator &al, const Location &a_loc, SymbolTable* a_parent_symtab, char* a_name, symbol_t** a_procs, size_t n_procs, accessType a_access) {
        CustomOperator_t *n;
        n = al.make_new<CustomOperator_t>();
        n->base.type = symbolType::CustomOperator;
        n->base.base.type = asrType::symbol;
        n->base.base.loc = a_loc;
        n->m_parent_symtab = a_parent_symtab;
        n->m_name = a_name;
        n->m_procs = a_procs;
        n->n_procs = n_procs;
        n->m_access = a_access;
        return (asr_t*)n;
    }

    struct ExternalSymbol_t // Constructor
    {
        const static symbolType class_type = symbolType::ExternalSymbol;
        typedef symbol_t parent_type;
        symbol_t base;
        SymbolTable* m_parent_symtab;
        char* m_name;
        symbol_t* m_external;
        char* m_module_name;
        char** m_scope_names; size_t n_scope_names; // Sequence
        char* m_original_name;
        accessType m_access;
    };
    static inline asr_t* make_ExternalSymbol_t(Allocator &al, const Location &a_loc, SymbolTable* a_parent_symtab, char* a_name, symbol_t* a_external, char* a_module_name, char** a_scope_names, size_t n_scope_names, char* a_original_name, accessType a_access) {
        ExternalSymbol_t *n;
        n = al.make_new<ExternalSymbol_t>();
        n->base.type = symbolType::ExternalSymbol;
        n->base.base.type = asrType::symbol;
        n->base.base.loc = a_loc;
        n->m_parent_symtab = a_parent_symtab;
        n->m_name = a_name;
        n->m_external = a_external;
        n->m_module_name = a_module_name;
        n->m_scope_names = a_scope_names;
        n->n_scope_names = n_scope_names;
        n->m_original_name = a_original_name;
        n->m_access = a_access;
        return (asr_t*)n;
    }

    struct DerivedType_t // Constructor
    {
        const static symbolType class_type = symbolType::DerivedType;
        typedef symbol_t parent_type;
        symbol_t base;
        SymbolTable* m_symtab;
        char* m_name;
        char** m_members; size_t n_members; // Sequence
        abiType m_abi;
        accessType m_access;
        symbol_t* m_parent;
    };
    static inline asr_t* make_DerivedType_t(Allocator &al, const Location &a_loc, SymbolTable* a_symtab, char* a_name, char** a_members, size_t n_members, abiType a_abi, accessType a_access, symbol_t* a_parent) {
        DerivedType_t *n;
        n = al.make_new<DerivedType_t>();
        n->base.type = symbolType::DerivedType;
        n->base.base.type = asrType::symbol;
        n->base.base.loc = a_loc;
        n->m_symtab = a_symtab;
        a_symtab->asr_owner = (asr_t*)n;
        n->m_name = a_name;
        n->m_members = a_members;
        n->n_members = n_members;
        n->m_abi = a_abi;
        n->m_access = a_access;
        n->m_parent = a_parent;
        return (asr_t*)n;
    }

    struct Variable_t // Constructor
    {
        const static symbolType class_type = symbolType::Variable;
        typedef symbol_t parent_type;
        symbol_t base;
        SymbolTable* m_parent_symtab;
        char* m_name;
        intentType m_intent;
        expr_t* m_symbolic_value;
        expr_t* m_value;
        storage_typeType m_storage;
        ttype_t* m_type;
        abiType m_abi;
        accessType m_access;
        presenceType m_presence;
        bool m_value_attr;
    };
    static inline asr_t* make_Variable_t(Allocator &al, const Location &a_loc, SymbolTable* a_parent_symtab, char* a_name, intentType a_intent, expr_t* a_symbolic_value, expr_t* a_value, storage_typeType a_storage, ttype_t* a_type, abiType a_abi, accessType a_access, presenceType a_presence, bool a_value_attr) {
        Variable_t *n;
        n = al.make_new<Variable_t>();
        n->base.type = symbolType::Variable;
        n->base.base.type = asrType::symbol;
        n->base.base.loc = a_loc;
        n->m_parent_symtab = a_parent_symtab;
        n->m_name = a_name;
        n->m_intent = a_intent;
        n->m_symbolic_value = a_symbolic_value;
        n->m_value = a_value;
        n->m_storage = a_storage;
        n->m_type = a_type;
        n->m_abi = a_abi;
        n->m_access = a_access;
        n->m_presence = a_presence;
        n->m_value_attr = a_value_attr;
        return (asr_t*)n;
    }

    struct ClassType_t // Constructor
    {
        const static symbolType class_type = symbolType::ClassType;
        typedef symbol_t parent_type;
        symbol_t base;
        SymbolTable* m_symtab;
        char* m_name;
        abiType m_abi;
        accessType m_access;
    };
    static inline asr_t* make_ClassType_t(Allocator &al, const Location &a_loc, SymbolTable* a_symtab, char* a_name, abiType a_abi, accessType a_access) {
        ClassType_t *n;
        n = al.make_new<ClassType_t>();
        n->base.type = symbolType::ClassType;
        n->base.base.type = asrType::symbol;
        n->base.base.loc = a_loc;
        n->m_symtab = a_symtab;
        a_symtab->asr_owner = (asr_t*)n;
        n->m_name = a_name;
        n->m_abi = a_abi;
        n->m_access = a_access;
        return (asr_t*)n;
    }

    struct ClassProcedure_t // Constructor
    {
        const static symbolType class_type = symbolType::ClassProcedure;
        typedef symbol_t parent_type;
        symbol_t base;
        SymbolTable* m_parent_symtab;
        char* m_name;
        char* m_proc_name;
        symbol_t* m_proc;
        abiType m_abi;
    };
    static inline asr_t* make_ClassProcedure_t(Allocator &al, const Location &a_loc, SymbolTable* a_parent_symtab, char* a_name, char* a_proc_name, symbol_t* a_proc, abiType a_abi) {
        ClassProcedure_t *n;
        n = al.make_new<ClassProcedure_t>();
        n->base.type = symbolType::ClassProcedure;
        n->base.base.type = asrType::symbol;
        n->base.base.loc = a_loc;
        n->m_parent_symtab = a_parent_symtab;
        n->m_name = a_name;
        n->m_proc_name = a_proc_name;
        n->m_proc = a_proc;
        n->m_abi = a_abi;
        return (asr_t*)n;
    }

    struct AssociateBlock_t // Constructor
    {
        const static symbolType class_type = symbolType::AssociateBlock;
        typedef symbol_t parent_type;
        symbol_t base;
        SymbolTable* m_symtab;
        char* m_name;
        stmt_t** m_body; size_t n_body; // Sequence
    };
    static inline asr_t* make_AssociateBlock_t(Allocator &al, const Location &a_loc, SymbolTable* a_symtab, char* a_name, stmt_t** a_body, size_t n_body) {
        AssociateBlock_t *n;
        n = al.make_new<AssociateBlock_t>();
        n->base.type = symbolType::AssociateBlock;
        n->base.base.type = asrType::symbol;
        n->base.base.loc = a_loc;
        n->m_symtab = a_symtab;
        a_symtab->asr_owner = (asr_t*)n;
        n->m_name = a_name;
        n->m_body = a_body;
        n->n_body = n_body;
        return (asr_t*)n;
    }

    struct Block_t // Constructor
    {
        const static symbolType class_type = symbolType::Block;
        typedef symbol_t parent_type;
        symbol_t base;
        SymbolTable* m_symtab;
        char* m_name;
        stmt_t** m_body; size_t n_body; // Sequence
    };
    static inline asr_t* make_Block_t(Allocator &al, const Location &a_loc, SymbolTable* a_symtab, char* a_name, stmt_t** a_body, size_t n_body) {
        Block_t *n;
        n = al.make_new<Block_t>();
        n->base.type = symbolType::Block;
        n->base.base.type = asrType::symbol;
        n->base.base.loc = a_loc;
        n->m_symtab = a_symtab;
        a_symtab->asr_owner = (asr_t*)n;
        n->m_name = a_name;
        n->m_body = a_body;
        n->n_body = n_body;
        return (asr_t*)n;
    }



enum stmtType // Types
{
    Allocate, Assign, Assignment, Associate, Cycle, ExplicitDeallocate, ImplicitDeallocate, DoConcurrentLoop, DoLoop, ErrorStop, Exit, ForAllSingle, GoTo, GoToTarget, If, IfArithmetic, Print, FileOpen, FileClose, FileRead, FileRewind, FileInquire, FileWrite, Return, Select, Stop, Assert, SubroutineCall, Where, WhileLoop, Nullify, Flush, ListAppend, AssociateBlockCall, BlockCall, SetInsert, SetRemove, ListInsert, ListRemove, DictInsert
};

struct stmt_t // Sum
{
    const static asrType class_type = asrType::stmt;
    asr_t base;
    stmtType type;
};

    struct Allocate_t // Constructor
    {
        const static stmtType class_type = stmtType::Allocate;
        typedef stmt_t parent_type;
        stmt_t base;
        alloc_arg_t* m_args; size_t n_args; // Sequence
        expr_t* m_stat;
        expr_t* m_errmsg;
        expr_t* m_source;
    };
    static inline asr_t* make_Allocate_t(Allocator &al, const Location &a_loc, alloc_arg_t* a_args, size_t n_args, expr_t* a_stat, expr_t* a_errmsg, expr_t* a_source) {
        Allocate_t *n;
        n = al.make_new<Allocate_t>();
        n->base.type = stmtType::Allocate;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_args = a_args;
        n->n_args = n_args;
        n->m_stat = a_stat;
        n->m_errmsg = a_errmsg;
        n->m_source = a_source;
        return (asr_t*)n;
    }

    struct Assign_t // Constructor
    {
        const static stmtType class_type = stmtType::Assign;
        typedef stmt_t parent_type;
        stmt_t base;
        int64_t m_label;
        char* m_variable;
    };
    static inline asr_t* make_Assign_t(Allocator &al, const Location &a_loc, int64_t a_label, char* a_variable) {
        Assign_t *n;
        n = al.make_new<Assign_t>();
        n->base.type = stmtType::Assign;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_label = a_label;
        n->m_variable = a_variable;
        return (asr_t*)n;
    }

    struct Assignment_t // Constructor
    {
        const static stmtType class_type = stmtType::Assignment;
        typedef stmt_t parent_type;
        stmt_t base;
        expr_t* m_target;
        expr_t* m_value;
        stmt_t* m_overloaded;
    };
    static inline asr_t* make_Assignment_t(Allocator &al, const Location &a_loc, expr_t* a_target, expr_t* a_value, stmt_t* a_overloaded) {
        Assignment_t *n;
        n = al.make_new<Assignment_t>();
        n->base.type = stmtType::Assignment;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_target = a_target;
        n->m_value = a_value;
        n->m_overloaded = a_overloaded;
        return (asr_t*)n;
    }

    struct Associate_t // Constructor
    {
        const static stmtType class_type = stmtType::Associate;
        typedef stmt_t parent_type;
        stmt_t base;
        expr_t* m_target;
        expr_t* m_value;
    };
    static inline asr_t* make_Associate_t(Allocator &al, const Location &a_loc, expr_t* a_target, expr_t* a_value) {
        Associate_t *n;
        n = al.make_new<Associate_t>();
        n->base.type = stmtType::Associate;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_target = a_target;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct Cycle_t // Constructor
    {
        const static stmtType class_type = stmtType::Cycle;
        typedef stmt_t parent_type;
        stmt_t base;
    };
    static inline asr_t* make_Cycle_t(Allocator &al, const Location &a_loc) {
        Cycle_t *n;
        n = al.make_new<Cycle_t>();
        n->base.type = stmtType::Cycle;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        return (asr_t*)n;
    }

    struct ExplicitDeallocate_t // Constructor
    {
        const static stmtType class_type = stmtType::ExplicitDeallocate;
        typedef stmt_t parent_type;
        stmt_t base;
        symbol_t** m_vars; size_t n_vars; // Sequence
    };
    static inline asr_t* make_ExplicitDeallocate_t(Allocator &al, const Location &a_loc, symbol_t** a_vars, size_t n_vars) {
        ExplicitDeallocate_t *n;
        n = al.make_new<ExplicitDeallocate_t>();
        n->base.type = stmtType::ExplicitDeallocate;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_vars = a_vars;
        n->n_vars = n_vars;
        return (asr_t*)n;
    }

    struct ImplicitDeallocate_t // Constructor
    {
        const static stmtType class_type = stmtType::ImplicitDeallocate;
        typedef stmt_t parent_type;
        stmt_t base;
        symbol_t** m_vars; size_t n_vars; // Sequence
    };
    static inline asr_t* make_ImplicitDeallocate_t(Allocator &al, const Location &a_loc, symbol_t** a_vars, size_t n_vars) {
        ImplicitDeallocate_t *n;
        n = al.make_new<ImplicitDeallocate_t>();
        n->base.type = stmtType::ImplicitDeallocate;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_vars = a_vars;
        n->n_vars = n_vars;
        return (asr_t*)n;
    }

    struct DoConcurrentLoop_t // Constructor
    {
        const static stmtType class_type = stmtType::DoConcurrentLoop;
        typedef stmt_t parent_type;
        stmt_t base;
        do_loop_head_t m_head;
        stmt_t** m_body; size_t n_body; // Sequence
    };
    static inline asr_t* make_DoConcurrentLoop_t(Allocator &al, const Location &a_loc, do_loop_head_t a_head, stmt_t** a_body, size_t n_body) {
        DoConcurrentLoop_t *n;
        n = al.make_new<DoConcurrentLoop_t>();
        n->base.type = stmtType::DoConcurrentLoop;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_head = a_head;
        n->m_body = a_body;
        n->n_body = n_body;
        return (asr_t*)n;
    }

    struct DoLoop_t // Constructor
    {
        const static stmtType class_type = stmtType::DoLoop;
        typedef stmt_t parent_type;
        stmt_t base;
        do_loop_head_t m_head;
        stmt_t** m_body; size_t n_body; // Sequence
    };
    static inline asr_t* make_DoLoop_t(Allocator &al, const Location &a_loc, do_loop_head_t a_head, stmt_t** a_body, size_t n_body) {
        DoLoop_t *n;
        n = al.make_new<DoLoop_t>();
        n->base.type = stmtType::DoLoop;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_head = a_head;
        n->m_body = a_body;
        n->n_body = n_body;
        return (asr_t*)n;
    }

    struct ErrorStop_t // Constructor
    {
        const static stmtType class_type = stmtType::ErrorStop;
        typedef stmt_t parent_type;
        stmt_t base;
        expr_t* m_code;
    };
    static inline asr_t* make_ErrorStop_t(Allocator &al, const Location &a_loc, expr_t* a_code) {
        ErrorStop_t *n;
        n = al.make_new<ErrorStop_t>();
        n->base.type = stmtType::ErrorStop;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_code = a_code;
        return (asr_t*)n;
    }

    struct Exit_t // Constructor
    {
        const static stmtType class_type = stmtType::Exit;
        typedef stmt_t parent_type;
        stmt_t base;
    };
    static inline asr_t* make_Exit_t(Allocator &al, const Location &a_loc) {
        Exit_t *n;
        n = al.make_new<Exit_t>();
        n->base.type = stmtType::Exit;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        return (asr_t*)n;
    }

    struct ForAllSingle_t // Constructor
    {
        const static stmtType class_type = stmtType::ForAllSingle;
        typedef stmt_t parent_type;
        stmt_t base;
        do_loop_head_t m_head;
        stmt_t* m_assign_stmt;
    };
    static inline asr_t* make_ForAllSingle_t(Allocator &al, const Location &a_loc, do_loop_head_t a_head, stmt_t* a_assign_stmt) {
        ForAllSingle_t *n;
        n = al.make_new<ForAllSingle_t>();
        n->base.type = stmtType::ForAllSingle;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_head = a_head;
        n->m_assign_stmt = a_assign_stmt;
        return (asr_t*)n;
    }

    struct GoTo_t // Constructor
    {
        const static stmtType class_type = stmtType::GoTo;
        typedef stmt_t parent_type;
        stmt_t base;
        int64_t m_target_id;
    };
    static inline asr_t* make_GoTo_t(Allocator &al, const Location &a_loc, int64_t a_target_id) {
        GoTo_t *n;
        n = al.make_new<GoTo_t>();
        n->base.type = stmtType::GoTo;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_target_id = a_target_id;
        return (asr_t*)n;
    }

    struct GoToTarget_t // Constructor
    {
        const static stmtType class_type = stmtType::GoToTarget;
        typedef stmt_t parent_type;
        stmt_t base;
        int64_t m_id;
    };
    static inline asr_t* make_GoToTarget_t(Allocator &al, const Location &a_loc, int64_t a_id) {
        GoToTarget_t *n;
        n = al.make_new<GoToTarget_t>();
        n->base.type = stmtType::GoToTarget;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_id = a_id;
        return (asr_t*)n;
    }

    struct If_t // Constructor
    {
        const static stmtType class_type = stmtType::If;
        typedef stmt_t parent_type;
        stmt_t base;
        expr_t* m_test;
        stmt_t** m_body; size_t n_body; // Sequence
        stmt_t** m_orelse; size_t n_orelse; // Sequence
    };
    static inline asr_t* make_If_t(Allocator &al, const Location &a_loc, expr_t* a_test, stmt_t** a_body, size_t n_body, stmt_t** a_orelse, size_t n_orelse) {
        If_t *n;
        n = al.make_new<If_t>();
        n->base.type = stmtType::If;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_test = a_test;
        n->m_body = a_body;
        n->n_body = n_body;
        n->m_orelse = a_orelse;
        n->n_orelse = n_orelse;
        return (asr_t*)n;
    }

    struct IfArithmetic_t // Constructor
    {
        const static stmtType class_type = stmtType::IfArithmetic;
        typedef stmt_t parent_type;
        stmt_t base;
        expr_t* m_test;
        int64_t m_lt_label;
        int64_t m_eq_label;
        int64_t m_gt_label;
    };
    static inline asr_t* make_IfArithmetic_t(Allocator &al, const Location &a_loc, expr_t* a_test, int64_t a_lt_label, int64_t a_eq_label, int64_t a_gt_label) {
        IfArithmetic_t *n;
        n = al.make_new<IfArithmetic_t>();
        n->base.type = stmtType::IfArithmetic;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_test = a_test;
        n->m_lt_label = a_lt_label;
        n->m_eq_label = a_eq_label;
        n->m_gt_label = a_gt_label;
        return (asr_t*)n;
    }

    struct Print_t // Constructor
    {
        const static stmtType class_type = stmtType::Print;
        typedef stmt_t parent_type;
        stmt_t base;
        expr_t* m_fmt;
        expr_t** m_values; size_t n_values; // Sequence
    };
    static inline asr_t* make_Print_t(Allocator &al, const Location &a_loc, expr_t* a_fmt, expr_t** a_values, size_t n_values) {
        Print_t *n;
        n = al.make_new<Print_t>();
        n->base.type = stmtType::Print;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_fmt = a_fmt;
        n->m_values = a_values;
        n->n_values = n_values;
        return (asr_t*)n;
    }

    struct FileOpen_t // Constructor
    {
        const static stmtType class_type = stmtType::FileOpen;
        typedef stmt_t parent_type;
        stmt_t base;
        int64_t m_label;
        expr_t* m_newunit;
        expr_t* m_filename;
        expr_t* m_status;
    };
    static inline asr_t* make_FileOpen_t(Allocator &al, const Location &a_loc, int64_t a_label, expr_t* a_newunit, expr_t* a_filename, expr_t* a_status) {
        FileOpen_t *n;
        n = al.make_new<FileOpen_t>();
        n->base.type = stmtType::FileOpen;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_label = a_label;
        n->m_newunit = a_newunit;
        n->m_filename = a_filename;
        n->m_status = a_status;
        return (asr_t*)n;
    }

    struct FileClose_t // Constructor
    {
        const static stmtType class_type = stmtType::FileClose;
        typedef stmt_t parent_type;
        stmt_t base;
        int64_t m_label;
        expr_t* m_unit;
        expr_t* m_iostat;
        expr_t* m_iomsg;
        expr_t* m_err;
        expr_t* m_status;
    };
    static inline asr_t* make_FileClose_t(Allocator &al, const Location &a_loc, int64_t a_label, expr_t* a_unit, expr_t* a_iostat, expr_t* a_iomsg, expr_t* a_err, expr_t* a_status) {
        FileClose_t *n;
        n = al.make_new<FileClose_t>();
        n->base.type = stmtType::FileClose;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_label = a_label;
        n->m_unit = a_unit;
        n->m_iostat = a_iostat;
        n->m_iomsg = a_iomsg;
        n->m_err = a_err;
        n->m_status = a_status;
        return (asr_t*)n;
    }

    struct FileRead_t // Constructor
    {
        const static stmtType class_type = stmtType::FileRead;
        typedef stmt_t parent_type;
        stmt_t base;
        int64_t m_label;
        expr_t* m_unit;
        expr_t* m_fmt;
        expr_t* m_iomsg;
        expr_t* m_iostat;
        expr_t* m_id;
        expr_t** m_values; size_t n_values; // Sequence
    };
    static inline asr_t* make_FileRead_t(Allocator &al, const Location &a_loc, int64_t a_label, expr_t* a_unit, expr_t* a_fmt, expr_t* a_iomsg, expr_t* a_iostat, expr_t* a_id, expr_t** a_values, size_t n_values) {
        FileRead_t *n;
        n = al.make_new<FileRead_t>();
        n->base.type = stmtType::FileRead;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_label = a_label;
        n->m_unit = a_unit;
        n->m_fmt = a_fmt;
        n->m_iomsg = a_iomsg;
        n->m_iostat = a_iostat;
        n->m_id = a_id;
        n->m_values = a_values;
        n->n_values = n_values;
        return (asr_t*)n;
    }

    struct FileRewind_t // Constructor
    {
        const static stmtType class_type = stmtType::FileRewind;
        typedef stmt_t parent_type;
        stmt_t base;
        int64_t m_label;
        expr_t* m_unit;
        expr_t* m_iostat;
        expr_t* m_err;
    };
    static inline asr_t* make_FileRewind_t(Allocator &al, const Location &a_loc, int64_t a_label, expr_t* a_unit, expr_t* a_iostat, expr_t* a_err) {
        FileRewind_t *n;
        n = al.make_new<FileRewind_t>();
        n->base.type = stmtType::FileRewind;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_label = a_label;
        n->m_unit = a_unit;
        n->m_iostat = a_iostat;
        n->m_err = a_err;
        return (asr_t*)n;
    }

    struct FileInquire_t // Constructor
    {
        const static stmtType class_type = stmtType::FileInquire;
        typedef stmt_t parent_type;
        stmt_t base;
        int64_t m_label;
        expr_t* m_unit;
        expr_t* m_file;
        expr_t* m_iostat;
        expr_t* m_err;
        expr_t* m_exist;
        expr_t* m_opened;
        expr_t* m_number;
        expr_t* m_named;
        expr_t* m_name;
        expr_t* m_access;
        expr_t* m_sequential;
        expr_t* m_direct;
        expr_t* m_form;
        expr_t* m_formatted;
        expr_t* m_unformatted;
        expr_t* m_recl;
        expr_t* m_nextrec;
        expr_t* m_blank;
        expr_t* m_position;
        expr_t* m_action;
        expr_t* m_read;
        expr_t* m_write;
        expr_t* m_readwrite;
        expr_t* m_delim;
        expr_t* m_pad;
        expr_t* m_flen;
        expr_t* m_blocksize;
        expr_t* m_convert;
        expr_t* m_carriagecontrol;
        expr_t* m_iolength;
    };
    static inline asr_t* make_FileInquire_t(Allocator &al, const Location &a_loc, int64_t a_label, expr_t* a_unit, expr_t* a_file, expr_t* a_iostat, expr_t* a_err, expr_t* a_exist, expr_t* a_opened, expr_t* a_number, expr_t* a_named, expr_t* a_name, expr_t* a_access, expr_t* a_sequential, expr_t* a_direct, expr_t* a_form, expr_t* a_formatted, expr_t* a_unformatted, expr_t* a_recl, expr_t* a_nextrec, expr_t* a_blank, expr_t* a_position, expr_t* a_action, expr_t* a_read, expr_t* a_write, expr_t* a_readwrite, expr_t* a_delim, expr_t* a_pad, expr_t* a_flen, expr_t* a_blocksize, expr_t* a_convert, expr_t* a_carriagecontrol, expr_t* a_iolength) {
        FileInquire_t *n;
        n = al.make_new<FileInquire_t>();
        n->base.type = stmtType::FileInquire;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_label = a_label;
        n->m_unit = a_unit;
        n->m_file = a_file;
        n->m_iostat = a_iostat;
        n->m_err = a_err;
        n->m_exist = a_exist;
        n->m_opened = a_opened;
        n->m_number = a_number;
        n->m_named = a_named;
        n->m_name = a_name;
        n->m_access = a_access;
        n->m_sequential = a_sequential;
        n->m_direct = a_direct;
        n->m_form = a_form;
        n->m_formatted = a_formatted;
        n->m_unformatted = a_unformatted;
        n->m_recl = a_recl;
        n->m_nextrec = a_nextrec;
        n->m_blank = a_blank;
        n->m_position = a_position;
        n->m_action = a_action;
        n->m_read = a_read;
        n->m_write = a_write;
        n->m_readwrite = a_readwrite;
        n->m_delim = a_delim;
        n->m_pad = a_pad;
        n->m_flen = a_flen;
        n->m_blocksize = a_blocksize;
        n->m_convert = a_convert;
        n->m_carriagecontrol = a_carriagecontrol;
        n->m_iolength = a_iolength;
        return (asr_t*)n;
    }

    struct FileWrite_t // Constructor
    {
        const static stmtType class_type = stmtType::FileWrite;
        typedef stmt_t parent_type;
        stmt_t base;
        int64_t m_label;
        expr_t* m_unit;
        expr_t* m_fmt;
        expr_t* m_iomsg;
        expr_t* m_iostat;
        expr_t* m_id;
        expr_t** m_values; size_t n_values; // Sequence
    };
    static inline asr_t* make_FileWrite_t(Allocator &al, const Location &a_loc, int64_t a_label, expr_t* a_unit, expr_t* a_fmt, expr_t* a_iomsg, expr_t* a_iostat, expr_t* a_id, expr_t** a_values, size_t n_values) {
        FileWrite_t *n;
        n = al.make_new<FileWrite_t>();
        n->base.type = stmtType::FileWrite;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_label = a_label;
        n->m_unit = a_unit;
        n->m_fmt = a_fmt;
        n->m_iomsg = a_iomsg;
        n->m_iostat = a_iostat;
        n->m_id = a_id;
        n->m_values = a_values;
        n->n_values = n_values;
        return (asr_t*)n;
    }

    struct Return_t // Constructor
    {
        const static stmtType class_type = stmtType::Return;
        typedef stmt_t parent_type;
        stmt_t base;
    };
    static inline asr_t* make_Return_t(Allocator &al, const Location &a_loc) {
        Return_t *n;
        n = al.make_new<Return_t>();
        n->base.type = stmtType::Return;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        return (asr_t*)n;
    }

    struct Select_t // Constructor
    {
        const static stmtType class_type = stmtType::Select;
        typedef stmt_t parent_type;
        stmt_t base;
        expr_t* m_test;
        case_stmt_t** m_body; size_t n_body; // Sequence
        stmt_t** m_default; size_t n_default; // Sequence
    };
    static inline asr_t* make_Select_t(Allocator &al, const Location &a_loc, expr_t* a_test, case_stmt_t** a_body, size_t n_body, stmt_t** a_default, size_t n_default) {
        Select_t *n;
        n = al.make_new<Select_t>();
        n->base.type = stmtType::Select;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_test = a_test;
        n->m_body = a_body;
        n->n_body = n_body;
        n->m_default = a_default;
        n->n_default = n_default;
        return (asr_t*)n;
    }

    struct Stop_t // Constructor
    {
        const static stmtType class_type = stmtType::Stop;
        typedef stmt_t parent_type;
        stmt_t base;
        expr_t* m_code;
    };
    static inline asr_t* make_Stop_t(Allocator &al, const Location &a_loc, expr_t* a_code) {
        Stop_t *n;
        n = al.make_new<Stop_t>();
        n->base.type = stmtType::Stop;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_code = a_code;
        return (asr_t*)n;
    }

    struct Assert_t // Constructor
    {
        const static stmtType class_type = stmtType::Assert;
        typedef stmt_t parent_type;
        stmt_t base;
        expr_t* m_test;
        expr_t* m_msg;
    };
    static inline asr_t* make_Assert_t(Allocator &al, const Location &a_loc, expr_t* a_test, expr_t* a_msg) {
        Assert_t *n;
        n = al.make_new<Assert_t>();
        n->base.type = stmtType::Assert;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_test = a_test;
        n->m_msg = a_msg;
        return (asr_t*)n;
    }

    struct SubroutineCall_t // Constructor
    {
        const static stmtType class_type = stmtType::SubroutineCall;
        typedef stmt_t parent_type;
        stmt_t base;
        symbol_t* m_name;
        symbol_t* m_original_name;
        call_arg_t* m_args; size_t n_args; // Sequence
        expr_t* m_dt;
    };
    static inline asr_t* make_SubroutineCall_t(Allocator &al, const Location &a_loc, symbol_t* a_name, symbol_t* a_original_name, call_arg_t* a_args, size_t n_args, expr_t* a_dt) {
        SubroutineCall_t *n;
        n = al.make_new<SubroutineCall_t>();
        n->base.type = stmtType::SubroutineCall;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_name = a_name;
        n->m_original_name = a_original_name;
        n->m_args = a_args;
        n->n_args = n_args;
        n->m_dt = a_dt;
        return (asr_t*)n;
    }

    struct Where_t // Constructor
    {
        const static stmtType class_type = stmtType::Where;
        typedef stmt_t parent_type;
        stmt_t base;
        expr_t* m_test;
        stmt_t** m_body; size_t n_body; // Sequence
        stmt_t** m_orelse; size_t n_orelse; // Sequence
    };
    static inline asr_t* make_Where_t(Allocator &al, const Location &a_loc, expr_t* a_test, stmt_t** a_body, size_t n_body, stmt_t** a_orelse, size_t n_orelse) {
        Where_t *n;
        n = al.make_new<Where_t>();
        n->base.type = stmtType::Where;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_test = a_test;
        n->m_body = a_body;
        n->n_body = n_body;
        n->m_orelse = a_orelse;
        n->n_orelse = n_orelse;
        return (asr_t*)n;
    }

    struct WhileLoop_t // Constructor
    {
        const static stmtType class_type = stmtType::WhileLoop;
        typedef stmt_t parent_type;
        stmt_t base;
        expr_t* m_test;
        stmt_t** m_body; size_t n_body; // Sequence
    };
    static inline asr_t* make_WhileLoop_t(Allocator &al, const Location &a_loc, expr_t* a_test, stmt_t** a_body, size_t n_body) {
        WhileLoop_t *n;
        n = al.make_new<WhileLoop_t>();
        n->base.type = stmtType::WhileLoop;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_test = a_test;
        n->m_body = a_body;
        n->n_body = n_body;
        return (asr_t*)n;
    }

    struct Nullify_t // Constructor
    {
        const static stmtType class_type = stmtType::Nullify;
        typedef stmt_t parent_type;
        stmt_t base;
        symbol_t** m_vars; size_t n_vars; // Sequence
    };
    static inline asr_t* make_Nullify_t(Allocator &al, const Location &a_loc, symbol_t** a_vars, size_t n_vars) {
        Nullify_t *n;
        n = al.make_new<Nullify_t>();
        n->base.type = stmtType::Nullify;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_vars = a_vars;
        n->n_vars = n_vars;
        return (asr_t*)n;
    }

    struct Flush_t // Constructor
    {
        const static stmtType class_type = stmtType::Flush;
        typedef stmt_t parent_type;
        stmt_t base;
        int64_t m_label;
        expr_t* m_unit;
        expr_t* m_err;
        expr_t* m_iomsg;
        expr_t* m_iostat;
    };
    static inline asr_t* make_Flush_t(Allocator &al, const Location &a_loc, int64_t a_label, expr_t* a_unit, expr_t* a_err, expr_t* a_iomsg, expr_t* a_iostat) {
        Flush_t *n;
        n = al.make_new<Flush_t>();
        n->base.type = stmtType::Flush;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_label = a_label;
        n->m_unit = a_unit;
        n->m_err = a_err;
        n->m_iomsg = a_iomsg;
        n->m_iostat = a_iostat;
        return (asr_t*)n;
    }

    struct ListAppend_t // Constructor
    {
        const static stmtType class_type = stmtType::ListAppend;
        typedef stmt_t parent_type;
        stmt_t base;
        symbol_t* m_a;
        expr_t* m_ele;
    };
    static inline asr_t* make_ListAppend_t(Allocator &al, const Location &a_loc, symbol_t* a_a, expr_t* a_ele) {
        ListAppend_t *n;
        n = al.make_new<ListAppend_t>();
        n->base.type = stmtType::ListAppend;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_a = a_a;
        n->m_ele = a_ele;
        return (asr_t*)n;
    }

    struct AssociateBlockCall_t // Constructor
    {
        const static stmtType class_type = stmtType::AssociateBlockCall;
        typedef stmt_t parent_type;
        stmt_t base;
        symbol_t* m_m;
    };
    static inline asr_t* make_AssociateBlockCall_t(Allocator &al, const Location &a_loc, symbol_t* a_m) {
        AssociateBlockCall_t *n;
        n = al.make_new<AssociateBlockCall_t>();
        n->base.type = stmtType::AssociateBlockCall;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_m = a_m;
        return (asr_t*)n;
    }

    struct BlockCall_t // Constructor
    {
        const static stmtType class_type = stmtType::BlockCall;
        typedef stmt_t parent_type;
        stmt_t base;
        symbol_t* m_m;
    };
    static inline asr_t* make_BlockCall_t(Allocator &al, const Location &a_loc, symbol_t* a_m) {
        BlockCall_t *n;
        n = al.make_new<BlockCall_t>();
        n->base.type = stmtType::BlockCall;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_m = a_m;
        return (asr_t*)n;
    }

    struct SetInsert_t // Constructor
    {
        const static stmtType class_type = stmtType::SetInsert;
        typedef stmt_t parent_type;
        stmt_t base;
        symbol_t* m_a;
        expr_t* m_ele;
    };
    static inline asr_t* make_SetInsert_t(Allocator &al, const Location &a_loc, symbol_t* a_a, expr_t* a_ele) {
        SetInsert_t *n;
        n = al.make_new<SetInsert_t>();
        n->base.type = stmtType::SetInsert;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_a = a_a;
        n->m_ele = a_ele;
        return (asr_t*)n;
    }

    struct SetRemove_t // Constructor
    {
        const static stmtType class_type = stmtType::SetRemove;
        typedef stmt_t parent_type;
        stmt_t base;
        symbol_t* m_a;
        expr_t* m_ele;
    };
    static inline asr_t* make_SetRemove_t(Allocator &al, const Location &a_loc, symbol_t* a_a, expr_t* a_ele) {
        SetRemove_t *n;
        n = al.make_new<SetRemove_t>();
        n->base.type = stmtType::SetRemove;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_a = a_a;
        n->m_ele = a_ele;
        return (asr_t*)n;
    }

    struct ListInsert_t // Constructor
    {
        const static stmtType class_type = stmtType::ListInsert;
        typedef stmt_t parent_type;
        stmt_t base;
        symbol_t* m_a;
        expr_t* m_pos;
        expr_t* m_ele;
    };
    static inline asr_t* make_ListInsert_t(Allocator &al, const Location &a_loc, symbol_t* a_a, expr_t* a_pos, expr_t* a_ele) {
        ListInsert_t *n;
        n = al.make_new<ListInsert_t>();
        n->base.type = stmtType::ListInsert;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_a = a_a;
        n->m_pos = a_pos;
        n->m_ele = a_ele;
        return (asr_t*)n;
    }

    struct ListRemove_t // Constructor
    {
        const static stmtType class_type = stmtType::ListRemove;
        typedef stmt_t parent_type;
        stmt_t base;
        symbol_t* m_a;
        expr_t* m_ele;
    };
    static inline asr_t* make_ListRemove_t(Allocator &al, const Location &a_loc, symbol_t* a_a, expr_t* a_ele) {
        ListRemove_t *n;
        n = al.make_new<ListRemove_t>();
        n->base.type = stmtType::ListRemove;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_a = a_a;
        n->m_ele = a_ele;
        return (asr_t*)n;
    }

    struct DictInsert_t // Constructor
    {
        const static stmtType class_type = stmtType::DictInsert;
        typedef stmt_t parent_type;
        stmt_t base;
        symbol_t* m_a;
        expr_t* m_key;
        expr_t* m_value;
    };
    static inline asr_t* make_DictInsert_t(Allocator &al, const Location &a_loc, symbol_t* a_a, expr_t* a_key, expr_t* a_value) {
        DictInsert_t *n;
        n = al.make_new<DictInsert_t>();
        n->base.type = stmtType::DictInsert;
        n->base.base.type = asrType::stmt;
        n->base.base.loc = a_loc;
        n->m_a = a_a;
        n->m_key = a_key;
        n->m_value = a_value;
        return (asr_t*)n;
    }



enum exprType // Types
{
    BoolOp, BinOp, StrOp, UnaryOp, ComplexConstructor, NamedExpr, Compare, IfExp, FunctionCall, DerivedTypeConstructor, ImpliedDoLoop, IntegerConstant, IntegerBOZ, RealConstant, ComplexConstant, LogicalConstant, ListConstant, ListLen, ListConcat, ArrayConstant, SetConstant, SetLen, TupleConstant, TupleLen, StringConstant, StringConcat, StringLen, DictConstant, DictLen, Var, ArrayRef, ArraySize, ArrayBound, ArrayTranspose, ArrayMatMul, ArrayPack, Transfer, DerivedRef, Cast, ComplexRe, ComplexIm, DictItem
};

struct expr_t // Sum
{
    const static asrType class_type = asrType::expr;
    asr_t base;
    exprType type;
};

    struct BoolOp_t // Constructor
    {
        const static exprType class_type = exprType::BoolOp;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_left;
        boolopType m_op;
        expr_t* m_right;
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_BoolOp_t(Allocator &al, const Location &a_loc, expr_t* a_left, boolopType a_op, expr_t* a_right, ttype_t* a_type, expr_t* a_value) {
        BoolOp_t *n;
        n = al.make_new<BoolOp_t>();
        n->base.type = exprType::BoolOp;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_left = a_left;
        n->m_op = a_op;
        n->m_right = a_right;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct BinOp_t // Constructor
    {
        const static exprType class_type = exprType::BinOp;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_left;
        binopType m_op;
        expr_t* m_right;
        ttype_t* m_type;
        expr_t* m_value;
        expr_t* m_overloaded;
    };
    static inline asr_t* make_BinOp_t(Allocator &al, const Location &a_loc, expr_t* a_left, binopType a_op, expr_t* a_right, ttype_t* a_type, expr_t* a_value, expr_t* a_overloaded) {
        BinOp_t *n;
        n = al.make_new<BinOp_t>();
        n->base.type = exprType::BinOp;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_left = a_left;
        n->m_op = a_op;
        n->m_right = a_right;
        n->m_type = a_type;
        n->m_value = a_value;
        n->m_overloaded = a_overloaded;
        return (asr_t*)n;
    }

    struct StrOp_t // Constructor
    {
        const static exprType class_type = exprType::StrOp;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_left;
        stropType m_op;
        expr_t* m_right;
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_StrOp_t(Allocator &al, const Location &a_loc, expr_t* a_left, stropType a_op, expr_t* a_right, ttype_t* a_type, expr_t* a_value) {
        StrOp_t *n;
        n = al.make_new<StrOp_t>();
        n->base.type = exprType::StrOp;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_left = a_left;
        n->m_op = a_op;
        n->m_right = a_right;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct UnaryOp_t // Constructor
    {
        const static exprType class_type = exprType::UnaryOp;
        typedef expr_t parent_type;
        expr_t base;
        unaryopType m_op;
        expr_t* m_operand;
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_UnaryOp_t(Allocator &al, const Location &a_loc, unaryopType a_op, expr_t* a_operand, ttype_t* a_type, expr_t* a_value) {
        UnaryOp_t *n;
        n = al.make_new<UnaryOp_t>();
        n->base.type = exprType::UnaryOp;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_op = a_op;
        n->m_operand = a_operand;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct ComplexConstructor_t // Constructor
    {
        const static exprType class_type = exprType::ComplexConstructor;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_re;
        expr_t* m_im;
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_ComplexConstructor_t(Allocator &al, const Location &a_loc, expr_t* a_re, expr_t* a_im, ttype_t* a_type, expr_t* a_value) {
        ComplexConstructor_t *n;
        n = al.make_new<ComplexConstructor_t>();
        n->base.type = exprType::ComplexConstructor;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_re = a_re;
        n->m_im = a_im;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct NamedExpr_t // Constructor
    {
        const static exprType class_type = exprType::NamedExpr;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_target;
        expr_t* m_value;
        ttype_t* m_type;
    };
    static inline asr_t* make_NamedExpr_t(Allocator &al, const Location &a_loc, expr_t* a_target, expr_t* a_value, ttype_t* a_type) {
        NamedExpr_t *n;
        n = al.make_new<NamedExpr_t>();
        n->base.type = exprType::NamedExpr;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_target = a_target;
        n->m_value = a_value;
        n->m_type = a_type;
        return (asr_t*)n;
    }

    struct Compare_t // Constructor
    {
        const static exprType class_type = exprType::Compare;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_left;
        cmpopType m_op;
        expr_t* m_right;
        ttype_t* m_type;
        expr_t* m_value;
        expr_t* m_overloaded;
    };
    static inline asr_t* make_Compare_t(Allocator &al, const Location &a_loc, expr_t* a_left, cmpopType a_op, expr_t* a_right, ttype_t* a_type, expr_t* a_value, expr_t* a_overloaded) {
        Compare_t *n;
        n = al.make_new<Compare_t>();
        n->base.type = exprType::Compare;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_left = a_left;
        n->m_op = a_op;
        n->m_right = a_right;
        n->m_type = a_type;
        n->m_value = a_value;
        n->m_overloaded = a_overloaded;
        return (asr_t*)n;
    }

    struct IfExp_t // Constructor
    {
        const static exprType class_type = exprType::IfExp;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_test;
        expr_t* m_body;
        expr_t* m_orelse;
        ttype_t* m_type;
    };
    static inline asr_t* make_IfExp_t(Allocator &al, const Location &a_loc, expr_t* a_test, expr_t* a_body, expr_t* a_orelse, ttype_t* a_type) {
        IfExp_t *n;
        n = al.make_new<IfExp_t>();
        n->base.type = exprType::IfExp;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_test = a_test;
        n->m_body = a_body;
        n->m_orelse = a_orelse;
        n->m_type = a_type;
        return (asr_t*)n;
    }

    struct FunctionCall_t // Constructor
    {
        const static exprType class_type = exprType::FunctionCall;
        typedef expr_t parent_type;
        expr_t base;
        symbol_t* m_name;
        symbol_t* m_original_name;
        call_arg_t* m_args; size_t n_args; // Sequence
        ttype_t* m_type;
        expr_t* m_value;
        expr_t* m_dt;
    };
    static inline asr_t* make_FunctionCall_t(Allocator &al, const Location &a_loc, symbol_t* a_name, symbol_t* a_original_name, call_arg_t* a_args, size_t n_args, ttype_t* a_type, expr_t* a_value, expr_t* a_dt) {
        FunctionCall_t *n;
        n = al.make_new<FunctionCall_t>();
        n->base.type = exprType::FunctionCall;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_name = a_name;
        n->m_original_name = a_original_name;
        n->m_args = a_args;
        n->n_args = n_args;
        n->m_type = a_type;
        n->m_value = a_value;
        n->m_dt = a_dt;
        return (asr_t*)n;
    }

    struct DerivedTypeConstructor_t // Constructor
    {
        const static exprType class_type = exprType::DerivedTypeConstructor;
        typedef expr_t parent_type;
        expr_t base;
        symbol_t* m_dt_sym;
        expr_t** m_args; size_t n_args; // Sequence
        ttype_t* m_type;
    };
    static inline asr_t* make_DerivedTypeConstructor_t(Allocator &al, const Location &a_loc, symbol_t* a_dt_sym, expr_t** a_args, size_t n_args, ttype_t* a_type) {
        DerivedTypeConstructor_t *n;
        n = al.make_new<DerivedTypeConstructor_t>();
        n->base.type = exprType::DerivedTypeConstructor;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_dt_sym = a_dt_sym;
        n->m_args = a_args;
        n->n_args = n_args;
        n->m_type = a_type;
        return (asr_t*)n;
    }

    struct ImpliedDoLoop_t // Constructor
    {
        const static exprType class_type = exprType::ImpliedDoLoop;
        typedef expr_t parent_type;
        expr_t base;
        expr_t** m_values; size_t n_values; // Sequence
        expr_t* m_var;
        expr_t* m_start;
        expr_t* m_end;
        expr_t* m_increment;
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_ImpliedDoLoop_t(Allocator &al, const Location &a_loc, expr_t** a_values, size_t n_values, expr_t* a_var, expr_t* a_start, expr_t* a_end, expr_t* a_increment, ttype_t* a_type, expr_t* a_value) {
        ImpliedDoLoop_t *n;
        n = al.make_new<ImpliedDoLoop_t>();
        n->base.type = exprType::ImpliedDoLoop;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_values = a_values;
        n->n_values = n_values;
        n->m_var = a_var;
        n->m_start = a_start;
        n->m_end = a_end;
        n->m_increment = a_increment;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct IntegerConstant_t // Constructor
    {
        const static exprType class_type = exprType::IntegerConstant;
        typedef expr_t parent_type;
        expr_t base;
        int64_t m_n;
        ttype_t* m_type;
    };
    static inline asr_t* make_IntegerConstant_t(Allocator &al, const Location &a_loc, int64_t a_n, ttype_t* a_type) {
        IntegerConstant_t *n;
        n = al.make_new<IntegerConstant_t>();
        n->base.type = exprType::IntegerConstant;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_n = a_n;
        n->m_type = a_type;
        return (asr_t*)n;
    }

    struct IntegerBOZ_t // Constructor
    {
        const static exprType class_type = exprType::IntegerBOZ;
        typedef expr_t parent_type;
        expr_t base;
        int64_t m_v;
        integerbozType m_intboz_type;
        ttype_t* m_type;
    };
    static inline asr_t* make_IntegerBOZ_t(Allocator &al, const Location &a_loc, int64_t a_v, integerbozType a_intboz_type, ttype_t* a_type) {
        IntegerBOZ_t *n;
        n = al.make_new<IntegerBOZ_t>();
        n->base.type = exprType::IntegerBOZ;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_v = a_v;
        n->m_intboz_type = a_intboz_type;
        n->m_type = a_type;
        return (asr_t*)n;
    }

    struct RealConstant_t // Constructor
    {
        const static exprType class_type = exprType::RealConstant;
        typedef expr_t parent_type;
        expr_t base;
        double m_r;
        ttype_t* m_type;
    };
    static inline asr_t* make_RealConstant_t(Allocator &al, const Location &a_loc, double a_r, ttype_t* a_type) {
        RealConstant_t *n;
        n = al.make_new<RealConstant_t>();
        n->base.type = exprType::RealConstant;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_r = a_r;
        n->m_type = a_type;
        return (asr_t*)n;
    }

    struct ComplexConstant_t // Constructor
    {
        const static exprType class_type = exprType::ComplexConstant;
        typedef expr_t parent_type;
        expr_t base;
        double m_re;
        double m_im;
        ttype_t* m_type;
    };
    static inline asr_t* make_ComplexConstant_t(Allocator &al, const Location &a_loc, double a_re, double a_im, ttype_t* a_type) {
        ComplexConstant_t *n;
        n = al.make_new<ComplexConstant_t>();
        n->base.type = exprType::ComplexConstant;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_re = a_re;
        n->m_im = a_im;
        n->m_type = a_type;
        return (asr_t*)n;
    }

    struct LogicalConstant_t // Constructor
    {
        const static exprType class_type = exprType::LogicalConstant;
        typedef expr_t parent_type;
        expr_t base;
        bool m_value;
        ttype_t* m_type;
    };
    static inline asr_t* make_LogicalConstant_t(Allocator &al, const Location &a_loc, bool a_value, ttype_t* a_type) {
        LogicalConstant_t *n;
        n = al.make_new<LogicalConstant_t>();
        n->base.type = exprType::LogicalConstant;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_value = a_value;
        n->m_type = a_type;
        return (asr_t*)n;
    }

    struct ListConstant_t // Constructor
    {
        const static exprType class_type = exprType::ListConstant;
        typedef expr_t parent_type;
        expr_t base;
        expr_t** m_args; size_t n_args; // Sequence
        ttype_t* m_type;
    };
    static inline asr_t* make_ListConstant_t(Allocator &al, const Location &a_loc, expr_t** a_args, size_t n_args, ttype_t* a_type) {
        ListConstant_t *n;
        n = al.make_new<ListConstant_t>();
        n->base.type = exprType::ListConstant;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_args = a_args;
        n->n_args = n_args;
        n->m_type = a_type;
        return (asr_t*)n;
    }

    struct ListLen_t // Constructor
    {
        const static exprType class_type = exprType::ListLen;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_arg;
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_ListLen_t(Allocator &al, const Location &a_loc, expr_t* a_arg, ttype_t* a_type, expr_t* a_value) {
        ListLen_t *n;
        n = al.make_new<ListLen_t>();
        n->base.type = exprType::ListLen;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_arg = a_arg;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct ListConcat_t // Constructor
    {
        const static exprType class_type = exprType::ListConcat;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_left;
        expr_t* m_right;
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_ListConcat_t(Allocator &al, const Location &a_loc, expr_t* a_left, expr_t* a_right, ttype_t* a_type, expr_t* a_value) {
        ListConcat_t *n;
        n = al.make_new<ListConcat_t>();
        n->base.type = exprType::ListConcat;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_left = a_left;
        n->m_right = a_right;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct ArrayConstant_t // Constructor
    {
        const static exprType class_type = exprType::ArrayConstant;
        typedef expr_t parent_type;
        expr_t base;
        expr_t** m_args; size_t n_args; // Sequence
        ttype_t* m_type;
    };
    static inline asr_t* make_ArrayConstant_t(Allocator &al, const Location &a_loc, expr_t** a_args, size_t n_args, ttype_t* a_type) {
        ArrayConstant_t *n;
        n = al.make_new<ArrayConstant_t>();
        n->base.type = exprType::ArrayConstant;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_args = a_args;
        n->n_args = n_args;
        n->m_type = a_type;
        return (asr_t*)n;
    }

    struct SetConstant_t // Constructor
    {
        const static exprType class_type = exprType::SetConstant;
        typedef expr_t parent_type;
        expr_t base;
        expr_t** m_elements; size_t n_elements; // Sequence
        ttype_t* m_type;
    };
    static inline asr_t* make_SetConstant_t(Allocator &al, const Location &a_loc, expr_t** a_elements, size_t n_elements, ttype_t* a_type) {
        SetConstant_t *n;
        n = al.make_new<SetConstant_t>();
        n->base.type = exprType::SetConstant;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_elements = a_elements;
        n->n_elements = n_elements;
        n->m_type = a_type;
        return (asr_t*)n;
    }

    struct SetLen_t // Constructor
    {
        const static exprType class_type = exprType::SetLen;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_arg;
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_SetLen_t(Allocator &al, const Location &a_loc, expr_t* a_arg, ttype_t* a_type, expr_t* a_value) {
        SetLen_t *n;
        n = al.make_new<SetLen_t>();
        n->base.type = exprType::SetLen;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_arg = a_arg;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct TupleConstant_t // Constructor
    {
        const static exprType class_type = exprType::TupleConstant;
        typedef expr_t parent_type;
        expr_t base;
        expr_t** m_elements; size_t n_elements; // Sequence
        ttype_t* m_type;
    };
    static inline asr_t* make_TupleConstant_t(Allocator &al, const Location &a_loc, expr_t** a_elements, size_t n_elements, ttype_t* a_type) {
        TupleConstant_t *n;
        n = al.make_new<TupleConstant_t>();
        n->base.type = exprType::TupleConstant;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_elements = a_elements;
        n->n_elements = n_elements;
        n->m_type = a_type;
        return (asr_t*)n;
    }

    struct TupleLen_t // Constructor
    {
        const static exprType class_type = exprType::TupleLen;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_arg;
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_TupleLen_t(Allocator &al, const Location &a_loc, expr_t* a_arg, ttype_t* a_type, expr_t* a_value) {
        TupleLen_t *n;
        n = al.make_new<TupleLen_t>();
        n->base.type = exprType::TupleLen;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_arg = a_arg;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct StringConstant_t // Constructor
    {
        const static exprType class_type = exprType::StringConstant;
        typedef expr_t parent_type;
        expr_t base;
        char* m_s;
        ttype_t* m_type;
    };
    static inline asr_t* make_StringConstant_t(Allocator &al, const Location &a_loc, char* a_s, ttype_t* a_type) {
        StringConstant_t *n;
        n = al.make_new<StringConstant_t>();
        n->base.type = exprType::StringConstant;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_s = a_s;
        n->m_type = a_type;
        return (asr_t*)n;
    }

    struct StringConcat_t // Constructor
    {
        const static exprType class_type = exprType::StringConcat;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_left;
        expr_t* m_right;
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_StringConcat_t(Allocator &al, const Location &a_loc, expr_t* a_left, expr_t* a_right, ttype_t* a_type, expr_t* a_value) {
        StringConcat_t *n;
        n = al.make_new<StringConcat_t>();
        n->base.type = exprType::StringConcat;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_left = a_left;
        n->m_right = a_right;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct StringLen_t // Constructor
    {
        const static exprType class_type = exprType::StringLen;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_arg;
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_StringLen_t(Allocator &al, const Location &a_loc, expr_t* a_arg, ttype_t* a_type, expr_t* a_value) {
        StringLen_t *n;
        n = al.make_new<StringLen_t>();
        n->base.type = exprType::StringLen;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_arg = a_arg;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct DictConstant_t // Constructor
    {
        const static exprType class_type = exprType::DictConstant;
        typedef expr_t parent_type;
        expr_t base;
        expr_t** m_keys; size_t n_keys; // Sequence
        expr_t** m_values; size_t n_values; // Sequence
        ttype_t* m_type;
    };
    static inline asr_t* make_DictConstant_t(Allocator &al, const Location &a_loc, expr_t** a_keys, size_t n_keys, expr_t** a_values, size_t n_values, ttype_t* a_type) {
        DictConstant_t *n;
        n = al.make_new<DictConstant_t>();
        n->base.type = exprType::DictConstant;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_keys = a_keys;
        n->n_keys = n_keys;
        n->m_values = a_values;
        n->n_values = n_values;
        n->m_type = a_type;
        return (asr_t*)n;
    }

    struct DictLen_t // Constructor
    {
        const static exprType class_type = exprType::DictLen;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_arg;
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_DictLen_t(Allocator &al, const Location &a_loc, expr_t* a_arg, ttype_t* a_type, expr_t* a_value) {
        DictLen_t *n;
        n = al.make_new<DictLen_t>();
        n->base.type = exprType::DictLen;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_arg = a_arg;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct Var_t // Constructor
    {
        const static exprType class_type = exprType::Var;
        typedef expr_t parent_type;
        expr_t base;
        symbol_t* m_v;
    };
    static inline asr_t* make_Var_t(Allocator &al, const Location &a_loc, symbol_t* a_v) {
        Var_t *n;
        n = al.make_new<Var_t>();
        n->base.type = exprType::Var;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_v = a_v;
        return (asr_t*)n;
    }

    struct ArrayRef_t // Constructor
    {
        const static exprType class_type = exprType::ArrayRef;
        typedef expr_t parent_type;
        expr_t base;
        symbol_t* m_v;
        array_index_t* m_args; size_t n_args; // Sequence
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_ArrayRef_t(Allocator &al, const Location &a_loc, symbol_t* a_v, array_index_t* a_args, size_t n_args, ttype_t* a_type, expr_t* a_value) {
        ArrayRef_t *n;
        n = al.make_new<ArrayRef_t>();
        n->base.type = exprType::ArrayRef;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_v = a_v;
        n->m_args = a_args;
        n->n_args = n_args;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct ArraySize_t // Constructor
    {
        const static exprType class_type = exprType::ArraySize;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_v;
        expr_t* m_dim;
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_ArraySize_t(Allocator &al, const Location &a_loc, expr_t* a_v, expr_t* a_dim, ttype_t* a_type, expr_t* a_value) {
        ArraySize_t *n;
        n = al.make_new<ArraySize_t>();
        n->base.type = exprType::ArraySize;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_v = a_v;
        n->m_dim = a_dim;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct ArrayBound_t // Constructor
    {
        const static exprType class_type = exprType::ArrayBound;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_v;
        expr_t* m_dim;
        ttype_t* m_type;
        arrayboundType m_bound;
        expr_t* m_value;
    };
    static inline asr_t* make_ArrayBound_t(Allocator &al, const Location &a_loc, expr_t* a_v, expr_t* a_dim, ttype_t* a_type, arrayboundType a_bound, expr_t* a_value) {
        ArrayBound_t *n;
        n = al.make_new<ArrayBound_t>();
        n->base.type = exprType::ArrayBound;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_v = a_v;
        n->m_dim = a_dim;
        n->m_type = a_type;
        n->m_bound = a_bound;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct ArrayTranspose_t // Constructor
    {
        const static exprType class_type = exprType::ArrayTranspose;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_matrix;
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_ArrayTranspose_t(Allocator &al, const Location &a_loc, expr_t* a_matrix, ttype_t* a_type, expr_t* a_value) {
        ArrayTranspose_t *n;
        n = al.make_new<ArrayTranspose_t>();
        n->base.type = exprType::ArrayTranspose;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_matrix = a_matrix;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct ArrayMatMul_t // Constructor
    {
        const static exprType class_type = exprType::ArrayMatMul;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_matrix_a;
        expr_t* m_matrix_b;
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_ArrayMatMul_t(Allocator &al, const Location &a_loc, expr_t* a_matrix_a, expr_t* a_matrix_b, ttype_t* a_type, expr_t* a_value) {
        ArrayMatMul_t *n;
        n = al.make_new<ArrayMatMul_t>();
        n->base.type = exprType::ArrayMatMul;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_matrix_a = a_matrix_a;
        n->m_matrix_b = a_matrix_b;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct ArrayPack_t // Constructor
    {
        const static exprType class_type = exprType::ArrayPack;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_array;
        expr_t* m_mask;
        expr_t* m_vector;
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_ArrayPack_t(Allocator &al, const Location &a_loc, expr_t* a_array, expr_t* a_mask, expr_t* a_vector, ttype_t* a_type, expr_t* a_value) {
        ArrayPack_t *n;
        n = al.make_new<ArrayPack_t>();
        n->base.type = exprType::ArrayPack;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_array = a_array;
        n->m_mask = a_mask;
        n->m_vector = a_vector;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct Transfer_t // Constructor
    {
        const static exprType class_type = exprType::Transfer;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_source;
        expr_t* m_mold;
        expr_t* m_size;
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_Transfer_t(Allocator &al, const Location &a_loc, expr_t* a_source, expr_t* a_mold, expr_t* a_size, ttype_t* a_type, expr_t* a_value) {
        Transfer_t *n;
        n = al.make_new<Transfer_t>();
        n->base.type = exprType::Transfer;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_source = a_source;
        n->m_mold = a_mold;
        n->m_size = a_size;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct DerivedRef_t // Constructor
    {
        const static exprType class_type = exprType::DerivedRef;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_v;
        symbol_t* m_m;
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_DerivedRef_t(Allocator &al, const Location &a_loc, expr_t* a_v, symbol_t* a_m, ttype_t* a_type, expr_t* a_value) {
        DerivedRef_t *n;
        n = al.make_new<DerivedRef_t>();
        n->base.type = exprType::DerivedRef;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_v = a_v;
        n->m_m = a_m;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct Cast_t // Constructor
    {
        const static exprType class_type = exprType::Cast;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_arg;
        cast_kindType m_kind;
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_Cast_t(Allocator &al, const Location &a_loc, expr_t* a_arg, cast_kindType a_kind, ttype_t* a_type, expr_t* a_value) {
        Cast_t *n;
        n = al.make_new<Cast_t>();
        n->base.type = exprType::Cast;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_arg = a_arg;
        n->m_kind = a_kind;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct ComplexRe_t // Constructor
    {
        const static exprType class_type = exprType::ComplexRe;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_arg;
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_ComplexRe_t(Allocator &al, const Location &a_loc, expr_t* a_arg, ttype_t* a_type, expr_t* a_value) {
        ComplexRe_t *n;
        n = al.make_new<ComplexRe_t>();
        n->base.type = exprType::ComplexRe;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_arg = a_arg;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct ComplexIm_t // Constructor
    {
        const static exprType class_type = exprType::ComplexIm;
        typedef expr_t parent_type;
        expr_t base;
        expr_t* m_arg;
        ttype_t* m_type;
        expr_t* m_value;
    };
    static inline asr_t* make_ComplexIm_t(Allocator &al, const Location &a_loc, expr_t* a_arg, ttype_t* a_type, expr_t* a_value) {
        ComplexIm_t *n;
        n = al.make_new<ComplexIm_t>();
        n->base.type = exprType::ComplexIm;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_arg = a_arg;
        n->m_type = a_type;
        n->m_value = a_value;
        return (asr_t*)n;
    }

    struct DictItem_t // Constructor
    {
        const static exprType class_type = exprType::DictItem;
        typedef expr_t parent_type;
        expr_t base;
        symbol_t* m_a;
        expr_t* m_key;
        expr_t* m_default;
        ttype_t* m_type;
    };
    static inline asr_t* make_DictItem_t(Allocator &al, const Location &a_loc, symbol_t* a_a, expr_t* a_key, expr_t* a_default, ttype_t* a_type) {
        DictItem_t *n;
        n = al.make_new<DictItem_t>();
        n->base.type = exprType::DictItem;
        n->base.base.type = asrType::expr;
        n->base.base.loc = a_loc;
        n->m_a = a_a;
        n->m_key = a_key;
        n->m_default = a_default;
        n->m_type = a_type;
        return (asr_t*)n;
    }



enum ttypeType // Types
{
    Integer, Real, Complex, Character, Logical, Set, List, Tuple, Derived, Class, Dict, Pointer
};

struct ttype_t // Sum
{
    const static asrType class_type = asrType::ttype;
    asr_t base;
    ttypeType type;
};

    struct Integer_t // Constructor
    {
        const static ttypeType class_type = ttypeType::Integer;
        typedef ttype_t parent_type;
        ttype_t base;
        int64_t m_kind;
        dimension_t* m_dims; size_t n_dims; // Sequence
    };
    static inline asr_t* make_Integer_t(Allocator &al, const Location &a_loc, int64_t a_kind, dimension_t* a_dims, size_t n_dims) {
        Integer_t *n;
        n = al.make_new<Integer_t>();
        n->base.type = ttypeType::Integer;
        n->base.base.type = asrType::ttype;
        n->base.base.loc = a_loc;
        n->m_kind = a_kind;
        n->m_dims = a_dims;
        n->n_dims = n_dims;
        return (asr_t*)n;
    }

    struct Real_t // Constructor
    {
        const static ttypeType class_type = ttypeType::Real;
        typedef ttype_t parent_type;
        ttype_t base;
        int64_t m_kind;
        dimension_t* m_dims; size_t n_dims; // Sequence
    };
    static inline asr_t* make_Real_t(Allocator &al, const Location &a_loc, int64_t a_kind, dimension_t* a_dims, size_t n_dims) {
        Real_t *n;
        n = al.make_new<Real_t>();
        n->base.type = ttypeType::Real;
        n->base.base.type = asrType::ttype;
        n->base.base.loc = a_loc;
        n->m_kind = a_kind;
        n->m_dims = a_dims;
        n->n_dims = n_dims;
        return (asr_t*)n;
    }

    struct Complex_t // Constructor
    {
        const static ttypeType class_type = ttypeType::Complex;
        typedef ttype_t parent_type;
        ttype_t base;
        int64_t m_kind;
        dimension_t* m_dims; size_t n_dims; // Sequence
    };
    static inline asr_t* make_Complex_t(Allocator &al, const Location &a_loc, int64_t a_kind, dimension_t* a_dims, size_t n_dims) {
        Complex_t *n;
        n = al.make_new<Complex_t>();
        n->base.type = ttypeType::Complex;
        n->base.base.type = asrType::ttype;
        n->base.base.loc = a_loc;
        n->m_kind = a_kind;
        n->m_dims = a_dims;
        n->n_dims = n_dims;
        return (asr_t*)n;
    }

    struct Character_t // Constructor
    {
        const static ttypeType class_type = ttypeType::Character;
        typedef ttype_t parent_type;
        ttype_t base;
        int64_t m_kind;
        int64_t m_len;
        expr_t* m_len_expr;
        dimension_t* m_dims; size_t n_dims; // Sequence
    };
    static inline asr_t* make_Character_t(Allocator &al, const Location &a_loc, int64_t a_kind, int64_t a_len, expr_t* a_len_expr, dimension_t* a_dims, size_t n_dims) {
        Character_t *n;
        n = al.make_new<Character_t>();
        n->base.type = ttypeType::Character;
        n->base.base.type = asrType::ttype;
        n->base.base.loc = a_loc;
        n->m_kind = a_kind;
        n->m_len = a_len;
        n->m_len_expr = a_len_expr;
        n->m_dims = a_dims;
        n->n_dims = n_dims;
        return (asr_t*)n;
    }

    struct Logical_t // Constructor
    {
        const static ttypeType class_type = ttypeType::Logical;
        typedef ttype_t parent_type;
        ttype_t base;
        int64_t m_kind;
        dimension_t* m_dims; size_t n_dims; // Sequence
    };
    static inline asr_t* make_Logical_t(Allocator &al, const Location &a_loc, int64_t a_kind, dimension_t* a_dims, size_t n_dims) {
        Logical_t *n;
        n = al.make_new<Logical_t>();
        n->base.type = ttypeType::Logical;
        n->base.base.type = asrType::ttype;
        n->base.base.loc = a_loc;
        n->m_kind = a_kind;
        n->m_dims = a_dims;
        n->n_dims = n_dims;
        return (asr_t*)n;
    }

    struct Set_t // Constructor
    {
        const static ttypeType class_type = ttypeType::Set;
        typedef ttype_t parent_type;
        ttype_t base;
        ttype_t* m_type;
    };
    static inline asr_t* make_Set_t(Allocator &al, const Location &a_loc, ttype_t* a_type) {
        Set_t *n;
        n = al.make_new<Set_t>();
        n->base.type = ttypeType::Set;
        n->base.base.type = asrType::ttype;
        n->base.base.loc = a_loc;
        n->m_type = a_type;
        return (asr_t*)n;
    }

    struct List_t // Constructor
    {
        const static ttypeType class_type = ttypeType::List;
        typedef ttype_t parent_type;
        ttype_t base;
        ttype_t* m_type;
    };
    static inline asr_t* make_List_t(Allocator &al, const Location &a_loc, ttype_t* a_type) {
        List_t *n;
        n = al.make_new<List_t>();
        n->base.type = ttypeType::List;
        n->base.base.type = asrType::ttype;
        n->base.base.loc = a_loc;
        n->m_type = a_type;
        return (asr_t*)n;
    }

    struct Tuple_t // Constructor
    {
        const static ttypeType class_type = ttypeType::Tuple;
        typedef ttype_t parent_type;
        ttype_t base;
        ttype_t** m_type; size_t n_type; // Sequence
    };
    static inline asr_t* make_Tuple_t(Allocator &al, const Location &a_loc, ttype_t** a_type, size_t n_type) {
        Tuple_t *n;
        n = al.make_new<Tuple_t>();
        n->base.type = ttypeType::Tuple;
        n->base.base.type = asrType::ttype;
        n->base.base.loc = a_loc;
        n->m_type = a_type;
        n->n_type = n_type;
        return (asr_t*)n;
    }

    struct Derived_t // Constructor
    {
        const static ttypeType class_type = ttypeType::Derived;
        typedef ttype_t parent_type;
        ttype_t base;
        symbol_t* m_derived_type;
        dimension_t* m_dims; size_t n_dims; // Sequence
    };
    static inline asr_t* make_Derived_t(Allocator &al, const Location &a_loc, symbol_t* a_derived_type, dimension_t* a_dims, size_t n_dims) {
        Derived_t *n;
        n = al.make_new<Derived_t>();
        n->base.type = ttypeType::Derived;
        n->base.base.type = asrType::ttype;
        n->base.base.loc = a_loc;
        n->m_derived_type = a_derived_type;
        n->m_dims = a_dims;
        n->n_dims = n_dims;
        return (asr_t*)n;
    }

    struct Class_t // Constructor
    {
        const static ttypeType class_type = ttypeType::Class;
        typedef ttype_t parent_type;
        ttype_t base;
        symbol_t* m_class_type;
        dimension_t* m_dims; size_t n_dims; // Sequence
    };
    static inline asr_t* make_Class_t(Allocator &al, const Location &a_loc, symbol_t* a_class_type, dimension_t* a_dims, size_t n_dims) {
        Class_t *n;
        n = al.make_new<Class_t>();
        n->base.type = ttypeType::Class;
        n->base.base.type = asrType::ttype;
        n->base.base.loc = a_loc;
        n->m_class_type = a_class_type;
        n->m_dims = a_dims;
        n->n_dims = n_dims;
        return (asr_t*)n;
    }

    struct Dict_t // Constructor
    {
        const static ttypeType class_type = ttypeType::Dict;
        typedef ttype_t parent_type;
        ttype_t base;
        ttype_t* m_key_type;
        ttype_t* m_value_type;
    };
    static inline asr_t* make_Dict_t(Allocator &al, const Location &a_loc, ttype_t* a_key_type, ttype_t* a_value_type) {
        Dict_t *n;
        n = al.make_new<Dict_t>();
        n->base.type = ttypeType::Dict;
        n->base.base.type = asrType::ttype;
        n->base.base.loc = a_loc;
        n->m_key_type = a_key_type;
        n->m_value_type = a_value_type;
        return (asr_t*)n;
    }

    struct Pointer_t // Constructor
    {
        const static ttypeType class_type = ttypeType::Pointer;
        typedef ttype_t parent_type;
        ttype_t base;
        ttype_t* m_type;
    };
    static inline asr_t* make_Pointer_t(Allocator &al, const Location &a_loc, ttype_t* a_type) {
        Pointer_t *n;
        n = al.make_new<Pointer_t>();
        n->base.type = ttypeType::Pointer;
        n->base.base.type = asrType::ttype;
        n->base.base.loc = a_loc;
        n->m_type = a_type;
        return (asr_t*)n;
    }



enum attributeType // Types
{
    Attribute
};

struct attribute_t // Sum
{
    const static asrType class_type = asrType::attribute;
    asr_t base;
    attributeType type;
};

    struct Attribute_t // Constructor
    {
        const static attributeType class_type = attributeType::Attribute;
        typedef attribute_t parent_type;
        attribute_t base;
        char* m_name;
        attribute_arg_t* m_args; size_t n_args; // Sequence
    };
    static inline asr_t* make_Attribute_t(Allocator &al, const Location &a_loc, char* a_name, attribute_arg_t* a_args, size_t n_args) {
        Attribute_t *n;
        n = al.make_new<Attribute_t>();
        n->base.type = attributeType::Attribute;
        n->base.base.type = asrType::attribute;
        n->base.base.loc = a_loc;
        n->m_name = a_name;
        n->m_args = a_args;
        n->n_args = n_args;
        return (asr_t*)n;
    }



enum tbindType // Types
{
    Bind
};

struct tbind_t // Sum
{
    const static asrType class_type = asrType::tbind;
    asr_t base;
    tbindType type;
};

    struct Bind_t // Constructor
    {
        const static tbindType class_type = tbindType::Bind;
        typedef tbind_t parent_type;
        tbind_t base;
        char* m_lang;
        char* m_name;
    };
    static inline asr_t* make_Bind_t(Allocator &al, const Location &a_loc, char* a_lang, char* a_name) {
        Bind_t *n;
        n = al.make_new<Bind_t>();
        n->base.type = tbindType::Bind;
        n->base.base.type = asrType::tbind;
        n->base.base.loc = a_loc;
        n->m_lang = a_lang;
        n->m_name = a_name;
        return (asr_t*)n;
    }



enum case_stmtType // Types
{
    CaseStmt, CaseStmt_Range
};

struct case_stmt_t // Sum
{
    const static asrType class_type = asrType::case_stmt;
    asr_t base;
    case_stmtType type;
};

    struct CaseStmt_t // Constructor
    {
        const static case_stmtType class_type = case_stmtType::CaseStmt;
        typedef case_stmt_t parent_type;
        case_stmt_t base;
        expr_t** m_test; size_t n_test; // Sequence
        stmt_t** m_body; size_t n_body; // Sequence
    };
    static inline asr_t* make_CaseStmt_t(Allocator &al, const Location &a_loc, expr_t** a_test, size_t n_test, stmt_t** a_body, size_t n_body) {
        CaseStmt_t *n;
        n = al.make_new<CaseStmt_t>();
        n->base.type = case_stmtType::CaseStmt;
        n->base.base.type = asrType::case_stmt;
        n->base.base.loc = a_loc;
        n->m_test = a_test;
        n->n_test = n_test;
        n->m_body = a_body;
        n->n_body = n_body;
        return (asr_t*)n;
    }

    struct CaseStmt_Range_t // Constructor
    {
        const static case_stmtType class_type = case_stmtType::CaseStmt_Range;
        typedef case_stmt_t parent_type;
        case_stmt_t base;
        expr_t* m_start;
        expr_t* m_end;
        stmt_t** m_body; size_t n_body; // Sequence
    };
    static inline asr_t* make_CaseStmt_Range_t(Allocator &al, const Location &a_loc, expr_t* a_start, expr_t* a_end, stmt_t** a_body, size_t n_body) {
        CaseStmt_Range_t *n;
        n = al.make_new<CaseStmt_Range_t>();
        n->base.type = case_stmtType::CaseStmt_Range;
        n->base.base.type = asrType::case_stmt;
        n->base.base.loc = a_loc;
        n->m_start = a_start;
        n->m_end = a_end;
        n->m_body = a_body;
        n->n_body = n_body;
        return (asr_t*)n;
    }





/******************************************************************************/
// Visitor functions

template <class Visitor>
static void visit_unit_t(const unit_t &x, Visitor &v) {
    LFORTRAN_ASSERT(x.base.type == asrType::unit)
    switch (x.type) {
        case unitType::TranslationUnit: { v.visit_TranslationUnit((const TranslationUnit_t &)x); return; }
    }
}

template <class Visitor>
static void visit_symbol_t(const symbol_t &x, Visitor &v) {
    LFORTRAN_ASSERT(x.base.type == asrType::symbol)
    switch (x.type) {
        case symbolType::Program: { v.visit_Program((const Program_t &)x); return; }
        case symbolType::Module: { v.visit_Module((const Module_t &)x); return; }
        case symbolType::Subroutine: { v.visit_Subroutine((const Subroutine_t &)x); return; }
        case symbolType::Function: { v.visit_Function((const Function_t &)x); return; }
        case symbolType::GenericProcedure: { v.visit_GenericProcedure((const GenericProcedure_t &)x); return; }
        case symbolType::CustomOperator: { v.visit_CustomOperator((const CustomOperator_t &)x); return; }
        case symbolType::ExternalSymbol: { v.visit_ExternalSymbol((const ExternalSymbol_t &)x); return; }
        case symbolType::DerivedType: { v.visit_DerivedType((const DerivedType_t &)x); return; }
        case symbolType::Variable: { v.visit_Variable((const Variable_t &)x); return; }
        case symbolType::ClassType: { v.visit_ClassType((const ClassType_t &)x); return; }
        case symbolType::ClassProcedure: { v.visit_ClassProcedure((const ClassProcedure_t &)x); return; }
        case symbolType::AssociateBlock: { v.visit_AssociateBlock((const AssociateBlock_t &)x); return; }
        case symbolType::Block: { v.visit_Block((const Block_t &)x); return; }
    }
}

template <class Visitor>
static void visit_stmt_t(const stmt_t &x, Visitor &v) {
    LFORTRAN_ASSERT(x.base.type == asrType::stmt)
    switch (x.type) {
        case stmtType::Allocate: { v.visit_Allocate((const Allocate_t &)x); return; }
        case stmtType::Assign: { v.visit_Assign((const Assign_t &)x); return; }
        case stmtType::Assignment: { v.visit_Assignment((const Assignment_t &)x); return; }
        case stmtType::Associate: { v.visit_Associate((const Associate_t &)x); return; }
        case stmtType::Cycle: { v.visit_Cycle((const Cycle_t &)x); return; }
        case stmtType::ExplicitDeallocate: { v.visit_ExplicitDeallocate((const ExplicitDeallocate_t &)x); return; }
        case stmtType::ImplicitDeallocate: { v.visit_ImplicitDeallocate((const ImplicitDeallocate_t &)x); return; }
        case stmtType::DoConcurrentLoop: { v.visit_DoConcurrentLoop((const DoConcurrentLoop_t &)x); return; }
        case stmtType::DoLoop: { v.visit_DoLoop((const DoLoop_t &)x); return; }
        case stmtType::ErrorStop: { v.visit_ErrorStop((const ErrorStop_t &)x); return; }
        case stmtType::Exit: { v.visit_Exit((const Exit_t &)x); return; }
        case stmtType::ForAllSingle: { v.visit_ForAllSingle((const ForAllSingle_t &)x); return; }
        case stmtType::GoTo: { v.visit_GoTo((const GoTo_t &)x); return; }
        case stmtType::GoToTarget: { v.visit_GoToTarget((const GoToTarget_t &)x); return; }
        case stmtType::If: { v.visit_If((const If_t &)x); return; }
        case stmtType::IfArithmetic: { v.visit_IfArithmetic((const IfArithmetic_t &)x); return; }
        case stmtType::Print: { v.visit_Print((const Print_t &)x); return; }
        case stmtType::FileOpen: { v.visit_FileOpen((const FileOpen_t &)x); return; }
        case stmtType::FileClose: { v.visit_FileClose((const FileClose_t &)x); return; }
        case stmtType::FileRead: { v.visit_FileRead((const FileRead_t &)x); return; }
        case stmtType::FileRewind: { v.visit_FileRewind((const FileRewind_t &)x); return; }
        case stmtType::FileInquire: { v.visit_FileInquire((const FileInquire_t &)x); return; }
        case stmtType::FileWrite: { v.visit_FileWrite((const FileWrite_t &)x); return; }
        case stmtType::Return: { v.visit_Return((const Return_t &)x); return; }
        case stmtType::Select: { v.visit_Select((const Select_t &)x); return; }
        case stmtType::Stop: { v.visit_Stop((const Stop_t &)x); return; }
        case stmtType::Assert: { v.visit_Assert((const Assert_t &)x); return; }
        case stmtType::SubroutineCall: { v.visit_SubroutineCall((const SubroutineCall_t &)x); return; }
        case stmtType::Where: { v.visit_Where((const Where_t &)x); return; }
        case stmtType::WhileLoop: { v.visit_WhileLoop((const WhileLoop_t &)x); return; }
        case stmtType::Nullify: { v.visit_Nullify((const Nullify_t &)x); return; }
        case stmtType::Flush: { v.visit_Flush((const Flush_t &)x); return; }
        case stmtType::ListAppend: { v.visit_ListAppend((const ListAppend_t &)x); return; }
        case stmtType::AssociateBlockCall: { v.visit_AssociateBlockCall((const AssociateBlockCall_t &)x); return; }
        case stmtType::BlockCall: { v.visit_BlockCall((const BlockCall_t &)x); return; }
        case stmtType::SetInsert: { v.visit_SetInsert((const SetInsert_t &)x); return; }
        case stmtType::SetRemove: { v.visit_SetRemove((const SetRemove_t &)x); return; }
        case stmtType::ListInsert: { v.visit_ListInsert((const ListInsert_t &)x); return; }
        case stmtType::ListRemove: { v.visit_ListRemove((const ListRemove_t &)x); return; }
        case stmtType::DictInsert: { v.visit_DictInsert((const DictInsert_t &)x); return; }
    }
}

template <class Visitor>
static void visit_expr_t(const expr_t &x, Visitor &v) {
    LFORTRAN_ASSERT(x.base.type == asrType::expr)
    switch (x.type) {
        case exprType::BoolOp: { v.visit_BoolOp((const BoolOp_t &)x); return; }
        case exprType::BinOp: { v.visit_BinOp((const BinOp_t &)x); return; }
        case exprType::StrOp: { v.visit_StrOp((const StrOp_t &)x); return; }
        case exprType::UnaryOp: { v.visit_UnaryOp((const UnaryOp_t &)x); return; }
        case exprType::ComplexConstructor: { v.visit_ComplexConstructor((const ComplexConstructor_t &)x); return; }
        case exprType::NamedExpr: { v.visit_NamedExpr((const NamedExpr_t &)x); return; }
        case exprType::Compare: { v.visit_Compare((const Compare_t &)x); return; }
        case exprType::IfExp: { v.visit_IfExp((const IfExp_t &)x); return; }
        case exprType::FunctionCall: { v.visit_FunctionCall((const FunctionCall_t &)x); return; }
        case exprType::DerivedTypeConstructor: { v.visit_DerivedTypeConstructor((const DerivedTypeConstructor_t &)x); return; }
        case exprType::ImpliedDoLoop: { v.visit_ImpliedDoLoop((const ImpliedDoLoop_t &)x); return; }
        case exprType::IntegerConstant: { v.visit_IntegerConstant((const IntegerConstant_t &)x); return; }
        case exprType::IntegerBOZ: { v.visit_IntegerBOZ((const IntegerBOZ_t &)x); return; }
        case exprType::RealConstant: { v.visit_RealConstant((const RealConstant_t &)x); return; }
        case exprType::ComplexConstant: { v.visit_ComplexConstant((const ComplexConstant_t &)x); return; }
        case exprType::LogicalConstant: { v.visit_LogicalConstant((const LogicalConstant_t &)x); return; }
        case exprType::ListConstant: { v.visit_ListConstant((const ListConstant_t &)x); return; }
        case exprType::ListLen: { v.visit_ListLen((const ListLen_t &)x); return; }
        case exprType::ListConcat: { v.visit_ListConcat((const ListConcat_t &)x); return; }
        case exprType::ArrayConstant: { v.visit_ArrayConstant((const ArrayConstant_t &)x); return; }
        case exprType::SetConstant: { v.visit_SetConstant((const SetConstant_t &)x); return; }
        case exprType::SetLen: { v.visit_SetLen((const SetLen_t &)x); return; }
        case exprType::TupleConstant: { v.visit_TupleConstant((const TupleConstant_t &)x); return; }
        case exprType::TupleLen: { v.visit_TupleLen((const TupleLen_t &)x); return; }
        case exprType::StringConstant: { v.visit_StringConstant((const StringConstant_t &)x); return; }
        case exprType::StringConcat: { v.visit_StringConcat((const StringConcat_t &)x); return; }
        case exprType::StringLen: { v.visit_StringLen((const StringLen_t &)x); return; }
        case exprType::DictConstant: { v.visit_DictConstant((const DictConstant_t &)x); return; }
        case exprType::DictLen: { v.visit_DictLen((const DictLen_t &)x); return; }
        case exprType::Var: { v.visit_Var((const Var_t &)x); return; }
        case exprType::ArrayRef: { v.visit_ArrayRef((const ArrayRef_t &)x); return; }
        case exprType::ArraySize: { v.visit_ArraySize((const ArraySize_t &)x); return; }
        case exprType::ArrayBound: { v.visit_ArrayBound((const ArrayBound_t &)x); return; }
        case exprType::ArrayTranspose: { v.visit_ArrayTranspose((const ArrayTranspose_t &)x); return; }
        case exprType::ArrayMatMul: { v.visit_ArrayMatMul((const ArrayMatMul_t &)x); return; }
        case exprType::ArrayPack: { v.visit_ArrayPack((const ArrayPack_t &)x); return; }
        case exprType::Transfer: { v.visit_Transfer((const Transfer_t &)x); return; }
        case exprType::DerivedRef: { v.visit_DerivedRef((const DerivedRef_t &)x); return; }
        case exprType::Cast: { v.visit_Cast((const Cast_t &)x); return; }
        case exprType::ComplexRe: { v.visit_ComplexRe((const ComplexRe_t &)x); return; }
        case exprType::ComplexIm: { v.visit_ComplexIm((const ComplexIm_t &)x); return; }
        case exprType::DictItem: { v.visit_DictItem((const DictItem_t &)x); return; }
    }
}

template <class Visitor>
static void visit_ttype_t(const ttype_t &x, Visitor &v) {
    LFORTRAN_ASSERT(x.base.type == asrType::ttype)
    switch (x.type) {
        case ttypeType::Integer: { v.visit_Integer((const Integer_t &)x); return; }
        case ttypeType::Real: { v.visit_Real((const Real_t &)x); return; }
        case ttypeType::Complex: { v.visit_Complex((const Complex_t &)x); return; }
        case ttypeType::Character: { v.visit_Character((const Character_t &)x); return; }
        case ttypeType::Logical: { v.visit_Logical((const Logical_t &)x); return; }
        case ttypeType::Set: { v.visit_Set((const Set_t &)x); return; }
        case ttypeType::List: { v.visit_List((const List_t &)x); return; }
        case ttypeType::Tuple: { v.visit_Tuple((const Tuple_t &)x); return; }
        case ttypeType::Derived: { v.visit_Derived((const Derived_t &)x); return; }
        case ttypeType::Class: { v.visit_Class((const Class_t &)x); return; }
        case ttypeType::Dict: { v.visit_Dict((const Dict_t &)x); return; }
        case ttypeType::Pointer: { v.visit_Pointer((const Pointer_t &)x); return; }
    }
}

template <class Visitor>
static void visit_attribute_t(const attribute_t &x, Visitor &v) {
    LFORTRAN_ASSERT(x.base.type == asrType::attribute)
    switch (x.type) {
        case attributeType::Attribute: { v.visit_Attribute((const Attribute_t &)x); return; }
    }
}

template <class Visitor>
static void visit_tbind_t(const tbind_t &x, Visitor &v) {
    LFORTRAN_ASSERT(x.base.type == asrType::tbind)
    switch (x.type) {
        case tbindType::Bind: { v.visit_Bind((const Bind_t &)x); return; }
    }
}

template <class Visitor>
static void visit_case_stmt_t(const case_stmt_t &x, Visitor &v) {
    LFORTRAN_ASSERT(x.base.type == asrType::case_stmt)
    switch (x.type) {
        case case_stmtType::CaseStmt: { v.visit_CaseStmt((const CaseStmt_t &)x); return; }
        case case_stmtType::CaseStmt_Range: { v.visit_CaseStmt_Range((const CaseStmt_Range_t &)x); return; }
    }
}



template <class Visitor>
static void visit_asr_t(const asr_t &x, Visitor &v) {
    switch (x.type) {
        case asrType::unit: { v.visit_unit((const unit_t &)x); return; }
        case asrType::symbol: { v.visit_symbol((const symbol_t &)x); return; }
        case asrType::stmt: { v.visit_stmt((const stmt_t &)x); return; }
        case asrType::expr: { v.visit_expr((const expr_t &)x); return; }
        case asrType::ttype: { v.visit_ttype((const ttype_t &)x); return; }
        case asrType::attribute: { v.visit_attribute((const attribute_t &)x); return; }
        case asrType::tbind: { v.visit_tbind((const tbind_t &)x); return; }
        case asrType::case_stmt: { v.visit_case_stmt((const case_stmt_t &)x); return; }
    }
}



/******************************************************************************/
// Visitor base class

template <class Derived>
class BaseVisitor
{
private:
    Derived& self() { return static_cast<Derived&>(*this); }
public:
    void visit_asr(const asr_t &b) { visit_asr_t(b, self()); }
    void visit_unit(const unit_t &b) { visit_unit_t(b, self()); }
        void visit_TranslationUnit(const TranslationUnit_t & /* x */) { throw LFortran::LFortranException("visit_TranslationUnit() not implemented"); }
    void visit_symbol(const symbol_t &b) { visit_symbol_t(b, self()); }
        void visit_Program(const Program_t & /* x */) { throw LFortran::LFortranException("visit_Program() not implemented"); }
        void visit_Module(const Module_t & /* x */) { throw LFortran::LFortranException("visit_Module() not implemented"); }
        void visit_Subroutine(const Subroutine_t & /* x */) { throw LFortran::LFortranException("visit_Subroutine() not implemented"); }
        void visit_Function(const Function_t & /* x */) { throw LFortran::LFortranException("visit_Function() not implemented"); }
        void visit_GenericProcedure(const GenericProcedure_t & /* x */) { throw LFortran::LFortranException("visit_GenericProcedure() not implemented"); }
        void visit_CustomOperator(const CustomOperator_t & /* x */) { throw LFortran::LFortranException("visit_CustomOperator() not implemented"); }
        void visit_ExternalSymbol(const ExternalSymbol_t & /* x */) { throw LFortran::LFortranException("visit_ExternalSymbol() not implemented"); }
        void visit_DerivedType(const DerivedType_t & /* x */) { throw LFortran::LFortranException("visit_DerivedType() not implemented"); }
        void visit_Variable(const Variable_t & /* x */) { throw LFortran::LFortranException("visit_Variable() not implemented"); }
        void visit_ClassType(const ClassType_t & /* x */) { throw LFortran::LFortranException("visit_ClassType() not implemented"); }
        void visit_ClassProcedure(const ClassProcedure_t & /* x */) { throw LFortran::LFortranException("visit_ClassProcedure() not implemented"); }
        void visit_AssociateBlock(const AssociateBlock_t & /* x */) { throw LFortran::LFortranException("visit_AssociateBlock() not implemented"); }
        void visit_Block(const Block_t & /* x */) { throw LFortran::LFortranException("visit_Block() not implemented"); }
    void visit_stmt(const stmt_t &b) { visit_stmt_t(b, self()); }
        void visit_Allocate(const Allocate_t & /* x */) { throw LFortran::LFortranException("visit_Allocate() not implemented"); }
        void visit_Assign(const Assign_t & /* x */) { throw LFortran::LFortranException("visit_Assign() not implemented"); }
        void visit_Assignment(const Assignment_t & /* x */) { throw LFortran::LFortranException("visit_Assignment() not implemented"); }
        void visit_Associate(const Associate_t & /* x */) { throw LFortran::LFortranException("visit_Associate() not implemented"); }
        void visit_Cycle(const Cycle_t & /* x */) { throw LFortran::LFortranException("visit_Cycle() not implemented"); }
        void visit_ExplicitDeallocate(const ExplicitDeallocate_t & /* x */) { throw LFortran::LFortranException("visit_ExplicitDeallocate() not implemented"); }
        void visit_ImplicitDeallocate(const ImplicitDeallocate_t & /* x */) { throw LFortran::LFortranException("visit_ImplicitDeallocate() not implemented"); }
        void visit_DoConcurrentLoop(const DoConcurrentLoop_t & /* x */) { throw LFortran::LFortranException("visit_DoConcurrentLoop() not implemented"); }
        void visit_DoLoop(const DoLoop_t & /* x */) { throw LFortran::LFortranException("visit_DoLoop() not implemented"); }
        void visit_ErrorStop(const ErrorStop_t & /* x */) { throw LFortran::LFortranException("visit_ErrorStop() not implemented"); }
        void visit_Exit(const Exit_t & /* x */) { throw LFortran::LFortranException("visit_Exit() not implemented"); }
        void visit_ForAllSingle(const ForAllSingle_t & /* x */) { throw LFortran::LFortranException("visit_ForAllSingle() not implemented"); }
        void visit_GoTo(const GoTo_t & /* x */) { throw LFortran::LFortranException("visit_GoTo() not implemented"); }
        void visit_GoToTarget(const GoToTarget_t & /* x */) { throw LFortran::LFortranException("visit_GoToTarget() not implemented"); }
        void visit_If(const If_t & /* x */) { throw LFortran::LFortranException("visit_If() not implemented"); }
        void visit_IfArithmetic(const IfArithmetic_t & /* x */) { throw LFortran::LFortranException("visit_IfArithmetic() not implemented"); }
        void visit_Print(const Print_t & /* x */) { throw LFortran::LFortranException("visit_Print() not implemented"); }
        void visit_FileOpen(const FileOpen_t & /* x */) { throw LFortran::LFortranException("visit_FileOpen() not implemented"); }
        void visit_FileClose(const FileClose_t & /* x */) { throw LFortran::LFortranException("visit_FileClose() not implemented"); }
        void visit_FileRead(const FileRead_t & /* x */) { throw LFortran::LFortranException("visit_FileRead() not implemented"); }
        void visit_FileRewind(const FileRewind_t & /* x */) { throw LFortran::LFortranException("visit_FileRewind() not implemented"); }
        void visit_FileInquire(const FileInquire_t & /* x */) { throw LFortran::LFortranException("visit_FileInquire() not implemented"); }
        void visit_FileWrite(const FileWrite_t & /* x */) { throw LFortran::LFortranException("visit_FileWrite() not implemented"); }
        void visit_Return(const Return_t & /* x */) { throw LFortran::LFortranException("visit_Return() not implemented"); }
        void visit_Select(const Select_t & /* x */) { throw LFortran::LFortranException("visit_Select() not implemented"); }
        void visit_Stop(const Stop_t & /* x */) { throw LFortran::LFortranException("visit_Stop() not implemented"); }
        void visit_Assert(const Assert_t & /* x */) { throw LFortran::LFortranException("visit_Assert() not implemented"); }
        void visit_SubroutineCall(const SubroutineCall_t & /* x */) { throw LFortran::LFortranException("visit_SubroutineCall() not implemented"); }
        void visit_Where(const Where_t & /* x */) { throw LFortran::LFortranException("visit_Where() not implemented"); }
        void visit_WhileLoop(const WhileLoop_t & /* x */) { throw LFortran::LFortranException("visit_WhileLoop() not implemented"); }
        void visit_Nullify(const Nullify_t & /* x */) { throw LFortran::LFortranException("visit_Nullify() not implemented"); }
        void visit_Flush(const Flush_t & /* x */) { throw LFortran::LFortranException("visit_Flush() not implemented"); }
        void visit_ListAppend(const ListAppend_t & /* x */) { throw LFortran::LFortranException("visit_ListAppend() not implemented"); }
        void visit_AssociateBlockCall(const AssociateBlockCall_t & /* x */) { throw LFortran::LFortranException("visit_AssociateBlockCall() not implemented"); }
        void visit_BlockCall(const BlockCall_t & /* x */) { throw LFortran::LFortranException("visit_BlockCall() not implemented"); }
        void visit_SetInsert(const SetInsert_t & /* x */) { throw LFortran::LFortranException("visit_SetInsert() not implemented"); }
        void visit_SetRemove(const SetRemove_t & /* x */) { throw LFortran::LFortranException("visit_SetRemove() not implemented"); }
        void visit_ListInsert(const ListInsert_t & /* x */) { throw LFortran::LFortranException("visit_ListInsert() not implemented"); }
        void visit_ListRemove(const ListRemove_t & /* x */) { throw LFortran::LFortranException("visit_ListRemove() not implemented"); }
        void visit_DictInsert(const DictInsert_t & /* x */) { throw LFortran::LFortranException("visit_DictInsert() not implemented"); }
    void visit_expr(const expr_t &b) { visit_expr_t(b, self()); }
        void visit_BoolOp(const BoolOp_t & /* x */) { throw LFortran::LFortranException("visit_BoolOp() not implemented"); }
        void visit_BinOp(const BinOp_t & /* x */) { throw LFortran::LFortranException("visit_BinOp() not implemented"); }
        void visit_StrOp(const StrOp_t & /* x */) { throw LFortran::LFortranException("visit_StrOp() not implemented"); }
        void visit_UnaryOp(const UnaryOp_t & /* x */) { throw LFortran::LFortranException("visit_UnaryOp() not implemented"); }
        void visit_ComplexConstructor(const ComplexConstructor_t & /* x */) { throw LFortran::LFortranException("visit_ComplexConstructor() not implemented"); }
        void visit_NamedExpr(const NamedExpr_t & /* x */) { throw LFortran::LFortranException("visit_NamedExpr() not implemented"); }
        void visit_Compare(const Compare_t & /* x */) { throw LFortran::LFortranException("visit_Compare() not implemented"); }
        void visit_IfExp(const IfExp_t & /* x */) { throw LFortran::LFortranException("visit_IfExp() not implemented"); }
        void visit_FunctionCall(const FunctionCall_t & /* x */) { throw LFortran::LFortranException("visit_FunctionCall() not implemented"); }
        void visit_DerivedTypeConstructor(const DerivedTypeConstructor_t & /* x */) { throw LFortran::LFortranException("visit_DerivedTypeConstructor() not implemented"); }
        void visit_ImpliedDoLoop(const ImpliedDoLoop_t & /* x */) { throw LFortran::LFortranException("visit_ImpliedDoLoop() not implemented"); }
        void visit_IntegerConstant(const IntegerConstant_t & /* x */) { throw LFortran::LFortranException("visit_IntegerConstant() not implemented"); }
        void visit_IntegerBOZ(const IntegerBOZ_t & /* x */) { throw LFortran::LFortranException("visit_IntegerBOZ() not implemented"); }
        void visit_RealConstant(const RealConstant_t & /* x */) { throw LFortran::LFortranException("visit_RealConstant() not implemented"); }
        void visit_ComplexConstant(const ComplexConstant_t & /* x */) { throw LFortran::LFortranException("visit_ComplexConstant() not implemented"); }
        void visit_LogicalConstant(const LogicalConstant_t & /* x */) { throw LFortran::LFortranException("visit_LogicalConstant() not implemented"); }
        void visit_ListConstant(const ListConstant_t & /* x */) { throw LFortran::LFortranException("visit_ListConstant() not implemented"); }
        void visit_ListLen(const ListLen_t & /* x */) { throw LFortran::LFortranException("visit_ListLen() not implemented"); }
        void visit_ListConcat(const ListConcat_t & /* x */) { throw LFortran::LFortranException("visit_ListConcat() not implemented"); }
        void visit_ArrayConstant(const ArrayConstant_t & /* x */) { throw LFortran::LFortranException("visit_ArrayConstant() not implemented"); }
        void visit_SetConstant(const SetConstant_t & /* x */) { throw LFortran::LFortranException("visit_SetConstant() not implemented"); }
        void visit_SetLen(const SetLen_t & /* x */) { throw LFortran::LFortranException("visit_SetLen() not implemented"); }
        void visit_TupleConstant(const TupleConstant_t & /* x */) { throw LFortran::LFortranException("visit_TupleConstant() not implemented"); }
        void visit_TupleLen(const TupleLen_t & /* x */) { throw LFortran::LFortranException("visit_TupleLen() not implemented"); }
        void visit_StringConstant(const StringConstant_t & /* x */) { throw LFortran::LFortranException("visit_StringConstant() not implemented"); }
        void visit_StringConcat(const StringConcat_t & /* x */) { throw LFortran::LFortranException("visit_StringConcat() not implemented"); }
        void visit_StringLen(const StringLen_t & /* x */) { throw LFortran::LFortranException("visit_StringLen() not implemented"); }
        void visit_DictConstant(const DictConstant_t & /* x */) { throw LFortran::LFortranException("visit_DictConstant() not implemented"); }
        void visit_DictLen(const DictLen_t & /* x */) { throw LFortran::LFortranException("visit_DictLen() not implemented"); }
        void visit_Var(const Var_t & /* x */) { throw LFortran::LFortranException("visit_Var() not implemented"); }
        void visit_ArrayRef(const ArrayRef_t & /* x */) { throw LFortran::LFortranException("visit_ArrayRef() not implemented"); }
        void visit_ArraySize(const ArraySize_t & /* x */) { throw LFortran::LFortranException("visit_ArraySize() not implemented"); }
        void visit_ArrayBound(const ArrayBound_t & /* x */) { throw LFortran::LFortranException("visit_ArrayBound() not implemented"); }
        void visit_ArrayTranspose(const ArrayTranspose_t & /* x */) { throw LFortran::LFortranException("visit_ArrayTranspose() not implemented"); }
        void visit_ArrayMatMul(const ArrayMatMul_t & /* x */) { throw LFortran::LFortranException("visit_ArrayMatMul() not implemented"); }
        void visit_ArrayPack(const ArrayPack_t & /* x */) { throw LFortran::LFortranException("visit_ArrayPack() not implemented"); }
        void visit_Transfer(const Transfer_t & /* x */) { throw LFortran::LFortranException("visit_Transfer() not implemented"); }
        void visit_DerivedRef(const DerivedRef_t & /* x */) { throw LFortran::LFortranException("visit_DerivedRef() not implemented"); }
        void visit_Cast(const Cast_t & /* x */) { throw LFortran::LFortranException("visit_Cast() not implemented"); }
        void visit_ComplexRe(const ComplexRe_t & /* x */) { throw LFortran::LFortranException("visit_ComplexRe() not implemented"); }
        void visit_ComplexIm(const ComplexIm_t & /* x */) { throw LFortran::LFortranException("visit_ComplexIm() not implemented"); }
        void visit_DictItem(const DictItem_t & /* x */) { throw LFortran::LFortranException("visit_DictItem() not implemented"); }
    void visit_ttype(const ttype_t &b) { visit_ttype_t(b, self()); }
        void visit_Integer(const Integer_t & /* x */) { throw LFortran::LFortranException("visit_Integer() not implemented"); }
        void visit_Real(const Real_t & /* x */) { throw LFortran::LFortranException("visit_Real() not implemented"); }
        void visit_Complex(const Complex_t & /* x */) { throw LFortran::LFortranException("visit_Complex() not implemented"); }
        void visit_Character(const Character_t & /* x */) { throw LFortran::LFortranException("visit_Character() not implemented"); }
        void visit_Logical(const Logical_t & /* x */) { throw LFortran::LFortranException("visit_Logical() not implemented"); }
        void visit_Set(const Set_t & /* x */) { throw LFortran::LFortranException("visit_Set() not implemented"); }
        void visit_List(const List_t & /* x */) { throw LFortran::LFortranException("visit_List() not implemented"); }
        void visit_Tuple(const Tuple_t & /* x */) { throw LFortran::LFortranException("visit_Tuple() not implemented"); }
        void visit_Derived(const Derived_t & /* x */) { throw LFortran::LFortranException("visit_Derived() not implemented"); }
        void visit_Class(const Class_t & /* x */) { throw LFortran::LFortranException("visit_Class() not implemented"); }
        void visit_Dict(const Dict_t & /* x */) { throw LFortran::LFortranException("visit_Dict() not implemented"); }
        void visit_Pointer(const Pointer_t & /* x */) { throw LFortran::LFortranException("visit_Pointer() not implemented"); }
    void visit_attribute(const attribute_t &b) { visit_attribute_t(b, self()); }
        void visit_Attribute(const Attribute_t & /* x */) { throw LFortran::LFortranException("visit_Attribute() not implemented"); }
    void visit_tbind(const tbind_t &b) { visit_tbind_t(b, self()); }
        void visit_Bind(const Bind_t & /* x */) { throw LFortran::LFortranException("visit_Bind() not implemented"); }
    void visit_case_stmt(const case_stmt_t &b) { visit_case_stmt_t(b, self()); }
        void visit_CaseStmt(const CaseStmt_t & /* x */) { throw LFortran::LFortranException("visit_CaseStmt() not implemented"); }
        void visit_CaseStmt_Range(const CaseStmt_Range_t & /* x */) { throw LFortran::LFortranException("visit_CaseStmt_Range() not implemented"); }
};


/******************************************************************************/
// Walk Visitor base class

template <class Derived>
class BaseWalkVisitor : public BaseVisitor<Derived>
{
private:
    Derived& self() { return static_cast<Derived&>(*this); }
public:
    void visit_TranslationUnit(const TranslationUnit_t &x) {
        for (auto &a : x.m_global_scope->get_scope()) {
            this->visit_symbol(*a.second);
        }
    }
    void visit_Program(const Program_t &x) {
        for (auto &a : x.m_symtab->get_scope()) {
            this->visit_symbol(*a.second);
        }
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
    }
    void visit_Module(const Module_t &x) {
        for (auto &a : x.m_symtab->get_scope()) {
            this->visit_symbol(*a.second);
        }
    }
    void visit_Subroutine(const Subroutine_t &x) {
        for (auto &a : x.m_symtab->get_scope()) {
            this->visit_symbol(*a.second);
        }
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_expr(*x.m_args[i]);
        }
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
    }
    void visit_Function(const Function_t &x) {
        for (auto &a : x.m_symtab->get_scope()) {
            this->visit_symbol(*a.second);
        }
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_expr(*x.m_args[i]);
        }
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
        self().visit_expr(*x.m_return_var);
    }
    void visit_GenericProcedure(const GenericProcedure_t &x) {
        for (size_t i=0; i<x.n_procs; i++) {
        }
    }
    void visit_CustomOperator(const CustomOperator_t &x) {
        for (size_t i=0; i<x.n_procs; i++) {
        }
    }
    void visit_ExternalSymbol(const ExternalSymbol_t &x) {
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_DerivedType(const DerivedType_t &x) {
        for (auto &a : x.m_symtab->get_scope()) {
            this->visit_symbol(*a.second);
        }
    }
    void visit_Variable(const Variable_t &x) {
        if (x.m_symbolic_value)
            self().visit_expr(*x.m_symbolic_value);
        if (x.m_value)
            self().visit_expr(*x.m_value);
        self().visit_ttype(*x.m_type);
    }
    void visit_ClassType(const ClassType_t &x) {
        for (auto &a : x.m_symtab->get_scope()) {
            this->visit_symbol(*a.second);
        }
    }
    void visit_ClassProcedure(const ClassProcedure_t &x) {
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_AssociateBlock(const AssociateBlock_t &x) {
        for (auto &a : x.m_symtab->get_scope()) {
            this->visit_symbol(*a.second);
        }
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
    }
    void visit_Block(const Block_t &x) {
        for (auto &a : x.m_symtab->get_scope()) {
            this->visit_symbol(*a.second);
        }
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
    }
    void visit_Allocate(const Allocate_t &x) {
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_alloc_arg(x.m_args[i]);
        }
        if (x.m_stat)
            self().visit_expr(*x.m_stat);
        if (x.m_errmsg)
            self().visit_expr(*x.m_errmsg);
        if (x.m_source)
            self().visit_expr(*x.m_source);
    }
    void visit_Assign(const Assign_t &x) {
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_Assignment(const Assignment_t &x) {
        self().visit_expr(*x.m_target);
        self().visit_expr(*x.m_value);
        if (x.m_overloaded)
            self().visit_stmt(*x.m_overloaded);
    }
    void visit_Associate(const Associate_t &x) {
        self().visit_expr(*x.m_target);
        self().visit_expr(*x.m_value);
    }
    void visit_Cycle(const Cycle_t &x) {
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_ExplicitDeallocate(const ExplicitDeallocate_t &x) {
        for (size_t i=0; i<x.n_vars; i++) {
        }
    }
    void visit_ImplicitDeallocate(const ImplicitDeallocate_t &x) {
        for (size_t i=0; i<x.n_vars; i++) {
        }
    }
    void visit_DoConcurrentLoop(const DoConcurrentLoop_t &x) {
        self().visit_do_loop_head(x.m_head);
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
    }
    void visit_DoLoop(const DoLoop_t &x) {
        self().visit_do_loop_head(x.m_head);
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
    }
    void visit_ErrorStop(const ErrorStop_t &x) {
        if (x.m_code)
            self().visit_expr(*x.m_code);
    }
    void visit_Exit(const Exit_t &x) {
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_ForAllSingle(const ForAllSingle_t &x) {
        self().visit_do_loop_head(x.m_head);
        self().visit_stmt(*x.m_assign_stmt);
    }
    void visit_GoTo(const GoTo_t &x) {
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_GoToTarget(const GoToTarget_t &x) {
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_If(const If_t &x) {
        self().visit_expr(*x.m_test);
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
        for (size_t i=0; i<x.n_orelse; i++) {
            self().visit_stmt(*x.m_orelse[i]);
        }
    }
    void visit_IfArithmetic(const IfArithmetic_t &x) {
        self().visit_expr(*x.m_test);
    }
    void visit_Print(const Print_t &x) {
        if (x.m_fmt)
            self().visit_expr(*x.m_fmt);
        for (size_t i=0; i<x.n_values; i++) {
            self().visit_expr(*x.m_values[i]);
        }
    }
    void visit_FileOpen(const FileOpen_t &x) {
        if (x.m_newunit)
            self().visit_expr(*x.m_newunit);
        if (x.m_filename)
            self().visit_expr(*x.m_filename);
        if (x.m_status)
            self().visit_expr(*x.m_status);
    }
    void visit_FileClose(const FileClose_t &x) {
        if (x.m_unit)
            self().visit_expr(*x.m_unit);
        if (x.m_iostat)
            self().visit_expr(*x.m_iostat);
        if (x.m_iomsg)
            self().visit_expr(*x.m_iomsg);
        if (x.m_err)
            self().visit_expr(*x.m_err);
        if (x.m_status)
            self().visit_expr(*x.m_status);
    }
    void visit_FileRead(const FileRead_t &x) {
        if (x.m_unit)
            self().visit_expr(*x.m_unit);
        if (x.m_fmt)
            self().visit_expr(*x.m_fmt);
        if (x.m_iomsg)
            self().visit_expr(*x.m_iomsg);
        if (x.m_iostat)
            self().visit_expr(*x.m_iostat);
        if (x.m_id)
            self().visit_expr(*x.m_id);
        for (size_t i=0; i<x.n_values; i++) {
            self().visit_expr(*x.m_values[i]);
        }
    }
    void visit_FileRewind(const FileRewind_t &x) {
        if (x.m_unit)
            self().visit_expr(*x.m_unit);
        if (x.m_iostat)
            self().visit_expr(*x.m_iostat);
        if (x.m_err)
            self().visit_expr(*x.m_err);
    }
    void visit_FileInquire(const FileInquire_t &x) {
        if (x.m_unit)
            self().visit_expr(*x.m_unit);
        if (x.m_file)
            self().visit_expr(*x.m_file);
        if (x.m_iostat)
            self().visit_expr(*x.m_iostat);
        if (x.m_err)
            self().visit_expr(*x.m_err);
        if (x.m_exist)
            self().visit_expr(*x.m_exist);
        if (x.m_opened)
            self().visit_expr(*x.m_opened);
        if (x.m_number)
            self().visit_expr(*x.m_number);
        if (x.m_named)
            self().visit_expr(*x.m_named);
        if (x.m_name)
            self().visit_expr(*x.m_name);
        if (x.m_access)
            self().visit_expr(*x.m_access);
        if (x.m_sequential)
            self().visit_expr(*x.m_sequential);
        if (x.m_direct)
            self().visit_expr(*x.m_direct);
        if (x.m_form)
            self().visit_expr(*x.m_form);
        if (x.m_formatted)
            self().visit_expr(*x.m_formatted);
        if (x.m_unformatted)
            self().visit_expr(*x.m_unformatted);
        if (x.m_recl)
            self().visit_expr(*x.m_recl);
        if (x.m_nextrec)
            self().visit_expr(*x.m_nextrec);
        if (x.m_blank)
            self().visit_expr(*x.m_blank);
        if (x.m_position)
            self().visit_expr(*x.m_position);
        if (x.m_action)
            self().visit_expr(*x.m_action);
        if (x.m_read)
            self().visit_expr(*x.m_read);
        if (x.m_write)
            self().visit_expr(*x.m_write);
        if (x.m_readwrite)
            self().visit_expr(*x.m_readwrite);
        if (x.m_delim)
            self().visit_expr(*x.m_delim);
        if (x.m_pad)
            self().visit_expr(*x.m_pad);
        if (x.m_flen)
            self().visit_expr(*x.m_flen);
        if (x.m_blocksize)
            self().visit_expr(*x.m_blocksize);
        if (x.m_convert)
            self().visit_expr(*x.m_convert);
        if (x.m_carriagecontrol)
            self().visit_expr(*x.m_carriagecontrol);
        if (x.m_iolength)
            self().visit_expr(*x.m_iolength);
    }
    void visit_FileWrite(const FileWrite_t &x) {
        if (x.m_unit)
            self().visit_expr(*x.m_unit);
        if (x.m_fmt)
            self().visit_expr(*x.m_fmt);
        if (x.m_iomsg)
            self().visit_expr(*x.m_iomsg);
        if (x.m_iostat)
            self().visit_expr(*x.m_iostat);
        if (x.m_id)
            self().visit_expr(*x.m_id);
        for (size_t i=0; i<x.n_values; i++) {
            self().visit_expr(*x.m_values[i]);
        }
    }
    void visit_Return(const Return_t &x) {
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_Select(const Select_t &x) {
        self().visit_expr(*x.m_test);
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_case_stmt(*x.m_body[i]);
        }
        for (size_t i=0; i<x.n_default; i++) {
            self().visit_stmt(*x.m_default[i]);
        }
    }
    void visit_Stop(const Stop_t &x) {
        if (x.m_code)
            self().visit_expr(*x.m_code);
    }
    void visit_Assert(const Assert_t &x) {
        self().visit_expr(*x.m_test);
        if (x.m_msg)
            self().visit_expr(*x.m_msg);
    }
    void visit_SubroutineCall(const SubroutineCall_t &x) {
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_call_arg(x.m_args[i]);
        }
        if (x.m_dt)
            self().visit_expr(*x.m_dt);
    }
    void visit_Where(const Where_t &x) {
        self().visit_expr(*x.m_test);
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
        for (size_t i=0; i<x.n_orelse; i++) {
            self().visit_stmt(*x.m_orelse[i]);
        }
    }
    void visit_WhileLoop(const WhileLoop_t &x) {
        self().visit_expr(*x.m_test);
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
    }
    void visit_Nullify(const Nullify_t &x) {
        for (size_t i=0; i<x.n_vars; i++) {
        }
    }
    void visit_Flush(const Flush_t &x) {
        self().visit_expr(*x.m_unit);
        if (x.m_err)
            self().visit_expr(*x.m_err);
        if (x.m_iomsg)
            self().visit_expr(*x.m_iomsg);
        if (x.m_iostat)
            self().visit_expr(*x.m_iostat);
    }
    void visit_ListAppend(const ListAppend_t &x) {
        self().visit_expr(*x.m_ele);
    }
    void visit_AssociateBlockCall(const AssociateBlockCall_t &x) {
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_BlockCall(const BlockCall_t &x) {
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_SetInsert(const SetInsert_t &x) {
        self().visit_expr(*x.m_ele);
    }
    void visit_SetRemove(const SetRemove_t &x) {
        self().visit_expr(*x.m_ele);
    }
    void visit_ListInsert(const ListInsert_t &x) {
        self().visit_expr(*x.m_pos);
        self().visit_expr(*x.m_ele);
    }
    void visit_ListRemove(const ListRemove_t &x) {
        self().visit_expr(*x.m_ele);
    }
    void visit_DictInsert(const DictInsert_t &x) {
        self().visit_expr(*x.m_key);
        self().visit_expr(*x.m_value);
    }
    void visit_BoolOp(const BoolOp_t &x) {
        self().visit_expr(*x.m_left);
        self().visit_expr(*x.m_right);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_BinOp(const BinOp_t &x) {
        self().visit_expr(*x.m_left);
        self().visit_expr(*x.m_right);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
        if (x.m_overloaded)
            self().visit_expr(*x.m_overloaded);
    }
    void visit_StrOp(const StrOp_t &x) {
        self().visit_expr(*x.m_left);
        self().visit_expr(*x.m_right);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_UnaryOp(const UnaryOp_t &x) {
        self().visit_expr(*x.m_operand);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_ComplexConstructor(const ComplexConstructor_t &x) {
        self().visit_expr(*x.m_re);
        self().visit_expr(*x.m_im);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_NamedExpr(const NamedExpr_t &x) {
        self().visit_expr(*x.m_target);
        self().visit_expr(*x.m_value);
        self().visit_ttype(*x.m_type);
    }
    void visit_Compare(const Compare_t &x) {
        self().visit_expr(*x.m_left);
        self().visit_expr(*x.m_right);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
        if (x.m_overloaded)
            self().visit_expr(*x.m_overloaded);
    }
    void visit_IfExp(const IfExp_t &x) {
        self().visit_expr(*x.m_test);
        self().visit_expr(*x.m_body);
        self().visit_expr(*x.m_orelse);
        self().visit_ttype(*x.m_type);
    }
    void visit_FunctionCall(const FunctionCall_t &x) {
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_call_arg(x.m_args[i]);
        }
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
        if (x.m_dt)
            self().visit_expr(*x.m_dt);
    }
    void visit_DerivedTypeConstructor(const DerivedTypeConstructor_t &x) {
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_expr(*x.m_args[i]);
        }
        self().visit_ttype(*x.m_type);
    }
    void visit_ImpliedDoLoop(const ImpliedDoLoop_t &x) {
        for (size_t i=0; i<x.n_values; i++) {
            self().visit_expr(*x.m_values[i]);
        }
        self().visit_expr(*x.m_var);
        self().visit_expr(*x.m_start);
        self().visit_expr(*x.m_end);
        if (x.m_increment)
            self().visit_expr(*x.m_increment);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_IntegerConstant(const IntegerConstant_t &x) {
        self().visit_ttype(*x.m_type);
    }
    void visit_IntegerBOZ(const IntegerBOZ_t &x) {
        if (x.m_type)
            self().visit_ttype(*x.m_type);
    }
    void visit_RealConstant(const RealConstant_t &x) {
        self().visit_ttype(*x.m_type);
    }
    void visit_ComplexConstant(const ComplexConstant_t &x) {
        self().visit_ttype(*x.m_type);
    }
    void visit_LogicalConstant(const LogicalConstant_t &x) {
        self().visit_ttype(*x.m_type);
    }
    void visit_ListConstant(const ListConstant_t &x) {
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_expr(*x.m_args[i]);
        }
        self().visit_ttype(*x.m_type);
    }
    void visit_ListLen(const ListLen_t &x) {
        self().visit_expr(*x.m_arg);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_ListConcat(const ListConcat_t &x) {
        self().visit_expr(*x.m_left);
        self().visit_expr(*x.m_right);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_ArrayConstant(const ArrayConstant_t &x) {
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_expr(*x.m_args[i]);
        }
        self().visit_ttype(*x.m_type);
    }
    void visit_SetConstant(const SetConstant_t &x) {
        for (size_t i=0; i<x.n_elements; i++) {
            self().visit_expr(*x.m_elements[i]);
        }
        self().visit_ttype(*x.m_type);
    }
    void visit_SetLen(const SetLen_t &x) {
        self().visit_expr(*x.m_arg);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_TupleConstant(const TupleConstant_t &x) {
        for (size_t i=0; i<x.n_elements; i++) {
            self().visit_expr(*x.m_elements[i]);
        }
        self().visit_ttype(*x.m_type);
    }
    void visit_TupleLen(const TupleLen_t &x) {
        self().visit_expr(*x.m_arg);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_StringConstant(const StringConstant_t &x) {
        self().visit_ttype(*x.m_type);
    }
    void visit_StringConcat(const StringConcat_t &x) {
        self().visit_expr(*x.m_left);
        self().visit_expr(*x.m_right);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_StringLen(const StringLen_t &x) {
        self().visit_expr(*x.m_arg);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_DictConstant(const DictConstant_t &x) {
        for (size_t i=0; i<x.n_keys; i++) {
            self().visit_expr(*x.m_keys[i]);
        }
        for (size_t i=0; i<x.n_values; i++) {
            self().visit_expr(*x.m_values[i]);
        }
        self().visit_ttype(*x.m_type);
    }
    void visit_DictLen(const DictLen_t &x) {
        self().visit_expr(*x.m_arg);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_Var(const Var_t &x) {
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_ArrayRef(const ArrayRef_t &x) {
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_array_index(x.m_args[i]);
        }
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_ArraySize(const ArraySize_t &x) {
        self().visit_expr(*x.m_v);
        if (x.m_dim)
            self().visit_expr(*x.m_dim);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_ArrayBound(const ArrayBound_t &x) {
        self().visit_expr(*x.m_v);
        if (x.m_dim)
            self().visit_expr(*x.m_dim);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_ArrayTranspose(const ArrayTranspose_t &x) {
        self().visit_expr(*x.m_matrix);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_ArrayMatMul(const ArrayMatMul_t &x) {
        self().visit_expr(*x.m_matrix_a);
        self().visit_expr(*x.m_matrix_b);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_ArrayPack(const ArrayPack_t &x) {
        self().visit_expr(*x.m_array);
        self().visit_expr(*x.m_mask);
        if (x.m_vector)
            self().visit_expr(*x.m_vector);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_Transfer(const Transfer_t &x) {
        self().visit_expr(*x.m_source);
        self().visit_expr(*x.m_mold);
        if (x.m_size)
            self().visit_expr(*x.m_size);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_DerivedRef(const DerivedRef_t &x) {
        self().visit_expr(*x.m_v);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_Cast(const Cast_t &x) {
        self().visit_expr(*x.m_arg);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_ComplexRe(const ComplexRe_t &x) {
        self().visit_expr(*x.m_arg);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_ComplexIm(const ComplexIm_t &x) {
        self().visit_expr(*x.m_arg);
        self().visit_ttype(*x.m_type);
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_DictItem(const DictItem_t &x) {
        self().visit_expr(*x.m_key);
        if (x.m_default)
            self().visit_expr(*x.m_default);
        self().visit_ttype(*x.m_type);
    }
    void visit_Integer(const Integer_t &x) {
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
        }
    }
    void visit_Real(const Real_t &x) {
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
        }
    }
    void visit_Complex(const Complex_t &x) {
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
        }
    }
    void visit_Character(const Character_t &x) {
        if (x.m_len_expr)
            self().visit_expr(*x.m_len_expr);
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
        }
    }
    void visit_Logical(const Logical_t &x) {
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
        }
    }
    void visit_Set(const Set_t &x) {
        self().visit_ttype(*x.m_type);
    }
    void visit_List(const List_t &x) {
        self().visit_ttype(*x.m_type);
    }
    void visit_Tuple(const Tuple_t &x) {
        for (size_t i=0; i<x.n_type; i++) {
            self().visit_ttype(*x.m_type[i]);
        }
    }
    void visit_Derived(const Derived_t &x) {
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
        }
    }
    void visit_Class(const Class_t &x) {
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
        }
    }
    void visit_Dict(const Dict_t &x) {
        self().visit_ttype(*x.m_key_type);
        self().visit_ttype(*x.m_value_type);
    }
    void visit_Pointer(const Pointer_t &x) {
        self().visit_ttype(*x.m_type);
    }
    void visit_dimension(const dimension_t &x) {
        if (x.m_start)
            self().visit_expr(*x.m_start);
        if (x.m_end)
            self().visit_expr(*x.m_end);
    }
    void visit_alloc_arg(const alloc_arg_t &x) {
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
        }
    }
    void visit_Attribute(const Attribute_t &x) {
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_attribute_arg(x.m_args[i]);
        }
    }
    void visit_attribute_arg(const attribute_arg_t &x) {
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_call_arg(const call_arg_t &x) {
        if (x.m_value)
            self().visit_expr(*x.m_value);
    }
    void visit_Bind(const Bind_t &x) {
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_array_index(const array_index_t &x) {
        if (x.m_left)
            self().visit_expr(*x.m_left);
        if (x.m_right)
            self().visit_expr(*x.m_right);
        if (x.m_step)
            self().visit_expr(*x.m_step);
    }
    void visit_do_loop_head(const do_loop_head_t &x) {
        if (x.m_v)
            self().visit_expr(*x.m_v);
        if (x.m_start)
            self().visit_expr(*x.m_start);
        if (x.m_end)
            self().visit_expr(*x.m_end);
        if (x.m_increment)
            self().visit_expr(*x.m_increment);
    }
    void visit_CaseStmt(const CaseStmt_t &x) {
        for (size_t i=0; i<x.n_test; i++) {
            self().visit_expr(*x.m_test[i]);
        }
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
    }
    void visit_CaseStmt_Range(const CaseStmt_Range_t &x) {
        if (x.m_start)
            self().visit_expr(*x.m_start);
        if (x.m_end)
            self().visit_expr(*x.m_end);
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
    }
};


/******************************************************************************/
// Pickle Visitor base class

template <class Derived>
class PickleBaseVisitor : public BaseVisitor<Derived>
{
private:
    Derived& self() { return static_cast<Derived&>(*this); }
public:
    std::string s, indtd;
    bool use_colors;
    bool indent, start_line = true;
    int indent_level = 0, indent_spaces = 3, lvl = 0;
    int tmp = 0, tmp1 = 0, tmp2 = 2;
    int curly[2000], round[2000];
public:
    PickleBaseVisitor() : use_colors(false), indent(false) { s.reserve(100000); }
    void inc_indent() {
        indent_level++;
        indtd = std::string(indent_level*indent_spaces, ' ');
    }
    void dec_indent() {
        indent_level--;
        LFORTRAN_ASSERT(indent_level >= 0);
        indtd = std::string(indent_level*indent_spaces, ' ');
    }
    void visit_TranslationUnit(const TranslationUnit_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("TranslationUnit");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        if(indent) {
            s.append("\n"+indtd);
            round[++tmp1] = indent_level;
            inc_indent();
        }
        s.append("(");
        if (use_colors) {
            s.append(color(fg::yellow));
        }
        s.append("SymbolTable");
        if (use_colors) {
            s.append(color(fg::reset));
        }
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_global_scope->get_counter());
        s.append(" ");
        if(indent) {
            s.append("\n"+indtd);
            curly[++tmp] = indent_level;
            tmp2 = 1;
            inc_indent();
        }
        s.append("{");
        {
            size_t i = 0;
            for (auto &a : x.m_global_scope->get_scope()) {
        if(indent) {
            s.append("\n"+indtd);
            inc_indent();
        }
              s.append(a.first + ": ");
                this->visit_symbol(*a.second);
                if (i < x.m_global_scope->get_scope().size()-1) { 
                    s.append(", ");
                    if(indent) {
                            for(int times = 0; times < tmp2; times++)
                             dec_indent();
                     }
                }
                i++;
            }
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - curly[tmp] ); times++)
                dec_indent();
            curly[tmp] = 0; if(tmp > 1) tmp--;
            tmp2++;
            s.append("\n"+indtd);
        }
        s.append("})");
        if(indent) dec_indent();
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append("[");
        for (size_t i=0; i<x.n_items; i++) {
            self().visit_asr(*x.m_items[i]);
            if (i < x.n_items-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_Program(const Program_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Program");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        if(indent) {
            s.append("\n"+indtd);
            round[++tmp1] = indent_level;
            inc_indent();
        }
        s.append("(");
        if (use_colors) {
            s.append(color(fg::yellow));
        }
        s.append("SymbolTable");
        if (use_colors) {
            s.append(color(fg::reset));
        }
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_symtab->get_counter());
        s.append(" ");
        if(indent) {
            s.append("\n"+indtd);
            curly[++tmp] = indent_level;
            tmp2 = 1;
            inc_indent();
        }
        s.append("{");
        {
            size_t i = 0;
            for (auto &a : x.m_symtab->get_scope()) {
        if(indent) {
            s.append("\n"+indtd);
            inc_indent();
        }
              s.append(a.first + ": ");
                this->visit_symbol(*a.second);
                if (i < x.m_symtab->get_scope().size()-1) { 
                    s.append(", ");
                    if(indent) {
                            for(int times = 0; times < tmp2; times++)
                             dec_indent();
                     }
                }
                i++;
            }
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - curly[tmp] ); times++)
                dec_indent();
            curly[tmp] = 0; if(tmp > 1) tmp--;
            tmp2++;
            s.append("\n"+indtd);
        }
        s.append("})");
        if(indent) dec_indent();
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_name);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_dependencies; i++) {
            s.append(x.m_dependencies[i]);
            if (i < x.n_dependencies-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
            if (i < x.n_body-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Module(const Module_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Module");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        if(indent) {
            s.append("\n"+indtd);
            round[++tmp1] = indent_level;
            inc_indent();
        }
        s.append("(");
        if (use_colors) {
            s.append(color(fg::yellow));
        }
        s.append("SymbolTable");
        if (use_colors) {
            s.append(color(fg::reset));
        }
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_symtab->get_counter());
        s.append(" ");
        if(indent) {
            s.append("\n"+indtd);
            curly[++tmp] = indent_level;
            tmp2 = 1;
            inc_indent();
        }
        s.append("{");
        {
            size_t i = 0;
            for (auto &a : x.m_symtab->get_scope()) {
        if(indent) {
            s.append("\n"+indtd);
            inc_indent();
        }
              s.append(a.first + ": ");
                this->visit_symbol(*a.second);
                if (i < x.m_symtab->get_scope().size()-1) { 
                    s.append(", ");
                    if(indent) {
                            for(int times = 0; times < tmp2; times++)
                             dec_indent();
                     }
                }
                i++;
            }
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - curly[tmp] ); times++)
                dec_indent();
            curly[tmp] = 0; if(tmp > 1) tmp--;
            tmp2++;
            s.append("\n"+indtd);
        }
        s.append("})");
        if(indent) dec_indent();
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_name);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_dependencies; i++) {
            s.append(x.m_dependencies[i]);
            if (i < x.n_dependencies-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        if (x.m_loaded_from_mod) {
            s.append(".true.");
        } else {
            s.append(".false.");
        }
        s.append(" ");
        if (x.m_intrinsic) {
            s.append(".true.");
        } else {
            s.append(".false.");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_Subroutine(const Subroutine_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Subroutine");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        if(indent) {
            s.append("\n"+indtd);
            round[++tmp1] = indent_level;
            inc_indent();
        }
        s.append("(");
        if (use_colors) {
            s.append(color(fg::yellow));
        }
        s.append("SymbolTable");
        if (use_colors) {
            s.append(color(fg::reset));
        }
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_symtab->get_counter());
        s.append(" ");
        if(indent) {
            s.append("\n"+indtd);
            curly[++tmp] = indent_level;
            tmp2 = 1;
            inc_indent();
        }
        s.append("{");
        {
            size_t i = 0;
            for (auto &a : x.m_symtab->get_scope()) {
        if(indent) {
            s.append("\n"+indtd);
            inc_indent();
        }
              s.append(a.first + ": ");
                this->visit_symbol(*a.second);
                if (i < x.m_symtab->get_scope().size()-1) { 
                    s.append(", ");
                    if(indent) {
                            for(int times = 0; times < tmp2; times++)
                             dec_indent();
                     }
                }
                i++;
            }
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - curly[tmp] ); times++)
                dec_indent();
            curly[tmp] = 0; if(tmp > 1) tmp--;
            tmp2++;
            s.append("\n"+indtd);
        }
        s.append("})");
        if(indent) dec_indent();
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_name);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_expr(*x.m_args[i]);
            if (i < x.n_args-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
            if (i < x.n_body-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        visit_abiType(x.m_abi);
        s.append(" ");
        visit_accessType(x.m_access);
        s.append(" ");
        visit_deftypeType(x.m_deftype);
        s.append(" ");
        if (x.m_bindc_name) {
            s.append("\"" + std::string(x.m_bindc_name) + "\"");
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_pure) {
            s.append(".true.");
        } else {
            s.append(".false.");
        }
        s.append(" ");
        if (x.m_module) {
            s.append(".true.");
        } else {
            s.append(".false.");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Function(const Function_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Function");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        if(indent) {
            s.append("\n"+indtd);
            round[++tmp1] = indent_level;
            inc_indent();
        }
        s.append("(");
        if (use_colors) {
            s.append(color(fg::yellow));
        }
        s.append("SymbolTable");
        if (use_colors) {
            s.append(color(fg::reset));
        }
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_symtab->get_counter());
        s.append(" ");
        if(indent) {
            s.append("\n"+indtd);
            curly[++tmp] = indent_level;
            tmp2 = 1;
            inc_indent();
        }
        s.append("{");
        {
            size_t i = 0;
            for (auto &a : x.m_symtab->get_scope()) {
        if(indent) {
            s.append("\n"+indtd);
            inc_indent();
        }
              s.append(a.first + ": ");
                this->visit_symbol(*a.second);
                if (i < x.m_symtab->get_scope().size()-1) { 
                    s.append(", ");
                    if(indent) {
                            for(int times = 0; times < tmp2; times++)
                             dec_indent();
                     }
                }
                i++;
            }
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - curly[tmp] ); times++)
                dec_indent();
            curly[tmp] = 0; if(tmp > 1) tmp--;
            tmp2++;
            s.append("\n"+indtd);
        }
        s.append("})");
        if(indent) dec_indent();
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_name);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_expr(*x.m_args[i]);
            if (i < x.n_args-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
            if (i < x.n_body-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        self().visit_expr(*x.m_return_var);
        s.append(" ");
        visit_abiType(x.m_abi);
        s.append(" ");
        visit_accessType(x.m_access);
        s.append(" ");
        visit_deftypeType(x.m_deftype);
        s.append(" ");
        if (x.m_bindc_name) {
            s.append("\"" + std::string(x.m_bindc_name) + "\"");
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_GenericProcedure(const GenericProcedure_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("GenericProcedure");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_parent_symtab->get_counter());
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_name);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_procs; i++) {
            self().visit_symbol(*x.m_procs[i]);
            if (i < x.n_procs-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        visit_accessType(x.m_access);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_CustomOperator(const CustomOperator_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("CustomOperator");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_parent_symtab->get_counter());
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_name);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_procs; i++) {
            self().visit_symbol(*x.m_procs[i]);
            if (i < x.n_procs-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        visit_accessType(x.m_access);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ExternalSymbol(const ExternalSymbol_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ExternalSymbol");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_parent_symtab->get_counter());
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_name);
        s.append(" ");
        self().visit_symbol(*x.m_external);
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_module_name);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_scope_names; i++) {
            s.append(x.m_scope_names[i]);
            if (i < x.n_scope_names-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_original_name);
        s.append(" ");
        visit_accessType(x.m_access);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_DerivedType(const DerivedType_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("DerivedType");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        if(indent) {
            s.append("\n"+indtd);
            round[++tmp1] = indent_level;
            inc_indent();
        }
        s.append("(");
        if (use_colors) {
            s.append(color(fg::yellow));
        }
        s.append("SymbolTable");
        if (use_colors) {
            s.append(color(fg::reset));
        }
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_symtab->get_counter());
        s.append(" ");
        if(indent) {
            s.append("\n"+indtd);
            curly[++tmp] = indent_level;
            tmp2 = 1;
            inc_indent();
        }
        s.append("{");
        {
            size_t i = 0;
            for (auto &a : x.m_symtab->get_scope()) {
        if(indent) {
            s.append("\n"+indtd);
            inc_indent();
        }
              s.append(a.first + ": ");
                this->visit_symbol(*a.second);
                if (i < x.m_symtab->get_scope().size()-1) { 
                    s.append(", ");
                    if(indent) {
                            for(int times = 0; times < tmp2; times++)
                             dec_indent();
                     }
                }
                i++;
            }
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - curly[tmp] ); times++)
                dec_indent();
            curly[tmp] = 0; if(tmp > 1) tmp--;
            tmp2++;
            s.append("\n"+indtd);
        }
        s.append("})");
        if(indent) dec_indent();
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_name);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_members; i++) {
            s.append(x.m_members[i]);
            if (i < x.n_members-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        visit_abiType(x.m_abi);
        s.append(" ");
        visit_accessType(x.m_access);
        s.append(" ");
        if (x.m_parent) {
            self().visit_symbol(*x.m_parent);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Variable(const Variable_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Variable");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_parent_symtab->get_counter());
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_name);
        s.append(" ");
        visit_intentType(x.m_intent);
        s.append(" ");
        if (x.m_symbolic_value) {
            self().visit_expr(*x.m_symbolic_value);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        s.append(" ");
        visit_storage_typeType(x.m_storage);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        visit_abiType(x.m_abi);
        s.append(" ");
        visit_accessType(x.m_access);
        s.append(" ");
        visit_presenceType(x.m_presence);
        s.append(" ");
        if (x.m_value_attr) {
            s.append(".true.");
        } else {
            s.append(".false.");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ClassType(const ClassType_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ClassType");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        if(indent) {
            s.append("\n"+indtd);
            round[++tmp1] = indent_level;
            inc_indent();
        }
        s.append("(");
        if (use_colors) {
            s.append(color(fg::yellow));
        }
        s.append("SymbolTable");
        if (use_colors) {
            s.append(color(fg::reset));
        }
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_symtab->get_counter());
        s.append(" ");
        if(indent) {
            s.append("\n"+indtd);
            curly[++tmp] = indent_level;
            tmp2 = 1;
            inc_indent();
        }
        s.append("{");
        {
            size_t i = 0;
            for (auto &a : x.m_symtab->get_scope()) {
        if(indent) {
            s.append("\n"+indtd);
            inc_indent();
        }
              s.append(a.first + ": ");
                this->visit_symbol(*a.second);
                if (i < x.m_symtab->get_scope().size()-1) { 
                    s.append(", ");
                    if(indent) {
                            for(int times = 0; times < tmp2; times++)
                             dec_indent();
                     }
                }
                i++;
            }
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - curly[tmp] ); times++)
                dec_indent();
            curly[tmp] = 0; if(tmp > 1) tmp--;
            tmp2++;
            s.append("\n"+indtd);
        }
        s.append("})");
        if(indent) dec_indent();
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_name);
        s.append(" ");
        visit_abiType(x.m_abi);
        s.append(" ");
        visit_accessType(x.m_access);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_ClassProcedure(const ClassProcedure_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ClassProcedure");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_parent_symtab->get_counter());
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_name);
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_proc_name);
        s.append(" ");
        self().visit_symbol(*x.m_proc);
        s.append(" ");
        visit_abiType(x.m_abi);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_AssociateBlock(const AssociateBlock_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("AssociateBlock");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        if(indent) {
            s.append("\n"+indtd);
            round[++tmp1] = indent_level;
            inc_indent();
        }
        s.append("(");
        if (use_colors) {
            s.append(color(fg::yellow));
        }
        s.append("SymbolTable");
        if (use_colors) {
            s.append(color(fg::reset));
        }
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_symtab->get_counter());
        s.append(" ");
        if(indent) {
            s.append("\n"+indtd);
            curly[++tmp] = indent_level;
            tmp2 = 1;
            inc_indent();
        }
        s.append("{");
        {
            size_t i = 0;
            for (auto &a : x.m_symtab->get_scope()) {
        if(indent) {
            s.append("\n"+indtd);
            inc_indent();
        }
              s.append(a.first + ": ");
                this->visit_symbol(*a.second);
                if (i < x.m_symtab->get_scope().size()-1) { 
                    s.append(", ");
                    if(indent) {
                            for(int times = 0; times < tmp2; times++)
                             dec_indent();
                     }
                }
                i++;
            }
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - curly[tmp] ); times++)
                dec_indent();
            curly[tmp] = 0; if(tmp > 1) tmp--;
            tmp2++;
            s.append("\n"+indtd);
        }
        s.append("})");
        if(indent) dec_indent();
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_name);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
            if (i < x.n_body-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Block(const Block_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Block");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        if(indent) {
            s.append("\n"+indtd);
            round[++tmp1] = indent_level;
            inc_indent();
        }
        s.append("(");
        if (use_colors) {
            s.append(color(fg::yellow));
        }
        s.append("SymbolTable");
        if (use_colors) {
            s.append(color(fg::reset));
        }
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_symtab->get_counter());
        s.append(" ");
        if(indent) {
            s.append("\n"+indtd);
            curly[++tmp] = indent_level;
            tmp2 = 1;
            inc_indent();
        }
        s.append("{");
        {
            size_t i = 0;
            for (auto &a : x.m_symtab->get_scope()) {
        if(indent) {
            s.append("\n"+indtd);
            inc_indent();
        }
              s.append(a.first + ": ");
                this->visit_symbol(*a.second);
                if (i < x.m_symtab->get_scope().size()-1) { 
                    s.append(", ");
                    if(indent) {
                            for(int times = 0; times < tmp2; times++)
                             dec_indent();
                     }
                }
                i++;
            }
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - curly[tmp] ); times++)
                dec_indent();
            curly[tmp] = 0; if(tmp > 1) tmp--;
            tmp2++;
            s.append("\n"+indtd);
        }
        s.append("})");
        if(indent) dec_indent();
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_name);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
            if (i < x.n_body-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_storage_typeType(const storage_typeType &x) {
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::green));
        }
        switch (x) {
            case (storage_typeType::Default) : {
                if(indent) s.append("\n"+indtd);
                s.append("Default");
             break; }
            case (storage_typeType::Save) : {
                if(indent) s.append("\n"+indtd);
                s.append("Save");
             break; }
            case (storage_typeType::Parameter) : {
                if(indent) s.append("\n"+indtd);
                s.append("Parameter");
             break; }
            case (storage_typeType::Allocatable) : {
                if(indent) s.append("\n"+indtd);
                s.append("Allocatable");
             break; }
        }
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
    }
    void visit_accessType(const accessType &x) {
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::green));
        }
        switch (x) {
            case (accessType::Public) : {
                if(indent) s.append("\n"+indtd);
                s.append("Public");
             break; }
            case (accessType::Private) : {
                if(indent) s.append("\n"+indtd);
                s.append("Private");
             break; }
        }
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
    }
    void visit_intentType(const intentType &x) {
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::green));
        }
        switch (x) {
            case (intentType::Local) : {
                if(indent) s.append("\n"+indtd);
                s.append("Local");
             break; }
            case (intentType::In) : {
                if(indent) s.append("\n"+indtd);
                s.append("In");
             break; }
            case (intentType::Out) : {
                if(indent) s.append("\n"+indtd);
                s.append("Out");
             break; }
            case (intentType::InOut) : {
                if(indent) s.append("\n"+indtd);
                s.append("InOut");
             break; }
            case (intentType::ReturnVar) : {
                if(indent) s.append("\n"+indtd);
                s.append("ReturnVar");
             break; }
            case (intentType::Unspecified) : {
                if(indent) s.append("\n"+indtd);
                s.append("Unspecified");
             break; }
        }
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
    }
    void visit_deftypeType(const deftypeType &x) {
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::green));
        }
        switch (x) {
            case (deftypeType::Implementation) : {
                if(indent) s.append("\n"+indtd);
                s.append("Implementation");
             break; }
            case (deftypeType::Interface) : {
                if(indent) s.append("\n"+indtd);
                s.append("Interface");
             break; }
        }
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
    }
    void visit_presenceType(const presenceType &x) {
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::green));
        }
        switch (x) {
            case (presenceType::Required) : {
                if(indent) s.append("\n"+indtd);
                s.append("Required");
             break; }
            case (presenceType::Optional) : {
                if(indent) s.append("\n"+indtd);
                s.append("Optional");
             break; }
        }
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
    }
    void visit_abiType(const abiType &x) {
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::green));
        }
        switch (x) {
            case (abiType::Source) : {
                if(indent) s.append("\n"+indtd);
                s.append("Source");
             break; }
            case (abiType::LFortranModule) : {
                if(indent) s.append("\n"+indtd);
                s.append("LFortranModule");
             break; }
            case (abiType::GFortranModule) : {
                if(indent) s.append("\n"+indtd);
                s.append("GFortranModule");
             break; }
            case (abiType::BindC) : {
                if(indent) s.append("\n"+indtd);
                s.append("BindC");
             break; }
            case (abiType::Interactive) : {
                if(indent) s.append("\n"+indtd);
                s.append("Interactive");
             break; }
            case (abiType::Intrinsic) : {
                if(indent) s.append("\n"+indtd);
                s.append("Intrinsic");
             break; }
        }
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
    }
    void visit_Allocate(const Allocate_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Allocate");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_alloc_arg(x.m_args[i]);
            if (i < x.n_args-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        if (x.m_stat) {
            self().visit_expr(*x.m_stat);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_errmsg) {
            self().visit_expr(*x.m_errmsg);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_source) {
            self().visit_expr(*x.m_source);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Assign(const Assign_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Assign");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append(std::to_string(x.m_label));
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_variable);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_Assignment(const Assignment_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("=");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_target);
        s.append(" ");
        self().visit_expr(*x.m_value);
        s.append(" ");
        if (x.m_overloaded) {
            self().visit_stmt(*x.m_overloaded);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Associate(const Associate_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("=>");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_target);
        s.append(" ");
        self().visit_expr(*x.m_value);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Cycle(const Cycle_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Cycle");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_ExplicitDeallocate(const ExplicitDeallocate_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ExplicitDeallocate");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_vars; i++) {
            self().visit_symbol(*x.m_vars[i]);
            if (i < x.n_vars-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ImplicitDeallocate(const ImplicitDeallocate_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ImplicitDeallocate");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_vars; i++) {
            self().visit_symbol(*x.m_vars[i]);
            if (i < x.n_vars-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_DoConcurrentLoop(const DoConcurrentLoop_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("DoConcurrentLoop");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_do_loop_head(x.m_head);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
            if (i < x.n_body-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_DoLoop(const DoLoop_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("DoLoop");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_do_loop_head(x.m_head);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
            if (i < x.n_body-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ErrorStop(const ErrorStop_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ErrorStop");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        if (x.m_code) {
            self().visit_expr(*x.m_code);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Exit(const Exit_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Exit");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_ForAllSingle(const ForAllSingle_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ForAllSingle");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_do_loop_head(x.m_head);
        s.append(" ");
        self().visit_stmt(*x.m_assign_stmt);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_GoTo(const GoTo_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("GoTo");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append(std::to_string(x.m_target_id));
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_GoToTarget(const GoToTarget_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("GoToTarget");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append(std::to_string(x.m_id));
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_If(const If_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("If");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_test);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
            if (i < x.n_body-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_orelse; i++) {
            self().visit_stmt(*x.m_orelse[i]);
            if (i < x.n_orelse-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_IfArithmetic(const IfArithmetic_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("IfArithmetic");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_test);
        s.append(" ");
        s.append(std::to_string(x.m_lt_label));
        s.append(" ");
        s.append(std::to_string(x.m_eq_label));
        s.append(" ");
        s.append(std::to_string(x.m_gt_label));
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Print(const Print_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Print");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        if (x.m_fmt) {
            self().visit_expr(*x.m_fmt);
        } else {
            s.append("()");
        }
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_values; i++) {
            self().visit_expr(*x.m_values[i]);
            if (i < x.n_values-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_FileOpen(const FileOpen_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("FileOpen");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append(std::to_string(x.m_label));
        s.append(" ");
        if (x.m_newunit) {
            self().visit_expr(*x.m_newunit);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_filename) {
            self().visit_expr(*x.m_filename);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_status) {
            self().visit_expr(*x.m_status);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_FileClose(const FileClose_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("FileClose");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append(std::to_string(x.m_label));
        s.append(" ");
        if (x.m_unit) {
            self().visit_expr(*x.m_unit);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_iostat) {
            self().visit_expr(*x.m_iostat);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_iomsg) {
            self().visit_expr(*x.m_iomsg);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_err) {
            self().visit_expr(*x.m_err);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_status) {
            self().visit_expr(*x.m_status);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_FileRead(const FileRead_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("FileRead");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append(std::to_string(x.m_label));
        s.append(" ");
        if (x.m_unit) {
            self().visit_expr(*x.m_unit);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_fmt) {
            self().visit_expr(*x.m_fmt);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_iomsg) {
            self().visit_expr(*x.m_iomsg);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_iostat) {
            self().visit_expr(*x.m_iostat);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_id) {
            self().visit_expr(*x.m_id);
        } else {
            s.append("()");
        }
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_values; i++) {
            self().visit_expr(*x.m_values[i]);
            if (i < x.n_values-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_FileRewind(const FileRewind_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("FileRewind");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append(std::to_string(x.m_label));
        s.append(" ");
        if (x.m_unit) {
            self().visit_expr(*x.m_unit);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_iostat) {
            self().visit_expr(*x.m_iostat);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_err) {
            self().visit_expr(*x.m_err);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_FileInquire(const FileInquire_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("FileInquire");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append(std::to_string(x.m_label));
        s.append(" ");
        if (x.m_unit) {
            self().visit_expr(*x.m_unit);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_file) {
            self().visit_expr(*x.m_file);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_iostat) {
            self().visit_expr(*x.m_iostat);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_err) {
            self().visit_expr(*x.m_err);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_exist) {
            self().visit_expr(*x.m_exist);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_opened) {
            self().visit_expr(*x.m_opened);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_number) {
            self().visit_expr(*x.m_number);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_named) {
            self().visit_expr(*x.m_named);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_name) {
            self().visit_expr(*x.m_name);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_access) {
            self().visit_expr(*x.m_access);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_sequential) {
            self().visit_expr(*x.m_sequential);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_direct) {
            self().visit_expr(*x.m_direct);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_form) {
            self().visit_expr(*x.m_form);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_formatted) {
            self().visit_expr(*x.m_formatted);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_unformatted) {
            self().visit_expr(*x.m_unformatted);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_recl) {
            self().visit_expr(*x.m_recl);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_nextrec) {
            self().visit_expr(*x.m_nextrec);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_blank) {
            self().visit_expr(*x.m_blank);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_position) {
            self().visit_expr(*x.m_position);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_action) {
            self().visit_expr(*x.m_action);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_read) {
            self().visit_expr(*x.m_read);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_write) {
            self().visit_expr(*x.m_write);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_readwrite) {
            self().visit_expr(*x.m_readwrite);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_delim) {
            self().visit_expr(*x.m_delim);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_pad) {
            self().visit_expr(*x.m_pad);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_flen) {
            self().visit_expr(*x.m_flen);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_blocksize) {
            self().visit_expr(*x.m_blocksize);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_convert) {
            self().visit_expr(*x.m_convert);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_carriagecontrol) {
            self().visit_expr(*x.m_carriagecontrol);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_iolength) {
            self().visit_expr(*x.m_iolength);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_FileWrite(const FileWrite_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("FileWrite");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append(std::to_string(x.m_label));
        s.append(" ");
        if (x.m_unit) {
            self().visit_expr(*x.m_unit);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_fmt) {
            self().visit_expr(*x.m_fmt);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_iomsg) {
            self().visit_expr(*x.m_iomsg);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_iostat) {
            self().visit_expr(*x.m_iostat);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_id) {
            self().visit_expr(*x.m_id);
        } else {
            s.append("()");
        }
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_values; i++) {
            self().visit_expr(*x.m_values[i]);
            if (i < x.n_values-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Return(const Return_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Return");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_Select(const Select_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Select");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_test);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_case_stmt(*x.m_body[i]);
            if (i < x.n_body-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_default; i++) {
            self().visit_stmt(*x.m_default[i]);
            if (i < x.n_default-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Stop(const Stop_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Stop");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        if (x.m_code) {
            self().visit_expr(*x.m_code);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Assert(const Assert_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Assert");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_test);
        s.append(" ");
        if (x.m_msg) {
            self().visit_expr(*x.m_msg);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_SubroutineCall(const SubroutineCall_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("SubroutineCall");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_symbol(*x.m_name);
        s.append(" ");
        if (x.m_original_name) {
            self().visit_symbol(*x.m_original_name);
        } else {
            s.append("()");
        }
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_call_arg(x.m_args[i]);
            if (i < x.n_args-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        if (x.m_dt) {
            self().visit_expr(*x.m_dt);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Where(const Where_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Where");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_test);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
            if (i < x.n_body-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_orelse; i++) {
            self().visit_stmt(*x.m_orelse[i]);
            if (i < x.n_orelse-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_WhileLoop(const WhileLoop_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("WhileLoop");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_test);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
            if (i < x.n_body-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Nullify(const Nullify_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Nullify");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_vars; i++) {
            self().visit_symbol(*x.m_vars[i]);
            if (i < x.n_vars-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Flush(const Flush_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Flush");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append(std::to_string(x.m_label));
        s.append(" ");
        self().visit_expr(*x.m_unit);
        s.append(" ");
        if (x.m_err) {
            self().visit_expr(*x.m_err);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_iomsg) {
            self().visit_expr(*x.m_iomsg);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_iostat) {
            self().visit_expr(*x.m_iostat);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ListAppend(const ListAppend_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ListAppend");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_symbol(*x.m_a);
        s.append(" ");
        self().visit_expr(*x.m_ele);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_AssociateBlockCall(const AssociateBlockCall_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("AssociateBlockCall");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_symbol(*x.m_m);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_BlockCall(const BlockCall_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("BlockCall");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_symbol(*x.m_m);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_SetInsert(const SetInsert_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("SetInsert");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_symbol(*x.m_a);
        s.append(" ");
        self().visit_expr(*x.m_ele);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_SetRemove(const SetRemove_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("SetRemove");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_symbol(*x.m_a);
        s.append(" ");
        self().visit_expr(*x.m_ele);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ListInsert(const ListInsert_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ListInsert");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_symbol(*x.m_a);
        s.append(" ");
        self().visit_expr(*x.m_pos);
        s.append(" ");
        self().visit_expr(*x.m_ele);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ListRemove(const ListRemove_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ListRemove");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_symbol(*x.m_a);
        s.append(" ");
        self().visit_expr(*x.m_ele);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_DictInsert(const DictInsert_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("DictInsert");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_symbol(*x.m_a);
        s.append(" ");
        self().visit_expr(*x.m_key);
        s.append(" ");
        self().visit_expr(*x.m_value);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_BoolOp(const BoolOp_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("BoolOp");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_left);
        s.append(" ");
        visit_boolopType(x.m_op);
        s.append(" ");
        self().visit_expr(*x.m_right);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_BinOp(const BinOp_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("BinOp");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_left);
        s.append(" ");
        visit_binopType(x.m_op);
        s.append(" ");
        self().visit_expr(*x.m_right);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_overloaded) {
            self().visit_expr(*x.m_overloaded);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_StrOp(const StrOp_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("StrOp");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_left);
        s.append(" ");
        visit_stropType(x.m_op);
        s.append(" ");
        self().visit_expr(*x.m_right);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_UnaryOp(const UnaryOp_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("UnaryOp");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        visit_unaryopType(x.m_op);
        s.append(" ");
        self().visit_expr(*x.m_operand);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ComplexConstructor(const ComplexConstructor_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ComplexConstructor");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_re);
        s.append(" ");
        self().visit_expr(*x.m_im);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_NamedExpr(const NamedExpr_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("NamedExpr");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_target);
        s.append(" ");
        self().visit_expr(*x.m_value);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Compare(const Compare_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Compare");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_left);
        s.append(" ");
        visit_cmpopType(x.m_op);
        s.append(" ");
        self().visit_expr(*x.m_right);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_overloaded) {
            self().visit_expr(*x.m_overloaded);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_IfExp(const IfExp_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("IfExp");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_test);
        s.append(" ");
        self().visit_expr(*x.m_body);
        s.append(" ");
        self().visit_expr(*x.m_orelse);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_FunctionCall(const FunctionCall_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("FunctionCall");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_symbol(*x.m_name);
        s.append(" ");
        if (x.m_original_name) {
            self().visit_symbol(*x.m_original_name);
        } else {
            s.append("()");
        }
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_call_arg(x.m_args[i]);
            if (i < x.n_args-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_dt) {
            self().visit_expr(*x.m_dt);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_DerivedTypeConstructor(const DerivedTypeConstructor_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("DerivedTypeConstructor");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_symbol(*x.m_dt_sym);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_expr(*x.m_args[i]);
            if (i < x.n_args-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        self().visit_ttype(*x.m_type);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ImpliedDoLoop(const ImpliedDoLoop_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ImpliedDoLoop");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_values; i++) {
            self().visit_expr(*x.m_values[i]);
            if (i < x.n_values-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        self().visit_expr(*x.m_var);
        s.append(" ");
        self().visit_expr(*x.m_start);
        s.append(" ");
        self().visit_expr(*x.m_end);
        s.append(" ");
        if (x.m_increment) {
            self().visit_expr(*x.m_increment);
        } else {
            s.append("()");
        }
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_IntegerConstant(const IntegerConstant_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("IntegerConstant");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append(std::to_string(x.m_n));
        s.append(" ");
        self().visit_ttype(*x.m_type);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_IntegerBOZ(const IntegerBOZ_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("IntegerBOZ");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append(std::to_string(x.m_v));
        s.append(" ");
        visit_integerbozType(x.m_intboz_type);
        s.append(" ");
        if (x.m_type) {
            self().visit_ttype(*x.m_type);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_RealConstant(const RealConstant_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("RealConstant");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append(std::to_string(x.m_r));
        s.append(" ");
        self().visit_ttype(*x.m_type);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ComplexConstant(const ComplexConstant_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ComplexConstant");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append(std::to_string(x.m_re));
        s.append(" ");
        s.append(std::to_string(x.m_im));
        s.append(" ");
        self().visit_ttype(*x.m_type);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_LogicalConstant(const LogicalConstant_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("LogicalConstant");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        if (x.m_value) {
            s.append(".true.");
        } else {
            s.append(".false.");
        }
        s.append(" ");
        self().visit_ttype(*x.m_type);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ListConstant(const ListConstant_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ListConstant");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_expr(*x.m_args[i]);
            if (i < x.n_args-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        self().visit_ttype(*x.m_type);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ListLen(const ListLen_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ListLen");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_arg);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ListConcat(const ListConcat_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ListConcat");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_left);
        s.append(" ");
        self().visit_expr(*x.m_right);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ArrayConstant(const ArrayConstant_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ArrayConstant");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_expr(*x.m_args[i]);
            if (i < x.n_args-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        self().visit_ttype(*x.m_type);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_SetConstant(const SetConstant_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("SetConstant");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_elements; i++) {
            self().visit_expr(*x.m_elements[i]);
            if (i < x.n_elements-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        self().visit_ttype(*x.m_type);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_SetLen(const SetLen_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("SetLen");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_arg);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_TupleConstant(const TupleConstant_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("TupleConstant");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_elements; i++) {
            self().visit_expr(*x.m_elements[i]);
            if (i < x.n_elements-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        self().visit_ttype(*x.m_type);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_TupleLen(const TupleLen_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("TupleLen");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_arg);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_StringConstant(const StringConstant_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("StringConstant");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append("\"" + std::string(x.m_s) + "\"");
        s.append(" ");
        self().visit_ttype(*x.m_type);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_StringConcat(const StringConcat_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("StringConcat");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_left);
        s.append(" ");
        self().visit_expr(*x.m_right);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_StringLen(const StringLen_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("StringLen");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_arg);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_DictConstant(const DictConstant_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("DictConstant");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_keys; i++) {
            self().visit_expr(*x.m_keys[i]);
            if (i < x.n_keys-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_values; i++) {
            self().visit_expr(*x.m_values[i]);
            if (i < x.n_values-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        self().visit_ttype(*x.m_type);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_DictLen(const DictLen_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("DictLen");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_arg);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Var(const Var_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Var");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_symbol(*x.m_v);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ArrayRef(const ArrayRef_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ArrayRef");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_symbol(*x.m_v);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_array_index(x.m_args[i]);
            if (i < x.n_args-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ArraySize(const ArraySize_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ArraySize");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_v);
        s.append(" ");
        if (x.m_dim) {
            self().visit_expr(*x.m_dim);
        } else {
            s.append("()");
        }
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ArrayBound(const ArrayBound_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ArrayBound");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_v);
        s.append(" ");
        if (x.m_dim) {
            self().visit_expr(*x.m_dim);
        } else {
            s.append("()");
        }
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        visit_arrayboundType(x.m_bound);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ArrayTranspose(const ArrayTranspose_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ArrayTranspose");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_matrix);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ArrayMatMul(const ArrayMatMul_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ArrayMatMul");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_matrix_a);
        s.append(" ");
        self().visit_expr(*x.m_matrix_b);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ArrayPack(const ArrayPack_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ArrayPack");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_array);
        s.append(" ");
        self().visit_expr(*x.m_mask);
        s.append(" ");
        if (x.m_vector) {
            self().visit_expr(*x.m_vector);
        } else {
            s.append("()");
        }
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Transfer(const Transfer_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Transfer");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_source);
        s.append(" ");
        self().visit_expr(*x.m_mold);
        s.append(" ");
        if (x.m_size) {
            self().visit_expr(*x.m_size);
        } else {
            s.append("()");
        }
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_DerivedRef(const DerivedRef_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("DerivedRef");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_v);
        s.append(" ");
        self().visit_symbol(*x.m_m);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Cast(const Cast_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Cast");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_arg);
        s.append(" ");
        visit_cast_kindType(x.m_kind);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ComplexRe(const ComplexRe_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ComplexRe");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_arg);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_ComplexIm(const ComplexIm_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("ComplexIm");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_expr(*x.m_arg);
        s.append(" ");
        self().visit_ttype(*x.m_type);
        s.append(" ");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_DictItem(const DictItem_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("DictItem");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_symbol(*x.m_a);
        s.append(" ");
        self().visit_expr(*x.m_key);
        s.append(" ");
        if (x.m_default) {
            self().visit_expr(*x.m_default);
        } else {
            s.append("()");
        }
        s.append(" ");
        self().visit_ttype(*x.m_type);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Integer(const Integer_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Integer");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append(std::to_string(x.m_kind));
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
            if (i < x.n_dims-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Real(const Real_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Real");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append(std::to_string(x.m_kind));
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
            if (i < x.n_dims-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Complex(const Complex_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Complex");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append(std::to_string(x.m_kind));
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
            if (i < x.n_dims-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Character(const Character_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Character");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append(std::to_string(x.m_kind));
        s.append(" ");
        s.append(std::to_string(x.m_len));
        s.append(" ");
        if (x.m_len_expr) {
            self().visit_expr(*x.m_len_expr);
        } else {
            s.append("()");
        }
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
            if (i < x.n_dims-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Logical(const Logical_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Logical");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append(std::to_string(x.m_kind));
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
            if (i < x.n_dims-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Set(const Set_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Set");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_ttype(*x.m_type);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_List(const List_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("List");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_ttype(*x.m_type);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Tuple(const Tuple_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Tuple");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_type; i++) {
            self().visit_ttype(*x.m_type[i]);
            if (i < x.n_type-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Derived(const Derived_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Derived");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_symbol(*x.m_derived_type);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
            if (i < x.n_dims-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Class(const Class_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Class");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_symbol(*x.m_class_type);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
            if (i < x.n_dims-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Dict(const Dict_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Dict");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_ttype(*x.m_key_type);
        s.append(" ");
        self().visit_ttype(*x.m_value_type);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Pointer(const Pointer_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Pointer");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        self().visit_ttype(*x.m_type);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_boolopType(const boolopType &x) {
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::green));
        }
        switch (x) {
            case (boolopType::And) : {
                if(indent) s.append("\n"+indtd);
                s.append("And");
             break; }
            case (boolopType::Or) : {
                if(indent) s.append("\n"+indtd);
                s.append("Or");
             break; }
            case (boolopType::Xor) : {
                if(indent) s.append("\n"+indtd);
                s.append("Xor");
             break; }
            case (boolopType::NEqv) : {
                if(indent) s.append("\n"+indtd);
                s.append("NEqv");
             break; }
            case (boolopType::Eqv) : {
                if(indent) s.append("\n"+indtd);
                s.append("Eqv");
             break; }
        }
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
    }
    void visit_binopType(const binopType &x) {
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::green));
        }
        switch (x) {
            case (binopType::Add) : {
                if(indent) s.append("\n"+indtd);
                s.append("Add");
             break; }
            case (binopType::Sub) : {
                if(indent) s.append("\n"+indtd);
                s.append("Sub");
             break; }
            case (binopType::Mul) : {
                if(indent) s.append("\n"+indtd);
                s.append("Mul");
             break; }
            case (binopType::Div) : {
                if(indent) s.append("\n"+indtd);
                s.append("Div");
             break; }
            case (binopType::Pow) : {
                if(indent) s.append("\n"+indtd);
                s.append("Pow");
             break; }
        }
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
    }
    void visit_unaryopType(const unaryopType &x) {
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::green));
        }
        switch (x) {
            case (unaryopType::Invert) : {
                if(indent) s.append("\n"+indtd);
                s.append("Invert");
             break; }
            case (unaryopType::Not) : {
                if(indent) s.append("\n"+indtd);
                s.append("Not");
             break; }
            case (unaryopType::UAdd) : {
                if(indent) s.append("\n"+indtd);
                s.append("UAdd");
             break; }
            case (unaryopType::USub) : {
                if(indent) s.append("\n"+indtd);
                s.append("USub");
             break; }
        }
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
    }
    void visit_stropType(const stropType &x) {
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::green));
        }
        switch (x) {
            case (stropType::Repeat) : {
                if(indent) s.append("\n"+indtd);
                s.append("Repeat");
             break; }
        }
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
    }
    void visit_cmpopType(const cmpopType &x) {
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::green));
        }
        switch (x) {
            case (cmpopType::Eq) : {
                if(indent) s.append("\n"+indtd);
                s.append("Eq");
             break; }
            case (cmpopType::NotEq) : {
                if(indent) s.append("\n"+indtd);
                s.append("NotEq");
             break; }
            case (cmpopType::Lt) : {
                if(indent) s.append("\n"+indtd);
                s.append("Lt");
             break; }
            case (cmpopType::LtE) : {
                if(indent) s.append("\n"+indtd);
                s.append("LtE");
             break; }
            case (cmpopType::Gt) : {
                if(indent) s.append("\n"+indtd);
                s.append("Gt");
             break; }
            case (cmpopType::GtE) : {
                if(indent) s.append("\n"+indtd);
                s.append("GtE");
             break; }
        }
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
    }
    void visit_integerbozType(const integerbozType &x) {
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::green));
        }
        switch (x) {
            case (integerbozType::Binary) : {
                if(indent) s.append("\n"+indtd);
                s.append("Binary");
             break; }
            case (integerbozType::Hex) : {
                if(indent) s.append("\n"+indtd);
                s.append("Hex");
             break; }
            case (integerbozType::Octal) : {
                if(indent) s.append("\n"+indtd);
                s.append("Octal");
             break; }
        }
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
    }
    void visit_arrayboundType(const arrayboundType &x) {
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::green));
        }
        switch (x) {
            case (arrayboundType::LBound) : {
                if(indent) s.append("\n"+indtd);
                s.append("LBound");
             break; }
            case (arrayboundType::UBound) : {
                if(indent) s.append("\n"+indtd);
                s.append("UBound");
             break; }
        }
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
    }
    void visit_cast_kindType(const cast_kindType &x) {
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::green));
        }
        switch (x) {
            case (cast_kindType::RealToInteger) : {
                if(indent) s.append("\n"+indtd);
                s.append("RealToInteger");
             break; }
            case (cast_kindType::IntegerToReal) : {
                if(indent) s.append("\n"+indtd);
                s.append("IntegerToReal");
             break; }
            case (cast_kindType::RealToReal) : {
                if(indent) s.append("\n"+indtd);
                s.append("RealToReal");
             break; }
            case (cast_kindType::IntegerToInteger) : {
                if(indent) s.append("\n"+indtd);
                s.append("IntegerToInteger");
             break; }
            case (cast_kindType::RealToComplex) : {
                if(indent) s.append("\n"+indtd);
                s.append("RealToComplex");
             break; }
            case (cast_kindType::IntegerToComplex) : {
                if(indent) s.append("\n"+indtd);
                s.append("IntegerToComplex");
             break; }
            case (cast_kindType::IntegerToLogical) : {
                if(indent) s.append("\n"+indtd);
                s.append("IntegerToLogical");
             break; }
            case (cast_kindType::ComplexToComplex) : {
                if(indent) s.append("\n"+indtd);
                s.append("ComplexToComplex");
             break; }
            case (cast_kindType::ComplexToReal) : {
                if(indent) s.append("\n"+indtd);
                s.append("ComplexToReal");
             break; }
            case (cast_kindType::LogicalToInteger) : {
                if(indent) s.append("\n"+indtd);
                s.append("LogicalToInteger");
             break; }
        }
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
    }
    void visit_dimension(const dimension_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (x.m_start) {
            self().visit_expr(*x.m_start);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_end) {
            self().visit_expr(*x.m_end);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_alloc_arg(const alloc_arg_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        self().visit_symbol(*x.m_a);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
            if (i < x.n_dims-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Attribute(const Attribute_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Attribute");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_name);
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_attribute_arg(x.m_args[i]);
            if (i < x.n_args-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_attribute_arg(const attribute_arg_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if(indent) s.append("\n"+indtd);
        s.append(x.m_arg);
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_call_arg(const call_arg_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (x.m_value) {
            self().visit_expr(*x.m_value);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_Bind(const Bind_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("Bind");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append("\"" + std::string(x.m_lang) + "\"");
        s.append(" ");
        s.append("\"" + std::string(x.m_name) + "\"");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_array_index(const array_index_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (x.m_left) {
            self().visit_expr(*x.m_left);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_right) {
            self().visit_expr(*x.m_right);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_step) {
            self().visit_expr(*x.m_step);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_do_loop_head(const do_loop_head_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (x.m_v) {
            self().visit_expr(*x.m_v);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_start) {
            self().visit_expr(*x.m_start);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_end) {
            self().visit_expr(*x.m_end);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_increment) {
            self().visit_expr(*x.m_increment);
        } else {
            s.append("()");
        }
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_CaseStmt(const CaseStmt_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("CaseStmt");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_test; i++) {
            self().visit_expr(*x.m_test[i]);
            if (i < x.n_test-1) s.append(" ");
        }
        s.append("]");
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
            if (i < x.n_body-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
    void visit_CaseStmt_Range(const CaseStmt_Range_t &x) {
        if(indent) {
            tmp1++;
            round[tmp1] = indent_level;
            if(start_line) {
                start_line = false;
                s.append(indtd);
                inc_indent();
            } else {
                s.append("\n"+indtd);
                inc_indent();
            }
        }
        s.append("(");
        if (use_colors) {
            s.append(color(style::bold));
            s.append(color(fg::magenta));
        }
        s.append("CaseStmt_Range");
        if (use_colors) {
            s.append(color(fg::reset));
            s.append(color(style::reset));
        }
        s.append(" ");
        if (x.m_start) {
            self().visit_expr(*x.m_start);
        } else {
            s.append("()");
        }
        s.append(" ");
        if (x.m_end) {
            self().visit_expr(*x.m_end);
        } else {
            s.append("()");
        }
        s.append(" ");
        s.append("[");
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
            if (i < x.n_body-1) s.append(" ");
        }
        s.append("]");
        if(indent) {
            lvl = indent_level;
            for(int times = 0; times < (lvl - round[tmp1] ); times++)
                dec_indent();
            round[tmp1] = 0; if(tmp1 > 1) tmp1--;}
        s.append(")");
    }
};


/******************************************************************************/
// Serialization Visitor base class

template <class Derived>
class SerializationBaseVisitor : public BaseVisitor<Derived>
{
private:
    Derived& self() { return static_cast<Derived&>(*this); }
public:
    void visit_TranslationUnit(const TranslationUnit_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_global_scope->counter);
        self().write_int64(x.m_global_scope->get_scope().size());
        for (auto &a : x.m_global_scope->get_scope()) {
            if (ASR::is_a<ASR::Subroutine_t>(*a.second) || ASR::is_a<ASR::Function_t>(*a.second)) {
                continue;
            }
            self().write_string(a.first);
            this->visit_symbol(*a.second);
        }
        for (auto &a : x.m_global_scope->get_scope()) {
            if (ASR::is_a<ASR::Subroutine_t>(*a.second) || ASR::is_a<ASR::Function_t>(*a.second)) {
                self().write_string(a.first);
                this->visit_symbol(*a.second);
            }
        }
        self().write_int64(x.n_items);
        for (size_t i=0; i<x.n_items; i++) {
            self().write_int8(x.m_items[i]->type);
            self().visit_asr(*x.m_items[i]);
        }
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_Program(const Program_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_symtab->counter);
        self().write_int64(x.m_symtab->get_scope().size());
        for (auto &a : x.m_symtab->get_scope()) {
            if (ASR::is_a<ASR::Subroutine_t>(*a.second) || ASR::is_a<ASR::Function_t>(*a.second)) {
                continue;
            }
            self().write_string(a.first);
            this->visit_symbol(*a.second);
        }
        for (auto &a : x.m_symtab->get_scope()) {
            if (ASR::is_a<ASR::Subroutine_t>(*a.second) || ASR::is_a<ASR::Function_t>(*a.second)) {
                self().write_string(a.first);
                this->visit_symbol(*a.second);
            }
        }
        self().write_string(x.m_name);
        self().write_int64(x.n_dependencies);
        for (size_t i=0; i<x.n_dependencies; i++) {
            self().write_string(x.m_dependencies[i]);
        }
        self().write_int64(x.n_body);
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
    }
    void visit_Module(const Module_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_symtab->counter);
        self().write_int64(x.m_symtab->get_scope().size());
        for (auto &a : x.m_symtab->get_scope()) {
            if (ASR::is_a<ASR::Subroutine_t>(*a.second) || ASR::is_a<ASR::Function_t>(*a.second)) {
                continue;
            }
            self().write_string(a.first);
            this->visit_symbol(*a.second);
        }
        for (auto &a : x.m_symtab->get_scope()) {
            if (ASR::is_a<ASR::Subroutine_t>(*a.second) || ASR::is_a<ASR::Function_t>(*a.second)) {
                self().write_string(a.first);
                this->visit_symbol(*a.second);
            }
        }
        self().write_string(x.m_name);
        self().write_int64(x.n_dependencies);
        for (size_t i=0; i<x.n_dependencies; i++) {
            self().write_string(x.m_dependencies[i]);
        }
        if (x.m_loaded_from_mod) {
            self().write_bool(true);
        } else {
            self().write_bool(false);
        }
        if (x.m_intrinsic) {
            self().write_bool(true);
        } else {
            self().write_bool(false);
        }
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_Subroutine(const Subroutine_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_symtab->counter);
        self().write_int64(x.m_symtab->get_scope().size());
        for (auto &a : x.m_symtab->get_scope()) {
            if (ASR::is_a<ASR::Subroutine_t>(*a.second) || ASR::is_a<ASR::Function_t>(*a.second)) {
                continue;
            }
            self().write_string(a.first);
            this->visit_symbol(*a.second);
        }
        for (auto &a : x.m_symtab->get_scope()) {
            if (ASR::is_a<ASR::Subroutine_t>(*a.second) || ASR::is_a<ASR::Function_t>(*a.second)) {
                self().write_string(a.first);
                this->visit_symbol(*a.second);
            }
        }
        self().write_string(x.m_name);
        self().write_int64(x.n_args);
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_expr(*x.m_args[i]);
        }
        self().write_int64(x.n_body);
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
        visit_abiType(x.m_abi);
        visit_accessType(x.m_access);
        visit_deftypeType(x.m_deftype);
        if (x.m_bindc_name) {
            self().write_bool(true);
            self().write_string(x.m_bindc_name);
        } else {
            self().write_bool(false);
        }
        if (x.m_pure) {
            self().write_bool(true);
        } else {
            self().write_bool(false);
        }
        if (x.m_module) {
            self().write_bool(true);
        } else {
            self().write_bool(false);
        }
    }
    void visit_Function(const Function_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_symtab->counter);
        self().write_int64(x.m_symtab->get_scope().size());
        for (auto &a : x.m_symtab->get_scope()) {
            if (ASR::is_a<ASR::Subroutine_t>(*a.second) || ASR::is_a<ASR::Function_t>(*a.second)) {
                continue;
            }
            self().write_string(a.first);
            this->visit_symbol(*a.second);
        }
        for (auto &a : x.m_symtab->get_scope()) {
            if (ASR::is_a<ASR::Subroutine_t>(*a.second) || ASR::is_a<ASR::Function_t>(*a.second)) {
                self().write_string(a.first);
                this->visit_symbol(*a.second);
            }
        }
        self().write_string(x.m_name);
        self().write_int64(x.n_args);
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_expr(*x.m_args[i]);
        }
        self().write_int64(x.n_body);
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
        self().visit_expr(*x.m_return_var);
        visit_abiType(x.m_abi);
        visit_accessType(x.m_access);
        visit_deftypeType(x.m_deftype);
        if (x.m_bindc_name) {
            self().write_bool(true);
            self().write_string(x.m_bindc_name);
        } else {
            self().write_bool(false);
        }
    }
    void visit_GenericProcedure(const GenericProcedure_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_parent_symtab->counter);
        self().write_string(x.m_name);
        self().write_int64(x.n_procs);
        for (size_t i=0; i<x.n_procs; i++) {
            self().write_symbol(*x.m_procs[i]);
        }
        visit_accessType(x.m_access);
    }
    void visit_CustomOperator(const CustomOperator_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_parent_symtab->counter);
        self().write_string(x.m_name);
        self().write_int64(x.n_procs);
        for (size_t i=0; i<x.n_procs; i++) {
            self().write_symbol(*x.m_procs[i]);
        }
        visit_accessType(x.m_access);
    }
    void visit_ExternalSymbol(const ExternalSymbol_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_parent_symtab->counter);
        self().write_string(x.m_name);
        // We skip the symbol for ExternalSymbol
        self().write_string(x.m_module_name);
        self().write_int64(x.n_scope_names);
        for (size_t i=0; i<x.n_scope_names; i++) {
            self().write_string(x.m_scope_names[i]);
        }
        self().write_string(x.m_original_name);
        visit_accessType(x.m_access);
    }
    void visit_DerivedType(const DerivedType_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_symtab->counter);
        self().write_int64(x.m_symtab->get_scope().size());
        for (auto &a : x.m_symtab->get_scope()) {
            if (ASR::is_a<ASR::Subroutine_t>(*a.second) || ASR::is_a<ASR::Function_t>(*a.second)) {
                continue;
            }
            self().write_string(a.first);
            this->visit_symbol(*a.second);
        }
        for (auto &a : x.m_symtab->get_scope()) {
            if (ASR::is_a<ASR::Subroutine_t>(*a.second) || ASR::is_a<ASR::Function_t>(*a.second)) {
                self().write_string(a.first);
                this->visit_symbol(*a.second);
            }
        }
        self().write_string(x.m_name);
        self().write_int64(x.n_members);
        for (size_t i=0; i<x.n_members; i++) {
            self().write_string(x.m_members[i]);
        }
        visit_abiType(x.m_abi);
        visit_accessType(x.m_access);
        if (x.m_parent) {
            self().write_bool(true);
            self().write_symbol(*x.m_parent);
        } else {
            self().write_bool(false);
        }
    }
    void visit_Variable(const Variable_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_parent_symtab->counter);
        self().write_string(x.m_name);
        visit_intentType(x.m_intent);
        if (x.m_symbolic_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_symbolic_value);
        } else {
            self().write_bool(false);
        }
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
        visit_storage_typeType(x.m_storage);
        self().visit_ttype(*x.m_type);
        visit_abiType(x.m_abi);
        visit_accessType(x.m_access);
        visit_presenceType(x.m_presence);
        if (x.m_value_attr) {
            self().write_bool(true);
        } else {
            self().write_bool(false);
        }
    }
    void visit_ClassType(const ClassType_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_symtab->counter);
        self().write_int64(x.m_symtab->get_scope().size());
        for (auto &a : x.m_symtab->get_scope()) {
            if (ASR::is_a<ASR::Subroutine_t>(*a.second) || ASR::is_a<ASR::Function_t>(*a.second)) {
                continue;
            }
            self().write_string(a.first);
            this->visit_symbol(*a.second);
        }
        for (auto &a : x.m_symtab->get_scope()) {
            if (ASR::is_a<ASR::Subroutine_t>(*a.second) || ASR::is_a<ASR::Function_t>(*a.second)) {
                self().write_string(a.first);
                this->visit_symbol(*a.second);
            }
        }
        self().write_string(x.m_name);
        visit_abiType(x.m_abi);
        visit_accessType(x.m_access);
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_ClassProcedure(const ClassProcedure_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_parent_symtab->counter);
        self().write_string(x.m_name);
        self().write_string(x.m_proc_name);
        self().write_symbol(*x.m_proc);
        visit_abiType(x.m_abi);
    }
    void visit_AssociateBlock(const AssociateBlock_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_symtab->counter);
        self().write_int64(x.m_symtab->get_scope().size());
        for (auto &a : x.m_symtab->get_scope()) {
            if (ASR::is_a<ASR::Subroutine_t>(*a.second) || ASR::is_a<ASR::Function_t>(*a.second)) {
                continue;
            }
            self().write_string(a.first);
            this->visit_symbol(*a.second);
        }
        for (auto &a : x.m_symtab->get_scope()) {
            if (ASR::is_a<ASR::Subroutine_t>(*a.second) || ASR::is_a<ASR::Function_t>(*a.second)) {
                self().write_string(a.first);
                this->visit_symbol(*a.second);
            }
        }
        self().write_string(x.m_name);
        self().write_int64(x.n_body);
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
    }
    void visit_Block(const Block_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_symtab->counter);
        self().write_int64(x.m_symtab->get_scope().size());
        for (auto &a : x.m_symtab->get_scope()) {
            if (ASR::is_a<ASR::Subroutine_t>(*a.second) || ASR::is_a<ASR::Function_t>(*a.second)) {
                continue;
            }
            self().write_string(a.first);
            this->visit_symbol(*a.second);
        }
        for (auto &a : x.m_symtab->get_scope()) {
            if (ASR::is_a<ASR::Subroutine_t>(*a.second) || ASR::is_a<ASR::Function_t>(*a.second)) {
                self().write_string(a.first);
                this->visit_symbol(*a.second);
            }
        }
        self().write_string(x.m_name);
        self().write_int64(x.n_body);
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
    }
    void visit_storage_typeType(const storage_typeType &x) {
        self().write_int8(x);
    }
    void visit_accessType(const accessType &x) {
        self().write_int8(x);
    }
    void visit_intentType(const intentType &x) {
        self().write_int8(x);
    }
    void visit_deftypeType(const deftypeType &x) {
        self().write_int8(x);
    }
    void visit_presenceType(const presenceType &x) {
        self().write_int8(x);
    }
    void visit_abiType(const abiType &x) {
        self().write_int8(x);
    }
    void visit_Allocate(const Allocate_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.n_args);
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_alloc_arg(x.m_args[i]);
        }
        if (x.m_stat) {
            self().write_bool(true);
            self().visit_expr(*x.m_stat);
        } else {
            self().write_bool(false);
        }
        if (x.m_errmsg) {
            self().write_bool(true);
            self().visit_expr(*x.m_errmsg);
        } else {
            self().write_bool(false);
        }
        if (x.m_source) {
            self().write_bool(true);
            self().visit_expr(*x.m_source);
        } else {
            self().write_bool(false);
        }
    }
    void visit_Assign(const Assign_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_label);
        self().write_string(x.m_variable);
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_Assignment(const Assignment_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_target);
        self().visit_expr(*x.m_value);
        if (x.m_overloaded) {
            self().write_bool(true);
            self().visit_stmt(*x.m_overloaded);
        } else {
            self().write_bool(false);
        }
    }
    void visit_Associate(const Associate_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_target);
        self().visit_expr(*x.m_value);
    }
    void visit_Cycle(const Cycle_t &x) {
        self().write_int8(x.base.type);
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_ExplicitDeallocate(const ExplicitDeallocate_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.n_vars);
        for (size_t i=0; i<x.n_vars; i++) {
            self().write_symbol(*x.m_vars[i]);
        }
    }
    void visit_ImplicitDeallocate(const ImplicitDeallocate_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.n_vars);
        for (size_t i=0; i<x.n_vars; i++) {
            self().write_symbol(*x.m_vars[i]);
        }
    }
    void visit_DoConcurrentLoop(const DoConcurrentLoop_t &x) {
        self().write_int8(x.base.type);
        self().visit_do_loop_head(x.m_head);
        self().write_int64(x.n_body);
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
    }
    void visit_DoLoop(const DoLoop_t &x) {
        self().write_int8(x.base.type);
        self().visit_do_loop_head(x.m_head);
        self().write_int64(x.n_body);
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
    }
    void visit_ErrorStop(const ErrorStop_t &x) {
        self().write_int8(x.base.type);
        if (x.m_code) {
            self().write_bool(true);
            self().visit_expr(*x.m_code);
        } else {
            self().write_bool(false);
        }
    }
    void visit_Exit(const Exit_t &x) {
        self().write_int8(x.base.type);
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_ForAllSingle(const ForAllSingle_t &x) {
        self().write_int8(x.base.type);
        self().visit_do_loop_head(x.m_head);
        self().visit_stmt(*x.m_assign_stmt);
    }
    void visit_GoTo(const GoTo_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_target_id);
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_GoToTarget(const GoToTarget_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_id);
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_If(const If_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_test);
        self().write_int64(x.n_body);
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
        self().write_int64(x.n_orelse);
        for (size_t i=0; i<x.n_orelse; i++) {
            self().visit_stmt(*x.m_orelse[i]);
        }
    }
    void visit_IfArithmetic(const IfArithmetic_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_test);
        self().write_int64(x.m_lt_label);
        self().write_int64(x.m_eq_label);
        self().write_int64(x.m_gt_label);
    }
    void visit_Print(const Print_t &x) {
        self().write_int8(x.base.type);
        if (x.m_fmt) {
            self().write_bool(true);
            self().visit_expr(*x.m_fmt);
        } else {
            self().write_bool(false);
        }
        self().write_int64(x.n_values);
        for (size_t i=0; i<x.n_values; i++) {
            self().visit_expr(*x.m_values[i]);
        }
    }
    void visit_FileOpen(const FileOpen_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_label);
        if (x.m_newunit) {
            self().write_bool(true);
            self().visit_expr(*x.m_newunit);
        } else {
            self().write_bool(false);
        }
        if (x.m_filename) {
            self().write_bool(true);
            self().visit_expr(*x.m_filename);
        } else {
            self().write_bool(false);
        }
        if (x.m_status) {
            self().write_bool(true);
            self().visit_expr(*x.m_status);
        } else {
            self().write_bool(false);
        }
    }
    void visit_FileClose(const FileClose_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_label);
        if (x.m_unit) {
            self().write_bool(true);
            self().visit_expr(*x.m_unit);
        } else {
            self().write_bool(false);
        }
        if (x.m_iostat) {
            self().write_bool(true);
            self().visit_expr(*x.m_iostat);
        } else {
            self().write_bool(false);
        }
        if (x.m_iomsg) {
            self().write_bool(true);
            self().visit_expr(*x.m_iomsg);
        } else {
            self().write_bool(false);
        }
        if (x.m_err) {
            self().write_bool(true);
            self().visit_expr(*x.m_err);
        } else {
            self().write_bool(false);
        }
        if (x.m_status) {
            self().write_bool(true);
            self().visit_expr(*x.m_status);
        } else {
            self().write_bool(false);
        }
    }
    void visit_FileRead(const FileRead_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_label);
        if (x.m_unit) {
            self().write_bool(true);
            self().visit_expr(*x.m_unit);
        } else {
            self().write_bool(false);
        }
        if (x.m_fmt) {
            self().write_bool(true);
            self().visit_expr(*x.m_fmt);
        } else {
            self().write_bool(false);
        }
        if (x.m_iomsg) {
            self().write_bool(true);
            self().visit_expr(*x.m_iomsg);
        } else {
            self().write_bool(false);
        }
        if (x.m_iostat) {
            self().write_bool(true);
            self().visit_expr(*x.m_iostat);
        } else {
            self().write_bool(false);
        }
        if (x.m_id) {
            self().write_bool(true);
            self().visit_expr(*x.m_id);
        } else {
            self().write_bool(false);
        }
        self().write_int64(x.n_values);
        for (size_t i=0; i<x.n_values; i++) {
            self().visit_expr(*x.m_values[i]);
        }
    }
    void visit_FileRewind(const FileRewind_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_label);
        if (x.m_unit) {
            self().write_bool(true);
            self().visit_expr(*x.m_unit);
        } else {
            self().write_bool(false);
        }
        if (x.m_iostat) {
            self().write_bool(true);
            self().visit_expr(*x.m_iostat);
        } else {
            self().write_bool(false);
        }
        if (x.m_err) {
            self().write_bool(true);
            self().visit_expr(*x.m_err);
        } else {
            self().write_bool(false);
        }
    }
    void visit_FileInquire(const FileInquire_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_label);
        if (x.m_unit) {
            self().write_bool(true);
            self().visit_expr(*x.m_unit);
        } else {
            self().write_bool(false);
        }
        if (x.m_file) {
            self().write_bool(true);
            self().visit_expr(*x.m_file);
        } else {
            self().write_bool(false);
        }
        if (x.m_iostat) {
            self().write_bool(true);
            self().visit_expr(*x.m_iostat);
        } else {
            self().write_bool(false);
        }
        if (x.m_err) {
            self().write_bool(true);
            self().visit_expr(*x.m_err);
        } else {
            self().write_bool(false);
        }
        if (x.m_exist) {
            self().write_bool(true);
            self().visit_expr(*x.m_exist);
        } else {
            self().write_bool(false);
        }
        if (x.m_opened) {
            self().write_bool(true);
            self().visit_expr(*x.m_opened);
        } else {
            self().write_bool(false);
        }
        if (x.m_number) {
            self().write_bool(true);
            self().visit_expr(*x.m_number);
        } else {
            self().write_bool(false);
        }
        if (x.m_named) {
            self().write_bool(true);
            self().visit_expr(*x.m_named);
        } else {
            self().write_bool(false);
        }
        if (x.m_name) {
            self().write_bool(true);
            self().visit_expr(*x.m_name);
        } else {
            self().write_bool(false);
        }
        if (x.m_access) {
            self().write_bool(true);
            self().visit_expr(*x.m_access);
        } else {
            self().write_bool(false);
        }
        if (x.m_sequential) {
            self().write_bool(true);
            self().visit_expr(*x.m_sequential);
        } else {
            self().write_bool(false);
        }
        if (x.m_direct) {
            self().write_bool(true);
            self().visit_expr(*x.m_direct);
        } else {
            self().write_bool(false);
        }
        if (x.m_form) {
            self().write_bool(true);
            self().visit_expr(*x.m_form);
        } else {
            self().write_bool(false);
        }
        if (x.m_formatted) {
            self().write_bool(true);
            self().visit_expr(*x.m_formatted);
        } else {
            self().write_bool(false);
        }
        if (x.m_unformatted) {
            self().write_bool(true);
            self().visit_expr(*x.m_unformatted);
        } else {
            self().write_bool(false);
        }
        if (x.m_recl) {
            self().write_bool(true);
            self().visit_expr(*x.m_recl);
        } else {
            self().write_bool(false);
        }
        if (x.m_nextrec) {
            self().write_bool(true);
            self().visit_expr(*x.m_nextrec);
        } else {
            self().write_bool(false);
        }
        if (x.m_blank) {
            self().write_bool(true);
            self().visit_expr(*x.m_blank);
        } else {
            self().write_bool(false);
        }
        if (x.m_position) {
            self().write_bool(true);
            self().visit_expr(*x.m_position);
        } else {
            self().write_bool(false);
        }
        if (x.m_action) {
            self().write_bool(true);
            self().visit_expr(*x.m_action);
        } else {
            self().write_bool(false);
        }
        if (x.m_read) {
            self().write_bool(true);
            self().visit_expr(*x.m_read);
        } else {
            self().write_bool(false);
        }
        if (x.m_write) {
            self().write_bool(true);
            self().visit_expr(*x.m_write);
        } else {
            self().write_bool(false);
        }
        if (x.m_readwrite) {
            self().write_bool(true);
            self().visit_expr(*x.m_readwrite);
        } else {
            self().write_bool(false);
        }
        if (x.m_delim) {
            self().write_bool(true);
            self().visit_expr(*x.m_delim);
        } else {
            self().write_bool(false);
        }
        if (x.m_pad) {
            self().write_bool(true);
            self().visit_expr(*x.m_pad);
        } else {
            self().write_bool(false);
        }
        if (x.m_flen) {
            self().write_bool(true);
            self().visit_expr(*x.m_flen);
        } else {
            self().write_bool(false);
        }
        if (x.m_blocksize) {
            self().write_bool(true);
            self().visit_expr(*x.m_blocksize);
        } else {
            self().write_bool(false);
        }
        if (x.m_convert) {
            self().write_bool(true);
            self().visit_expr(*x.m_convert);
        } else {
            self().write_bool(false);
        }
        if (x.m_carriagecontrol) {
            self().write_bool(true);
            self().visit_expr(*x.m_carriagecontrol);
        } else {
            self().write_bool(false);
        }
        if (x.m_iolength) {
            self().write_bool(true);
            self().visit_expr(*x.m_iolength);
        } else {
            self().write_bool(false);
        }
    }
    void visit_FileWrite(const FileWrite_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_label);
        if (x.m_unit) {
            self().write_bool(true);
            self().visit_expr(*x.m_unit);
        } else {
            self().write_bool(false);
        }
        if (x.m_fmt) {
            self().write_bool(true);
            self().visit_expr(*x.m_fmt);
        } else {
            self().write_bool(false);
        }
        if (x.m_iomsg) {
            self().write_bool(true);
            self().visit_expr(*x.m_iomsg);
        } else {
            self().write_bool(false);
        }
        if (x.m_iostat) {
            self().write_bool(true);
            self().visit_expr(*x.m_iostat);
        } else {
            self().write_bool(false);
        }
        if (x.m_id) {
            self().write_bool(true);
            self().visit_expr(*x.m_id);
        } else {
            self().write_bool(false);
        }
        self().write_int64(x.n_values);
        for (size_t i=0; i<x.n_values; i++) {
            self().visit_expr(*x.m_values[i]);
        }
    }
    void visit_Return(const Return_t &x) {
        self().write_int8(x.base.type);
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_Select(const Select_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_test);
        self().write_int64(x.n_body);
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_case_stmt(*x.m_body[i]);
        }
        self().write_int64(x.n_default);
        for (size_t i=0; i<x.n_default; i++) {
            self().visit_stmt(*x.m_default[i]);
        }
    }
    void visit_Stop(const Stop_t &x) {
        self().write_int8(x.base.type);
        if (x.m_code) {
            self().write_bool(true);
            self().visit_expr(*x.m_code);
        } else {
            self().write_bool(false);
        }
    }
    void visit_Assert(const Assert_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_test);
        if (x.m_msg) {
            self().write_bool(true);
            self().visit_expr(*x.m_msg);
        } else {
            self().write_bool(false);
        }
    }
    void visit_SubroutineCall(const SubroutineCall_t &x) {
        self().write_int8(x.base.type);
        self().write_symbol(*x.m_name);
        if (x.m_original_name) {
            self().write_bool(true);
            self().write_symbol(*x.m_original_name);
        } else {
            self().write_bool(false);
        }
        self().write_int64(x.n_args);
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_call_arg(x.m_args[i]);
        }
        if (x.m_dt) {
            self().write_bool(true);
            self().visit_expr(*x.m_dt);
        } else {
            self().write_bool(false);
        }
    }
    void visit_Where(const Where_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_test);
        self().write_int64(x.n_body);
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
        self().write_int64(x.n_orelse);
        for (size_t i=0; i<x.n_orelse; i++) {
            self().visit_stmt(*x.m_orelse[i]);
        }
    }
    void visit_WhileLoop(const WhileLoop_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_test);
        self().write_int64(x.n_body);
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
    }
    void visit_Nullify(const Nullify_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.n_vars);
        for (size_t i=0; i<x.n_vars; i++) {
            self().write_symbol(*x.m_vars[i]);
        }
    }
    void visit_Flush(const Flush_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_label);
        self().visit_expr(*x.m_unit);
        if (x.m_err) {
            self().write_bool(true);
            self().visit_expr(*x.m_err);
        } else {
            self().write_bool(false);
        }
        if (x.m_iomsg) {
            self().write_bool(true);
            self().visit_expr(*x.m_iomsg);
        } else {
            self().write_bool(false);
        }
        if (x.m_iostat) {
            self().write_bool(true);
            self().visit_expr(*x.m_iostat);
        } else {
            self().write_bool(false);
        }
    }
    void visit_ListAppend(const ListAppend_t &x) {
        self().write_int8(x.base.type);
        self().write_symbol(*x.m_a);
        self().visit_expr(*x.m_ele);
    }
    void visit_AssociateBlockCall(const AssociateBlockCall_t &x) {
        self().write_int8(x.base.type);
        self().write_symbol(*x.m_m);
    }
    void visit_BlockCall(const BlockCall_t &x) {
        self().write_int8(x.base.type);
        self().write_symbol(*x.m_m);
    }
    void visit_SetInsert(const SetInsert_t &x) {
        self().write_int8(x.base.type);
        self().write_symbol(*x.m_a);
        self().visit_expr(*x.m_ele);
    }
    void visit_SetRemove(const SetRemove_t &x) {
        self().write_int8(x.base.type);
        self().write_symbol(*x.m_a);
        self().visit_expr(*x.m_ele);
    }
    void visit_ListInsert(const ListInsert_t &x) {
        self().write_int8(x.base.type);
        self().write_symbol(*x.m_a);
        self().visit_expr(*x.m_pos);
        self().visit_expr(*x.m_ele);
    }
    void visit_ListRemove(const ListRemove_t &x) {
        self().write_int8(x.base.type);
        self().write_symbol(*x.m_a);
        self().visit_expr(*x.m_ele);
    }
    void visit_DictInsert(const DictInsert_t &x) {
        self().write_int8(x.base.type);
        self().write_symbol(*x.m_a);
        self().visit_expr(*x.m_key);
        self().visit_expr(*x.m_value);
    }
    void visit_BoolOp(const BoolOp_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_left);
        visit_boolopType(x.m_op);
        self().visit_expr(*x.m_right);
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_BinOp(const BinOp_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_left);
        visit_binopType(x.m_op);
        self().visit_expr(*x.m_right);
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
        if (x.m_overloaded) {
            self().write_bool(true);
            self().visit_expr(*x.m_overloaded);
        } else {
            self().write_bool(false);
        }
    }
    void visit_StrOp(const StrOp_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_left);
        visit_stropType(x.m_op);
        self().visit_expr(*x.m_right);
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_UnaryOp(const UnaryOp_t &x) {
        self().write_int8(x.base.type);
        visit_unaryopType(x.m_op);
        self().visit_expr(*x.m_operand);
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_ComplexConstructor(const ComplexConstructor_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_re);
        self().visit_expr(*x.m_im);
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_NamedExpr(const NamedExpr_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_target);
        self().visit_expr(*x.m_value);
        self().visit_ttype(*x.m_type);
    }
    void visit_Compare(const Compare_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_left);
        visit_cmpopType(x.m_op);
        self().visit_expr(*x.m_right);
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
        if (x.m_overloaded) {
            self().write_bool(true);
            self().visit_expr(*x.m_overloaded);
        } else {
            self().write_bool(false);
        }
    }
    void visit_IfExp(const IfExp_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_test);
        self().visit_expr(*x.m_body);
        self().visit_expr(*x.m_orelse);
        self().visit_ttype(*x.m_type);
    }
    void visit_FunctionCall(const FunctionCall_t &x) {
        self().write_int8(x.base.type);
        self().write_symbol(*x.m_name);
        if (x.m_original_name) {
            self().write_bool(true);
            self().write_symbol(*x.m_original_name);
        } else {
            self().write_bool(false);
        }
        self().write_int64(x.n_args);
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_call_arg(x.m_args[i]);
        }
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
        if (x.m_dt) {
            self().write_bool(true);
            self().visit_expr(*x.m_dt);
        } else {
            self().write_bool(false);
        }
    }
    void visit_DerivedTypeConstructor(const DerivedTypeConstructor_t &x) {
        self().write_int8(x.base.type);
        self().write_symbol(*x.m_dt_sym);
        self().write_int64(x.n_args);
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_expr(*x.m_args[i]);
        }
        self().visit_ttype(*x.m_type);
    }
    void visit_ImpliedDoLoop(const ImpliedDoLoop_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.n_values);
        for (size_t i=0; i<x.n_values; i++) {
            self().visit_expr(*x.m_values[i]);
        }
        self().visit_expr(*x.m_var);
        self().visit_expr(*x.m_start);
        self().visit_expr(*x.m_end);
        if (x.m_increment) {
            self().write_bool(true);
            self().visit_expr(*x.m_increment);
        } else {
            self().write_bool(false);
        }
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_IntegerConstant(const IntegerConstant_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_n);
        self().visit_ttype(*x.m_type);
    }
    void visit_IntegerBOZ(const IntegerBOZ_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_v);
        visit_integerbozType(x.m_intboz_type);
        if (x.m_type) {
            self().write_bool(true);
            self().visit_ttype(*x.m_type);
        } else {
            self().write_bool(false);
        }
    }
    void visit_RealConstant(const RealConstant_t &x) {
        self().write_int8(x.base.type);
        self().write_float64(x.m_r);
        self().visit_ttype(*x.m_type);
    }
    void visit_ComplexConstant(const ComplexConstant_t &x) {
        self().write_int8(x.base.type);
        self().write_float64(x.m_re);
        self().write_float64(x.m_im);
        self().visit_ttype(*x.m_type);
    }
    void visit_LogicalConstant(const LogicalConstant_t &x) {
        self().write_int8(x.base.type);
        if (x.m_value) {
            self().write_bool(true);
        } else {
            self().write_bool(false);
        }
        self().visit_ttype(*x.m_type);
    }
    void visit_ListConstant(const ListConstant_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.n_args);
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_expr(*x.m_args[i]);
        }
        self().visit_ttype(*x.m_type);
    }
    void visit_ListLen(const ListLen_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_arg);
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_ListConcat(const ListConcat_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_left);
        self().visit_expr(*x.m_right);
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_ArrayConstant(const ArrayConstant_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.n_args);
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_expr(*x.m_args[i]);
        }
        self().visit_ttype(*x.m_type);
    }
    void visit_SetConstant(const SetConstant_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.n_elements);
        for (size_t i=0; i<x.n_elements; i++) {
            self().visit_expr(*x.m_elements[i]);
        }
        self().visit_ttype(*x.m_type);
    }
    void visit_SetLen(const SetLen_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_arg);
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_TupleConstant(const TupleConstant_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.n_elements);
        for (size_t i=0; i<x.n_elements; i++) {
            self().visit_expr(*x.m_elements[i]);
        }
        self().visit_ttype(*x.m_type);
    }
    void visit_TupleLen(const TupleLen_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_arg);
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_StringConstant(const StringConstant_t &x) {
        self().write_int8(x.base.type);
        self().write_string(x.m_s);
        self().visit_ttype(*x.m_type);
    }
    void visit_StringConcat(const StringConcat_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_left);
        self().visit_expr(*x.m_right);
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_StringLen(const StringLen_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_arg);
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_DictConstant(const DictConstant_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.n_keys);
        for (size_t i=0; i<x.n_keys; i++) {
            self().visit_expr(*x.m_keys[i]);
        }
        self().write_int64(x.n_values);
        for (size_t i=0; i<x.n_values; i++) {
            self().visit_expr(*x.m_values[i]);
        }
        self().visit_ttype(*x.m_type);
    }
    void visit_DictLen(const DictLen_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_arg);
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_Var(const Var_t &x) {
        self().write_int8(x.base.type);
        self().write_symbol(*x.m_v);
    }
    void visit_ArrayRef(const ArrayRef_t &x) {
        self().write_int8(x.base.type);
        self().write_symbol(*x.m_v);
        self().write_int64(x.n_args);
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_array_index(x.m_args[i]);
        }
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_ArraySize(const ArraySize_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_v);
        if (x.m_dim) {
            self().write_bool(true);
            self().visit_expr(*x.m_dim);
        } else {
            self().write_bool(false);
        }
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_ArrayBound(const ArrayBound_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_v);
        if (x.m_dim) {
            self().write_bool(true);
            self().visit_expr(*x.m_dim);
        } else {
            self().write_bool(false);
        }
        self().visit_ttype(*x.m_type);
        visit_arrayboundType(x.m_bound);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_ArrayTranspose(const ArrayTranspose_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_matrix);
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_ArrayMatMul(const ArrayMatMul_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_matrix_a);
        self().visit_expr(*x.m_matrix_b);
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_ArrayPack(const ArrayPack_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_array);
        self().visit_expr(*x.m_mask);
        if (x.m_vector) {
            self().write_bool(true);
            self().visit_expr(*x.m_vector);
        } else {
            self().write_bool(false);
        }
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_Transfer(const Transfer_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_source);
        self().visit_expr(*x.m_mold);
        if (x.m_size) {
            self().write_bool(true);
            self().visit_expr(*x.m_size);
        } else {
            self().write_bool(false);
        }
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_DerivedRef(const DerivedRef_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_v);
        self().write_symbol(*x.m_m);
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_Cast(const Cast_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_arg);
        visit_cast_kindType(x.m_kind);
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_ComplexRe(const ComplexRe_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_arg);
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_ComplexIm(const ComplexIm_t &x) {
        self().write_int8(x.base.type);
        self().visit_expr(*x.m_arg);
        self().visit_ttype(*x.m_type);
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_DictItem(const DictItem_t &x) {
        self().write_int8(x.base.type);
        self().write_symbol(*x.m_a);
        self().visit_expr(*x.m_key);
        if (x.m_default) {
            self().write_bool(true);
            self().visit_expr(*x.m_default);
        } else {
            self().write_bool(false);
        }
        self().visit_ttype(*x.m_type);
    }
    void visit_Integer(const Integer_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_kind);
        self().write_int64(x.n_dims);
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
        }
    }
    void visit_Real(const Real_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_kind);
        self().write_int64(x.n_dims);
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
        }
    }
    void visit_Complex(const Complex_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_kind);
        self().write_int64(x.n_dims);
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
        }
    }
    void visit_Character(const Character_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_kind);
        self().write_int64(x.m_len);
        if (x.m_len_expr) {
            self().write_bool(true);
            self().visit_expr(*x.m_len_expr);
        } else {
            self().write_bool(false);
        }
        self().write_int64(x.n_dims);
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
        }
    }
    void visit_Logical(const Logical_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.m_kind);
        self().write_int64(x.n_dims);
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
        }
    }
    void visit_Set(const Set_t &x) {
        self().write_int8(x.base.type);
        self().visit_ttype(*x.m_type);
    }
    void visit_List(const List_t &x) {
        self().write_int8(x.base.type);
        self().visit_ttype(*x.m_type);
    }
    void visit_Tuple(const Tuple_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.n_type);
        for (size_t i=0; i<x.n_type; i++) {
            self().visit_ttype(*x.m_type[i]);
        }
    }
    void visit_Derived(const Derived_t &x) {
        self().write_int8(x.base.type);
        self().write_symbol(*x.m_derived_type);
        self().write_int64(x.n_dims);
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
        }
    }
    void visit_Class(const Class_t &x) {
        self().write_int8(x.base.type);
        self().write_symbol(*x.m_class_type);
        self().write_int64(x.n_dims);
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
        }
    }
    void visit_Dict(const Dict_t &x) {
        self().write_int8(x.base.type);
        self().visit_ttype(*x.m_key_type);
        self().visit_ttype(*x.m_value_type);
    }
    void visit_Pointer(const Pointer_t &x) {
        self().write_int8(x.base.type);
        self().visit_ttype(*x.m_type);
    }
    void visit_boolopType(const boolopType &x) {
        self().write_int8(x);
    }
    void visit_binopType(const binopType &x) {
        self().write_int8(x);
    }
    void visit_unaryopType(const unaryopType &x) {
        self().write_int8(x);
    }
    void visit_stropType(const stropType &x) {
        self().write_int8(x);
    }
    void visit_cmpopType(const cmpopType &x) {
        self().write_int8(x);
    }
    void visit_integerbozType(const integerbozType &x) {
        self().write_int8(x);
    }
    void visit_arrayboundType(const arrayboundType &x) {
        self().write_int8(x);
    }
    void visit_cast_kindType(const cast_kindType &x) {
        self().write_int8(x);
    }
    void visit_dimension(const dimension_t &x) {
        if (x.m_start) {
            self().write_bool(true);
            self().visit_expr(*x.m_start);
        } else {
            self().write_bool(false);
        }
        if (x.m_end) {
            self().write_bool(true);
            self().visit_expr(*x.m_end);
        } else {
            self().write_bool(false);
        }
    }
    void visit_alloc_arg(const alloc_arg_t &x) {
        self().write_symbol(*x.m_a);
        self().write_int64(x.n_dims);
        for (size_t i=0; i<x.n_dims; i++) {
            self().visit_dimension(x.m_dims[i]);
        }
    }
    void visit_Attribute(const Attribute_t &x) {
        self().write_int8(x.base.type);
        self().write_string(x.m_name);
        self().write_int64(x.n_args);
        for (size_t i=0; i<x.n_args; i++) {
            self().visit_attribute_arg(x.m_args[i]);
        }
    }
    void visit_attribute_arg(const attribute_arg_t &x) {
        self().write_string(x.m_arg);
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_call_arg(const call_arg_t &x) {
        if (x.m_value) {
            self().write_bool(true);
            self().visit_expr(*x.m_value);
        } else {
            self().write_bool(false);
        }
    }
    void visit_Bind(const Bind_t &x) {
        self().write_int8(x.base.type);
        self().write_string(x.m_lang);
        self().write_string(x.m_name);
        if ((bool&)x) { } // Suppress unused warning
    }
    void visit_array_index(const array_index_t &x) {
        if (x.m_left) {
            self().write_bool(true);
            self().visit_expr(*x.m_left);
        } else {
            self().write_bool(false);
        }
        if (x.m_right) {
            self().write_bool(true);
            self().visit_expr(*x.m_right);
        } else {
            self().write_bool(false);
        }
        if (x.m_step) {
            self().write_bool(true);
            self().visit_expr(*x.m_step);
        } else {
            self().write_bool(false);
        }
    }
    void visit_do_loop_head(const do_loop_head_t &x) {
        if (x.m_v) {
            self().write_bool(true);
            self().visit_expr(*x.m_v);
        } else {
            self().write_bool(false);
        }
        if (x.m_start) {
            self().write_bool(true);
            self().visit_expr(*x.m_start);
        } else {
            self().write_bool(false);
        }
        if (x.m_end) {
            self().write_bool(true);
            self().visit_expr(*x.m_end);
        } else {
            self().write_bool(false);
        }
        if (x.m_increment) {
            self().write_bool(true);
            self().visit_expr(*x.m_increment);
        } else {
            self().write_bool(false);
        }
    }
    void visit_CaseStmt(const CaseStmt_t &x) {
        self().write_int8(x.base.type);
        self().write_int64(x.n_test);
        for (size_t i=0; i<x.n_test; i++) {
            self().visit_expr(*x.m_test[i]);
        }
        self().write_int64(x.n_body);
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
    }
    void visit_CaseStmt_Range(const CaseStmt_Range_t &x) {
        self().write_int8(x.base.type);
        if (x.m_start) {
            self().write_bool(true);
            self().visit_expr(*x.m_start);
        } else {
            self().write_bool(false);
        }
        if (x.m_end) {
            self().write_bool(true);
            self().visit_expr(*x.m_end);
        } else {
            self().write_bool(false);
        }
        self().write_int64(x.n_body);
        for (size_t i=0; i<x.n_body; i++) {
            self().visit_stmt(*x.m_body[i]);
        }
    }
};


/******************************************************************************/
// Deserialization Visitor base class

template <class Derived>
class DeserializationBaseVisitor : public BaseVisitor<Derived>
{
private:
    Derived& self() { return static_cast<Derived&>(*this); }
public:
    Allocator &al;
    bool load_symtab_id;
    std::map<uint64_t,SymbolTable*> id_symtab_map;
    DeserializationBaseVisitor(Allocator &al, bool load_symtab_id) : al{al}, load_symtab_id{load_symtab_id} {}
    asr_t* deserialize_node() {
        uint8_t t = self().read_int8();
        ASR::asrType ty = static_cast<ASR::asrType>(t);
        switch (ty) {
            case (ASR::asrType::unit) : return self().deserialize_unit();
            case (ASR::asrType::symbol) : return self().deserialize_symbol();
            case (ASR::asrType::stmt) : return self().deserialize_stmt();
            case (ASR::asrType::expr) : return self().deserialize_expr();
            case (ASR::asrType::ttype) : return self().deserialize_ttype();
            case (ASR::asrType::attribute) : return self().deserialize_attribute();
            case (ASR::asrType::tbind) : return self().deserialize_tbind();
            case (ASR::asrType::case_stmt) : return self().deserialize_case_stmt();
            default : throw LFortranException("Unknown type in deserialize_node()");
        }
        throw LFortranException("Switch statement above was not exhaustive.");
    }
    asr_t* deserialize_TranslationUnit() {
        size_t n_items; // Sequence
        uint64_t m_global_scope_counter = self().read_int64();
        LFORTRAN_ASSERT(id_symtab_map.find(m_global_scope_counter) == id_symtab_map.end());
        SymbolTable *m_global_scope = al.make_new<SymbolTable>(nullptr);
        if (load_symtab_id) m_global_scope->counter = m_global_scope_counter;
        id_symtab_map[m_global_scope_counter] = m_global_scope;
        {
            size_t n = self().read_int64();
            for (size_t i=0; i<n; i++) {
                std::string name = self().read_string();
                symbol_t *sym = down_cast<symbol_t>(deserialize_symbol());
                self().symtab_insert_symbol(*m_global_scope, name, sym);
            }
        }
        n_items = self().read_int64();
        Vec<asr_t*> v_items;
        v_items.reserve(al, n_items);
        for (size_t i=0; i<n_items; i++) {
            v_items.push_back(al, self().deserialize_node());
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_TranslationUnit_t(al, loc, m_global_scope, v_items.p, v_items.n);
    }
    asr_t* deserialize_unit() {
        uint8_t t = self().read_int8();
        ASR::unitType ty = static_cast<ASR::unitType>(t);
        switch (ty) {
            case (ASR::unitType::TranslationUnit) : return self().deserialize_TranslationUnit();
            default : throw LFortranException("Unknown type in deserialize_unit()");
        }
        throw LFortranException("Switch statement above was not exhaustive.");
    }
    asr_t* deserialize_Program() {
        size_t n_dependencies; // Sequence
        size_t n_body; // Sequence
        uint64_t m_symtab_counter = self().read_int64();
        LFORTRAN_ASSERT(id_symtab_map.find(m_symtab_counter) == id_symtab_map.end());
        SymbolTable *m_symtab = al.make_new<SymbolTable>(nullptr);
        if (load_symtab_id) m_symtab->counter = m_symtab_counter;
        id_symtab_map[m_symtab_counter] = m_symtab;
        {
            size_t n = self().read_int64();
            for (size_t i=0; i<n; i++) {
                std::string name = self().read_string();
                symbol_t *sym = down_cast<symbol_t>(deserialize_symbol());
                self().symtab_insert_symbol(*m_symtab, name, sym);
            }
        }
        char *m_name;
        m_name = self().read_cstring();
        n_dependencies = self().read_int64();
        Vec<char*> v_dependencies;
        v_dependencies.reserve(al, n_dependencies);
        for (size_t i=0; i<n_dependencies; i++) {
            v_dependencies.push_back(al, self().read_cstring());
        }
        n_body = self().read_int64();
        Vec<stmt_t*> v_body;
        v_body.reserve(al, n_body);
        for (size_t i=0; i<n_body; i++) {
            v_body.push_back(al, ASR::down_cast<ASR::stmt_t>(self().deserialize_stmt()));
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Program_t(al, loc, m_symtab, m_name, v_dependencies.p, v_dependencies.n, v_body.p, v_body.n);
    }
    asr_t* deserialize_Module() {
        size_t n_dependencies; // Sequence
        uint64_t m_symtab_counter = self().read_int64();
        LFORTRAN_ASSERT(id_symtab_map.find(m_symtab_counter) == id_symtab_map.end());
        SymbolTable *m_symtab = al.make_new<SymbolTable>(nullptr);
        if (load_symtab_id) m_symtab->counter = m_symtab_counter;
        id_symtab_map[m_symtab_counter] = m_symtab;
        {
            size_t n = self().read_int64();
            for (size_t i=0; i<n; i++) {
                std::string name = self().read_string();
                symbol_t *sym = down_cast<symbol_t>(deserialize_symbol());
                self().symtab_insert_symbol(*m_symtab, name, sym);
            }
        }
        char *m_name;
        m_name = self().read_cstring();
        n_dependencies = self().read_int64();
        Vec<char*> v_dependencies;
        v_dependencies.reserve(al, n_dependencies);
        for (size_t i=0; i<n_dependencies; i++) {
            v_dependencies.push_back(al, self().read_cstring());
        }
        bool m_loaded_from_mod = self().read_bool();
        bool m_intrinsic = self().read_bool();
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Module_t(al, loc, m_symtab, m_name, v_dependencies.p, v_dependencies.n, m_loaded_from_mod, m_intrinsic);
    }
    asr_t* deserialize_Subroutine() {
        size_t n_args; // Sequence
        size_t n_body; // Sequence
        uint64_t m_symtab_counter = self().read_int64();
        LFORTRAN_ASSERT(id_symtab_map.find(m_symtab_counter) == id_symtab_map.end());
        SymbolTable *m_symtab = al.make_new<SymbolTable>(nullptr);
        if (load_symtab_id) m_symtab->counter = m_symtab_counter;
        id_symtab_map[m_symtab_counter] = m_symtab;
        {
            size_t n = self().read_int64();
            for (size_t i=0; i<n; i++) {
                std::string name = self().read_string();
                symbol_t *sym = down_cast<symbol_t>(deserialize_symbol());
                self().symtab_insert_symbol(*m_symtab, name, sym);
            }
        }
        char *m_name;
        m_name = self().read_cstring();
        n_args = self().read_int64();
        Vec<expr_t*> v_args;
        v_args.reserve(al, n_args);
        for (size_t i=0; i<n_args; i++) {
            v_args.push_back(al, ASR::down_cast<ASR::expr_t>(self().deserialize_expr()));
        }
        n_body = self().read_int64();
        Vec<stmt_t*> v_body;
        v_body.reserve(al, n_body);
        for (size_t i=0; i<n_body; i++) {
            v_body.push_back(al, ASR::down_cast<ASR::stmt_t>(self().deserialize_stmt()));
        }
        ASR::abiType m_abi = self().deserialize_abi();
        ASR::accessType m_access = self().deserialize_access();
        ASR::deftypeType m_deftype = self().deserialize_deftype();
        char *m_bindc_name;
        bool m_bindc_name_present = self().read_bool();
        if (m_bindc_name_present) {
        m_bindc_name = self().read_cstring();
        } else {
        m_bindc_name = nullptr;
        }
        bool m_pure = self().read_bool();
        bool m_module = self().read_bool();
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Subroutine_t(al, loc, m_symtab, m_name, v_args.p, v_args.n, v_body.p, v_body.n, m_abi, m_access, m_deftype, m_bindc_name, m_pure, m_module);
    }
    asr_t* deserialize_Function() {
        size_t n_args; // Sequence
        size_t n_body; // Sequence
        uint64_t m_symtab_counter = self().read_int64();
        LFORTRAN_ASSERT(id_symtab_map.find(m_symtab_counter) == id_symtab_map.end());
        SymbolTable *m_symtab = al.make_new<SymbolTable>(nullptr);
        if (load_symtab_id) m_symtab->counter = m_symtab_counter;
        id_symtab_map[m_symtab_counter] = m_symtab;
        {
            size_t n = self().read_int64();
            for (size_t i=0; i<n; i++) {
                std::string name = self().read_string();
                symbol_t *sym = down_cast<symbol_t>(deserialize_symbol());
                self().symtab_insert_symbol(*m_symtab, name, sym);
            }
        }
        char *m_name;
        m_name = self().read_cstring();
        n_args = self().read_int64();
        Vec<expr_t*> v_args;
        v_args.reserve(al, n_args);
        for (size_t i=0; i<n_args; i++) {
            v_args.push_back(al, ASR::down_cast<ASR::expr_t>(self().deserialize_expr()));
        }
        n_body = self().read_int64();
        Vec<stmt_t*> v_body;
        v_body.reserve(al, n_body);
        for (size_t i=0; i<n_body; i++) {
            v_body.push_back(al, ASR::down_cast<ASR::stmt_t>(self().deserialize_stmt()));
        }
        ASR::expr_t *m_return_var;
        m_return_var = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::abiType m_abi = self().deserialize_abi();
        ASR::accessType m_access = self().deserialize_access();
        ASR::deftypeType m_deftype = self().deserialize_deftype();
        char *m_bindc_name;
        bool m_bindc_name_present = self().read_bool();
        if (m_bindc_name_present) {
        m_bindc_name = self().read_cstring();
        } else {
        m_bindc_name = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Function_t(al, loc, m_symtab, m_name, v_args.p, v_args.n, v_body.p, v_body.n, m_return_var, m_abi, m_access, m_deftype, m_bindc_name);
    }
    asr_t* deserialize_GenericProcedure() {
        size_t n_procs; // Sequence
        uint64_t m_parent_symtab_counter = self().read_int64();
        LFORTRAN_ASSERT(id_symtab_map.find(m_parent_symtab_counter) != id_symtab_map.end());
        SymbolTable *m_parent_symtab = id_symtab_map[m_parent_symtab_counter];
        char *m_name;
        m_name = self().read_cstring();
        n_procs = self().read_int64();
        Vec<symbol_t*> v_procs;
        v_procs.reserve(al, n_procs);
        for (size_t i=0; i<n_procs; i++) {
            v_procs.push_back(al, self().read_symbol());
        }
        ASR::accessType m_access = self().deserialize_access();
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_GenericProcedure_t(al, loc, m_parent_symtab, m_name, v_procs.p, v_procs.n, m_access);
    }
    asr_t* deserialize_CustomOperator() {
        size_t n_procs; // Sequence
        uint64_t m_parent_symtab_counter = self().read_int64();
        LFORTRAN_ASSERT(id_symtab_map.find(m_parent_symtab_counter) != id_symtab_map.end());
        SymbolTable *m_parent_symtab = id_symtab_map[m_parent_symtab_counter];
        char *m_name;
        m_name = self().read_cstring();
        n_procs = self().read_int64();
        Vec<symbol_t*> v_procs;
        v_procs.reserve(al, n_procs);
        for (size_t i=0; i<n_procs; i++) {
            v_procs.push_back(al, self().read_symbol());
        }
        ASR::accessType m_access = self().deserialize_access();
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_CustomOperator_t(al, loc, m_parent_symtab, m_name, v_procs.p, v_procs.n, m_access);
    }
    asr_t* deserialize_ExternalSymbol() {
        size_t n_scope_names; // Sequence
        uint64_t m_parent_symtab_counter = self().read_int64();
        LFORTRAN_ASSERT(id_symtab_map.find(m_parent_symtab_counter) != id_symtab_map.end());
        SymbolTable *m_parent_symtab = id_symtab_map[m_parent_symtab_counter];
        char *m_name;
        m_name = self().read_cstring();
        ASR::symbol_t *m_external;
        // We skip the symbol for ExternalSymbol
        m_external = nullptr;
        char *m_module_name;
        m_module_name = self().read_cstring();
        n_scope_names = self().read_int64();
        Vec<char*> v_scope_names;
        v_scope_names.reserve(al, n_scope_names);
        for (size_t i=0; i<n_scope_names; i++) {
            v_scope_names.push_back(al, self().read_cstring());
        }
        char *m_original_name;
        m_original_name = self().read_cstring();
        ASR::accessType m_access = self().deserialize_access();
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ExternalSymbol_t(al, loc, m_parent_symtab, m_name, m_external, m_module_name, v_scope_names.p, v_scope_names.n, m_original_name, m_access);
    }
    asr_t* deserialize_DerivedType() {
        size_t n_members; // Sequence
        uint64_t m_symtab_counter = self().read_int64();
        LFORTRAN_ASSERT(id_symtab_map.find(m_symtab_counter) == id_symtab_map.end());
        SymbolTable *m_symtab = al.make_new<SymbolTable>(nullptr);
        if (load_symtab_id) m_symtab->counter = m_symtab_counter;
        id_symtab_map[m_symtab_counter] = m_symtab;
        {
            size_t n = self().read_int64();
            for (size_t i=0; i<n; i++) {
                std::string name = self().read_string();
                symbol_t *sym = down_cast<symbol_t>(deserialize_symbol());
                self().symtab_insert_symbol(*m_symtab, name, sym);
            }
        }
        char *m_name;
        m_name = self().read_cstring();
        n_members = self().read_int64();
        Vec<char*> v_members;
        v_members.reserve(al, n_members);
        for (size_t i=0; i<n_members; i++) {
            v_members.push_back(al, self().read_cstring());
        }
        ASR::abiType m_abi = self().deserialize_abi();
        ASR::accessType m_access = self().deserialize_access();
        ASR::symbol_t *m_parent;
        if (self().read_bool()) {
        m_parent = self().read_symbol();
        } else {
        m_parent = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_DerivedType_t(al, loc, m_symtab, m_name, v_members.p, v_members.n, m_abi, m_access, m_parent);
    }
    asr_t* deserialize_Variable() {
        uint64_t m_parent_symtab_counter = self().read_int64();
        LFORTRAN_ASSERT(id_symtab_map.find(m_parent_symtab_counter) != id_symtab_map.end());
        SymbolTable *m_parent_symtab = id_symtab_map[m_parent_symtab_counter];
        char *m_name;
        m_name = self().read_cstring();
        ASR::intentType m_intent = self().deserialize_intent();
        ASR::expr_t *m_symbolic_value;
        if (self().read_bool()) {
        m_symbolic_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_symbolic_value = nullptr;
        }
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        ASR::storage_typeType m_storage = self().deserialize_storage_type();
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::abiType m_abi = self().deserialize_abi();
        ASR::accessType m_access = self().deserialize_access();
        ASR::presenceType m_presence = self().deserialize_presence();
        bool m_value_attr = self().read_bool();
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Variable_t(al, loc, m_parent_symtab, m_name, m_intent, m_symbolic_value, m_value, m_storage, m_type, m_abi, m_access, m_presence, m_value_attr);
    }
    asr_t* deserialize_ClassType() {
        uint64_t m_symtab_counter = self().read_int64();
        LFORTRAN_ASSERT(id_symtab_map.find(m_symtab_counter) == id_symtab_map.end());
        SymbolTable *m_symtab = al.make_new<SymbolTable>(nullptr);
        if (load_symtab_id) m_symtab->counter = m_symtab_counter;
        id_symtab_map[m_symtab_counter] = m_symtab;
        {
            size_t n = self().read_int64();
            for (size_t i=0; i<n; i++) {
                std::string name = self().read_string();
                symbol_t *sym = down_cast<symbol_t>(deserialize_symbol());
                self().symtab_insert_symbol(*m_symtab, name, sym);
            }
        }
        char *m_name;
        m_name = self().read_cstring();
        ASR::abiType m_abi = self().deserialize_abi();
        ASR::accessType m_access = self().deserialize_access();
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ClassType_t(al, loc, m_symtab, m_name, m_abi, m_access);
    }
    asr_t* deserialize_ClassProcedure() {
        uint64_t m_parent_symtab_counter = self().read_int64();
        LFORTRAN_ASSERT(id_symtab_map.find(m_parent_symtab_counter) != id_symtab_map.end());
        SymbolTable *m_parent_symtab = id_symtab_map[m_parent_symtab_counter];
        char *m_name;
        m_name = self().read_cstring();
        char *m_proc_name;
        m_proc_name = self().read_cstring();
        ASR::symbol_t *m_proc;
        m_proc = self().read_symbol();
        ASR::abiType m_abi = self().deserialize_abi();
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ClassProcedure_t(al, loc, m_parent_symtab, m_name, m_proc_name, m_proc, m_abi);
    }
    asr_t* deserialize_AssociateBlock() {
        size_t n_body; // Sequence
        uint64_t m_symtab_counter = self().read_int64();
        LFORTRAN_ASSERT(id_symtab_map.find(m_symtab_counter) == id_symtab_map.end());
        SymbolTable *m_symtab = al.make_new<SymbolTable>(nullptr);
        if (load_symtab_id) m_symtab->counter = m_symtab_counter;
        id_symtab_map[m_symtab_counter] = m_symtab;
        {
            size_t n = self().read_int64();
            for (size_t i=0; i<n; i++) {
                std::string name = self().read_string();
                symbol_t *sym = down_cast<symbol_t>(deserialize_symbol());
                self().symtab_insert_symbol(*m_symtab, name, sym);
            }
        }
        char *m_name;
        m_name = self().read_cstring();
        n_body = self().read_int64();
        Vec<stmt_t*> v_body;
        v_body.reserve(al, n_body);
        for (size_t i=0; i<n_body; i++) {
            v_body.push_back(al, ASR::down_cast<ASR::stmt_t>(self().deserialize_stmt()));
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_AssociateBlock_t(al, loc, m_symtab, m_name, v_body.p, v_body.n);
    }
    asr_t* deserialize_Block() {
        size_t n_body; // Sequence
        uint64_t m_symtab_counter = self().read_int64();
        LFORTRAN_ASSERT(id_symtab_map.find(m_symtab_counter) == id_symtab_map.end());
        SymbolTable *m_symtab = al.make_new<SymbolTable>(nullptr);
        if (load_symtab_id) m_symtab->counter = m_symtab_counter;
        id_symtab_map[m_symtab_counter] = m_symtab;
        {
            size_t n = self().read_int64();
            for (size_t i=0; i<n; i++) {
                std::string name = self().read_string();
                symbol_t *sym = down_cast<symbol_t>(deserialize_symbol());
                self().symtab_insert_symbol(*m_symtab, name, sym);
            }
        }
        char *m_name;
        m_name = self().read_cstring();
        n_body = self().read_int64();
        Vec<stmt_t*> v_body;
        v_body.reserve(al, n_body);
        for (size_t i=0; i<n_body; i++) {
            v_body.push_back(al, ASR::down_cast<ASR::stmt_t>(self().deserialize_stmt()));
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Block_t(al, loc, m_symtab, m_name, v_body.p, v_body.n);
    }
    asr_t* deserialize_symbol() {
        uint8_t t = self().read_int8();
        ASR::symbolType ty = static_cast<ASR::symbolType>(t);
        switch (ty) {
            case (ASR::symbolType::Program) : return self().deserialize_Program();
            case (ASR::symbolType::Module) : return self().deserialize_Module();
            case (ASR::symbolType::Subroutine) : return self().deserialize_Subroutine();
            case (ASR::symbolType::Function) : return self().deserialize_Function();
            case (ASR::symbolType::GenericProcedure) : return self().deserialize_GenericProcedure();
            case (ASR::symbolType::CustomOperator) : return self().deserialize_CustomOperator();
            case (ASR::symbolType::ExternalSymbol) : return self().deserialize_ExternalSymbol();
            case (ASR::symbolType::DerivedType) : return self().deserialize_DerivedType();
            case (ASR::symbolType::Variable) : return self().deserialize_Variable();
            case (ASR::symbolType::ClassType) : return self().deserialize_ClassType();
            case (ASR::symbolType::ClassProcedure) : return self().deserialize_ClassProcedure();
            case (ASR::symbolType::AssociateBlock) : return self().deserialize_AssociateBlock();
            case (ASR::symbolType::Block) : return self().deserialize_Block();
            default : throw LFortranException("Unknown type in deserialize_symbol()");
        }
        throw LFortranException("Switch statement above was not exhaustive.");
    }
    storage_typeType deserialize_storage_type() {
        uint8_t t = self().read_int8();
        storage_typeType ty = static_cast<storage_typeType>(t);
        return ty;
    }
    accessType deserialize_access() {
        uint8_t t = self().read_int8();
        accessType ty = static_cast<accessType>(t);
        return ty;
    }
    intentType deserialize_intent() {
        uint8_t t = self().read_int8();
        intentType ty = static_cast<intentType>(t);
        return ty;
    }
    deftypeType deserialize_deftype() {
        uint8_t t = self().read_int8();
        deftypeType ty = static_cast<deftypeType>(t);
        return ty;
    }
    presenceType deserialize_presence() {
        uint8_t t = self().read_int8();
        presenceType ty = static_cast<presenceType>(t);
        return ty;
    }
    abiType deserialize_abi() {
        uint8_t t = self().read_int8();
        abiType ty = static_cast<abiType>(t);
        return ty;
    }
    asr_t* deserialize_Allocate() {
        size_t n_args; // Sequence
        n_args = self().read_int64();
        Vec<alloc_arg_t> v_args;
        v_args.reserve(al, n_args);
        for (size_t i=0; i<n_args; i++) {
            v_args.push_back(al, self().deserialize_alloc_arg());
        }
        ASR::expr_t *m_stat;
        if (self().read_bool()) {
        m_stat = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_stat = nullptr;
        }
        ASR::expr_t *m_errmsg;
        if (self().read_bool()) {
        m_errmsg = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_errmsg = nullptr;
        }
        ASR::expr_t *m_source;
        if (self().read_bool()) {
        m_source = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_source = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Allocate_t(al, loc, v_args.p, v_args.n, m_stat, m_errmsg, m_source);
    }
    asr_t* deserialize_Assign() {
        int64_t m_label = self().read_int64();
        char *m_variable;
        m_variable = self().read_cstring();
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Assign_t(al, loc, m_label, m_variable);
    }
    asr_t* deserialize_Assignment() {
        ASR::expr_t *m_target;
        m_target = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_value;
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::stmt_t *m_overloaded;
        if (self().read_bool()) {
        m_overloaded = ASR::down_cast<ASR::stmt_t>(self().deserialize_stmt());
        } else {
        m_overloaded = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Assignment_t(al, loc, m_target, m_value, m_overloaded);
    }
    asr_t* deserialize_Associate() {
        ASR::expr_t *m_target;
        m_target = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_value;
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Associate_t(al, loc, m_target, m_value);
    }
    asr_t* deserialize_Cycle() {
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Cycle_t(al, loc);
    }
    asr_t* deserialize_ExplicitDeallocate() {
        size_t n_vars; // Sequence
        n_vars = self().read_int64();
        Vec<symbol_t*> v_vars;
        v_vars.reserve(al, n_vars);
        for (size_t i=0; i<n_vars; i++) {
            v_vars.push_back(al, self().read_symbol());
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ExplicitDeallocate_t(al, loc, v_vars.p, v_vars.n);
    }
    asr_t* deserialize_ImplicitDeallocate() {
        size_t n_vars; // Sequence
        n_vars = self().read_int64();
        Vec<symbol_t*> v_vars;
        v_vars.reserve(al, n_vars);
        for (size_t i=0; i<n_vars; i++) {
            v_vars.push_back(al, self().read_symbol());
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ImplicitDeallocate_t(al, loc, v_vars.p, v_vars.n);
    }
    asr_t* deserialize_DoConcurrentLoop() {
        size_t n_body; // Sequence
        ASR::do_loop_head_t m_head = self().deserialize_do_loop_head();
        n_body = self().read_int64();
        Vec<stmt_t*> v_body;
        v_body.reserve(al, n_body);
        for (size_t i=0; i<n_body; i++) {
            v_body.push_back(al, ASR::down_cast<ASR::stmt_t>(self().deserialize_stmt()));
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_DoConcurrentLoop_t(al, loc, m_head, v_body.p, v_body.n);
    }
    asr_t* deserialize_DoLoop() {
        size_t n_body; // Sequence
        ASR::do_loop_head_t m_head = self().deserialize_do_loop_head();
        n_body = self().read_int64();
        Vec<stmt_t*> v_body;
        v_body.reserve(al, n_body);
        for (size_t i=0; i<n_body; i++) {
            v_body.push_back(al, ASR::down_cast<ASR::stmt_t>(self().deserialize_stmt()));
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_DoLoop_t(al, loc, m_head, v_body.p, v_body.n);
    }
    asr_t* deserialize_ErrorStop() {
        ASR::expr_t *m_code;
        if (self().read_bool()) {
        m_code = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_code = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ErrorStop_t(al, loc, m_code);
    }
    asr_t* deserialize_Exit() {
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Exit_t(al, loc);
    }
    asr_t* deserialize_ForAllSingle() {
        ASR::do_loop_head_t m_head = self().deserialize_do_loop_head();
        ASR::stmt_t *m_assign_stmt;
        m_assign_stmt = ASR::down_cast<ASR::stmt_t>(self().deserialize_stmt());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ForAllSingle_t(al, loc, m_head, m_assign_stmt);
    }
    asr_t* deserialize_GoTo() {
        int64_t m_target_id = self().read_int64();
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_GoTo_t(al, loc, m_target_id);
    }
    asr_t* deserialize_GoToTarget() {
        int64_t m_id = self().read_int64();
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_GoToTarget_t(al, loc, m_id);
    }
    asr_t* deserialize_If() {
        size_t n_body; // Sequence
        size_t n_orelse; // Sequence
        ASR::expr_t *m_test;
        m_test = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        n_body = self().read_int64();
        Vec<stmt_t*> v_body;
        v_body.reserve(al, n_body);
        for (size_t i=0; i<n_body; i++) {
            v_body.push_back(al, ASR::down_cast<ASR::stmt_t>(self().deserialize_stmt()));
        }
        n_orelse = self().read_int64();
        Vec<stmt_t*> v_orelse;
        v_orelse.reserve(al, n_orelse);
        for (size_t i=0; i<n_orelse; i++) {
            v_orelse.push_back(al, ASR::down_cast<ASR::stmt_t>(self().deserialize_stmt()));
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_If_t(al, loc, m_test, v_body.p, v_body.n, v_orelse.p, v_orelse.n);
    }
    asr_t* deserialize_IfArithmetic() {
        ASR::expr_t *m_test;
        m_test = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        int64_t m_lt_label = self().read_int64();
        int64_t m_eq_label = self().read_int64();
        int64_t m_gt_label = self().read_int64();
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_IfArithmetic_t(al, loc, m_test, m_lt_label, m_eq_label, m_gt_label);
    }
    asr_t* deserialize_Print() {
        size_t n_values; // Sequence
        ASR::expr_t *m_fmt;
        if (self().read_bool()) {
        m_fmt = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_fmt = nullptr;
        }
        n_values = self().read_int64();
        Vec<expr_t*> v_values;
        v_values.reserve(al, n_values);
        for (size_t i=0; i<n_values; i++) {
            v_values.push_back(al, ASR::down_cast<ASR::expr_t>(self().deserialize_expr()));
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Print_t(al, loc, m_fmt, v_values.p, v_values.n);
    }
    asr_t* deserialize_FileOpen() {
        int64_t m_label = self().read_int64();
        ASR::expr_t *m_newunit;
        if (self().read_bool()) {
        m_newunit = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_newunit = nullptr;
        }
        ASR::expr_t *m_filename;
        if (self().read_bool()) {
        m_filename = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_filename = nullptr;
        }
        ASR::expr_t *m_status;
        if (self().read_bool()) {
        m_status = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_status = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_FileOpen_t(al, loc, m_label, m_newunit, m_filename, m_status);
    }
    asr_t* deserialize_FileClose() {
        int64_t m_label = self().read_int64();
        ASR::expr_t *m_unit;
        if (self().read_bool()) {
        m_unit = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_unit = nullptr;
        }
        ASR::expr_t *m_iostat;
        if (self().read_bool()) {
        m_iostat = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_iostat = nullptr;
        }
        ASR::expr_t *m_iomsg;
        if (self().read_bool()) {
        m_iomsg = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_iomsg = nullptr;
        }
        ASR::expr_t *m_err;
        if (self().read_bool()) {
        m_err = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_err = nullptr;
        }
        ASR::expr_t *m_status;
        if (self().read_bool()) {
        m_status = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_status = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_FileClose_t(al, loc, m_label, m_unit, m_iostat, m_iomsg, m_err, m_status);
    }
    asr_t* deserialize_FileRead() {
        size_t n_values; // Sequence
        int64_t m_label = self().read_int64();
        ASR::expr_t *m_unit;
        if (self().read_bool()) {
        m_unit = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_unit = nullptr;
        }
        ASR::expr_t *m_fmt;
        if (self().read_bool()) {
        m_fmt = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_fmt = nullptr;
        }
        ASR::expr_t *m_iomsg;
        if (self().read_bool()) {
        m_iomsg = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_iomsg = nullptr;
        }
        ASR::expr_t *m_iostat;
        if (self().read_bool()) {
        m_iostat = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_iostat = nullptr;
        }
        ASR::expr_t *m_id;
        if (self().read_bool()) {
        m_id = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_id = nullptr;
        }
        n_values = self().read_int64();
        Vec<expr_t*> v_values;
        v_values.reserve(al, n_values);
        for (size_t i=0; i<n_values; i++) {
            v_values.push_back(al, ASR::down_cast<ASR::expr_t>(self().deserialize_expr()));
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_FileRead_t(al, loc, m_label, m_unit, m_fmt, m_iomsg, m_iostat, m_id, v_values.p, v_values.n);
    }
    asr_t* deserialize_FileRewind() {
        int64_t m_label = self().read_int64();
        ASR::expr_t *m_unit;
        if (self().read_bool()) {
        m_unit = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_unit = nullptr;
        }
        ASR::expr_t *m_iostat;
        if (self().read_bool()) {
        m_iostat = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_iostat = nullptr;
        }
        ASR::expr_t *m_err;
        if (self().read_bool()) {
        m_err = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_err = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_FileRewind_t(al, loc, m_label, m_unit, m_iostat, m_err);
    }
    asr_t* deserialize_FileInquire() {
        int64_t m_label = self().read_int64();
        ASR::expr_t *m_unit;
        if (self().read_bool()) {
        m_unit = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_unit = nullptr;
        }
        ASR::expr_t *m_file;
        if (self().read_bool()) {
        m_file = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_file = nullptr;
        }
        ASR::expr_t *m_iostat;
        if (self().read_bool()) {
        m_iostat = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_iostat = nullptr;
        }
        ASR::expr_t *m_err;
        if (self().read_bool()) {
        m_err = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_err = nullptr;
        }
        ASR::expr_t *m_exist;
        if (self().read_bool()) {
        m_exist = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_exist = nullptr;
        }
        ASR::expr_t *m_opened;
        if (self().read_bool()) {
        m_opened = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_opened = nullptr;
        }
        ASR::expr_t *m_number;
        if (self().read_bool()) {
        m_number = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_number = nullptr;
        }
        ASR::expr_t *m_named;
        if (self().read_bool()) {
        m_named = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_named = nullptr;
        }
        ASR::expr_t *m_name;
        if (self().read_bool()) {
        m_name = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_name = nullptr;
        }
        ASR::expr_t *m_access;
        if (self().read_bool()) {
        m_access = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_access = nullptr;
        }
        ASR::expr_t *m_sequential;
        if (self().read_bool()) {
        m_sequential = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_sequential = nullptr;
        }
        ASR::expr_t *m_direct;
        if (self().read_bool()) {
        m_direct = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_direct = nullptr;
        }
        ASR::expr_t *m_form;
        if (self().read_bool()) {
        m_form = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_form = nullptr;
        }
        ASR::expr_t *m_formatted;
        if (self().read_bool()) {
        m_formatted = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_formatted = nullptr;
        }
        ASR::expr_t *m_unformatted;
        if (self().read_bool()) {
        m_unformatted = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_unformatted = nullptr;
        }
        ASR::expr_t *m_recl;
        if (self().read_bool()) {
        m_recl = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_recl = nullptr;
        }
        ASR::expr_t *m_nextrec;
        if (self().read_bool()) {
        m_nextrec = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_nextrec = nullptr;
        }
        ASR::expr_t *m_blank;
        if (self().read_bool()) {
        m_blank = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_blank = nullptr;
        }
        ASR::expr_t *m_position;
        if (self().read_bool()) {
        m_position = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_position = nullptr;
        }
        ASR::expr_t *m_action;
        if (self().read_bool()) {
        m_action = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_action = nullptr;
        }
        ASR::expr_t *m_read;
        if (self().read_bool()) {
        m_read = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_read = nullptr;
        }
        ASR::expr_t *m_write;
        if (self().read_bool()) {
        m_write = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_write = nullptr;
        }
        ASR::expr_t *m_readwrite;
        if (self().read_bool()) {
        m_readwrite = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_readwrite = nullptr;
        }
        ASR::expr_t *m_delim;
        if (self().read_bool()) {
        m_delim = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_delim = nullptr;
        }
        ASR::expr_t *m_pad;
        if (self().read_bool()) {
        m_pad = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_pad = nullptr;
        }
        ASR::expr_t *m_flen;
        if (self().read_bool()) {
        m_flen = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_flen = nullptr;
        }
        ASR::expr_t *m_blocksize;
        if (self().read_bool()) {
        m_blocksize = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_blocksize = nullptr;
        }
        ASR::expr_t *m_convert;
        if (self().read_bool()) {
        m_convert = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_convert = nullptr;
        }
        ASR::expr_t *m_carriagecontrol;
        if (self().read_bool()) {
        m_carriagecontrol = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_carriagecontrol = nullptr;
        }
        ASR::expr_t *m_iolength;
        if (self().read_bool()) {
        m_iolength = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_iolength = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_FileInquire_t(al, loc, m_label, m_unit, m_file, m_iostat, m_err, m_exist, m_opened, m_number, m_named, m_name, m_access, m_sequential, m_direct, m_form, m_formatted, m_unformatted, m_recl, m_nextrec, m_blank, m_position, m_action, m_read, m_write, m_readwrite, m_delim, m_pad, m_flen, m_blocksize, m_convert, m_carriagecontrol, m_iolength);
    }
    asr_t* deserialize_FileWrite() {
        size_t n_values; // Sequence
        int64_t m_label = self().read_int64();
        ASR::expr_t *m_unit;
        if (self().read_bool()) {
        m_unit = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_unit = nullptr;
        }
        ASR::expr_t *m_fmt;
        if (self().read_bool()) {
        m_fmt = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_fmt = nullptr;
        }
        ASR::expr_t *m_iomsg;
        if (self().read_bool()) {
        m_iomsg = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_iomsg = nullptr;
        }
        ASR::expr_t *m_iostat;
        if (self().read_bool()) {
        m_iostat = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_iostat = nullptr;
        }
        ASR::expr_t *m_id;
        if (self().read_bool()) {
        m_id = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_id = nullptr;
        }
        n_values = self().read_int64();
        Vec<expr_t*> v_values;
        v_values.reserve(al, n_values);
        for (size_t i=0; i<n_values; i++) {
            v_values.push_back(al, ASR::down_cast<ASR::expr_t>(self().deserialize_expr()));
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_FileWrite_t(al, loc, m_label, m_unit, m_fmt, m_iomsg, m_iostat, m_id, v_values.p, v_values.n);
    }
    asr_t* deserialize_Return() {
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Return_t(al, loc);
    }
    asr_t* deserialize_Select() {
        size_t n_body; // Sequence
        size_t n_default; // Sequence
        ASR::expr_t *m_test;
        m_test = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        n_body = self().read_int64();
        Vec<case_stmt_t*> v_body;
        v_body.reserve(al, n_body);
        for (size_t i=0; i<n_body; i++) {
            v_body.push_back(al, ASR::down_cast<ASR::case_stmt_t>(self().deserialize_case_stmt()));
        }
        n_default = self().read_int64();
        Vec<stmt_t*> v_default;
        v_default.reserve(al, n_default);
        for (size_t i=0; i<n_default; i++) {
            v_default.push_back(al, ASR::down_cast<ASR::stmt_t>(self().deserialize_stmt()));
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Select_t(al, loc, m_test, v_body.p, v_body.n, v_default.p, v_default.n);
    }
    asr_t* deserialize_Stop() {
        ASR::expr_t *m_code;
        if (self().read_bool()) {
        m_code = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_code = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Stop_t(al, loc, m_code);
    }
    asr_t* deserialize_Assert() {
        ASR::expr_t *m_test;
        m_test = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_msg;
        if (self().read_bool()) {
        m_msg = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_msg = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Assert_t(al, loc, m_test, m_msg);
    }
    asr_t* deserialize_SubroutineCall() {
        size_t n_args; // Sequence
        ASR::symbol_t *m_name;
        m_name = self().read_symbol();
        ASR::symbol_t *m_original_name;
        if (self().read_bool()) {
        m_original_name = self().read_symbol();
        } else {
        m_original_name = nullptr;
        }
        n_args = self().read_int64();
        Vec<call_arg_t> v_args;
        v_args.reserve(al, n_args);
        for (size_t i=0; i<n_args; i++) {
            v_args.push_back(al, self().deserialize_call_arg());
        }
        ASR::expr_t *m_dt;
        if (self().read_bool()) {
        m_dt = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_dt = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_SubroutineCall_t(al, loc, m_name, m_original_name, v_args.p, v_args.n, m_dt);
    }
    asr_t* deserialize_Where() {
        size_t n_body; // Sequence
        size_t n_orelse; // Sequence
        ASR::expr_t *m_test;
        m_test = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        n_body = self().read_int64();
        Vec<stmt_t*> v_body;
        v_body.reserve(al, n_body);
        for (size_t i=0; i<n_body; i++) {
            v_body.push_back(al, ASR::down_cast<ASR::stmt_t>(self().deserialize_stmt()));
        }
        n_orelse = self().read_int64();
        Vec<stmt_t*> v_orelse;
        v_orelse.reserve(al, n_orelse);
        for (size_t i=0; i<n_orelse; i++) {
            v_orelse.push_back(al, ASR::down_cast<ASR::stmt_t>(self().deserialize_stmt()));
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Where_t(al, loc, m_test, v_body.p, v_body.n, v_orelse.p, v_orelse.n);
    }
    asr_t* deserialize_WhileLoop() {
        size_t n_body; // Sequence
        ASR::expr_t *m_test;
        m_test = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        n_body = self().read_int64();
        Vec<stmt_t*> v_body;
        v_body.reserve(al, n_body);
        for (size_t i=0; i<n_body; i++) {
            v_body.push_back(al, ASR::down_cast<ASR::stmt_t>(self().deserialize_stmt()));
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_WhileLoop_t(al, loc, m_test, v_body.p, v_body.n);
    }
    asr_t* deserialize_Nullify() {
        size_t n_vars; // Sequence
        n_vars = self().read_int64();
        Vec<symbol_t*> v_vars;
        v_vars.reserve(al, n_vars);
        for (size_t i=0; i<n_vars; i++) {
            v_vars.push_back(al, self().read_symbol());
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Nullify_t(al, loc, v_vars.p, v_vars.n);
    }
    asr_t* deserialize_Flush() {
        int64_t m_label = self().read_int64();
        ASR::expr_t *m_unit;
        m_unit = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_err;
        if (self().read_bool()) {
        m_err = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_err = nullptr;
        }
        ASR::expr_t *m_iomsg;
        if (self().read_bool()) {
        m_iomsg = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_iomsg = nullptr;
        }
        ASR::expr_t *m_iostat;
        if (self().read_bool()) {
        m_iostat = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_iostat = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Flush_t(al, loc, m_label, m_unit, m_err, m_iomsg, m_iostat);
    }
    asr_t* deserialize_ListAppend() {
        ASR::symbol_t *m_a;
        m_a = self().read_symbol();
        ASR::expr_t *m_ele;
        m_ele = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ListAppend_t(al, loc, m_a, m_ele);
    }
    asr_t* deserialize_AssociateBlockCall() {
        ASR::symbol_t *m_m;
        m_m = self().read_symbol();
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_AssociateBlockCall_t(al, loc, m_m);
    }
    asr_t* deserialize_BlockCall() {
        ASR::symbol_t *m_m;
        m_m = self().read_symbol();
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_BlockCall_t(al, loc, m_m);
    }
    asr_t* deserialize_SetInsert() {
        ASR::symbol_t *m_a;
        m_a = self().read_symbol();
        ASR::expr_t *m_ele;
        m_ele = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_SetInsert_t(al, loc, m_a, m_ele);
    }
    asr_t* deserialize_SetRemove() {
        ASR::symbol_t *m_a;
        m_a = self().read_symbol();
        ASR::expr_t *m_ele;
        m_ele = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_SetRemove_t(al, loc, m_a, m_ele);
    }
    asr_t* deserialize_ListInsert() {
        ASR::symbol_t *m_a;
        m_a = self().read_symbol();
        ASR::expr_t *m_pos;
        m_pos = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_ele;
        m_ele = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ListInsert_t(al, loc, m_a, m_pos, m_ele);
    }
    asr_t* deserialize_ListRemove() {
        ASR::symbol_t *m_a;
        m_a = self().read_symbol();
        ASR::expr_t *m_ele;
        m_ele = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ListRemove_t(al, loc, m_a, m_ele);
    }
    asr_t* deserialize_DictInsert() {
        ASR::symbol_t *m_a;
        m_a = self().read_symbol();
        ASR::expr_t *m_key;
        m_key = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_value;
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_DictInsert_t(al, loc, m_a, m_key, m_value);
    }
    asr_t* deserialize_stmt() {
        uint8_t t = self().read_int8();
        ASR::stmtType ty = static_cast<ASR::stmtType>(t);
        switch (ty) {
            case (ASR::stmtType::Allocate) : return self().deserialize_Allocate();
            case (ASR::stmtType::Assign) : return self().deserialize_Assign();
            case (ASR::stmtType::Assignment) : return self().deserialize_Assignment();
            case (ASR::stmtType::Associate) : return self().deserialize_Associate();
            case (ASR::stmtType::Cycle) : return self().deserialize_Cycle();
            case (ASR::stmtType::ExplicitDeallocate) : return self().deserialize_ExplicitDeallocate();
            case (ASR::stmtType::ImplicitDeallocate) : return self().deserialize_ImplicitDeallocate();
            case (ASR::stmtType::DoConcurrentLoop) : return self().deserialize_DoConcurrentLoop();
            case (ASR::stmtType::DoLoop) : return self().deserialize_DoLoop();
            case (ASR::stmtType::ErrorStop) : return self().deserialize_ErrorStop();
            case (ASR::stmtType::Exit) : return self().deserialize_Exit();
            case (ASR::stmtType::ForAllSingle) : return self().deserialize_ForAllSingle();
            case (ASR::stmtType::GoTo) : return self().deserialize_GoTo();
            case (ASR::stmtType::GoToTarget) : return self().deserialize_GoToTarget();
            case (ASR::stmtType::If) : return self().deserialize_If();
            case (ASR::stmtType::IfArithmetic) : return self().deserialize_IfArithmetic();
            case (ASR::stmtType::Print) : return self().deserialize_Print();
            case (ASR::stmtType::FileOpen) : return self().deserialize_FileOpen();
            case (ASR::stmtType::FileClose) : return self().deserialize_FileClose();
            case (ASR::stmtType::FileRead) : return self().deserialize_FileRead();
            case (ASR::stmtType::FileRewind) : return self().deserialize_FileRewind();
            case (ASR::stmtType::FileInquire) : return self().deserialize_FileInquire();
            case (ASR::stmtType::FileWrite) : return self().deserialize_FileWrite();
            case (ASR::stmtType::Return) : return self().deserialize_Return();
            case (ASR::stmtType::Select) : return self().deserialize_Select();
            case (ASR::stmtType::Stop) : return self().deserialize_Stop();
            case (ASR::stmtType::Assert) : return self().deserialize_Assert();
            case (ASR::stmtType::SubroutineCall) : return self().deserialize_SubroutineCall();
            case (ASR::stmtType::Where) : return self().deserialize_Where();
            case (ASR::stmtType::WhileLoop) : return self().deserialize_WhileLoop();
            case (ASR::stmtType::Nullify) : return self().deserialize_Nullify();
            case (ASR::stmtType::Flush) : return self().deserialize_Flush();
            case (ASR::stmtType::ListAppend) : return self().deserialize_ListAppend();
            case (ASR::stmtType::AssociateBlockCall) : return self().deserialize_AssociateBlockCall();
            case (ASR::stmtType::BlockCall) : return self().deserialize_BlockCall();
            case (ASR::stmtType::SetInsert) : return self().deserialize_SetInsert();
            case (ASR::stmtType::SetRemove) : return self().deserialize_SetRemove();
            case (ASR::stmtType::ListInsert) : return self().deserialize_ListInsert();
            case (ASR::stmtType::ListRemove) : return self().deserialize_ListRemove();
            case (ASR::stmtType::DictInsert) : return self().deserialize_DictInsert();
            default : throw LFortranException("Unknown type in deserialize_stmt()");
        }
        throw LFortranException("Switch statement above was not exhaustive.");
    }
    asr_t* deserialize_BoolOp() {
        ASR::expr_t *m_left;
        m_left = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::boolopType m_op = self().deserialize_boolop();
        ASR::expr_t *m_right;
        m_right = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_BoolOp_t(al, loc, m_left, m_op, m_right, m_type, m_value);
    }
    asr_t* deserialize_BinOp() {
        ASR::expr_t *m_left;
        m_left = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::binopType m_op = self().deserialize_binop();
        ASR::expr_t *m_right;
        m_right = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        ASR::expr_t *m_overloaded;
        if (self().read_bool()) {
        m_overloaded = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_overloaded = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_BinOp_t(al, loc, m_left, m_op, m_right, m_type, m_value, m_overloaded);
    }
    asr_t* deserialize_StrOp() {
        ASR::expr_t *m_left;
        m_left = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::stropType m_op = self().deserialize_strop();
        ASR::expr_t *m_right;
        m_right = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_StrOp_t(al, loc, m_left, m_op, m_right, m_type, m_value);
    }
    asr_t* deserialize_UnaryOp() {
        ASR::unaryopType m_op = self().deserialize_unaryop();
        ASR::expr_t *m_operand;
        m_operand = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_UnaryOp_t(al, loc, m_op, m_operand, m_type, m_value);
    }
    asr_t* deserialize_ComplexConstructor() {
        ASR::expr_t *m_re;
        m_re = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_im;
        m_im = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ComplexConstructor_t(al, loc, m_re, m_im, m_type, m_value);
    }
    asr_t* deserialize_NamedExpr() {
        ASR::expr_t *m_target;
        m_target = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_value;
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_NamedExpr_t(al, loc, m_target, m_value, m_type);
    }
    asr_t* deserialize_Compare() {
        ASR::expr_t *m_left;
        m_left = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::cmpopType m_op = self().deserialize_cmpop();
        ASR::expr_t *m_right;
        m_right = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        ASR::expr_t *m_overloaded;
        if (self().read_bool()) {
        m_overloaded = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_overloaded = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Compare_t(al, loc, m_left, m_op, m_right, m_type, m_value, m_overloaded);
    }
    asr_t* deserialize_IfExp() {
        ASR::expr_t *m_test;
        m_test = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_body;
        m_body = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_orelse;
        m_orelse = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_IfExp_t(al, loc, m_test, m_body, m_orelse, m_type);
    }
    asr_t* deserialize_FunctionCall() {
        size_t n_args; // Sequence
        ASR::symbol_t *m_name;
        m_name = self().read_symbol();
        ASR::symbol_t *m_original_name;
        if (self().read_bool()) {
        m_original_name = self().read_symbol();
        } else {
        m_original_name = nullptr;
        }
        n_args = self().read_int64();
        Vec<call_arg_t> v_args;
        v_args.reserve(al, n_args);
        for (size_t i=0; i<n_args; i++) {
            v_args.push_back(al, self().deserialize_call_arg());
        }
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        ASR::expr_t *m_dt;
        if (self().read_bool()) {
        m_dt = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_dt = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_FunctionCall_t(al, loc, m_name, m_original_name, v_args.p, v_args.n, m_type, m_value, m_dt);
    }
    asr_t* deserialize_DerivedTypeConstructor() {
        size_t n_args; // Sequence
        ASR::symbol_t *m_dt_sym;
        m_dt_sym = self().read_symbol();
        n_args = self().read_int64();
        Vec<expr_t*> v_args;
        v_args.reserve(al, n_args);
        for (size_t i=0; i<n_args; i++) {
            v_args.push_back(al, ASR::down_cast<ASR::expr_t>(self().deserialize_expr()));
        }
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_DerivedTypeConstructor_t(al, loc, m_dt_sym, v_args.p, v_args.n, m_type);
    }
    asr_t* deserialize_ImpliedDoLoop() {
        size_t n_values; // Sequence
        n_values = self().read_int64();
        Vec<expr_t*> v_values;
        v_values.reserve(al, n_values);
        for (size_t i=0; i<n_values; i++) {
            v_values.push_back(al, ASR::down_cast<ASR::expr_t>(self().deserialize_expr()));
        }
        ASR::expr_t *m_var;
        m_var = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_start;
        m_start = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_end;
        m_end = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_increment;
        if (self().read_bool()) {
        m_increment = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_increment = nullptr;
        }
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ImpliedDoLoop_t(al, loc, v_values.p, v_values.n, m_var, m_start, m_end, m_increment, m_type, m_value);
    }
    asr_t* deserialize_IntegerConstant() {
        int64_t m_n = self().read_int64();
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_IntegerConstant_t(al, loc, m_n, m_type);
    }
    asr_t* deserialize_IntegerBOZ() {
        int64_t m_v = self().read_int64();
        ASR::integerbozType m_intboz_type = self().deserialize_integerboz();
        ASR::ttype_t *m_type;
        if (self().read_bool()) {
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        } else {
        m_type = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_IntegerBOZ_t(al, loc, m_v, m_intboz_type, m_type);
    }
    asr_t* deserialize_RealConstant() {
        double m_r = self().read_float64();
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_RealConstant_t(al, loc, m_r, m_type);
    }
    asr_t* deserialize_ComplexConstant() {
        double m_re = self().read_float64();
        double m_im = self().read_float64();
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ComplexConstant_t(al, loc, m_re, m_im, m_type);
    }
    asr_t* deserialize_LogicalConstant() {
        bool m_value = self().read_bool();
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_LogicalConstant_t(al, loc, m_value, m_type);
    }
    asr_t* deserialize_ListConstant() {
        size_t n_args; // Sequence
        n_args = self().read_int64();
        Vec<expr_t*> v_args;
        v_args.reserve(al, n_args);
        for (size_t i=0; i<n_args; i++) {
            v_args.push_back(al, ASR::down_cast<ASR::expr_t>(self().deserialize_expr()));
        }
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ListConstant_t(al, loc, v_args.p, v_args.n, m_type);
    }
    asr_t* deserialize_ListLen() {
        ASR::expr_t *m_arg;
        m_arg = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ListLen_t(al, loc, m_arg, m_type, m_value);
    }
    asr_t* deserialize_ListConcat() {
        ASR::expr_t *m_left;
        m_left = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_right;
        m_right = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ListConcat_t(al, loc, m_left, m_right, m_type, m_value);
    }
    asr_t* deserialize_ArrayConstant() {
        size_t n_args; // Sequence
        n_args = self().read_int64();
        Vec<expr_t*> v_args;
        v_args.reserve(al, n_args);
        for (size_t i=0; i<n_args; i++) {
            v_args.push_back(al, ASR::down_cast<ASR::expr_t>(self().deserialize_expr()));
        }
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ArrayConstant_t(al, loc, v_args.p, v_args.n, m_type);
    }
    asr_t* deserialize_SetConstant() {
        size_t n_elements; // Sequence
        n_elements = self().read_int64();
        Vec<expr_t*> v_elements;
        v_elements.reserve(al, n_elements);
        for (size_t i=0; i<n_elements; i++) {
            v_elements.push_back(al, ASR::down_cast<ASR::expr_t>(self().deserialize_expr()));
        }
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_SetConstant_t(al, loc, v_elements.p, v_elements.n, m_type);
    }
    asr_t* deserialize_SetLen() {
        ASR::expr_t *m_arg;
        m_arg = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_SetLen_t(al, loc, m_arg, m_type, m_value);
    }
    asr_t* deserialize_TupleConstant() {
        size_t n_elements; // Sequence
        n_elements = self().read_int64();
        Vec<expr_t*> v_elements;
        v_elements.reserve(al, n_elements);
        for (size_t i=0; i<n_elements; i++) {
            v_elements.push_back(al, ASR::down_cast<ASR::expr_t>(self().deserialize_expr()));
        }
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_TupleConstant_t(al, loc, v_elements.p, v_elements.n, m_type);
    }
    asr_t* deserialize_TupleLen() {
        ASR::expr_t *m_arg;
        m_arg = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_TupleLen_t(al, loc, m_arg, m_type, m_value);
    }
    asr_t* deserialize_StringConstant() {
        char *m_s;
        m_s = self().read_cstring();
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_StringConstant_t(al, loc, m_s, m_type);
    }
    asr_t* deserialize_StringConcat() {
        ASR::expr_t *m_left;
        m_left = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_right;
        m_right = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_StringConcat_t(al, loc, m_left, m_right, m_type, m_value);
    }
    asr_t* deserialize_StringLen() {
        ASR::expr_t *m_arg;
        m_arg = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_StringLen_t(al, loc, m_arg, m_type, m_value);
    }
    asr_t* deserialize_DictConstant() {
        size_t n_keys; // Sequence
        size_t n_values; // Sequence
        n_keys = self().read_int64();
        Vec<expr_t*> v_keys;
        v_keys.reserve(al, n_keys);
        for (size_t i=0; i<n_keys; i++) {
            v_keys.push_back(al, ASR::down_cast<ASR::expr_t>(self().deserialize_expr()));
        }
        n_values = self().read_int64();
        Vec<expr_t*> v_values;
        v_values.reserve(al, n_values);
        for (size_t i=0; i<n_values; i++) {
            v_values.push_back(al, ASR::down_cast<ASR::expr_t>(self().deserialize_expr()));
        }
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_DictConstant_t(al, loc, v_keys.p, v_keys.n, v_values.p, v_values.n, m_type);
    }
    asr_t* deserialize_DictLen() {
        ASR::expr_t *m_arg;
        m_arg = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_DictLen_t(al, loc, m_arg, m_type, m_value);
    }
    asr_t* deserialize_Var() {
        ASR::symbol_t *m_v;
        m_v = self().read_symbol();
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Var_t(al, loc, m_v);
    }
    asr_t* deserialize_ArrayRef() {
        size_t n_args; // Sequence
        ASR::symbol_t *m_v;
        m_v = self().read_symbol();
        n_args = self().read_int64();
        Vec<array_index_t> v_args;
        v_args.reserve(al, n_args);
        for (size_t i=0; i<n_args; i++) {
            v_args.push_back(al, self().deserialize_array_index());
        }
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ArrayRef_t(al, loc, m_v, v_args.p, v_args.n, m_type, m_value);
    }
    asr_t* deserialize_ArraySize() {
        ASR::expr_t *m_v;
        m_v = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_dim;
        if (self().read_bool()) {
        m_dim = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_dim = nullptr;
        }
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ArraySize_t(al, loc, m_v, m_dim, m_type, m_value);
    }
    asr_t* deserialize_ArrayBound() {
        ASR::expr_t *m_v;
        m_v = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_dim;
        if (self().read_bool()) {
        m_dim = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_dim = nullptr;
        }
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::arrayboundType m_bound = self().deserialize_arraybound();
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ArrayBound_t(al, loc, m_v, m_dim, m_type, m_bound, m_value);
    }
    asr_t* deserialize_ArrayTranspose() {
        ASR::expr_t *m_matrix;
        m_matrix = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ArrayTranspose_t(al, loc, m_matrix, m_type, m_value);
    }
    asr_t* deserialize_ArrayMatMul() {
        ASR::expr_t *m_matrix_a;
        m_matrix_a = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_matrix_b;
        m_matrix_b = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ArrayMatMul_t(al, loc, m_matrix_a, m_matrix_b, m_type, m_value);
    }
    asr_t* deserialize_ArrayPack() {
        ASR::expr_t *m_array;
        m_array = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_mask;
        m_mask = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_vector;
        if (self().read_bool()) {
        m_vector = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_vector = nullptr;
        }
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ArrayPack_t(al, loc, m_array, m_mask, m_vector, m_type, m_value);
    }
    asr_t* deserialize_Transfer() {
        ASR::expr_t *m_source;
        m_source = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_mold;
        m_mold = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_size;
        if (self().read_bool()) {
        m_size = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_size = nullptr;
        }
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Transfer_t(al, loc, m_source, m_mold, m_size, m_type, m_value);
    }
    asr_t* deserialize_DerivedRef() {
        ASR::expr_t *m_v;
        m_v = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::symbol_t *m_m;
        m_m = self().read_symbol();
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_DerivedRef_t(al, loc, m_v, m_m, m_type, m_value);
    }
    asr_t* deserialize_Cast() {
        ASR::expr_t *m_arg;
        m_arg = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::cast_kindType m_kind = self().deserialize_cast_kind();
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Cast_t(al, loc, m_arg, m_kind, m_type, m_value);
    }
    asr_t* deserialize_ComplexRe() {
        ASR::expr_t *m_arg;
        m_arg = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ComplexRe_t(al, loc, m_arg, m_type, m_value);
    }
    asr_t* deserialize_ComplexIm() {
        ASR::expr_t *m_arg;
        m_arg = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::expr_t *m_value;
        if (self().read_bool()) {
        m_value = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_value = nullptr;
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_ComplexIm_t(al, loc, m_arg, m_type, m_value);
    }
    asr_t* deserialize_DictItem() {
        ASR::symbol_t *m_a;
        m_a = self().read_symbol();
        ASR::expr_t *m_key;
        m_key = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        ASR::expr_t *m_default;
        if (self().read_bool()) {
        m_default = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_default = nullptr;
        }
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_DictItem_t(al, loc, m_a, m_key, m_default, m_type);
    }
    asr_t* deserialize_expr() {
        uint8_t t = self().read_int8();
        ASR::exprType ty = static_cast<ASR::exprType>(t);
        switch (ty) {
            case (ASR::exprType::BoolOp) : return self().deserialize_BoolOp();
            case (ASR::exprType::BinOp) : return self().deserialize_BinOp();
            case (ASR::exprType::StrOp) : return self().deserialize_StrOp();
            case (ASR::exprType::UnaryOp) : return self().deserialize_UnaryOp();
            case (ASR::exprType::ComplexConstructor) : return self().deserialize_ComplexConstructor();
            case (ASR::exprType::NamedExpr) : return self().deserialize_NamedExpr();
            case (ASR::exprType::Compare) : return self().deserialize_Compare();
            case (ASR::exprType::IfExp) : return self().deserialize_IfExp();
            case (ASR::exprType::FunctionCall) : return self().deserialize_FunctionCall();
            case (ASR::exprType::DerivedTypeConstructor) : return self().deserialize_DerivedTypeConstructor();
            case (ASR::exprType::ImpliedDoLoop) : return self().deserialize_ImpliedDoLoop();
            case (ASR::exprType::IntegerConstant) : return self().deserialize_IntegerConstant();
            case (ASR::exprType::IntegerBOZ) : return self().deserialize_IntegerBOZ();
            case (ASR::exprType::RealConstant) : return self().deserialize_RealConstant();
            case (ASR::exprType::ComplexConstant) : return self().deserialize_ComplexConstant();
            case (ASR::exprType::LogicalConstant) : return self().deserialize_LogicalConstant();
            case (ASR::exprType::ListConstant) : return self().deserialize_ListConstant();
            case (ASR::exprType::ListLen) : return self().deserialize_ListLen();
            case (ASR::exprType::ListConcat) : return self().deserialize_ListConcat();
            case (ASR::exprType::ArrayConstant) : return self().deserialize_ArrayConstant();
            case (ASR::exprType::SetConstant) : return self().deserialize_SetConstant();
            case (ASR::exprType::SetLen) : return self().deserialize_SetLen();
            case (ASR::exprType::TupleConstant) : return self().deserialize_TupleConstant();
            case (ASR::exprType::TupleLen) : return self().deserialize_TupleLen();
            case (ASR::exprType::StringConstant) : return self().deserialize_StringConstant();
            case (ASR::exprType::StringConcat) : return self().deserialize_StringConcat();
            case (ASR::exprType::StringLen) : return self().deserialize_StringLen();
            case (ASR::exprType::DictConstant) : return self().deserialize_DictConstant();
            case (ASR::exprType::DictLen) : return self().deserialize_DictLen();
            case (ASR::exprType::Var) : return self().deserialize_Var();
            case (ASR::exprType::ArrayRef) : return self().deserialize_ArrayRef();
            case (ASR::exprType::ArraySize) : return self().deserialize_ArraySize();
            case (ASR::exprType::ArrayBound) : return self().deserialize_ArrayBound();
            case (ASR::exprType::ArrayTranspose) : return self().deserialize_ArrayTranspose();
            case (ASR::exprType::ArrayMatMul) : return self().deserialize_ArrayMatMul();
            case (ASR::exprType::ArrayPack) : return self().deserialize_ArrayPack();
            case (ASR::exprType::Transfer) : return self().deserialize_Transfer();
            case (ASR::exprType::DerivedRef) : return self().deserialize_DerivedRef();
            case (ASR::exprType::Cast) : return self().deserialize_Cast();
            case (ASR::exprType::ComplexRe) : return self().deserialize_ComplexRe();
            case (ASR::exprType::ComplexIm) : return self().deserialize_ComplexIm();
            case (ASR::exprType::DictItem) : return self().deserialize_DictItem();
            default : throw LFortranException("Unknown type in deserialize_expr()");
        }
        throw LFortranException("Switch statement above was not exhaustive.");
    }
    asr_t* deserialize_Integer() {
        size_t n_dims; // Sequence
        int64_t m_kind = self().read_int64();
        n_dims = self().read_int64();
        Vec<dimension_t> v_dims;
        v_dims.reserve(al, n_dims);
        for (size_t i=0; i<n_dims; i++) {
            v_dims.push_back(al, self().deserialize_dimension());
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Integer_t(al, loc, m_kind, v_dims.p, v_dims.n);
    }
    asr_t* deserialize_Real() {
        size_t n_dims; // Sequence
        int64_t m_kind = self().read_int64();
        n_dims = self().read_int64();
        Vec<dimension_t> v_dims;
        v_dims.reserve(al, n_dims);
        for (size_t i=0; i<n_dims; i++) {
            v_dims.push_back(al, self().deserialize_dimension());
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Real_t(al, loc, m_kind, v_dims.p, v_dims.n);
    }
    asr_t* deserialize_Complex() {
        size_t n_dims; // Sequence
        int64_t m_kind = self().read_int64();
        n_dims = self().read_int64();
        Vec<dimension_t> v_dims;
        v_dims.reserve(al, n_dims);
        for (size_t i=0; i<n_dims; i++) {
            v_dims.push_back(al, self().deserialize_dimension());
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Complex_t(al, loc, m_kind, v_dims.p, v_dims.n);
    }
    asr_t* deserialize_Character() {
        size_t n_dims; // Sequence
        int64_t m_kind = self().read_int64();
        int64_t m_len = self().read_int64();
        ASR::expr_t *m_len_expr;
        if (self().read_bool()) {
        m_len_expr = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_len_expr = nullptr;
        }
        n_dims = self().read_int64();
        Vec<dimension_t> v_dims;
        v_dims.reserve(al, n_dims);
        for (size_t i=0; i<n_dims; i++) {
            v_dims.push_back(al, self().deserialize_dimension());
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Character_t(al, loc, m_kind, m_len, m_len_expr, v_dims.p, v_dims.n);
    }
    asr_t* deserialize_Logical() {
        size_t n_dims; // Sequence
        int64_t m_kind = self().read_int64();
        n_dims = self().read_int64();
        Vec<dimension_t> v_dims;
        v_dims.reserve(al, n_dims);
        for (size_t i=0; i<n_dims; i++) {
            v_dims.push_back(al, self().deserialize_dimension());
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Logical_t(al, loc, m_kind, v_dims.p, v_dims.n);
    }
    asr_t* deserialize_Set() {
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Set_t(al, loc, m_type);
    }
    asr_t* deserialize_List() {
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_List_t(al, loc, m_type);
    }
    asr_t* deserialize_Tuple() {
        size_t n_type; // Sequence
        n_type = self().read_int64();
        Vec<ttype_t*> v_type;
        v_type.reserve(al, n_type);
        for (size_t i=0; i<n_type; i++) {
            v_type.push_back(al, ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype()));
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Tuple_t(al, loc, v_type.p, v_type.n);
    }
    asr_t* deserialize_Derived() {
        size_t n_dims; // Sequence
        ASR::symbol_t *m_derived_type;
        m_derived_type = self().read_symbol();
        n_dims = self().read_int64();
        Vec<dimension_t> v_dims;
        v_dims.reserve(al, n_dims);
        for (size_t i=0; i<n_dims; i++) {
            v_dims.push_back(al, self().deserialize_dimension());
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Derived_t(al, loc, m_derived_type, v_dims.p, v_dims.n);
    }
    asr_t* deserialize_Class() {
        size_t n_dims; // Sequence
        ASR::symbol_t *m_class_type;
        m_class_type = self().read_symbol();
        n_dims = self().read_int64();
        Vec<dimension_t> v_dims;
        v_dims.reserve(al, n_dims);
        for (size_t i=0; i<n_dims; i++) {
            v_dims.push_back(al, self().deserialize_dimension());
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Class_t(al, loc, m_class_type, v_dims.p, v_dims.n);
    }
    asr_t* deserialize_Dict() {
        ASR::ttype_t *m_key_type;
        m_key_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        ASR::ttype_t *m_value_type;
        m_value_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Dict_t(al, loc, m_key_type, m_value_type);
    }
    asr_t* deserialize_Pointer() {
        ASR::ttype_t *m_type;
        m_type = ASR::down_cast<ASR::ttype_t>(self().deserialize_ttype());
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Pointer_t(al, loc, m_type);
    }
    asr_t* deserialize_ttype() {
        uint8_t t = self().read_int8();
        ASR::ttypeType ty = static_cast<ASR::ttypeType>(t);
        switch (ty) {
            case (ASR::ttypeType::Integer) : return self().deserialize_Integer();
            case (ASR::ttypeType::Real) : return self().deserialize_Real();
            case (ASR::ttypeType::Complex) : return self().deserialize_Complex();
            case (ASR::ttypeType::Character) : return self().deserialize_Character();
            case (ASR::ttypeType::Logical) : return self().deserialize_Logical();
            case (ASR::ttypeType::Set) : return self().deserialize_Set();
            case (ASR::ttypeType::List) : return self().deserialize_List();
            case (ASR::ttypeType::Tuple) : return self().deserialize_Tuple();
            case (ASR::ttypeType::Derived) : return self().deserialize_Derived();
            case (ASR::ttypeType::Class) : return self().deserialize_Class();
            case (ASR::ttypeType::Dict) : return self().deserialize_Dict();
            case (ASR::ttypeType::Pointer) : return self().deserialize_Pointer();
            default : throw LFortranException("Unknown type in deserialize_ttype()");
        }
        throw LFortranException("Switch statement above was not exhaustive.");
    }
    boolopType deserialize_boolop() {
        uint8_t t = self().read_int8();
        boolopType ty = static_cast<boolopType>(t);
        return ty;
    }
    binopType deserialize_binop() {
        uint8_t t = self().read_int8();
        binopType ty = static_cast<binopType>(t);
        return ty;
    }
    unaryopType deserialize_unaryop() {
        uint8_t t = self().read_int8();
        unaryopType ty = static_cast<unaryopType>(t);
        return ty;
    }
    stropType deserialize_strop() {
        uint8_t t = self().read_int8();
        stropType ty = static_cast<stropType>(t);
        return ty;
    }
    cmpopType deserialize_cmpop() {
        uint8_t t = self().read_int8();
        cmpopType ty = static_cast<cmpopType>(t);
        return ty;
    }
    integerbozType deserialize_integerboz() {
        uint8_t t = self().read_int8();
        integerbozType ty = static_cast<integerbozType>(t);
        return ty;
    }
    arrayboundType deserialize_arraybound() {
        uint8_t t = self().read_int8();
        arrayboundType ty = static_cast<arrayboundType>(t);
        return ty;
    }
    cast_kindType deserialize_cast_kind() {
        uint8_t t = self().read_int8();
        cast_kindType ty = static_cast<cast_kindType>(t);
        return ty;
    }
    dimension_t deserialize_dimension() {
        dimension_t x;
        {
            bool present=self().read_bool();
            if (present) {
                x.m_start = down_cast<expr_t>(deserialize_expr());
            } else {
                x.m_start = nullptr;
            }
        }
        {
            bool present=self().read_bool();
            if (present) {
                x.m_end = down_cast<expr_t>(deserialize_expr());
            } else {
                x.m_end = nullptr;
            }
        }
        return x;
    }
    alloc_arg_t deserialize_alloc_arg() {
        alloc_arg_t x;
        {
                x.m_a = self().read_symbol();
        }
        {
            uint64_t n = self().read_int64();
            Vec<dimension_t> v;
            v.reserve(al, n);
            for (uint64_t i=0; i<n; i++) {
                v.push_back(al, self().deserialize_dimension());
            }
            x.m_dims = v.p;
            x.n_dims = v.n;
        }
        return x;
    }
    asr_t* deserialize_Attribute() {
        size_t n_args; // Sequence
        char *m_name;
        m_name = self().read_cstring();
        n_args = self().read_int64();
        Vec<attribute_arg_t> v_args;
        v_args.reserve(al, n_args);
        for (size_t i=0; i<n_args; i++) {
            v_args.push_back(al, self().deserialize_attribute_arg());
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Attribute_t(al, loc, m_name, v_args.p, v_args.n);
    }
    asr_t* deserialize_attribute() {
        uint8_t t = self().read_int8();
        ASR::attributeType ty = static_cast<ASR::attributeType>(t);
        switch (ty) {
            case (ASR::attributeType::Attribute) : return self().deserialize_Attribute();
            default : throw LFortranException("Unknown type in deserialize_attribute()");
        }
        throw LFortranException("Switch statement above was not exhaustive.");
    }
    attribute_arg_t deserialize_attribute_arg() {
        attribute_arg_t x;
        {
                x.m_arg = self().read_cstring();
        }
        return x;
    }
    call_arg_t deserialize_call_arg() {
        call_arg_t x;
        {
            bool present=self().read_bool();
            if (present) {
                x.m_value = down_cast<expr_t>(deserialize_expr());
            } else {
                x.m_value = nullptr;
            }
        }
        return x;
    }
    asr_t* deserialize_Bind() {
        char *m_lang;
        m_lang = self().read_cstring();
        char *m_name;
        m_name = self().read_cstring();
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_Bind_t(al, loc, m_lang, m_name);
    }
    asr_t* deserialize_tbind() {
        uint8_t t = self().read_int8();
        ASR::tbindType ty = static_cast<ASR::tbindType>(t);
        switch (ty) {
            case (ASR::tbindType::Bind) : return self().deserialize_Bind();
            default : throw LFortranException("Unknown type in deserialize_tbind()");
        }
        throw LFortranException("Switch statement above was not exhaustive.");
    }
    array_index_t deserialize_array_index() {
        array_index_t x;
        {
            bool present=self().read_bool();
            if (present) {
                x.m_left = down_cast<expr_t>(deserialize_expr());
            } else {
                x.m_left = nullptr;
            }
        }
        {
            bool present=self().read_bool();
            if (present) {
                x.m_right = down_cast<expr_t>(deserialize_expr());
            } else {
                x.m_right = nullptr;
            }
        }
        {
            bool present=self().read_bool();
            if (present) {
                x.m_step = down_cast<expr_t>(deserialize_expr());
            } else {
                x.m_step = nullptr;
            }
        }
        return x;
    }
    do_loop_head_t deserialize_do_loop_head() {
        do_loop_head_t x;
        {
            bool present=self().read_bool();
            if (present) {
                x.m_v = down_cast<expr_t>(deserialize_expr());
            } else {
                x.m_v = nullptr;
            }
        }
        {
            bool present=self().read_bool();
            if (present) {
                x.m_start = down_cast<expr_t>(deserialize_expr());
            } else {
                x.m_start = nullptr;
            }
        }
        {
            bool present=self().read_bool();
            if (present) {
                x.m_end = down_cast<expr_t>(deserialize_expr());
            } else {
                x.m_end = nullptr;
            }
        }
        {
            bool present=self().read_bool();
            if (present) {
                x.m_increment = down_cast<expr_t>(deserialize_expr());
            } else {
                x.m_increment = nullptr;
            }
        }
        return x;
    }
    asr_t* deserialize_CaseStmt() {
        size_t n_test; // Sequence
        size_t n_body; // Sequence
        n_test = self().read_int64();
        Vec<expr_t*> v_test;
        v_test.reserve(al, n_test);
        for (size_t i=0; i<n_test; i++) {
            v_test.push_back(al, ASR::down_cast<ASR::expr_t>(self().deserialize_expr()));
        }
        n_body = self().read_int64();
        Vec<stmt_t*> v_body;
        v_body.reserve(al, n_body);
        for (size_t i=0; i<n_body; i++) {
            v_body.push_back(al, ASR::down_cast<ASR::stmt_t>(self().deserialize_stmt()));
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_CaseStmt_t(al, loc, v_test.p, v_test.n, v_body.p, v_body.n);
    }
    asr_t* deserialize_CaseStmt_Range() {
        size_t n_body; // Sequence
        ASR::expr_t *m_start;
        if (self().read_bool()) {
        m_start = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_start = nullptr;
        }
        ASR::expr_t *m_end;
        if (self().read_bool()) {
        m_end = ASR::down_cast<ASR::expr_t>(self().deserialize_expr());
        } else {
        m_end = nullptr;
        }
        n_body = self().read_int64();
        Vec<stmt_t*> v_body;
        v_body.reserve(al, n_body);
        for (size_t i=0; i<n_body; i++) {
            v_body.push_back(al, ASR::down_cast<ASR::stmt_t>(self().deserialize_stmt()));
        }
        Location loc;
        // FIXME: read loc from the stream
        loc.first=0; loc.last=0;
        return ASR::make_CaseStmt_Range_t(al, loc, m_start, m_end, v_body.p, v_body.n);
    }
    asr_t* deserialize_case_stmt() {
        uint8_t t = self().read_int8();
        ASR::case_stmtType ty = static_cast<ASR::case_stmtType>(t);
        switch (ty) {
            case (ASR::case_stmtType::CaseStmt) : return self().deserialize_CaseStmt();
            case (ASR::case_stmtType::CaseStmt_Range) : return self().deserialize_CaseStmt_Range();
            default : throw LFortranException("Unknown type in deserialize_case_stmt()");
        }
        throw LFortranException("Switch statement above was not exhaustive.");
    }
};


/******************************************************************************/
// Expression and statement Duplicator class

class ExprStmtDuplicator {
private:
    Allocator& al;

public:
    bool success;
    bool allow_procedure_calls;

    ExprStmtDuplicator(Allocator& al_) : al(al_), success(false), allow_procedure_calls(true) {}


    ASR::asr_t* duplicate_Allocate(Allocate_t* x) {
        expr_t* m_stat = duplicate_expr(x->m_stat);
        expr_t* m_errmsg = duplicate_expr(x->m_errmsg);
        expr_t* m_source = duplicate_expr(x->m_source);
        return make_Allocate_t(al, x->base.base.loc, x->m_args, x->n_args, m_stat, m_errmsg, m_source);
    }


    ASR::asr_t* duplicate_Assign(Assign_t* x) {
        return (asr_t*)x;
    }


    ASR::asr_t* duplicate_Assignment(Assignment_t* x) {
        expr_t* m_target = duplicate_expr(x->m_target);
        expr_t* m_value = duplicate_expr(x->m_value);
        stmt_t* m_overloaded = duplicate_stmt(x->m_overloaded);
        return make_Assignment_t(al, x->base.base.loc, m_target, m_value, m_overloaded);
    }


    ASR::asr_t* duplicate_Associate(Associate_t* x) {
        expr_t* m_target = duplicate_expr(x->m_target);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_Associate_t(al, x->base.base.loc, m_target, m_value);
    }


    ASR::asr_t* duplicate_Cycle(Cycle_t* x) {
        return (asr_t*)x;
    }


    ASR::asr_t* duplicate_ExplicitDeallocate(ExplicitDeallocate_t* x) {
        Vec<symbol_t*> m_vars;
        m_vars.reserve(al, x->n_vars);
        for (size_t i = 0; i < x->n_vars; i++) {
            m_vars.push_back(al, x->m_vars[i]);
        }
        return make_ExplicitDeallocate_t(al, x->base.base.loc, m_vars.p, x->n_vars);
    }


    ASR::asr_t* duplicate_ImplicitDeallocate(ImplicitDeallocate_t* x) {
        Vec<symbol_t*> m_vars;
        m_vars.reserve(al, x->n_vars);
        for (size_t i = 0; i < x->n_vars; i++) {
            m_vars.push_back(al, x->m_vars[i]);
        }
        return make_ImplicitDeallocate_t(al, x->base.base.loc, m_vars.p, x->n_vars);
    }


    ASR::asr_t* duplicate_DoConcurrentLoop(DoConcurrentLoop_t* x) {
        Vec<stmt_t*> m_body;
        m_body.reserve(al, x->n_body);
        for (size_t i = 0; i < x->n_body; i++) {
            m_body.push_back(al, duplicate_stmt(x->m_body[i]));
        }
        return make_DoConcurrentLoop_t(al, x->base.base.loc, x->m_head, m_body.p, x->n_body);
    }


    ASR::asr_t* duplicate_DoLoop(DoLoop_t* x) {
        Vec<stmt_t*> m_body;
        m_body.reserve(al, x->n_body);
        for (size_t i = 0; i < x->n_body; i++) {
            m_body.push_back(al, duplicate_stmt(x->m_body[i]));
        }
        return make_DoLoop_t(al, x->base.base.loc, x->m_head, m_body.p, x->n_body);
    }


    ASR::asr_t* duplicate_ErrorStop(ErrorStop_t* x) {
        expr_t* m_code = duplicate_expr(x->m_code);
        return make_ErrorStop_t(al, x->base.base.loc, m_code);
    }


    ASR::asr_t* duplicate_Exit(Exit_t* x) {
        return (asr_t*)x;
    }


    ASR::asr_t* duplicate_ForAllSingle(ForAllSingle_t* x) {
        stmt_t* m_assign_stmt = duplicate_stmt(x->m_assign_stmt);
        return make_ForAllSingle_t(al, x->base.base.loc, x->m_head, m_assign_stmt);
    }


    ASR::asr_t* duplicate_GoTo(GoTo_t* x) {
        return (asr_t*)x;
    }


    ASR::asr_t* duplicate_GoToTarget(GoToTarget_t* x) {
        return (asr_t*)x;
    }


    ASR::asr_t* duplicate_If(If_t* x) {
        expr_t* m_test = duplicate_expr(x->m_test);
        Vec<stmt_t*> m_body;
        m_body.reserve(al, x->n_body);
        for (size_t i = 0; i < x->n_body; i++) {
            m_body.push_back(al, duplicate_stmt(x->m_body[i]));
        }
        Vec<stmt_t*> m_orelse;
        m_orelse.reserve(al, x->n_orelse);
        for (size_t i = 0; i < x->n_orelse; i++) {
            m_orelse.push_back(al, duplicate_stmt(x->m_orelse[i]));
        }
        return make_If_t(al, x->base.base.loc, m_test, m_body.p, x->n_body, m_orelse.p, x->n_orelse);
    }


    ASR::asr_t* duplicate_IfArithmetic(IfArithmetic_t* x) {
        expr_t* m_test = duplicate_expr(x->m_test);
        return make_IfArithmetic_t(al, x->base.base.loc, m_test, x->m_lt_label, x->m_eq_label, x->m_gt_label);
    }


    ASR::asr_t* duplicate_Print(Print_t* x) {
        expr_t* m_fmt = duplicate_expr(x->m_fmt);
        Vec<expr_t*> m_values;
        m_values.reserve(al, x->n_values);
        for (size_t i = 0; i < x->n_values; i++) {
            m_values.push_back(al, duplicate_expr(x->m_values[i]));
        }
        return make_Print_t(al, x->base.base.loc, m_fmt, m_values.p, x->n_values);
    }


    ASR::asr_t* duplicate_FileOpen(FileOpen_t* x) {
        expr_t* m_newunit = duplicate_expr(x->m_newunit);
        expr_t* m_filename = duplicate_expr(x->m_filename);
        expr_t* m_status = duplicate_expr(x->m_status);
        return make_FileOpen_t(al, x->base.base.loc, x->m_label, m_newunit, m_filename, m_status);
    }


    ASR::asr_t* duplicate_FileClose(FileClose_t* x) {
        expr_t* m_unit = duplicate_expr(x->m_unit);
        expr_t* m_iostat = duplicate_expr(x->m_iostat);
        expr_t* m_iomsg = duplicate_expr(x->m_iomsg);
        expr_t* m_err = duplicate_expr(x->m_err);
        expr_t* m_status = duplicate_expr(x->m_status);
        return make_FileClose_t(al, x->base.base.loc, x->m_label, m_unit, m_iostat, m_iomsg, m_err, m_status);
    }


    ASR::asr_t* duplicate_FileRead(FileRead_t* x) {
        expr_t* m_unit = duplicate_expr(x->m_unit);
        expr_t* m_fmt = duplicate_expr(x->m_fmt);
        expr_t* m_iomsg = duplicate_expr(x->m_iomsg);
        expr_t* m_iostat = duplicate_expr(x->m_iostat);
        expr_t* m_id = duplicate_expr(x->m_id);
        Vec<expr_t*> m_values;
        m_values.reserve(al, x->n_values);
        for (size_t i = 0; i < x->n_values; i++) {
            m_values.push_back(al, duplicate_expr(x->m_values[i]));
        }
        return make_FileRead_t(al, x->base.base.loc, x->m_label, m_unit, m_fmt, m_iomsg, m_iostat, m_id, m_values.p, x->n_values);
    }


    ASR::asr_t* duplicate_FileRewind(FileRewind_t* x) {
        expr_t* m_unit = duplicate_expr(x->m_unit);
        expr_t* m_iostat = duplicate_expr(x->m_iostat);
        expr_t* m_err = duplicate_expr(x->m_err);
        return make_FileRewind_t(al, x->base.base.loc, x->m_label, m_unit, m_iostat, m_err);
    }


    ASR::asr_t* duplicate_FileInquire(FileInquire_t* x) {
        expr_t* m_unit = duplicate_expr(x->m_unit);
        expr_t* m_file = duplicate_expr(x->m_file);
        expr_t* m_iostat = duplicate_expr(x->m_iostat);
        expr_t* m_err = duplicate_expr(x->m_err);
        expr_t* m_exist = duplicate_expr(x->m_exist);
        expr_t* m_opened = duplicate_expr(x->m_opened);
        expr_t* m_number = duplicate_expr(x->m_number);
        expr_t* m_named = duplicate_expr(x->m_named);
        expr_t* m_name = duplicate_expr(x->m_name);
        expr_t* m_access = duplicate_expr(x->m_access);
        expr_t* m_sequential = duplicate_expr(x->m_sequential);
        expr_t* m_direct = duplicate_expr(x->m_direct);
        expr_t* m_form = duplicate_expr(x->m_form);
        expr_t* m_formatted = duplicate_expr(x->m_formatted);
        expr_t* m_unformatted = duplicate_expr(x->m_unformatted);
        expr_t* m_recl = duplicate_expr(x->m_recl);
        expr_t* m_nextrec = duplicate_expr(x->m_nextrec);
        expr_t* m_blank = duplicate_expr(x->m_blank);
        expr_t* m_position = duplicate_expr(x->m_position);
        expr_t* m_action = duplicate_expr(x->m_action);
        expr_t* m_read = duplicate_expr(x->m_read);
        expr_t* m_write = duplicate_expr(x->m_write);
        expr_t* m_readwrite = duplicate_expr(x->m_readwrite);
        expr_t* m_delim = duplicate_expr(x->m_delim);
        expr_t* m_pad = duplicate_expr(x->m_pad);
        expr_t* m_flen = duplicate_expr(x->m_flen);
        expr_t* m_blocksize = duplicate_expr(x->m_blocksize);
        expr_t* m_convert = duplicate_expr(x->m_convert);
        expr_t* m_carriagecontrol = duplicate_expr(x->m_carriagecontrol);
        expr_t* m_iolength = duplicate_expr(x->m_iolength);
        return make_FileInquire_t(al, x->base.base.loc, x->m_label, m_unit, m_file, m_iostat, m_err, m_exist, m_opened, m_number, m_named, m_name, m_access, m_sequential, m_direct, m_form, m_formatted, m_unformatted, m_recl, m_nextrec, m_blank, m_position, m_action, m_read, m_write, m_readwrite, m_delim, m_pad, m_flen, m_blocksize, m_convert, m_carriagecontrol, m_iolength);
    }


    ASR::asr_t* duplicate_FileWrite(FileWrite_t* x) {
        expr_t* m_unit = duplicate_expr(x->m_unit);
        expr_t* m_fmt = duplicate_expr(x->m_fmt);
        expr_t* m_iomsg = duplicate_expr(x->m_iomsg);
        expr_t* m_iostat = duplicate_expr(x->m_iostat);
        expr_t* m_id = duplicate_expr(x->m_id);
        Vec<expr_t*> m_values;
        m_values.reserve(al, x->n_values);
        for (size_t i = 0; i < x->n_values; i++) {
            m_values.push_back(al, duplicate_expr(x->m_values[i]));
        }
        return make_FileWrite_t(al, x->base.base.loc, x->m_label, m_unit, m_fmt, m_iomsg, m_iostat, m_id, m_values.p, x->n_values);
    }


    ASR::asr_t* duplicate_Return(Return_t* x) {
        return (asr_t*)x;
    }


    ASR::asr_t* duplicate_Select(Select_t* x) {
        expr_t* m_test = duplicate_expr(x->m_test);
        Vec<stmt_t*> m_default;
        m_default.reserve(al, x->n_default);
        for (size_t i = 0; i < x->n_default; i++) {
            m_default.push_back(al, duplicate_stmt(x->m_default[i]));
        }
        return make_Select_t(al, x->base.base.loc, m_test, x->m_body, x->n_body, m_default.p, x->n_default);
    }


    ASR::asr_t* duplicate_Stop(Stop_t* x) {
        expr_t* m_code = duplicate_expr(x->m_code);
        return make_Stop_t(al, x->base.base.loc, m_code);
    }


    ASR::asr_t* duplicate_Assert(Assert_t* x) {
        expr_t* m_test = duplicate_expr(x->m_test);
        expr_t* m_msg = duplicate_expr(x->m_msg);
        return make_Assert_t(al, x->base.base.loc, m_test, m_msg);
    }


    ASR::asr_t* duplicate_SubroutineCall(SubroutineCall_t* x) {
        symbol_t* m_name = x->m_name;
        symbol_t* m_original_name = x->m_original_name;
        Vec<call_arg_t> m_args;
        m_args.reserve(al, x->n_args);
        for (size_t i = 0; i < x->n_args; i++) {
            ASR::call_arg_t call_arg_copy;
            call_arg_copy.loc = x->m_args[i].loc;
            call_arg_copy.m_value = duplicate_expr(x->m_args[i].m_value);
            m_args.push_back(al, call_arg_copy);
        }
        expr_t* m_dt = duplicate_expr(x->m_dt);
        return make_SubroutineCall_t(al, x->base.base.loc, m_name, m_original_name, m_args.p, x->n_args, m_dt);
    }


    ASR::asr_t* duplicate_Where(Where_t* x) {
        expr_t* m_test = duplicate_expr(x->m_test);
        Vec<stmt_t*> m_body;
        m_body.reserve(al, x->n_body);
        for (size_t i = 0; i < x->n_body; i++) {
            m_body.push_back(al, duplicate_stmt(x->m_body[i]));
        }
        Vec<stmt_t*> m_orelse;
        m_orelse.reserve(al, x->n_orelse);
        for (size_t i = 0; i < x->n_orelse; i++) {
            m_orelse.push_back(al, duplicate_stmt(x->m_orelse[i]));
        }
        return make_Where_t(al, x->base.base.loc, m_test, m_body.p, x->n_body, m_orelse.p, x->n_orelse);
    }


    ASR::asr_t* duplicate_WhileLoop(WhileLoop_t* x) {
        expr_t* m_test = duplicate_expr(x->m_test);
        Vec<stmt_t*> m_body;
        m_body.reserve(al, x->n_body);
        for (size_t i = 0; i < x->n_body; i++) {
            m_body.push_back(al, duplicate_stmt(x->m_body[i]));
        }
        return make_WhileLoop_t(al, x->base.base.loc, m_test, m_body.p, x->n_body);
    }


    ASR::asr_t* duplicate_Nullify(Nullify_t* x) {
        Vec<symbol_t*> m_vars;
        m_vars.reserve(al, x->n_vars);
        for (size_t i = 0; i < x->n_vars; i++) {
            m_vars.push_back(al, x->m_vars[i]);
        }
        return make_Nullify_t(al, x->base.base.loc, m_vars.p, x->n_vars);
    }


    ASR::asr_t* duplicate_Flush(Flush_t* x) {
        expr_t* m_unit = duplicate_expr(x->m_unit);
        expr_t* m_err = duplicate_expr(x->m_err);
        expr_t* m_iomsg = duplicate_expr(x->m_iomsg);
        expr_t* m_iostat = duplicate_expr(x->m_iostat);
        return make_Flush_t(al, x->base.base.loc, x->m_label, m_unit, m_err, m_iomsg, m_iostat);
    }


    ASR::asr_t* duplicate_ListAppend(ListAppend_t* x) {
        symbol_t* m_a = x->m_a;
        expr_t* m_ele = duplicate_expr(x->m_ele);
        return make_ListAppend_t(al, x->base.base.loc, m_a, m_ele);
    }


    ASR::asr_t* duplicate_AssociateBlockCall(AssociateBlockCall_t* x) {
        symbol_t* m_m = x->m_m;
        return make_AssociateBlockCall_t(al, x->base.base.loc, m_m);
    }


    ASR::asr_t* duplicate_BlockCall(BlockCall_t* x) {
        symbol_t* m_m = x->m_m;
        return make_BlockCall_t(al, x->base.base.loc, m_m);
    }


    ASR::asr_t* duplicate_SetInsert(SetInsert_t* x) {
        symbol_t* m_a = x->m_a;
        expr_t* m_ele = duplicate_expr(x->m_ele);
        return make_SetInsert_t(al, x->base.base.loc, m_a, m_ele);
    }


    ASR::asr_t* duplicate_SetRemove(SetRemove_t* x) {
        symbol_t* m_a = x->m_a;
        expr_t* m_ele = duplicate_expr(x->m_ele);
        return make_SetRemove_t(al, x->base.base.loc, m_a, m_ele);
    }


    ASR::asr_t* duplicate_ListInsert(ListInsert_t* x) {
        symbol_t* m_a = x->m_a;
        expr_t* m_pos = duplicate_expr(x->m_pos);
        expr_t* m_ele = duplicate_expr(x->m_ele);
        return make_ListInsert_t(al, x->base.base.loc, m_a, m_pos, m_ele);
    }


    ASR::asr_t* duplicate_ListRemove(ListRemove_t* x) {
        symbol_t* m_a = x->m_a;
        expr_t* m_ele = duplicate_expr(x->m_ele);
        return make_ListRemove_t(al, x->base.base.loc, m_a, m_ele);
    }


    ASR::asr_t* duplicate_DictInsert(DictInsert_t* x) {
        symbol_t* m_a = x->m_a;
        expr_t* m_key = duplicate_expr(x->m_key);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_DictInsert_t(al, x->base.base.loc, m_a, m_key, m_value);
    }


    ASR::asr_t* duplicate_BoolOp(BoolOp_t* x) {
        expr_t* m_left = duplicate_expr(x->m_left);
        expr_t* m_right = duplicate_expr(x->m_right);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_BoolOp_t(al, x->base.base.loc, m_left, x->m_op, m_right, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_BinOp(BinOp_t* x) {
        expr_t* m_left = duplicate_expr(x->m_left);
        expr_t* m_right = duplicate_expr(x->m_right);
        expr_t* m_value = duplicate_expr(x->m_value);
        expr_t* m_overloaded = duplicate_expr(x->m_overloaded);
        return make_BinOp_t(al, x->base.base.loc, m_left, x->m_op, m_right, x->m_type, m_value, m_overloaded);
    }


    ASR::asr_t* duplicate_StrOp(StrOp_t* x) {
        expr_t* m_left = duplicate_expr(x->m_left);
        expr_t* m_right = duplicate_expr(x->m_right);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_StrOp_t(al, x->base.base.loc, m_left, x->m_op, m_right, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_UnaryOp(UnaryOp_t* x) {
        expr_t* m_operand = duplicate_expr(x->m_operand);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_UnaryOp_t(al, x->base.base.loc, x->m_op, m_operand, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_ComplexConstructor(ComplexConstructor_t* x) {
        expr_t* m_re = duplicate_expr(x->m_re);
        expr_t* m_im = duplicate_expr(x->m_im);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_ComplexConstructor_t(al, x->base.base.loc, m_re, m_im, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_NamedExpr(NamedExpr_t* x) {
        expr_t* m_target = duplicate_expr(x->m_target);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_NamedExpr_t(al, x->base.base.loc, m_target, m_value, x->m_type);
    }


    ASR::asr_t* duplicate_Compare(Compare_t* x) {
        expr_t* m_left = duplicate_expr(x->m_left);
        expr_t* m_right = duplicate_expr(x->m_right);
        expr_t* m_value = duplicate_expr(x->m_value);
        expr_t* m_overloaded = duplicate_expr(x->m_overloaded);
        return make_Compare_t(al, x->base.base.loc, m_left, x->m_op, m_right, x->m_type, m_value, m_overloaded);
    }


    ASR::asr_t* duplicate_IfExp(IfExp_t* x) {
        expr_t* m_test = duplicate_expr(x->m_test);
        expr_t* m_body = duplicate_expr(x->m_body);
        expr_t* m_orelse = duplicate_expr(x->m_orelse);
        return make_IfExp_t(al, x->base.base.loc, m_test, m_body, m_orelse, x->m_type);
    }


    ASR::asr_t* duplicate_FunctionCall(FunctionCall_t* x) {
        symbol_t* m_name = x->m_name;
        symbol_t* m_original_name = x->m_original_name;
        Vec<call_arg_t> m_args;
        m_args.reserve(al, x->n_args);
        for (size_t i = 0; i < x->n_args; i++) {
            ASR::call_arg_t call_arg_copy;
            call_arg_copy.loc = x->m_args[i].loc;
            call_arg_copy.m_value = duplicate_expr(x->m_args[i].m_value);
            m_args.push_back(al, call_arg_copy);
        }
        expr_t* m_value = duplicate_expr(x->m_value);
        expr_t* m_dt = duplicate_expr(x->m_dt);
        return make_FunctionCall_t(al, x->base.base.loc, m_name, m_original_name, m_args.p, x->n_args, x->m_type, m_value, m_dt);
    }


    ASR::asr_t* duplicate_DerivedTypeConstructor(DerivedTypeConstructor_t* x) {
        symbol_t* m_dt_sym = x->m_dt_sym;
        Vec<expr_t*> m_args;
        m_args.reserve(al, x->n_args);
        for (size_t i = 0; i < x->n_args; i++) {
            m_args.push_back(al, duplicate_expr(x->m_args[i]));
        }
        return make_DerivedTypeConstructor_t(al, x->base.base.loc, m_dt_sym, m_args.p, x->n_args, x->m_type);
    }


    ASR::asr_t* duplicate_ImpliedDoLoop(ImpliedDoLoop_t* x) {
        Vec<expr_t*> m_values;
        m_values.reserve(al, x->n_values);
        for (size_t i = 0; i < x->n_values; i++) {
            m_values.push_back(al, duplicate_expr(x->m_values[i]));
        }
        expr_t* m_var = duplicate_expr(x->m_var);
        expr_t* m_start = duplicate_expr(x->m_start);
        expr_t* m_end = duplicate_expr(x->m_end);
        expr_t* m_increment = duplicate_expr(x->m_increment);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_ImpliedDoLoop_t(al, x->base.base.loc, m_values.p, x->n_values, m_var, m_start, m_end, m_increment, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_IntegerConstant(IntegerConstant_t* x) {
        return (asr_t*)x;
    }


    ASR::asr_t* duplicate_IntegerBOZ(IntegerBOZ_t* x) {
        return (asr_t*)x;
    }


    ASR::asr_t* duplicate_RealConstant(RealConstant_t* x) {
        return (asr_t*)x;
    }


    ASR::asr_t* duplicate_ComplexConstant(ComplexConstant_t* x) {
        return (asr_t*)x;
    }


    ASR::asr_t* duplicate_LogicalConstant(LogicalConstant_t* x) {
        return (asr_t*)x;
    }


    ASR::asr_t* duplicate_ListConstant(ListConstant_t* x) {
        Vec<expr_t*> m_args;
        m_args.reserve(al, x->n_args);
        for (size_t i = 0; i < x->n_args; i++) {
            m_args.push_back(al, duplicate_expr(x->m_args[i]));
        }
        return make_ListConstant_t(al, x->base.base.loc, m_args.p, x->n_args, x->m_type);
    }


    ASR::asr_t* duplicate_ListLen(ListLen_t* x) {
        expr_t* m_arg = duplicate_expr(x->m_arg);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_ListLen_t(al, x->base.base.loc, m_arg, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_ListConcat(ListConcat_t* x) {
        expr_t* m_left = duplicate_expr(x->m_left);
        expr_t* m_right = duplicate_expr(x->m_right);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_ListConcat_t(al, x->base.base.loc, m_left, m_right, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_ArrayConstant(ArrayConstant_t* x) {
        Vec<expr_t*> m_args;
        m_args.reserve(al, x->n_args);
        for (size_t i = 0; i < x->n_args; i++) {
            m_args.push_back(al, duplicate_expr(x->m_args[i]));
        }
        return make_ArrayConstant_t(al, x->base.base.loc, m_args.p, x->n_args, x->m_type);
    }


    ASR::asr_t* duplicate_SetConstant(SetConstant_t* x) {
        Vec<expr_t*> m_elements;
        m_elements.reserve(al, x->n_elements);
        for (size_t i = 0; i < x->n_elements; i++) {
            m_elements.push_back(al, duplicate_expr(x->m_elements[i]));
        }
        return make_SetConstant_t(al, x->base.base.loc, m_elements.p, x->n_elements, x->m_type);
    }


    ASR::asr_t* duplicate_SetLen(SetLen_t* x) {
        expr_t* m_arg = duplicate_expr(x->m_arg);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_SetLen_t(al, x->base.base.loc, m_arg, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_TupleConstant(TupleConstant_t* x) {
        Vec<expr_t*> m_elements;
        m_elements.reserve(al, x->n_elements);
        for (size_t i = 0; i < x->n_elements; i++) {
            m_elements.push_back(al, duplicate_expr(x->m_elements[i]));
        }
        return make_TupleConstant_t(al, x->base.base.loc, m_elements.p, x->n_elements, x->m_type);
    }


    ASR::asr_t* duplicate_TupleLen(TupleLen_t* x) {
        expr_t* m_arg = duplicate_expr(x->m_arg);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_TupleLen_t(al, x->base.base.loc, m_arg, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_StringConstant(StringConstant_t* x) {
        return (asr_t*)x;
    }


    ASR::asr_t* duplicate_StringConcat(StringConcat_t* x) {
        expr_t* m_left = duplicate_expr(x->m_left);
        expr_t* m_right = duplicate_expr(x->m_right);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_StringConcat_t(al, x->base.base.loc, m_left, m_right, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_StringLen(StringLen_t* x) {
        expr_t* m_arg = duplicate_expr(x->m_arg);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_StringLen_t(al, x->base.base.loc, m_arg, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_DictConstant(DictConstant_t* x) {
        Vec<expr_t*> m_keys;
        m_keys.reserve(al, x->n_keys);
        for (size_t i = 0; i < x->n_keys; i++) {
            m_keys.push_back(al, duplicate_expr(x->m_keys[i]));
        }
        Vec<expr_t*> m_values;
        m_values.reserve(al, x->n_values);
        for (size_t i = 0; i < x->n_values; i++) {
            m_values.push_back(al, duplicate_expr(x->m_values[i]));
        }
        return make_DictConstant_t(al, x->base.base.loc, m_keys.p, x->n_keys, m_values.p, x->n_values, x->m_type);
    }


    ASR::asr_t* duplicate_DictLen(DictLen_t* x) {
        expr_t* m_arg = duplicate_expr(x->m_arg);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_DictLen_t(al, x->base.base.loc, m_arg, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_Var(Var_t* x) {
        symbol_t* m_v = x->m_v;
        return make_Var_t(al, x->base.base.loc, m_v);
    }


    ASR::asr_t* duplicate_ArrayRef(ArrayRef_t* x) {
        symbol_t* m_v = x->m_v;
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_ArrayRef_t(al, x->base.base.loc, m_v, x->m_args, x->n_args, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_ArraySize(ArraySize_t* x) {
        expr_t* m_v = duplicate_expr(x->m_v);
        expr_t* m_dim = duplicate_expr(x->m_dim);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_ArraySize_t(al, x->base.base.loc, m_v, m_dim, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_ArrayBound(ArrayBound_t* x) {
        expr_t* m_v = duplicate_expr(x->m_v);
        expr_t* m_dim = duplicate_expr(x->m_dim);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_ArrayBound_t(al, x->base.base.loc, m_v, m_dim, x->m_type, x->m_bound, m_value);
    }


    ASR::asr_t* duplicate_ArrayTranspose(ArrayTranspose_t* x) {
        expr_t* m_matrix = duplicate_expr(x->m_matrix);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_ArrayTranspose_t(al, x->base.base.loc, m_matrix, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_ArrayMatMul(ArrayMatMul_t* x) {
        expr_t* m_matrix_a = duplicate_expr(x->m_matrix_a);
        expr_t* m_matrix_b = duplicate_expr(x->m_matrix_b);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_ArrayMatMul_t(al, x->base.base.loc, m_matrix_a, m_matrix_b, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_ArrayPack(ArrayPack_t* x) {
        expr_t* m_array = duplicate_expr(x->m_array);
        expr_t* m_mask = duplicate_expr(x->m_mask);
        expr_t* m_vector = duplicate_expr(x->m_vector);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_ArrayPack_t(al, x->base.base.loc, m_array, m_mask, m_vector, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_Transfer(Transfer_t* x) {
        expr_t* m_source = duplicate_expr(x->m_source);
        expr_t* m_mold = duplicate_expr(x->m_mold);
        expr_t* m_size = duplicate_expr(x->m_size);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_Transfer_t(al, x->base.base.loc, m_source, m_mold, m_size, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_DerivedRef(DerivedRef_t* x) {
        expr_t* m_v = duplicate_expr(x->m_v);
        symbol_t* m_m = x->m_m;
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_DerivedRef_t(al, x->base.base.loc, m_v, m_m, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_Cast(Cast_t* x) {
        expr_t* m_arg = duplicate_expr(x->m_arg);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_Cast_t(al, x->base.base.loc, m_arg, x->m_kind, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_ComplexRe(ComplexRe_t* x) {
        expr_t* m_arg = duplicate_expr(x->m_arg);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_ComplexRe_t(al, x->base.base.loc, m_arg, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_ComplexIm(ComplexIm_t* x) {
        expr_t* m_arg = duplicate_expr(x->m_arg);
        expr_t* m_value = duplicate_expr(x->m_value);
        return make_ComplexIm_t(al, x->base.base.loc, m_arg, x->m_type, m_value);
    }


    ASR::asr_t* duplicate_DictItem(DictItem_t* x) {
        symbol_t* m_a = x->m_a;
        expr_t* m_key = duplicate_expr(x->m_key);
        expr_t* m_default = duplicate_expr(x->m_default);
        return make_DictItem_t(al, x->base.base.loc, m_a, m_key, m_default, x->m_type);
    }

    ASR::stmt_t* duplicate_stmt(ASR::stmt_t* x) {
        if( !x ) {
            return nullptr;
        }

        switch(x->type) {
            case ASR::stmtType::Allocate: {
                return down_cast<ASR::stmt_t>(duplicate_Allocate(down_cast<ASR::Allocate_t>(x)));
            }
            case ASR::stmtType::Assign: {
                return down_cast<ASR::stmt_t>(duplicate_Assign(down_cast<ASR::Assign_t>(x)));
            }
            case ASR::stmtType::Assignment: {
                return down_cast<ASR::stmt_t>(duplicate_Assignment(down_cast<ASR::Assignment_t>(x)));
            }
            case ASR::stmtType::Associate: {
                return down_cast<ASR::stmt_t>(duplicate_Associate(down_cast<ASR::Associate_t>(x)));
            }
            case ASR::stmtType::Cycle: {
                return down_cast<ASR::stmt_t>(duplicate_Cycle(down_cast<ASR::Cycle_t>(x)));
            }
            case ASR::stmtType::ExplicitDeallocate: {
                return down_cast<ASR::stmt_t>(duplicate_ExplicitDeallocate(down_cast<ASR::ExplicitDeallocate_t>(x)));
            }
            case ASR::stmtType::ImplicitDeallocate: {
                return down_cast<ASR::stmt_t>(duplicate_ImplicitDeallocate(down_cast<ASR::ImplicitDeallocate_t>(x)));
            }
            case ASR::stmtType::DoConcurrentLoop: {
                return down_cast<ASR::stmt_t>(duplicate_DoConcurrentLoop(down_cast<ASR::DoConcurrentLoop_t>(x)));
            }
            case ASR::stmtType::DoLoop: {
                return down_cast<ASR::stmt_t>(duplicate_DoLoop(down_cast<ASR::DoLoop_t>(x)));
            }
            case ASR::stmtType::ErrorStop: {
                return down_cast<ASR::stmt_t>(duplicate_ErrorStop(down_cast<ASR::ErrorStop_t>(x)));
            }
            case ASR::stmtType::Exit: {
                return down_cast<ASR::stmt_t>(duplicate_Exit(down_cast<ASR::Exit_t>(x)));
            }
            case ASR::stmtType::ForAllSingle: {
                return down_cast<ASR::stmt_t>(duplicate_ForAllSingle(down_cast<ASR::ForAllSingle_t>(x)));
            }
            case ASR::stmtType::GoTo: {
                return down_cast<ASR::stmt_t>(duplicate_GoTo(down_cast<ASR::GoTo_t>(x)));
            }
            case ASR::stmtType::GoToTarget: {
                return down_cast<ASR::stmt_t>(duplicate_GoToTarget(down_cast<ASR::GoToTarget_t>(x)));
            }
            case ASR::stmtType::If: {
                return down_cast<ASR::stmt_t>(duplicate_If(down_cast<ASR::If_t>(x)));
            }
            case ASR::stmtType::IfArithmetic: {
                return down_cast<ASR::stmt_t>(duplicate_IfArithmetic(down_cast<ASR::IfArithmetic_t>(x)));
            }
            case ASR::stmtType::Print: {
                return down_cast<ASR::stmt_t>(duplicate_Print(down_cast<ASR::Print_t>(x)));
            }
            case ASR::stmtType::FileOpen: {
                return down_cast<ASR::stmt_t>(duplicate_FileOpen(down_cast<ASR::FileOpen_t>(x)));
            }
            case ASR::stmtType::FileClose: {
                return down_cast<ASR::stmt_t>(duplicate_FileClose(down_cast<ASR::FileClose_t>(x)));
            }
            case ASR::stmtType::FileRead: {
                return down_cast<ASR::stmt_t>(duplicate_FileRead(down_cast<ASR::FileRead_t>(x)));
            }
            case ASR::stmtType::FileRewind: {
                return down_cast<ASR::stmt_t>(duplicate_FileRewind(down_cast<ASR::FileRewind_t>(x)));
            }
            case ASR::stmtType::FileInquire: {
                return down_cast<ASR::stmt_t>(duplicate_FileInquire(down_cast<ASR::FileInquire_t>(x)));
            }
            case ASR::stmtType::FileWrite: {
                return down_cast<ASR::stmt_t>(duplicate_FileWrite(down_cast<ASR::FileWrite_t>(x)));
            }
            case ASR::stmtType::Return: {
                return down_cast<ASR::stmt_t>(duplicate_Return(down_cast<ASR::Return_t>(x)));
            }
            case ASR::stmtType::Select: {
                return down_cast<ASR::stmt_t>(duplicate_Select(down_cast<ASR::Select_t>(x)));
            }
            case ASR::stmtType::Stop: {
                return down_cast<ASR::stmt_t>(duplicate_Stop(down_cast<ASR::Stop_t>(x)));
            }
            case ASR::stmtType::Assert: {
                return down_cast<ASR::stmt_t>(duplicate_Assert(down_cast<ASR::Assert_t>(x)));
            }
            case ASR::stmtType::SubroutineCall: {
                if( !allow_procedure_calls ) {
                    success = false;
                    return nullptr;
                }
                return down_cast<ASR::stmt_t>(duplicate_SubroutineCall(down_cast<ASR::SubroutineCall_t>(x)));
            }
            case ASR::stmtType::Where: {
                return down_cast<ASR::stmt_t>(duplicate_Where(down_cast<ASR::Where_t>(x)));
            }
            case ASR::stmtType::WhileLoop: {
                return down_cast<ASR::stmt_t>(duplicate_WhileLoop(down_cast<ASR::WhileLoop_t>(x)));
            }
            case ASR::stmtType::Nullify: {
                return down_cast<ASR::stmt_t>(duplicate_Nullify(down_cast<ASR::Nullify_t>(x)));
            }
            case ASR::stmtType::Flush: {
                return down_cast<ASR::stmt_t>(duplicate_Flush(down_cast<ASR::Flush_t>(x)));
            }
            case ASR::stmtType::ListAppend: {
                return down_cast<ASR::stmt_t>(duplicate_ListAppend(down_cast<ASR::ListAppend_t>(x)));
            }
            case ASR::stmtType::AssociateBlockCall: {
                return down_cast<ASR::stmt_t>(duplicate_AssociateBlockCall(down_cast<ASR::AssociateBlockCall_t>(x)));
            }
            case ASR::stmtType::BlockCall: {
                return down_cast<ASR::stmt_t>(duplicate_BlockCall(down_cast<ASR::BlockCall_t>(x)));
            }
            case ASR::stmtType::SetInsert: {
                return down_cast<ASR::stmt_t>(duplicate_SetInsert(down_cast<ASR::SetInsert_t>(x)));
            }
            case ASR::stmtType::SetRemove: {
                return down_cast<ASR::stmt_t>(duplicate_SetRemove(down_cast<ASR::SetRemove_t>(x)));
            }
            case ASR::stmtType::ListInsert: {
                return down_cast<ASR::stmt_t>(duplicate_ListInsert(down_cast<ASR::ListInsert_t>(x)));
            }
            case ASR::stmtType::ListRemove: {
                return down_cast<ASR::stmt_t>(duplicate_ListRemove(down_cast<ASR::ListRemove_t>(x)));
            }
            case ASR::stmtType::DictInsert: {
                return down_cast<ASR::stmt_t>(duplicate_DictInsert(down_cast<ASR::DictInsert_t>(x)));
            }
            default: {
                LFORTRAN_ASSERT_MSG(false, "Duplication of " + std::to_string(x->type) + " statement is not supported yet.");
            }
        }

        return nullptr;
    }

    ASR::expr_t* duplicate_expr(ASR::expr_t* x) {
        if( !x ) {
            return nullptr;
        }

        switch(x->type) {
            case ASR::exprType::BoolOp: {
                return down_cast<ASR::expr_t>(duplicate_BoolOp(down_cast<ASR::BoolOp_t>(x)));
            }
            case ASR::exprType::BinOp: {
                return down_cast<ASR::expr_t>(duplicate_BinOp(down_cast<ASR::BinOp_t>(x)));
            }
            case ASR::exprType::StrOp: {
                return down_cast<ASR::expr_t>(duplicate_StrOp(down_cast<ASR::StrOp_t>(x)));
            }
            case ASR::exprType::UnaryOp: {
                return down_cast<ASR::expr_t>(duplicate_UnaryOp(down_cast<ASR::UnaryOp_t>(x)));
            }
            case ASR::exprType::ComplexConstructor: {
                return down_cast<ASR::expr_t>(duplicate_ComplexConstructor(down_cast<ASR::ComplexConstructor_t>(x)));
            }
            case ASR::exprType::NamedExpr: {
                return down_cast<ASR::expr_t>(duplicate_NamedExpr(down_cast<ASR::NamedExpr_t>(x)));
            }
            case ASR::exprType::Compare: {
                return down_cast<ASR::expr_t>(duplicate_Compare(down_cast<ASR::Compare_t>(x)));
            }
            case ASR::exprType::IfExp: {
                return down_cast<ASR::expr_t>(duplicate_IfExp(down_cast<ASR::IfExp_t>(x)));
            }
            case ASR::exprType::FunctionCall: {
                if( !allow_procedure_calls ) {
                    success = false;
                    return nullptr;
                }
                return down_cast<ASR::expr_t>(duplicate_FunctionCall(down_cast<ASR::FunctionCall_t>(x)));
            }
            case ASR::exprType::DerivedTypeConstructor: {
                return down_cast<ASR::expr_t>(duplicate_DerivedTypeConstructor(down_cast<ASR::DerivedTypeConstructor_t>(x)));
            }
            case ASR::exprType::ImpliedDoLoop: {
                return down_cast<ASR::expr_t>(duplicate_ImpliedDoLoop(down_cast<ASR::ImpliedDoLoop_t>(x)));
            }
            case ASR::exprType::IntegerConstant: {
                return down_cast<ASR::expr_t>(duplicate_IntegerConstant(down_cast<ASR::IntegerConstant_t>(x)));
            }
            case ASR::exprType::IntegerBOZ: {
                return down_cast<ASR::expr_t>(duplicate_IntegerBOZ(down_cast<ASR::IntegerBOZ_t>(x)));
            }
            case ASR::exprType::RealConstant: {
                return down_cast<ASR::expr_t>(duplicate_RealConstant(down_cast<ASR::RealConstant_t>(x)));
            }
            case ASR::exprType::ComplexConstant: {
                return down_cast<ASR::expr_t>(duplicate_ComplexConstant(down_cast<ASR::ComplexConstant_t>(x)));
            }
            case ASR::exprType::LogicalConstant: {
                return down_cast<ASR::expr_t>(duplicate_LogicalConstant(down_cast<ASR::LogicalConstant_t>(x)));
            }
            case ASR::exprType::ListConstant: {
                return down_cast<ASR::expr_t>(duplicate_ListConstant(down_cast<ASR::ListConstant_t>(x)));
            }
            case ASR::exprType::ListLen: {
                return down_cast<ASR::expr_t>(duplicate_ListLen(down_cast<ASR::ListLen_t>(x)));
            }
            case ASR::exprType::ListConcat: {
                return down_cast<ASR::expr_t>(duplicate_ListConcat(down_cast<ASR::ListConcat_t>(x)));
            }
            case ASR::exprType::ArrayConstant: {
                return down_cast<ASR::expr_t>(duplicate_ArrayConstant(down_cast<ASR::ArrayConstant_t>(x)));
            }
            case ASR::exprType::SetConstant: {
                return down_cast<ASR::expr_t>(duplicate_SetConstant(down_cast<ASR::SetConstant_t>(x)));
            }
            case ASR::exprType::SetLen: {
                return down_cast<ASR::expr_t>(duplicate_SetLen(down_cast<ASR::SetLen_t>(x)));
            }
            case ASR::exprType::TupleConstant: {
                return down_cast<ASR::expr_t>(duplicate_TupleConstant(down_cast<ASR::TupleConstant_t>(x)));
            }
            case ASR::exprType::TupleLen: {
                return down_cast<ASR::expr_t>(duplicate_TupleLen(down_cast<ASR::TupleLen_t>(x)));
            }
            case ASR::exprType::StringConstant: {
                return down_cast<ASR::expr_t>(duplicate_StringConstant(down_cast<ASR::StringConstant_t>(x)));
            }
            case ASR::exprType::StringConcat: {
                return down_cast<ASR::expr_t>(duplicate_StringConcat(down_cast<ASR::StringConcat_t>(x)));
            }
            case ASR::exprType::StringLen: {
                return down_cast<ASR::expr_t>(duplicate_StringLen(down_cast<ASR::StringLen_t>(x)));
            }
            case ASR::exprType::DictConstant: {
                return down_cast<ASR::expr_t>(duplicate_DictConstant(down_cast<ASR::DictConstant_t>(x)));
            }
            case ASR::exprType::DictLen: {
                return down_cast<ASR::expr_t>(duplicate_DictLen(down_cast<ASR::DictLen_t>(x)));
            }
            case ASR::exprType::Var: {
                return down_cast<ASR::expr_t>(duplicate_Var(down_cast<ASR::Var_t>(x)));
            }
            case ASR::exprType::ArrayRef: {
                return down_cast<ASR::expr_t>(duplicate_ArrayRef(down_cast<ASR::ArrayRef_t>(x)));
            }
            case ASR::exprType::ArraySize: {
                return down_cast<ASR::expr_t>(duplicate_ArraySize(down_cast<ASR::ArraySize_t>(x)));
            }
            case ASR::exprType::ArrayBound: {
                return down_cast<ASR::expr_t>(duplicate_ArrayBound(down_cast<ASR::ArrayBound_t>(x)));
            }
            case ASR::exprType::ArrayTranspose: {
                return down_cast<ASR::expr_t>(duplicate_ArrayTranspose(down_cast<ASR::ArrayTranspose_t>(x)));
            }
            case ASR::exprType::ArrayMatMul: {
                return down_cast<ASR::expr_t>(duplicate_ArrayMatMul(down_cast<ASR::ArrayMatMul_t>(x)));
            }
            case ASR::exprType::ArrayPack: {
                return down_cast<ASR::expr_t>(duplicate_ArrayPack(down_cast<ASR::ArrayPack_t>(x)));
            }
            case ASR::exprType::Transfer: {
                return down_cast<ASR::expr_t>(duplicate_Transfer(down_cast<ASR::Transfer_t>(x)));
            }
            case ASR::exprType::DerivedRef: {
                return down_cast<ASR::expr_t>(duplicate_DerivedRef(down_cast<ASR::DerivedRef_t>(x)));
            }
            case ASR::exprType::Cast: {
                return down_cast<ASR::expr_t>(duplicate_Cast(down_cast<ASR::Cast_t>(x)));
            }
            case ASR::exprType::ComplexRe: {
                return down_cast<ASR::expr_t>(duplicate_ComplexRe(down_cast<ASR::ComplexRe_t>(x)));
            }
            case ASR::exprType::ComplexIm: {
                return down_cast<ASR::expr_t>(duplicate_ComplexIm(down_cast<ASR::ComplexIm_t>(x)));
            }
            case ASR::exprType::DictItem: {
                return down_cast<ASR::expr_t>(duplicate_DictItem(down_cast<ASR::DictItem_t>(x)));
            }
            default: {
                LFORTRAN_ASSERT_MSG(false, "Duplication of " + std::to_string(x->type) + " expression is not supported yet.");
            }
        }

        return nullptr;
    }

};


/******************************************************************************/
// Expression Replacer Base class

template <class Derived>
class BaseExprReplacer {
public:
    Derived& self() { return static_cast<Derived&>(*this); }

    ASR::expr_t** current_expr;
    ASR::expr_t** current_expr_copy;

    BaseExprReplacer() : current_expr(nullptr) {}


    void replace_BoolOp(BoolOp_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_left);
        self().replace_expr(x->m_left);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_right);
        self().replace_expr(x->m_right);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_BinOp(BinOp_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_left);
        self().replace_expr(x->m_left);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_right);
        self().replace_expr(x->m_right);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_overloaded);
        self().replace_expr(x->m_overloaded);
        current_expr = current_expr_copy;
    }


    void replace_StrOp(StrOp_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_left);
        self().replace_expr(x->m_left);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_right);
        self().replace_expr(x->m_right);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_UnaryOp(UnaryOp_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_operand);
        self().replace_expr(x->m_operand);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_ComplexConstructor(ComplexConstructor_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_re);
        self().replace_expr(x->m_re);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_im);
        self().replace_expr(x->m_im);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_NamedExpr(NamedExpr_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_target);
        self().replace_expr(x->m_target);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_Compare(Compare_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_left);
        self().replace_expr(x->m_left);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_right);
        self().replace_expr(x->m_right);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_overloaded);
        self().replace_expr(x->m_overloaded);
        current_expr = current_expr_copy;
    }


    void replace_IfExp(IfExp_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_test);
        self().replace_expr(x->m_test);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_body);
        self().replace_expr(x->m_body);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_orelse);
        self().replace_expr(x->m_orelse);
        current_expr = current_expr_copy;
    }


    void replace_FunctionCall(FunctionCall_t* x) {
        for (size_t i = 0; i < x->n_args; i++) {
            current_expr_copy = current_expr;
            current_expr = &(x->m_args[i].m_value);
            self().replace_expr(x->m_args[i].m_value);
            current_expr = current_expr_copy;
        }
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_dt);
        self().replace_expr(x->m_dt);
        current_expr = current_expr_copy;
    }


    void replace_DerivedTypeConstructor(DerivedTypeConstructor_t* x) {
        for (size_t i = 0; i < x->n_args; i++) {
        }
    }


    void replace_ImpliedDoLoop(ImpliedDoLoop_t* x) {
        for (size_t i = 0; i < x->n_values; i++) {
        }
        current_expr_copy = current_expr;
        current_expr = &(x->m_var);
        self().replace_expr(x->m_var);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_start);
        self().replace_expr(x->m_start);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_end);
        self().replace_expr(x->m_end);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_increment);
        self().replace_expr(x->m_increment);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_IntegerConstant(IntegerConstant_t* x) {
        if (x) { }
    }


    void replace_IntegerBOZ(IntegerBOZ_t* x) {
        if (x) { }
    }


    void replace_RealConstant(RealConstant_t* x) {
        if (x) { }
    }


    void replace_ComplexConstant(ComplexConstant_t* x) {
        if (x) { }
    }


    void replace_LogicalConstant(LogicalConstant_t* x) {
        if (x) { }
    }


    void replace_ListConstant(ListConstant_t* x) {
        for (size_t i = 0; i < x->n_args; i++) {
        }
    }


    void replace_ListLen(ListLen_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_arg);
        self().replace_expr(x->m_arg);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_ListConcat(ListConcat_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_left);
        self().replace_expr(x->m_left);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_right);
        self().replace_expr(x->m_right);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_ArrayConstant(ArrayConstant_t* x) {
        for (size_t i = 0; i < x->n_args; i++) {
        }
    }


    void replace_SetConstant(SetConstant_t* x) {
        for (size_t i = 0; i < x->n_elements; i++) {
        }
    }


    void replace_SetLen(SetLen_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_arg);
        self().replace_expr(x->m_arg);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_TupleConstant(TupleConstant_t* x) {
        for (size_t i = 0; i < x->n_elements; i++) {
        }
    }


    void replace_TupleLen(TupleLen_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_arg);
        self().replace_expr(x->m_arg);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_StringConstant(StringConstant_t* x) {
        if (x) { }
    }


    void replace_StringConcat(StringConcat_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_left);
        self().replace_expr(x->m_left);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_right);
        self().replace_expr(x->m_right);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_StringLen(StringLen_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_arg);
        self().replace_expr(x->m_arg);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_DictConstant(DictConstant_t* x) {
        for (size_t i = 0; i < x->n_keys; i++) {
        }
        for (size_t i = 0; i < x->n_values; i++) {
        }
    }


    void replace_DictLen(DictLen_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_arg);
        self().replace_expr(x->m_arg);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_Var(Var_t* x) {
        if (x) { }
    }


    void replace_ArrayRef(ArrayRef_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_ArraySize(ArraySize_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_v);
        self().replace_expr(x->m_v);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_dim);
        self().replace_expr(x->m_dim);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_ArrayBound(ArrayBound_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_v);
        self().replace_expr(x->m_v);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_dim);
        self().replace_expr(x->m_dim);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_ArrayTranspose(ArrayTranspose_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_matrix);
        self().replace_expr(x->m_matrix);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_ArrayMatMul(ArrayMatMul_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_matrix_a);
        self().replace_expr(x->m_matrix_a);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_matrix_b);
        self().replace_expr(x->m_matrix_b);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_ArrayPack(ArrayPack_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_array);
        self().replace_expr(x->m_array);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_mask);
        self().replace_expr(x->m_mask);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_vector);
        self().replace_expr(x->m_vector);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_Transfer(Transfer_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_source);
        self().replace_expr(x->m_source);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_mold);
        self().replace_expr(x->m_mold);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_size);
        self().replace_expr(x->m_size);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_DerivedRef(DerivedRef_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_v);
        self().replace_expr(x->m_v);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_Cast(Cast_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_arg);
        self().replace_expr(x->m_arg);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_ComplexRe(ComplexRe_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_arg);
        self().replace_expr(x->m_arg);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_ComplexIm(ComplexIm_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_arg);
        self().replace_expr(x->m_arg);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_value);
        self().replace_expr(x->m_value);
        current_expr = current_expr_copy;
    }


    void replace_DictItem(DictItem_t* x) {
        current_expr_copy = current_expr;
        current_expr = &(x->m_key);
        self().replace_expr(x->m_key);
        current_expr = current_expr_copy;
        current_expr_copy = current_expr;
        current_expr = &(x->m_default);
        self().replace_expr(x->m_default);
        current_expr = current_expr_copy;
    }

    void replace_expr(ASR::expr_t* x) {
        if( !x ) {
            return ;
        }

        switch(x->type) {
            case ASR::exprType::BoolOp: {
                self().replace_BoolOp(down_cast<ASR::BoolOp_t>(x));
                break;
            }
            case ASR::exprType::BinOp: {
                self().replace_BinOp(down_cast<ASR::BinOp_t>(x));
                break;
            }
            case ASR::exprType::StrOp: {
                self().replace_StrOp(down_cast<ASR::StrOp_t>(x));
                break;
            }
            case ASR::exprType::UnaryOp: {
                self().replace_UnaryOp(down_cast<ASR::UnaryOp_t>(x));
                break;
            }
            case ASR::exprType::ComplexConstructor: {
                self().replace_ComplexConstructor(down_cast<ASR::ComplexConstructor_t>(x));
                break;
            }
            case ASR::exprType::NamedExpr: {
                self().replace_NamedExpr(down_cast<ASR::NamedExpr_t>(x));
                break;
            }
            case ASR::exprType::Compare: {
                self().replace_Compare(down_cast<ASR::Compare_t>(x));
                break;
            }
            case ASR::exprType::IfExp: {
                self().replace_IfExp(down_cast<ASR::IfExp_t>(x));
                break;
            }
            case ASR::exprType::FunctionCall: {
                self().replace_FunctionCall(down_cast<ASR::FunctionCall_t>(x));
                break;
            }
            case ASR::exprType::DerivedTypeConstructor: {
                self().replace_DerivedTypeConstructor(down_cast<ASR::DerivedTypeConstructor_t>(x));
                break;
            }
            case ASR::exprType::ImpliedDoLoop: {
                self().replace_ImpliedDoLoop(down_cast<ASR::ImpliedDoLoop_t>(x));
                break;
            }
            case ASR::exprType::IntegerConstant: {
                self().replace_IntegerConstant(down_cast<ASR::IntegerConstant_t>(x));
                break;
            }
            case ASR::exprType::IntegerBOZ: {
                self().replace_IntegerBOZ(down_cast<ASR::IntegerBOZ_t>(x));
                break;
            }
            case ASR::exprType::RealConstant: {
                self().replace_RealConstant(down_cast<ASR::RealConstant_t>(x));
                break;
            }
            case ASR::exprType::ComplexConstant: {
                self().replace_ComplexConstant(down_cast<ASR::ComplexConstant_t>(x));
                break;
            }
            case ASR::exprType::LogicalConstant: {
                self().replace_LogicalConstant(down_cast<ASR::LogicalConstant_t>(x));
                break;
            }
            case ASR::exprType::ListConstant: {
                self().replace_ListConstant(down_cast<ASR::ListConstant_t>(x));
                break;
            }
            case ASR::exprType::ListLen: {
                self().replace_ListLen(down_cast<ASR::ListLen_t>(x));
                break;
            }
            case ASR::exprType::ListConcat: {
                self().replace_ListConcat(down_cast<ASR::ListConcat_t>(x));
                break;
            }
            case ASR::exprType::ArrayConstant: {
                self().replace_ArrayConstant(down_cast<ASR::ArrayConstant_t>(x));
                break;
            }
            case ASR::exprType::SetConstant: {
                self().replace_SetConstant(down_cast<ASR::SetConstant_t>(x));
                break;
            }
            case ASR::exprType::SetLen: {
                self().replace_SetLen(down_cast<ASR::SetLen_t>(x));
                break;
            }
            case ASR::exprType::TupleConstant: {
                self().replace_TupleConstant(down_cast<ASR::TupleConstant_t>(x));
                break;
            }
            case ASR::exprType::TupleLen: {
                self().replace_TupleLen(down_cast<ASR::TupleLen_t>(x));
                break;
            }
            case ASR::exprType::StringConstant: {
                self().replace_StringConstant(down_cast<ASR::StringConstant_t>(x));
                break;
            }
            case ASR::exprType::StringConcat: {
                self().replace_StringConcat(down_cast<ASR::StringConcat_t>(x));
                break;
            }
            case ASR::exprType::StringLen: {
                self().replace_StringLen(down_cast<ASR::StringLen_t>(x));
                break;
            }
            case ASR::exprType::DictConstant: {
                self().replace_DictConstant(down_cast<ASR::DictConstant_t>(x));
                break;
            }
            case ASR::exprType::DictLen: {
                self().replace_DictLen(down_cast<ASR::DictLen_t>(x));
                break;
            }
            case ASR::exprType::Var: {
                self().replace_Var(down_cast<ASR::Var_t>(x));
                break;
            }
            case ASR::exprType::ArrayRef: {
                self().replace_ArrayRef(down_cast<ASR::ArrayRef_t>(x));
                break;
            }
            case ASR::exprType::ArraySize: {
                self().replace_ArraySize(down_cast<ASR::ArraySize_t>(x));
                break;
            }
            case ASR::exprType::ArrayBound: {
                self().replace_ArrayBound(down_cast<ASR::ArrayBound_t>(x));
                break;
            }
            case ASR::exprType::ArrayTranspose: {
                self().replace_ArrayTranspose(down_cast<ASR::ArrayTranspose_t>(x));
                break;
            }
            case ASR::exprType::ArrayMatMul: {
                self().replace_ArrayMatMul(down_cast<ASR::ArrayMatMul_t>(x));
                break;
            }
            case ASR::exprType::ArrayPack: {
                self().replace_ArrayPack(down_cast<ASR::ArrayPack_t>(x));
                break;
            }
            case ASR::exprType::Transfer: {
                self().replace_Transfer(down_cast<ASR::Transfer_t>(x));
                break;
            }
            case ASR::exprType::DerivedRef: {
                self().replace_DerivedRef(down_cast<ASR::DerivedRef_t>(x));
                break;
            }
            case ASR::exprType::Cast: {
                self().replace_Cast(down_cast<ASR::Cast_t>(x));
                break;
            }
            case ASR::exprType::ComplexRe: {
                self().replace_ComplexRe(down_cast<ASR::ComplexRe_t>(x));
                break;
            }
            case ASR::exprType::ComplexIm: {
                self().replace_ComplexIm(down_cast<ASR::ComplexIm_t>(x));
                break;
            }
            case ASR::exprType::DictItem: {
                self().replace_DictItem(down_cast<ASR::DictItem_t>(x));
                break;
            }
            default: {
                LFORTRAN_ASSERT_MSG(false, "Duplication of " + std::to_string(x->type) + " expression is not supported yet.");
            }
        }

    }

};


} // namespace LFortran::ASR

#endif // LFORTRAN_ASR_H
